Package: scPipe
Title: Pipeline for single cell multi-omic data pre-processing
Date: 2022-10-12
Version: 2.12.0
Type: Package
Authors@R: c(person("Luyi", "Tian", role=c("aut"), email="tian.l@wehi.edu.au"), 
        person("Shian", "Su", role=c("aut", "cre"), email="su.s@wehi.edu.au"), 
        person("Shalin", "Naik", role=c("ctb"), email="naik.s@wehi.edu.au"), 
        person("Shani", "Amarasinghe", role=c("aut"), email="amarasinghe.s@wehi.edu.au"), 
        person("Oliver", "Voogd", role=c("aut"), email="voogd.o@wehi.edu.au"), 
        person("Phil", "Yang", role=c("aut"), email="yang.p@wehi.edu.au"), 
        person("Matthew", "Ritchie", role=c("ctb"), email="mritchie@wehi.edu.au"))
biocViews: ImmunoOncology, Software, Sequencing, RNASeq,
        GeneExpression, SingleCell, Visualization, SequenceMatching,
        Preprocessing, QualityControl, GenomeAnnotation, DataImport
Description: A preprocessing pipeline for single cell RNA-seq/ATAC-seq data that starts from the fastq files and produces a feature count matrix with associated quality control information. It can process fastq data generated by CEL-seq, MARS-seq, Drop-seq, Chromium 10x and SMART-seq protocols.
Depends: R (>= 4.2.0), SingleCellExperiment
LinkingTo: Rcpp, Rhtslib (>= 1.13.1), testthat
Imports: AnnotationDbi, basilisk, BiocGenerics, biomaRt, Biostrings,
        data.table, dplyr, DropletUtils, flexmix, GenomicRanges,
        GenomicAlignments, GGally, ggplot2, glue (>= 1.3.0), grDevices,
        graphics, hash, IRanges, magrittr, MASS, Matrix (>= 1.5.0),
        mclust, methods, MultiAssayExperiment, org.Hs.eg.db,
        org.Mm.eg.db, purrr, Rcpp (>= 0.11.3), reshape, reticulate,
        Rhtslib, rlang, robustbase, Rsamtools, Rsubread, rtracklayer,
        SummarizedExperiment, S4Vectors, scales, stats, stringr,
        tibble, tidyr, tools, utils, vctrs (>= 0.5.2)
SystemRequirements: C++11, GNU make
License: GPL (>= 2)
Encoding: UTF-8
RoxygenNote: 7.2.3
NeedsCompilation: yes
URL: https://github.com/LuyiTian/scPipe
BugReports: https://github.com/LuyiTian/scPipe
Suggests: BiocStyle, DT, GenomicFeatures, grid, igraph, kableExtra,
        knitr, locStra, plotly, rmarkdown, RColorBrewer, readr,
        reshape2, RANN, shiny, scater (>= 1.11.0), testthat, xml2, umap
VignetteBuilder: knitr
git_url: https://git.bioconductor.org/packages/scPipe
git_branch: RELEASE_3_23
git_last_commit: a390e00
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-28
Packaged: 2026-04-28 23:28:51 UTC; biocbuild
Author: Luyi Tian [aut],
  Shian Su [aut, cre],
  Shalin Naik [ctb],
  Shani Amarasinghe [aut],
  Oliver Voogd [aut],
  Phil Yang [aut],
  Matthew Ritchie [ctb]
Maintainer: Shian Su <su.s@wehi.edu.au>
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-29 05:00:02 UTC; unix

# Generated by roxygen2: do not edit by hand

export("QC_metrics<-")
export("UMI_dup_info<-")
export("demultiplex_info<-")
export("feature_info<-")
export("feature_type<-")
export("gene_id_type<-")
export(QC_metrics)
export(QC_metrics.sce)
export(TF.IDF.custom)
export(UMI_dup_info)
export(UMI_dup_info.sce)
export(anno_import)
export(anno_to_saf)
export(calculate_QC_metrics)
export(convert_geneid)
export(create_processed_report)
export(create_report)
export(create_sce_by_dir)
export(demultiplex_info)
export(demultiplex_info.sce)
export(detect_outlier)
export(feature_info)
export(feature_info.sce)
export(feature_type)
export(feature_type.sce)
export(gene_id_type)
export(gene_id_type.sce)
export(get_ercc_anno)
export(get_genes_by_GO)
export(get_read_str)
export(organism.sce)
export(plot_QC_pairs)
export(plot_UMI_dup)
export(plot_demultiplex)
export(plot_mapping)
export(remove_outliers)
export(sc_aligning)
export(sc_atac_bam_tagging)
export(sc_atac_cell_calling)
export(sc_atac_create_cell_qc_metrics)
export(sc_atac_create_fragments)
export(sc_atac_create_report)
export(sc_atac_create_sce)
export(sc_atac_emptydrops_cell_calling)
export(sc_atac_feature_counting)
export(sc_atac_filter_cell_calling)
export(sc_atac_peak_calling)
export(sc_atac_pipeline)
export(sc_atac_plot_cells_per_feature)
export(sc_atac_plot_features_per_cell)
export(sc_atac_plot_features_per_cell_ordered)
export(sc_atac_plot_fragments_cells_per_feature)
export(sc_atac_plot_fragments_features_per_cell)
export(sc_atac_plot_fragments_per_cell)
export(sc_atac_plot_fragments_per_feature)
export(sc_atac_remove_duplicates)
export(sc_atac_tfidf)
export(sc_atac_trim_barcode)
export(sc_correct_bam_bc)
export(sc_count_aligned_bam)
export(sc_demultiplex)
export(sc_demultiplex_and_count)
export(sc_detect_bc)
export(sc_exon_mapping)
export(sc_gene_counting)
export(sc_get_umap_data)
export(sc_integrate)
export(sc_interactive_umap_plot)
export(sc_mae_plot_umap)
export(sc_trim_barcode)
exportMethods("QC_metrics<-")
exportMethods("UMI_dup_info<-")
exportMethods("demultiplex_info<-")
exportMethods("feature_info<-")
exportMethods("feature_type<-")
exportMethods("gene_id_type<-")
exportMethods("organism<-")
exportMethods(QC_metrics)
exportMethods(UMI_dup_info)
exportMethods(demultiplex_info)
exportMethods(feature_info)
exportMethods(feature_type)
exportMethods(gene_id_type)
exportMethods(organism)
import(Rcpp)
import(Rhtslib)
import(SingleCellExperiment)
import(org.Hs.eg.db)
import(org.Mm.eg.db)
importClassesFrom(SingleCellExperiment,SingleCellExperiment)
importFrom(AnnotationDbi,mapIds)
importFrom(BiocGenerics,"organism<-")
importFrom(BiocGenerics,organism)
importFrom(GenomicRanges,mcols)
importFrom(Rsamtools,sortBam)
importFrom(S4Vectors,DataFrame)
importFrom(S4Vectors,SimpleList)
importFrom(SingleCellExperiment,SingleCellExperiment)
importFrom(SummarizedExperiment,"assay<-")
importFrom(SummarizedExperiment,"assays<-")
importFrom(SummarizedExperiment,"colData<-")
importFrom(SummarizedExperiment,"rowData<-")
importFrom(SummarizedExperiment,assay)
importFrom(SummarizedExperiment,assayNames)
importFrom(SummarizedExperiment,assays)
importFrom(SummarizedExperiment,colData)
importFrom(SummarizedExperiment,rowData)
importFrom(basilisk,BasiliskEnvironment)
importFrom(basilisk,basiliskRun)
importFrom(basilisk,basiliskStart)
importFrom(basilisk,basiliskStop)
importFrom(biomaRt,getBM)
importFrom(biomaRt,useDataset)
importFrom(biomaRt,useMart)
importFrom(data.table,":=")
importFrom(data.table,copy)
importFrom(data.table,fread)
importFrom(data.table,setkey)
importFrom(glue,glue)
importFrom(glue,glue_collapse)
importFrom(magrittr,"%>%")
importFrom(mclust,Mclust)
importFrom(mclust,mclustBIC)
importFrom(methods,as)
importFrom(methods,is)
importFrom(reshape,melt)
importFrom(rlang,.data)
importFrom(robustbase,covMcd)
importFrom(stats,coef)
importFrom(stats,complete.cases)
importFrom(stats,cov)
importFrom(stats,mahalanobis)
importFrom(stats,pchisq)
importFrom(stats,ppoints)
importFrom(stats,prcomp)
importFrom(stats,qchisq)
importFrom(stats,qnorm)
importFrom(stats,reorder)
importFrom(tibble,column_to_rownames)
importFrom(utils,head)
importFrom(utils,packageVersion)
importFrom(utils,read.csv)
importFrom(utils,write.table)
useDynLib(scPipe, .registration = TRUE)

# scPipe

[![Travis build status](https://travis-ci.org/LuyiTian/scPipe.svg?branch=master)](https://travis-ci.org/LuyiTian/scPipe)
[![Coverage
Status](https://codecov.io/gh/LuyiTian/scPipe/branch/master/graph/badge.svg)](https://codecov.io/gh/LuyiTian/scPipe)

<img src=inst/scPipe.png height="200">

scPipe is an R package that allows barcode demultiplexing, transcript mapping and quality control of raw sequencing data generated by 
* multiple 3 prime end sequencing protocols for scRNA-Seq including CEL-seq, MARS-seq, Chromium 10x and Drop-seq and,
* various scATAC-Seq platforms including sci_ATAC, sc-ATAC, 10X, etc. 

RNA-Seq module of scPipe produces a count matrix that is essential for downstream analysis along with a user-friendly HTML report that summarises data quality. These results can be used as input for downstream analyses including normalization, visualization and statistical testing.

ATAC-Seq module of scPipe contains capabilities of pre-processing scATAC-Seq data through scPipe. This functionality allows barcode demultiplexing, peak calling and quality control of raw sequencing data generated by multiple single-cell ATAC-Seq sequencing protocols including 10X, scATAC-Seq, dscATAC-Seq, dsciATAC-Seq, sciATAC-Seq, plate-based ATAC-Seq and scHTS-Seq.
ATAC-Seq module also produces a feature-barcode count matrix that is essential for downstream analysis along with a user-friendly HTML report that summarises data quality.

The scATAC-Seq preprocessing module of the package is under active development. Feel free to ask any questions or submit a pull request.

* [01/04/2021] scPipe now uses *samtools* to remove duplicates from scATAC-Seq data
* [01/04/2021] scPipe now uses *MACS3* for scATAC-Seq peak calling
* [01/04/2021] scPipe now uses *sinto* for scATAC-seq fragment file generation
* [15/02/2021] scPipe scATAC-Seq module also now uses the *SingleCellExperiment* class.
* [21/01/2021] now uses [macsr](https://github.com/macs3-project/MACSr) for peak calling (yet via R developmental version, R.0.1) 
* [03/01/2021] add the cell calling function to scATAC-Seq module for ScPipe (using the package [DropletUtils](https://bioconductor.org/packages/release/bioc/html/DropletUtils.html))
* [12/12/2020] complete 1st version of scATAC-Seq module for ScPipe
* [13/05/2020] initiate scATAC-Seq module for scPipe
* [21/09/2017] scPipe now uses the *SingleCellExperiment* class.

## Installation

### From Bioconductor

```
if (!requireNamespace("BiocManager", quietly=TRUE))
    install.packages("BiocManager")
BiocManager::install("scPipe")
```

### From GitHub (Developmental version)

```{r}
install.packages("devtools")
devtools::install_github("shaniAmare/scPipe")
```

## Getting started

The general workflow of scPipe is illustrated in the following figure:

<img src=inst/workflow_combined.png>

## Concept for scRNA-Seq preprocessing

* The `sc_trim_barcode` function will reformat each read and put the cell barcode and UMI sequence into the fastq read names: `@ACGATCGA_TAGAGC#SIMULATE_SEQ::002::000::0000::0
AAGACGTCTAAGGGCGGTGTACACCCTTTTGAGCAATGATTGCACAACCTGCGATCACCTTATACAGAATTAT+AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA`

* After alignment, the `sc_exon_mapping` function will put the cell barcode and UMI into the bam file with different tags, together with gene information: `AAAGTCAA_AACTCA#SIMULATE_SEQ::007::000::0013::10        0       ERCC-00171      142     40      73M     *       0       0       GCCTCGGGAATAAGCTGACGGTGACAAGGTTTCCCCCTAATCGAGACGCTGCAATAACACAGGGGCATACAGT AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA       HI:i:1  NH:i:1  NM:i:0  GE:Z:ERCC-00171 YC:Z:AAAGTCAA   YM:Z:AACTCA     YE:i:-364`. In this example the cell barcode is AAAGTCAA with tag `YC`, the UMI is AACTCA with tag `YM` and the gene that this read maps to is `ERCC-00171` with tag `GE`. This read is located 364 bp upstream of the transcription end site (TES), which is stored in the `YE` tag.

* The `sc_demultiplex` function will look for the cell barcode in BAM file (by default in the `YC` tag) and compare it against the known cell barcode annotation file, which is a csv file consisting of two columns. The first column is the cell name and second column is the cell barcode. For Chromium 10x and Drop-seq data we can run `sc_detect_bc` to find the barcodes and generate the cell barcode annotation file before running `sc_demultiplex`. An example barcode annotation file is available in the package from `system.file("extdata", "barcode_anno.csv", package = "scPipe")`. The output of `sc_demultiplex` will be multiple csv files corresponding to each cell. Each file has three columns, the first of which contains the gene id, the second column contains the UMI sequence and third column gives the relative location of the read to the TES. These files are used for `sc_gene_counting`.

For further examples see the vignette.

## Concept for scATAC-Seq preprocessing

* The `sc_atac_trim_barcode` function will reformat each read and put the cell barcode and UMI sequence into the fastq read names: `@ACGATCGA_TAGAGC#SIMULATE_SEQ::002::000::0000::0
AAGACGTCTAAGGGCGGTGTACACCCTTTTGAGCAATGATTGCACAACCTGCGATCACCTTATACAGAATTAT+AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA`

* The `sc_atac_aligning` function will align the reformatted fastq files and create bam files.

* After alignment, the `sc_atac_bam_tagging` function will put the cell barcode (and UMI, if available) into the bam file with different tags: `AAAGTCAA_AACTCA#SIMULATE_SEQ::007::000::0013::10        0       ERCC-00171      142     40      73M     *       0       0       GCCTCGGGAATAAGCTGACGGTGACAAGGTTTCCCCCTAATCGAGACGCTGCAATAACACAGGGGCATACAGT AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA       HI:i:1  NH:i:1  NM:i:0  GE:Z:ERCC-00171 YC:Z:AAAGTCAA   YM:Z:AACTCA     YE:i:-364`. In this example the cell barcode is AAAGTCAA with tag `YC`, the UMI is AACTCA with tag `YM` and the gene that this read maps to is `ERCC-00171` with tag `GE`. This read is located 364 bp upstream of the transcription end site (TES), which is stored in the `YE` tag.

* The `sc_atac_peak_calling` function can be used to call peaks using macsr via Linux/Mac environment(). However, macsr is yet not compatible with Windows.

* The `sc_atac_feature_counting` function will generate a feature-count matrix for the alignment and an input feature file (a genome, a bed file format of features, for example one generated through `sc_atac_peak_calling` or MACS2/3). If the feature file is a genome.fasta file a genome_bin approach is used to create the features. It would also generate quality statistics that will get stored in the `scPipe_atac_stats` folder within the working directory. Cell calling is a function implemented in this function to identify the "true" cells. 

* The function `sc_atac_create_sce` generates the `Single Cell Experiment` object from the feature-count matrix and the quality scores acquired throughout the pipeline. It also allows the user to generate a HTML report which can alternatively be created by the `sc_atac_create_report` function.

* The function `sc_atac_create_report` can be run within the `sc_atac_create_sce` or independently to create a report based on the quality statistics available through the processed pipeline.

A minimal example for scATAC-Seq module of scPipe is available [here](minimal_atac_example.md). For further examples see the relevant vignette.

## Acknowledgments
This package is inspired by the `scater`, `scran` and `scATAC=pro` packages. The idea to put cell barcode and UMI sequences into the BAM file is from [Drop-seq tools](http://mccarrolllab.com/dropseq/). Also some features of the scPipe-ATAC module were inspired by the [scATAC-pro](https://github.com/wbaopaul/scATAC-pro) and [SnapTools](https://github.com/r3fang/SnapTools) packages. We thank Dr Aaron Lun for suggestions on package development.

# biomaRt_func.R
ensembl_to_db <- list(ensembl_gene_id="ENSEMBL",
                    external_gene_name="SYMBOL",
                    entrezgene="ENTREZID")

#' Get genes related to certain GO terms from biomart database
#'
#' @param returns the gene id which is set as return. Default to be ensembl id
#' A possible list of attributes can be retrieved using the
#' function \code{listAttributes} from \code{biomaRt} package. The commonly used
#' id types are `external_gene_name`, `ensembl_gene_id` or `entrezgene`.
#' @param dataset Dataset you want to use. List of possible datasets can be
#' retrieved using the function \code{listDatasets} from \code{biomaRt} package.
#' @param go a vector of GO terms
#' @details Get genes related to certain GO terms from biomart database
#'
#' @return a vector of gene ids.
#'
#' @importFrom biomaRt useDataset getBM useMart
#' @importFrom AnnotationDbi mapIds
#'
#' @export
#' @examples
#' # get all genes under GO term GO:0005739 in mouse, return ensembl gene id
#' get_genes_by_GO(returns="ensembl_gene_id",
#'     dataset="mmusculus_gene_ensembl",
#'     go=c('GO:0005739'))
#'
get_genes_by_GO <- function(returns="ensembl_gene_id",
                            dataset="mmusculus_gene_ensembl",
                            go=NULL) {
    if (is.null(go)) {
        stop("must provide GO term. (i.e go=c('GO:0005739'))")
    }
    G_list <- tryCatch({mart <- useDataset(dataset, useMart("ensembl"))
                        G_list <- getBM(filters="go",
                        attributes=c(returns),
                        values=go,
                        mart=mart)
                        G_list},
                    error = function(e){
                        message("cannot connect to the ensembl database. ERROR:\n", e)
                        return(c())})

    if (length(G_list)==0) {
        if (dataset == "hsapiens_gene_ensembl") {
            message("Try to use org.Hs.eg.db.")
            #requireNamespace("org.Hs.eg.db", quietly = TRUE)
            if (returns %in% names(ensembl_to_db)) {
                tmp <- AnnotationDbi::mapIds(org.Hs.eg.db,keys=go,column=ensembl_to_db[returns][[1]], keytype="GO",multiVals = "list")
                return(unname(unlist(tmp)))
            } else {
                message("Unknown gene id type.")
            }
        } else if(dataset == "mmusculus_gene_ensembl") {
            message("Try to use org.Mm.eg.db.")
        #requireNamespace("org.Mm.eg.db", quietly = TRUE)
            if (returns %in% names(ensembl_to_db)){
                tmp <- AnnotationDbi::mapIds(org.Mm.eg.db,keys=go,column=ensembl_to_db[returns][[1]], keytype="GO",multiVals = "list")
                return(unname(unlist(tmp)))
            } else {
                message("Unknown gene id type.")
            }
        }

        return(c())
    }
    return(G_list[, returns])
}


#' convert the gene ids of a SingleCellExperiment object
#'
#' @param sce a SingleCellExperiment object
#' @param returns the gene id which is set as return. Default to be `external_gene_name`
#' A possible list of attributes can be retrieved using the
#' function \code{listAttributes} from \code{biomaRt} package. The commonly used
#' id types are `external_gene_name`, `ensembl_gene_id` or `entrezgene`.
#' @param all logic. For genes that cannot covert to new gene id, keep them with the old
#' id or delete them. The default is keep them.
#' @details convert the gene id of all datas in the SingleCellExperiment object
#'
#' @return sce with converted id
#'
#' @importFrom biomaRt useDataset getBM useMart
#' @importFrom utils head
#' @importFrom methods is
#' @importFrom AnnotationDbi mapIds
#' @import org.Hs.eg.db
#' @import org.Mm.eg.db
#'
#' @export
#' @examples
#' # the gene id in example data are `external_gene_name`
#' # the following example will convert it to `external_gene_name`.
#' data("sc_sample_data")
#' data("sc_sample_qc")
#' sce = SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
#' organism(sce) = "mmusculus_gene_ensembl"
#' gene_id_type(sce) = "ensembl_gene_id"
#' QC_metrics(sce) = sc_sample_qc
#' demultiplex_info(sce) = cell_barcode_matching
#' UMI_dup_info(sce) = UMI_duplication
#' head(rownames(sce))
#' sce = convert_geneid(sce, return="external_gene_name")
#' head(rownames(sce))
#'
convert_geneid <- function(sce,
                            returns="external_gene_name",
                            all=TRUE) {
    sce <- validObject(sce)
    if (returns == gene_id_type(sce)) {
        stop("SingleCellExperiment already has genes in such id type. (gene_id_type(sce) == returns)")
    }

    organism <- organism(sce)
    if (organism == "NA") {
        message("organism not provided.")
        return(sce)
    }
    G_list <- tryCatch({
            mart <- useDataset(organism, useMart("ensembl")) 
            G_list <- getBM(filters=gene_id_type(sce), attributes=c(gene_id_type(sce), returns, "description"), values=rownames(sce), mart=mart)
            G_list
        }, error = function(e){
            message("cannot connect to the ensembl database:\n", e)
                return(c())
        })

    if(length(G_list)==0){
        if(organism == "hsapiens_gene_ensembl"){
            message("Try to use org.Hs.eg.db.")
            if ((returns %in% names(ensembl_to_db)) & (gene_id_type(sce) %in% names(ensembl_to_db))){
                tmp <- AnnotationDbi::mapIds(org.Hs.eg.db,
                            keys=rownames(sce),
                            column=ensembl_to_db[returns][[1]], 
                            keytype=ensembl_to_db[gene_id_type(sce)][[1]],
                            multiVals = "first")
                G_list <- data.frame(description=rep(NA, length(tmp)))
                G_list[, gene_id_type(sce)] <- names(tmp)
                G_list[, returns] <- tmp
            }else{
                message("Unknown gene id type.")
            }
            }else if(organism == "mmusculus_gene_ensembl"){
                message("Try to use org.Mm.eg.db.")
                if (returns %in% names(ensembl_to_db)){
                    tmp <- AnnotationDbi::mapIds(org.Mm.eg.db,
                                keys=rownames(sce),
                                column=ensembl_to_db[returns][[1]], 
                                keytype=ensembl_to_db[gene_id_type(sce)][[1]],
                                multiVals = "first")
                    G_list <- data.frame(description=rep(NA, length(tmp)))
                    G_list[, gene_id_type(sce)] <- names(tmp)
                    G_list[, returns] <- tmp
            } else{
                message("Unknown gene id type.")
            }
        } else{
            return(sce)
        }
    }

    G_list <- G_list[match(rownames(sce), G_list[, gene_id_type(sce)]), ]
    na_num <- sum(is.na(G_list[, returns]))
    dup_ids <- duplicated(G_list[, returns]) | duplicated(G_list[, returns], fromLast=TRUE)
    dup_num <- (sum(dup_ids)-na_num-1)/2
    message("Number of NA in new gene id: ", na_num, ". Duplicated id: ", dup_num)
    if (dup_num>0) {
        message("First 5 duplicated:")
        message(head(G_list[dup_ids & !(is.na(G_list[, returns])), ]))
    }
    G_list[, returns][dup_ids] <- NA
    if (all | (na_num+dup_num==0)) {
        # replace NA with old id
        G_list[, returns][is.na(G_list[, returns])] <- rownames(sce)[is.na(G_list[, returns])]
        if (!(gene_id_type(sce) %in% colnames((sce@int_elementMetadata)))) {
            sce@int_elementMetadata[, gene_id_type(sce)] <- rownames(sce)
        }
        if (!(returns %in% colnames(sce@int_elementMetadata))) {
            sce@int_elementMetadata[, returns] <- G_list[, returns]
        }
        if (!("description" %in% colnames(sce@int_elementMetadata))) {
            sce@int_elementMetadata[, "description"] <- G_list[, "description"]
        }
        rownames(sce) <- G_list[, returns]
    }
    else {
        G_list <- G_list[!is.na(G_list[, returns]), ]
        sce <- sce[!is.na(G_list[, returns]), ]
        if (!(gene_id_type(sce) %in% colnames(sce@int_elementMetadata))) {
            sce@int_elementMetadata[, gene_id_type(sce)] <- rownames(sce)
        }
        if (!(returns %in% colnames(sce@int_elementMetadata))) {
            sce@int_elementMetadata[, returns] <- G_list[, returns]
        }
        if (!("description" %in% colnames(sce@int_elementMetadata))) {
            sce@int_elementMetadata[, "description"] <- G_list[, "description"]
        }
        rownames(sce) <- G_list[, returns]
    }
    gene_id_type(sce) <- returns
    return(sce)
}

#' sc_trim_barcode
#'
#' @description Reformat fastq files so barcode and UMI sequences are moved from
#'   the sequence into the read name.
#'
#'@details Positions used in this function are 0-indexed, so they start from 0
#'  rather than 1. The default read structure in this function represents
#'  CEL-seq paired-ended reads. This contains a transcript in the first read, a
#'  UMI in the first 6bp of the second read followed by a 8bp barcode. So the
#'  read structure will be : \code{list(bs1=-1, bl1=0, bs2=6, bl2=8, us=0,
#'  ul=6)}. \code{bs1=-1, bl1=0} indicates negative start position and zero
#'  length for the barcode on read one, this is used to denote "no barcode" on
#'  read one. \code{bs2=6, bl2=8} indicates there is a barcode in read two that
#'  starts at the 7th base with length 8bp. \code{us=0, ul=6} indicates a UMI
#'  from first base of read two and the length in 6bp.
#'
#'  For a typical Drop-seq experiment the read structure will be
#'  \code{list(bs1=-1, bl1=0, bs2=0, bl2=12, us=12, ul=8)}, which means the read
#'  one only contains transcript, the first 12bp in read two are cell barcode, followed
#'  by a 8bp UMI.
#'
#' @name sc_trim_barcode
#' @param outfq the output fastq file, which reformat the barcode and UMI into
#'   the read name. Files ending in \code{.gz} will be automatically compressed.
#' @param r1 read one for pair-end reads. This read should contain
#'   the transcript.
#' @param r2 read two for pair-end reads, NULL if single read.
#'   (default: NULL)
#' @param read_structure a list containing the read structure configuration:
#'   \itemize{
#'     \item{bs1}: starting position of barcode in read one. -1 if no barcode in
#'       read one.
#'     \item{bl1}: length of barcode in read one, if there is no
#'       barcode in read one this number is used for trimming beginning of read
#'       one.
#'     \item{bs2}: starting position of barcode in read two
#'     \item{bl2}: length of barcode in read two
#'     \item{us}: starting position of UMI
#'     \item{ul}: length of UMI
#'   }
#' @param filter_settings A list contains read filter settings:\itemize{
#'  \item{rmlow} whether to remove the low quality reads.
#'  \item{rmN} whether to remove reads that contains N in UMI or cell barcode.
#'  \item{minq} the minimum base pair quality that we allowed
#'  \item{numbq} the maximum number of base pair that have quality
#'  below \code{numbq}
#'  }
#' @export
#' @return generates a trimmed fastq file named \code{outfq}
#'
#' @examples
#' data_dir="celseq2_demo"
#' \dontrun{
#' # for the complete workflow, refer to the vignettes
#' ...
#' sc_trim_barcode(file.path(data_dir, "combined.fastq"),
#'    file.path(data_dir, "simu_R1.fastq"),
#'    file.path(data_dir, "simu_R2.fastq"))
#' ...
#' }
sc_trim_barcode <- function(outfq, r1, r2=NULL,
                            read_structure = list(
                                bs1=-1, bl1=0, bs2=6, bl2=8, us=0, ul=6),
                            filter_settings = list(
                                rmlow=TRUE, rmN=TRUE, minq=20, numbq=2)) {

    outdir <- regmatches(outfq, regexpr(".*/", outfq))
    
    stopifnot(rlang::is_string(outdir))
    
    if (!dir.exists(outdir)) {
        dir.create(outdir, recursive = TRUE)
    }

    if (filter_settings$rmlow) {
        i_rmlow <- 1
    }
    else {
        i_rmlow <- 0
    }
    if (filter_settings$rmN) {
        i_rmN <- 1
    }
    else {
        i_rmN <- 0
    }

    if (substr(outfq, nchar(outfq) - 2, nchar(outfq)) == ".gz") {
        write_gz <- TRUE
    }
    else {
        write_gz <- FALSE
    }

    if (!is.null(r2)) {
        if (!file.exists(r1)) {stop("read1 fastq file does not exists.")}
        if (!file.exists(r2)) {stop("read2 fastq file does not exists.")}

        # expand tilde to home path for downstream gzopen() call
        r1 <- path.expand(r1)
        r2 <- path.expand(r2)


        rcpp_sc_trim_barcode_paired(outfq, r1, r2,
                                    read_structure$bs1,
                                    read_structure$bl1,
                                    read_structure$bs2,
                                    read_structure$bl2,
                                    read_structure$us,
                                    read_structure$ul,
                                    i_rmlow,
                                    i_rmN,
                                    filter_settings$minq,
                                    filter_settings$numbq,
                                    write_gz)
    }
    else {
        stop("not implemented.")
    }
}


#' sc_exon_mapping
#'
#' @description Map aligned reads to exon annotation.
#' The result will be written into optional fields in bam file with different
#' tags. Following this link for more information regarding to bam file format:
#' http://samtools.github.io/hts-specs
#'
#' The function can accept multiple bam file as input, if multiple bam file is
#' provided and the `bc_len` is zero, then the function will use the barcode in
#' the `barcode_vector` to insert into the `bc` bam tag. So the length of
#' `barcode_vector` and the length of `inbam` should be the same
#' If this is the case then the `max_mis` argument in `sc_demultiplex`
#' should be zero. If `be_len` is larger than zero, then the function will
#' still seek for barcode in fastq headers with given length. In this case
#' each bam file is not treated as from a single cell.
#'
#'
#' @name sc_exon_mapping
#' @param inbam input aligned bam file. can have multiple files as input
#' @param outbam output bam filename
#' @param annofn single string or vector of gff3 annotation filenames,
#'   data.frame in SAF format or GRanges object containing complete gene_id
#'   metadata column.
#' @param bam_tags list defining BAM tags where mapping information is
#'   stored.
#'   \itemize{
#'     \item "am": mapping status tag
#'     \item "ge": gene id
#'     \item "bc": cell barcode tag
#'     \item "mb": molecular barcode tag
#'   }
#' @param bc_len total barcode length
#' @param barcode_vector a list of barcode if each individual bam is a single
#' cell. (default: NULL). The barcode should be of the same length for each
#' cell.
#' @param UMI_len UMI length
#' @param stnd TRUE to perform strand specific mapping. (default: TRUE)
#' @param fix_chr TRUE to add `chr` to chromosome names, MT to chrM. (default: FALSE)
#' @param nthreads number of threads to use. (default: 1)
#'
#' @export
#' @return generates a bam file with exons assigned
#' @examples
#' data_dir="celseq2_demo"
#' ERCCanno_fn = system.file("extdata", "ERCC92_anno.gff3",
#'     package = "scPipe")
#' \dontrun{
#' # for the complete workflow, refer to the vignettes
#' ...
#' sc_exon_mapping(file.path(data_dir, "out.aln.bam"),
#'                 file.path(data_dir, "out.map.bam"),
#'                 ERCCanno_fn)
#' ...
#' }
#'
sc_exon_mapping <- function(inbam, outbam, annofn,
                                bam_tags = list(am="YE", ge="GE", bc="BC", mb="OX"),
                                bc_len=8, barcode_vector="", UMI_len=6, stnd=TRUE, fix_chr=FALSE,
                                nthreads=1) {
    if (stnd) {
        i_stnd <- 1
    }
    else {
        i_stnd <- 0
    }
    if (fix_chr) {
        i_fix_chr <- 1
    }
    else {
        i_fix_chr <- 0
    }

    if (any(!file.exists(inbam))) {
        stop("At least one input bam file does not exist")
    } else {
        inbam <- path.expand(inbam)
    }

    outbam <- path.expand(outbam)

    # if (length(inbam) > 1) {
    #   stop("Only one bam file can be used as input")
    # }

    if (!is(annofn, "character") &&
        !is(annofn, "GRanges") &&
        !is(annofn, "data.frame")) {
        stop("'annofn' must be either character vector, GRanges, or data.frame object")
    }

    if (is(annofn, "character")) {
        if (any(!file.exists(annofn))) {
            stop("At least one genome annotation file does not exist")
        } else {
            annofn <- path.expand(annofn)
        }
        rcpp_sc_exon_mapping_df_anno(inbam, outbam, anno_import(annofn), bam_tags$am, bam_tags$ge, bam_tags$bc, bam_tags$mb, bc_len,
                                    barcode_vector, UMI_len, stnd, fix_chr, nthreads)
    } else if (is(annofn, "GRanges")) {
        rcpp_sc_exon_mapping_df_anno(inbam, outbam, anno_to_saf(annofn), bam_tags$am, bam_tags$ge, bam_tags$bc, bam_tags$mb, bc_len,
                                    barcode_vector, UMI_len, stnd, fix_chr, nthreads)
    } else if (is(annofn, "data.frame")) {
        validate_saf(annofn)
        rcpp_sc_exon_mapping_df_anno(inbam, outbam, annofn, bam_tags$am, bam_tags$ge, bam_tags$bc, bam_tags$mb, bc_len,
                                    barcode_vector, UMI_len, stnd, fix_chr, nthreads)
    }
}


#' sc_demultiplex
#'
#' @description Process bam file by cell barcode,
#' output to outdir/count/[cell_id].csv.
#' the output contains information for all reads that can be mapped to exons.
#' including the gene id, UMI of that read and the distance to transcript end
#' position.
#'
#' @param inbam input bam file. This should be the output of
#' \code{sc_exon_mapping}
#' @param outdir output folder
#' @param bc_anno barcode annotation, first column is cell id, second column
#' is cell barcode sequence
#' @param max_mis maximum mismatch allowed in barcode. (default: 1)
#' @param bam_tags list defining BAM tags where mapping information is
#'   stored.
#'   \itemize{
#'     \item "am": mapping status tag
#'     \item "ge": gene id
#'     \item "bc": cell barcode tag
#'     \item "mb": molecular barcode tag
#'   }
#' @param mito mitochondrial chromosome name.
#' This should be consistent with the chromosome names in the bam file.
#' @param has_UMI whether the protocol contains UMI (default: TRUE)
#' @param nthreads number of threads to use. (default: 1)
#'
#' @export
#' @return no return
#' @examples
#' data_dir="celseq2_demo"
#' barcode_annotation_fn = system.file("extdata", "barcode_anno.csv",
#'     package = "scPipe")
#' \dontrun{
#' # refer to the vignettes for the complete workflow
#' ...
#' sc_demultiplex(file.path(data_dir, "out.map.bam"),
#'     data_dir,
#'     barcode_annotation_fn,has_UMI=FALSE)
#' ...
#' }
#'
sc_demultiplex <- function(inbam, outdir, bc_anno,
                            max_mis=1,
                            bam_tags = list(am="YE", ge="GE", bc="BC", mb="OX"),
                            mito="MT",
                            has_UMI=TRUE,
                            nthreads = 1) {
    dir.create(file.path(outdir, "count"), showWarnings = FALSE)
    dir.create(file.path(outdir, "stat"), showWarnings = FALSE)

    if (!dir.exists(outdir))
        dir.create(outdir, recursive = TRUE)

    outdir <- path.expand(outdir)

    if (!file.exists(inbam)) {
        stop("input bam file does not exists.")
    } else {
        inbam <- path.expand(inbam)
    }
    if (!file.exists(bc_anno)) {
        stop("barcode annotation file does not exists.")
    } else {
        bc_anno <- path.expand(bc_anno)
    }
    rcpp_sc_demultiplex(inbam, outdir, bc_anno, max_mis,
                        bam_tags$am, bam_tags$ge, bam_tags$bc, bam_tags$mb,
                        mito, has_UMI, nthreads)
}


#' sc_correct_bam_bc
#'
#' @description update the cell barcode tag in bam file with corrected barcode
#' output to a new bam file. the function will be useful for methods
#' that use the cell barcode information from bam file, such as `Demuxlet`
#'
#' @param inbam input bam file. This should be the output of
#' \code{sc_exon_mapping}
#' @param outbam output bam file with updated cell barcode
#' @param bc_anno barcode annotation, first column is cell id, second column
#' is cell barcode sequence
#' @param max_mis maximum mismatch allowed in barcode. (default: 1)
#' @param bam_tags list defining BAM tags where mapping information is
#'   stored.
#'   \itemize{
#'     \item "am": mapping status tag
#'     \item "ge": gene id
#'     \item "bc": cell barcode tag
#'     \item "mb": molecular barcode tag
#'   }
#' @param mito mitochondrial chromosome name.
#' This should be consistent with the chromosome names in the bam file.
#' @param nthreads number of threads to use. (default: 1)
#'
#' @export
#' @return no return
#' @examples
#' data_dir="celseq2_demo"
#' barcode_annotation_fn = system.file("extdata", "barcode_anno.csv",
#'     package = "scPipe")
#' \dontrun{
#' # refer to the vignettes for the complete workflow
#' ...
#' sc_correct_bam_bc(file.path(data_dir, "out.map.bam"),
#'     file.path(data_dir, "out.map.clean.bam"),
#'     barcode_annotation_fn)
#' ...
#' }
#'
sc_correct_bam_bc <- function(inbam, outbam, bc_anno,
                            max_mis=1,
                            bam_tags = list(am="YE", ge="GE", bc="BC", mb="OX"),
                            mito="MT",
                            nthreads = 1) {

    if (!file.exists(inbam)) {
        stop("input bam file does not exists.")
    } else {
        inbam <- path.expand(inbam)
    }

    outbam <- path.expand(outbam)

    if (!file.exists(bc_anno)) {
        stop("barcode annotation file does not exists.")
    } else {
        bc_anno <- path.expand(bc_anno)
    }
    rcpp_sc_clean_bam(inbam, outbam, bc_anno, max_mis,
                        bam_tags$am, bam_tags$ge, bam_tags$bc, bam_tags$mb,
                        mito, nthreads)
}


#' sc_gene_counting
#'
#' @description Generate gene counts matrix with UMI deduplication
#'
#' @param outdir output folder containing \code{sc_demultiplex} output
#' @param bc_anno barcode annotation comma-separated-values, first column is
#'   cell id, second column is cell barcode sequence
#' @param UMI_cor correct UMI sequencing error: 0 means no correction, 1 means
#'   simple correction and merge UMI with distance 1. 2 means merge on both UMI
#'   alignment position match.
#' @param gene_fl whether to remove low abundance genes. A gene is considered to
#'   have low abundance if only one copy of one UMI is associated with it.
#'
#' @export
#' @return no return
#' @examples
#' data_dir="celseq2_demo"
#' barcode_annotation_fn = system.file("extdata", "barcode_anno.csv",
#' package = "scPipe")
#' \dontrun{
#' # refer to the vignettes for the complete workflow
#' ...
#' sc_gene_counting(data_dir, barcode_annotation_fn)
#' ...
#' }
#'
sc_gene_counting <- function(outdir, bc_anno, UMI_cor=2, gene_fl=FALSE) {
    if (!dir.exists(outdir))
        dir.create(outdir, recursive = TRUE)

    outdir <- path.expand(outdir)

    dir.create(file.path(outdir, "count"), showWarnings = FALSE)
    dir.create(file.path(outdir, "stat"), showWarnings = FALSE)

    if (gene_fl) {
        i_gene_fl <- 1
    } else {
        i_gene_fl <- 0
    }

    if (!file.exists(bc_anno)) {
        stop("barcode annotation file does not exists.")
    } else {
        bc_anno <- path.expand(bc_anno)
    }

    rcpp_sc_gene_counting(outdir, bc_anno, UMI_cor, i_gene_fl)
}



#' sc_detect_bc
#'
#' @description Detect cell barcode and generate the barcode annotation
#'
#' @param infq input fastq file, should be the output file of
#' \code{sc_trim_barcode}
#' @param outcsv output barcode annotation
#' @param prefix the prefix of cell name (default: `CELL_`)
#' @param bc_len the length of cell barcode, should be consistent with bl1+bl2
#' in \code{sc_trim_barcode}
#' @param max_reads the maximum of reads processed (default: 1,000,000)
#' @param min_count minimum counts to keep, barcode will be discarded if
#' it has lower count. Default value is 10. This should be set according
#' to \code{max_reads}.
#' @param number_of_cells number of cells kept in result. (default: 10000)
#' @param max_mismatch the maximum mismatch allowed. Barcodes within this
#' number will be considered as sequence error and merged. (default: 1)
#' @param white_list_file a file that list all the possible barcodes
#' each row is a barcode sequence. the list for 10x can be found at:
#' https://community.10xgenomics.com/t5/Data-Sharing/List-of-valid-cellular-barcodes/td-p/527
#' (default: NULL)
#' @export
#' @return no return
#' @examples
#' \dontrun{
#' # `sc_detect_bc`` should run before `sc_demultiplex` for
#' # Drop-seq or 10X protocols
#' sc_detect_bc("input.fastq","output.cell_index.csv",bc_len=8)
#' sc_demultiplex(...,"output.cell_index.csv")
#' }
#'
#'
sc_detect_bc <- function(infq, outcsv, prefix="CELL_", bc_len,
                            max_reads=1000000, min_count=10, number_of_cells=10000,
                            max_mismatch=1,white_list_file=NULL) {
    if (!file.exists(infq)) {
        stop("input fastq file does not exists.")
    } else {
        infq <- path.expand(infq)
    }
    outcsv <- path.expand(outcsv)
    if (!is.null(white_list_file)){
        if(!file.exists(white_list_file)){
            stop("input whitelist file does not exists.")
        }
    } else {
        white_list_file <- ""
    }
    if (max_reads=="all") {m_r <- -1}
    else {m_r<-max_reads}
    rcpp_sc_detect_bc(infq, outcsv, prefix, bc_len, m_r, number_of_cells,
        min_count, max_mismatch, white_list_file)
}

#' sc_demultiplex_and_count
#'
#' @description Wrapper to run \code{\link{sc_demultiplex}} and
#'   \code{\link{sc_gene_counting}} with a single command
#'
#' @inheritParams sc_demultiplex
#' @inheritParams sc_gene_counting
#'
#' @return no return
#'
#' @examples
#' \dontrun{
#' refer to the vignettes for the complete workflow, replace demultiplex and
#' count with single command:
#' ...
#' sc_demultiplex_and_count(
#'    file.path(data_dir, "out.map.bam"),
#'    data_dir,
#'    barcode_annotation_fn,
#'    has_UMI = FALSE
#' )
#' ...
#' }
#'
#' @export
#'
sc_demultiplex_and_count <- function(
    inbam, outdir, bc_anno,
    max_mis = 1,
    bam_tags = list(am="YE", ge="GE", bc="BC", mb="OX"),
    mito = "MT", has_UMI = TRUE, UMI_cor = 1, gene_fl = FALSE,
    nthreads = 1) {
    sc_demultiplex(
        inbam = inbam,
        outdir = outdir,
        bc_anno = bc_anno,
        max_mis = max_mis,
        bam_tags = bam_tags,
        mito = mito,
        has_UMI = has_UMI,
        nthreads = nthreads
    )

    sc_gene_counting(
        outdir = outdir,
        bc_anno = bc_anno,
        UMI_cor = UMI_cor,
        gene_fl = gene_fl
    )
}

#' sc_count_aligned_bam
#'
#' @description Wrapper to run \code{\link{sc_exon_mapping}},
#'   \code{\link{sc_demultiplex}} and \code{\link{sc_gene_counting}} with a
#'   single command
#'
#' @inheritParams sc_exon_mapping
#' @inheritParams sc_demultiplex
#' @inheritParams sc_gene_counting
#' @param keep_mapped_bam TRUE if feature mapped bam file should be retained.
#'
#' @return no return
#'
#' @examples
#' \dontrun{
#' sc_count_aligned_bam(
#'   inbam = "aligned.bam",
#'   outbam = "mapped.bam",
#'   annofn = c("MusMusculus-GRCm38p4-UCSC.gff3", "ERCC92_anno.gff3"),
#'   outdir = "output",
#'   bc_anno = "barcodes.csv"
#' )
#' }
#'
#' @export
#'
sc_count_aligned_bam <- function(
    inbam, outbam, annofn,
    bam_tags = list(am="YE", ge="GE", bc="BC", mb="OX"),
    bc_len = 8, UMI_len = 6,
    stnd = TRUE, fix_chr = FALSE,
    outdir,
    bc_anno,
    max_mis = 1,
    mito = "MT",
    has_UMI = TRUE, UMI_cor = 1, gene_fl = FALSE,
    keep_mapped_bam = TRUE,
    nthreads = 1
    ) {
    sc_exon_mapping(
        inbam = inbam,
        outbam = outbam,
        annofn = annofn,
        bam_tags = bam_tags,
        bc_len = bc_len,
        UMI_len = UMI_len,
        stnd = stnd,
        fix_chr = fix_chr,
        nthreads = nthreads
    )

    sc_demultiplex_and_count(
        inbam = outbam,
        outdir = outdir,
        bc_anno = bc_anno,
        max_mis = max_mis,
        bam_tags = bam_tags,
        mito = mito,
        has_UMI = has_UMI,
        UMI_cor = UMI_cor,
        gene_fl = gene_fl,
        nthreads = nthreads
    )

    if (!keep_mapped_bam) {
        unlink(outbam)
    }
}

###############################
# Peak calling Using MACS3
###############################

#' sc_atac_peak_calling()
#'
#' @param inbam The input tagged, sorted, duplicate-free input BAM file
#' @param ref The reference genome file
#' @param genome_size The size of the genome
#' @param output_folder The path of the output folder
#'
#' @returns None (invisible `NULL`)
#' @examples
#' \dontrun{
#' sc_atac_peak_calling(
#'     inbam,
#'     reference) 
#' }
#' @export
#' 
#' @importFrom basilisk basiliskRun basiliskStart basiliskStop
sc_atac_peak_calling <- function(inbam, 
                                ref = NULL, 
                                genome_size = NULL,
                                output_folder = NULL) {
  
    if (is.null(ref) && is.null(genome_size)) {
        stop("No genome or genome size was specified. Must specify at least one!")
    } else if (is.null(genome_size)) {
        if (is.null(ref) || !file.exists(ref)) {
            stop("Specified reference file doesn't exist!")
        }
        genome_size <- signif(file.info(ref)$size, 2)
        message("Using genome size of", genome_size)
    }
    
    
    if(is.null(output_folder)) {
        output_folder <- file.path(getwd(), "scPipe-atac-output")
    }
    
    if (!dir.exists(output_folder)){
        dir.create(output_folder,recursive=TRUE)
        message("Output directory does not exist. Created: ", output_folder)
    }
    
    # Create log folder/file
    log_and_stats_folder <- file.path(output_folder, "scPipe_atac_stats")
    dir.create(log_and_stats_folder, showWarnings = FALSE)
    log_file <- file.path(log_and_stats_folder, "log_file.txt")
    if(!file.exists(log_file)) file.create(log_file)
    write(
        c(
        "sc_atac_peak_calling starts at ",
        as.character(Sys.time()),
        "\n"
        ),
        file = log_file, append = TRUE)
    
    # MACSr::callpeak(inbam, 
    #                 nomodel         = TRUE, 
    #                 shift           = 100, 
    #                 extsize         = 200, 
    #                 gsize           = genome_size, 
    #                 call_summits    = TRUE, 
    #                 store_bdg       = TRUE,
    #                 do_SPMR         = TRUE,
    #                 cutoff_analysis = TRUE,
    #                 outdir          = output_folder)
    
    proc <- basilisk::basiliskStart(scPipe_env)
    on.exit(basilisk::basiliskStop(proc))

    basilisk::basiliskRun(proc, function(inbam, genome_size, output) {
        macs <- reticulate::import("MACS3")
        system2("macs3", c("callpeak", "-t", 
                        inbam, 
                        "--nomodel", 
                        "--shift", 100, 
                        "--extsize", 200, 
                        "--g", genome_size, 
                        "--outdir", output_folder,
                        "--bdg", 
                        "--SPMR", 
                        "--cutoff-analysis",
                        "--call-summits"))
    }, inbam=inbam, genome_size=genome_size, output=output_folder)
    
    write(
        c(
        "sc_atac_peak_calling finishes at ",
        as.character(Sys.time()),
        "\n"
        ), file = log_file, append = TRUE)
}

#' Detect outliers based on robust linear regression of QQ plot
#' @param x a vector of mahalanobis distance
#' @param df degree of freedom for chi-square distribution
#' @param conf confidence for linear regression
#'
#' @return cell names of outliers
#' @importFrom stats ppoints qchisq coef qnorm 
.qq_outliers_robust <- function(x, df, conf) {
    n <- length(x)
    P <- stats::ppoints(n)
    z <- stats::qchisq(P, df=df)
    ord.x <- x[order(x)]
    coef <- stats::coef(MASS::rlm(ord.x ~ z, maxit=2000))
    a <- coef[1]
    b <- coef[2]
    zz <- stats::qnorm(1 - (1 - conf)/2)
    SE <- (b/stats::dchisq(z, df))*sqrt(P*(1 - P)/n)
    fit.value <- a + b*z
    upper <- fit.value + zz*SE
    thr <- max(ord.x[ord.x < upper])
    return(names(x[x > thr]))
}


#' Detect outliers based on QC metrics
#'
#' @description This algorithm will try to find \code{comp} number of components
#'   in quality control metrics using a Gaussian mixture model. Outlier
#'   detection is performed on the component with the most genes detected. The
#'   rest of the components will be considered poor quality cells. More cells
#'   will be classified low quality as you increase \code{comp}.
#
#' @param sce a \code{SingleCellExperiment} object containing QC metrics.
#' @param comp the number of component used in GMM. Depending on the quality of
#'  the experiment.
#' @param sel_col a vector of column names which indicate the columns to use
#'  for QC.
#' By default it will be the statistics generated by `calculate_QC_metrics()`
#' @param type only looking at low quality cells (`low`) or possible
#' doublets (`high`) or both (`both`)
#' @param conf confidence interval for linear regression at lower and
#' upper tails.Usually, this is smaller for lower tail because we hope
#' to pick out more low quality cells than doublets.
#' @param batch whether to perform quality control separately for each batch.
#' Default is FALSE. If set to TRUE then you should have a column called
#' `batch` in the `colData(sce)`.
#'
#' @details detect outlier using Mahalanobis distances
#'
#' @return an updated \code{SingleCellExperiment} object with an `outlier`
#' column in \code{colData}
#'
#' @importFrom stats cov pchisq mahalanobis complete.cases qchisq
#' @importFrom robustbase covMcd
#' @importFrom mclust Mclust mclustBIC
#' @importFrom glue glue glue_collapse
#'
#' @export
#' @examples
#' data("sc_sample_data")
#' data("sc_sample_qc")
#' sce = SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
#' organism(sce) = "mmusculus_gene_ensembl"
#' gene_id_type(sce) = "ensembl_gene_id"
#' QC_metrics(sce) = sc_sample_qc
#' demultiplex_info(sce) = cell_barcode_matching
#' UMI_dup_info(sce) = UMI_duplication
#' # the sample qc data already run through function `calculate_QC_metrics`
#' # for a new sce please run `calculate_QC_metrics` before `detect_outlier`
#' sce = detect_outlier(sce)
#' table(QC_metrics(sce)$outliers)
#'
detect_outlier <- function(sce,
                        comp=1,
                        sel_col=NULL,
                        type=c("low", "both", "high"),
                        conf=c(0.9, 0.99),
                        batch=FALSE) {
    type <- match.arg(type)
    sce <- validObject(sce)
    # check format:
    if (is.null(sel_col)) {
        sel_col <- c("number_of_genes", "total_count_per_cell",
                    "non_mt_percent", "non_ERCC_percent")
    }
    if (!all(sel_col %in% colnames(QC_metrics(sce)))) {
        missing <- sel_col[!(sel_col %in% colnames(QC_metrics(sce)))]

        missing_vars <- glue_collapse(glue("'{missing}'"), sep = ",", last = " and ")
        message(glue("the following QC metrics not found in colData from sce: {missing_vars}"))

        if (any(c("number_of_genes", "total_count_per_cell") %in% missing_vars)) {
            stop("the quality control metrics should at least contain the number of
                genes(`number_of_genes`) and counts per cell(`total_count_per_cell`)!")
        }
    }
    qc_cols <- colnames(QC_metrics(sce)) %in% sel_col
    x_all <- as.data.frame(QC_metrics(sce)[, qc_cols])
    x_all$total_count_per_cell <- log2(x_all$total_count_per_cell + 1)

    if (!all(stats::complete.cases(x_all))) {
        stop("NAs found in the selected columns, check the quality control matrix")
    }
    if (is.null(dim(x_all))) {
        QC_metrics(sce)$outliers <- FALSE
        return(sce)
    }
    if (!batch) {
        batch_info <- rep(1, nrow(x_all))
    } else {
        if ("batch" %in% colnames(colData(sce))) {
            batch_info <- colData(sce)$batch
        } else {
            stop("did not find column `batch` in colData(sce).")
        }
    }

    if(any(rowSums(x_all) == 0)){
        stop("Some cells have zero counts. remove these cells before detecting outlier.")
    }else if(any(colSums(x_all) == 0)){
        stop("Some QC metrics are all zeros, check your quality control matrix by QC_metrics(sce).")
    }
    outliers <- rep(FALSE, nrow(x_all))

    for (a_batch in unique(batch_info)) {
        x <- x_all[batch_info == a_batch, ]
        dist <- mahalanobis(x, center=colMeans(x), cov=cov(x))
        keep <- !(dist > qchisq(0.99, ncol(x)))
        mod <- Mclust(x[keep, ],
                    G=comp,
                    modelNames=c("EEE","EVV", "VVV"),
                    verbose=FALSE)

        if (comp == 1) {
            covr <- covMcd(x, alpha=0.7)
            dist <- mahalanobis(x,
                                center=covr$center,
                                cov=covr$cov)
            mean_diff <- sign(t(x) - covr$center)
            QC_sign <- c(-1, 1)[as.factor(apply(mean_diff, 2, function(t) {sum(t) > 0}))]
            neg_dist <- dist[QC_sign == -1]
            pos_dist <- dist[QC_sign == 1]
            if (type == "both") {
                outlier_cells <- .qq_outliers_robust(neg_dist, ncol(x), conf[1])
                outlier_cells <- c(outlier_cells,
                                .qq_outliers_robust(pos_dist, ncol(x), conf[2]))
            } else if (type == "low") {
                outlier_cells <- .qq_outliers_robust(neg_dist, ncol(x), conf[1])
            } else if (type == "high") {
                outlier_cells <- .qq_outliers_robust(pos_dist, ncol(x), conf[2])
            }
        } else {
            ord_fst <- c(seq_len(comp))[order(mod$parameters$mean[1, ], decreasing = TRUE)]
            poor_comp <- ord_fst[2:comp]
            good_comp <- ord_fst[1]
            keep1 <- rep(TRUE, nrow(x))
            keep1[keep][mod$classification %in% poor_comp] <- FALSE
            keep1[!keep] <- FALSE
            sub_x <- x[keep1, ]
            covr <- covMcd(sub_x, alpha=0.7)
            sub_dist <- mahalanobis(sub_x,
                                    center=covr$center,
                                    cov=covr$cov)

            mean_diff <- sign(t(sub_x) - covr$center)
            QC_sign <- c(-1, 1)[as.factor(apply(mean_diff, 2, function(t) {sum(t) > 0}))]
            neg_dist <- sub_dist[QC_sign == -1]
            pos_dist <- sub_dist[QC_sign == 1]

            outlier_cells <- .qq_outliers_robust(neg_dist, ncol(sub_x), conf[1])
            outlier_cells <- c(outlier_cells,
                                .qq_outliers_robust(pos_dist, ncol(sub_x), conf[2]))
            outlier_cells <- c(outlier_cells, rownames(x[!keep1, ]))

            if (!(type == "both")) {
                mean_diff <- sign(t(x)-mod$parameters$mean[, good_comp])
                QC_sign <- c(-1, 1)[as.factor(apply(mean_diff, 2, function(t) {sum(t) > 0}))]
                is_outlier_cell <- rownames(x) %in% outlier_cells
                if (type == "low") {
                    outlier_cells <- rownames(x)[is_outlier_cell & (QC_sign == -1)]
                } else if (type == "high") {
                    outlier_cells <- rownames(x)[is_outlier_cell & (QC_sign == 1)]
                }
            }
        }

        if (any(rownames(x) %in% outlier_cells)) {
            outliers[batch_info == a_batch][rownames(x)  %in% outlier_cells] <- TRUE
        }
    }
    QC_metrics(sce)$outliers <- as.factor(outliers)
    return(sce)
}


#' Calculate QC metrics from gene count matrix
#'
#' @param sce a \code{SingleCellExperiment} object containing gene counts
#' @details get QC metrics using gene count matrix.
#' The QC statistics added are \itemize{
#'  \item{number_of_genes} number of genes detected.
#'  \item{total_count_per_cell} sum of read number after UMI deduplication.
#'  \item{non_mt_percent} 1 - percentage of mitochondrial gene counts.
#'  Mitochondrial genes are retrived by GO term GO:0005739
#'  \item{non_ERCC_percent} ratio of exon counts to ERCC counts
#'  \item{non_ribo_percent} 1 - percentage of ribosomal gene counts
#' ribosomal genes are retrived by GO term GO:0005840.
#' }
#' @return an \code{SingleCellExperiment} with updated QC metrics
#'
#' @importFrom SummarizedExperiment assay assay<- assays assays<- assayNames rowData rowData<- colData colData<-
#'
#' @export
#' @examples
#' data("sc_sample_data")
#' data("sc_sample_qc")
#' sce <- SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
#' organism(sce) <- "mmusculus_gene_ensembl"
#' gene_id_type(sce) <- "ensembl_gene_id"
#' QC_metrics(sce) <- sc_sample_qc
#' demultiplex_info(sce) <- cell_barcode_matching
#' UMI_dup_info(sce) <- UMI_duplication
#'
#' # The sample qc data already run through function `calculate_QC_metrics`.
#' # So we delete these columns and run `calculate_QC_metrics` to get them again:
#' colnames(colnames(QC_metrics(sce)))
#' QC_metrics(sce) <- QC_metrics(sce)[,c("unaligned","aligned_unmapped","mapped_to_exon")]
#' sce = calculate_QC_metrics(sce)
#' colnames(QC_metrics(sce))
#'
calculate_QC_metrics <- function(sce) {
    sce <- validObject(sce) # check the sce object
    if (!("counts" %in% names(assays(sce)))){
        stop("counts not in names(assays(sce)). Cannot find count data.")
    }

    # get gene number
    gene_number <- colSums(assay(sce,"counts")>0)
    if (all(gene_number == 0)) {
        stop("all genes have zero count. Check the expression matrix.")
    }else{
        if (!("scPipe" %in% names(sce@metadata))){
            sce@metadata[["scPipe"]] <- list(QC_cols=c("number_of_genes"))
        } else if (!("QC_cols" %in% names(sce@metadata$scPipe))){
            sce@metadata[["scPipe"]] <- list(QC_cols=c("number_of_genes"))
            QC_metrics(sce) <- DataFrame(row.names = colnames(sce))
            # create a empty QC metrics if not exists
        }
        QC_metrics(sce)$number_of_genes <- gene_number
    }

    # get total count per cell
    QC_metrics(sce)$total_count_per_cell <- colSums(assay(sce,"counts"))

    # get ERCC ratio
    if(any(grepl("^ERCC-", rownames(sce)))){
        exon_count <- colSums(assay(sce,"counts")[!grepl("^ERCC-", rownames(sce)), , drop = FALSE])
        ERCC_count <- colSums(assay(sce,"counts")[grepl("^ERCC-", rownames(sce)), , drop = FALSE])
        QC_metrics(sce)$non_ERCC_percent <- exon_count/(ERCC_count+exon_count+1e-5)
    }else{
        message("no ERCC spike-in. Skip `non_ERCC_percent`")
        }

    # get mt percentage
    if (!is.na(gene_id_type(sce))) {
        mt_genes <- get_genes_by_GO(returns=gene_id_type(sce),
                                    dataset=organism(sce),
                                    go=c("GO:0005739"))
        if (length(mt_genes)>0) {
            if (sum(rownames(sce) %in% mt_genes)>1) {
                mt_count <- colSums(assay(sce,"counts")[rownames(sce) %in% mt_genes, , drop = FALSE])
                QC_metrics(sce)$non_mt_percent <- (colSums(assay(sce,"counts"))-
                                                    mt_count)/(colSums(assay(sce,"counts"))+1e-5)
                # add 1e-5 to make sure they are not NA
            }
        }
    }
    else {
        message("no gene_id_type, skip `non_mt_percent`")
    }

    # get ribosomal percentage
    if (!is.na(gene_id_type(sce))) {
        ribo_genes <- get_genes_by_GO(returns=gene_id_type(sce),
                                dataset=organism(sce),
                                go=c("GO:0005840"))
        if (length(ribo_genes)>0) {
            if (sum(rownames(sce) %in% ribo_genes)>1) {
                ribo_count <- colSums(assay(sce,"counts")[rownames(sce) %in% ribo_genes, ])
                QC_metrics(sce)$non_ribo_percent <- (colSums(assay(sce,"counts"))-
                                                    ribo_count)/(colSums(assay(sce,"counts"))+0.01)
            }
        }
    }
    else {
        message("no gene_id_type, skip `non_ribo_percent`")
    }
    return(sce)
}


#' Plot GGAlly pairs plot of QC statistics from \code{SingleCellExperiment}
#' object
#'
#' @param sce a \code{SingleCellExperiment} object
#' @param sel_col a vector of column names which indicate the columns to use for
#'   plot. By default it will be the statistics generated by
#'   `calculate_QC_metrics()`
#' @export
#' @return a ggplot2 object
#' @examples
#' data("sc_sample_data")
#' data("sc_sample_qc")
#' sce = SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
#' organism(sce) = "mmusculus_gene_ensembl"
#' gene_id_type(sce) = "ensembl_gene_id"
#' QC_metrics(sce) = sc_sample_qc
#' demultiplex_info(sce) = cell_barcode_matching
#' UMI_dup_info(sce) = UMI_duplication
#' sce = detect_outlier(sce)
#'
#' plot_QC_pairs(sce)
#'
plot_QC_pairs <- function(sce, sel_col=NULL) {
    sce <- validObject(sce) # check the sce object
    if (is.null(sel_col)) {
        sel_col <- c("number_of_genes", "total_count_per_cell", "non_mt_percent",
                    "non_ERCC_percent", "non_ribo_percent", "outliers")
    }
    if (!any(sel_col %in% colnames(QC_metrics(sce)))){
        stop("`sel_col` not in colnames(QC_metrics(sce))).")
    }
    x <- as.data.frame(QC_metrics(sce)[, colnames(QC_metrics(sce)) %in% sel_col])


    if ("outliers" %in% colnames(x)) {
        plot_output <- GGally::ggpairs(
        x,
        mapping = ggplot2::aes_string(colour = "outliers"),
        upper = list(continuous = GGally::wrap("cor", size=3, hjust=0.8))
        )
    }
    else {
        plot_output <- GGally::ggpairs(x)
    }

    return(plot_output)
}

#' Plot mapping statistics for \code{SingleCellExperiment} object.
#'
#' @param sce a \code{SingleCellExperiment} object
#' @param sel_col a vector of column names, indicating the columns to use for
#'   plot. by default it will be the mapping result.
#' @param percentage TRUE to convert the number of reads to percentage
#' @param dataname the name of this dataset, used as plot title
#'
#' @importFrom reshape melt
#' @importFrom stats prcomp reorder
#' @return a ggplot2 object
#' @export
#'
#' @examples
#' data("sc_sample_data")
#' data("sc_sample_qc")
#' sce = SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
#' organism(sce) = "mmusculus_gene_ensembl"
#' gene_id_type(sce) = "ensembl_gene_id"
#' QC_metrics(sce) = sc_sample_qc
#' demultiplex_info(sce) = cell_barcode_matching
#' UMI_dup_info(sce) = UMI_duplication
#'
#' plot_mapping(sce,percentage=TRUE,dataname="sc_sample")
#'
plot_mapping <- function(sce,
                        sel_col=NULL,
                        percentage=FALSE,
                        dataname="") {
    sce <- validObject(sce) # check the sce object
    if (is.null(sel_col)) {
        sel_col <- c("unaligned", "aligned_unmapped", "ambiguous_mapping",
                    "mapped_to_ERCC", "mapped_to_intron", "mapped_to_exon")
    }
    x <- as.data.frame(QC_metrics(sce)[, sel_col])


    mapping_stat <- x
    mapping_stat$sample_name <- stats::reorder(rownames(mapping_stat), mapping_stat$mapped_to_exon)
    mapping_stat_prop <- as.data.frame(prop.table(as.matrix(mapping_stat[, sapply(mapping_stat, is.numeric)]), 1))
    mapping_stat_prop$sample_name <- mapping_stat$sample_name
    dat.m <- melt(mapping_stat, id.vars="sample_name")
    dat.m1 <- melt(mapping_stat_prop, id.vars="sample_name")

    if (!percentage) {
        p <- ggplot2::ggplot(dat.m, ggplot2::aes_string(x="sample_name", y="value", fill="variable")) + ggplot2::scale_fill_brewer(palette="Set1")+
        ggplot2::geom_bar(stat="identity", width=1)+
        ggplot2::ylab("number of reads")+
        ggplot2::xlab("cell sorted by number of reads mapped to exon")+
        ggplot2::theme(axis.title.x=ggplot2::element_blank(), axis.text.x=ggplot2::element_blank())

        if (dataname != "") {
            p <- p + ggplot2::ggtitle(paste0("Overall mapping statistics of ", dataname, " (number of reads)"))
        } else {
            p <- p + ggplot2::ggtitle(paste0("Overall mapping statistics (number of reads)"))
        }
    }
    else {
        p <- ggplot2::ggplot(dat.m1, ggplot2::aes_string(x="sample_name", y="value", fill="variable")) + ggplot2::scale_fill_brewer(palette="Set1")+
        ggplot2::geom_bar(stat="identity", width=1)+
        ggplot2::ylab("percentage of reads")+
        ggplot2::xlab("cell sorted by number of reads mapped_to_exon")+
        ggplot2::scale_y_continuous(labels=scales::percent_format())+
        ggplot2::theme(axis.title.x=ggplot2::element_blank(), axis.text.x=ggplot2::element_blank())

        if (dataname != "") {
            p <- p + ggplot2::ggtitle(paste0("Overall mapping statistics of ", dataname, " (percentage)"))
        } else {
            p <- p + ggplot2::ggtitle(paste0("Overall mapping statistics (percentage)"))
        }
    }

    return(p)
}



#' plot_demultiplex
#'
#'@description Plot cell barcode demultiplexing result for the
#'  \code{SingleCellExperiment}. The barcode demultiplexing result is shown
#'  using a barplot, with the bars indicating proportions of total reads.
#'  Barcode matches and mismatches are summarised along with whether or not the
#'  read mapped to the genome. High proportion of genome aligned reads with no
#'  barcode match may indicate barcode integration failure.
#'
#' @param sce a \code{SingleCellExperiment} object
#'
#' @return a ggplot2 bar chart
#'
#' @importFrom rlang .data
#' @export
#'
#' @examples
#' data("sc_sample_data")
#' data("sc_sample_qc")
#' sce = SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
#' organism(sce) = "mmusculus_gene_ensembl"
#' gene_id_type(sce) = "ensembl_gene_id"
#' QC_metrics(sce) = sc_sample_qc
#' demultiplex_info(sce) = cell_barcode_matching
#' UMI_dup_info(sce) = UMI_duplication
#'
#' plot_demultiplex(sce)
#'
plot_demultiplex <- function(sce){
    sce <- validObject(sce) # check the sce object
    demultiplex_stat <- demultiplex_info(sce)
    demultiplex_stat[,"count"] <- demultiplex_stat[,"count"]/sum(demultiplex_stat[,"count"])
    if (is.null(demultiplex_stat)){
        stop("`demultiplex_stat` does not exists in sce. demultiplex_info(sce) == NULL)")
    }
    demultiplex_stat$label_y <- demultiplex_stat[,"count"]+0.05
    demultiplex_stat$label_tx <- scales::percent(demultiplex_stat[,"count"])
    #kp = demultiplex_stat[,"count"]/sum(demultiplex_stat[,"count"])>0.2
    #label_tx[!kp] = ""

    blank_theme <- ggplot2::theme_minimal()+
        ggplot2::theme(
        axis.text.x = ggplot2::element_text(angle = 30, hjust=1,size=12),
        panel.border = ggplot2::element_blank(),
        plot.title = ggplot2::element_text(size=14, face="bold"),
        legend.position = "none"
        )

    p <- ggplot2::ggplot(data=demultiplex_stat, ggplot2::aes(x=.data$status, y=.data$count,
                                            fill=.data$status))+
        ggplot2::geom_bar(width = 1, stat = "identity")+
        #coord_polar("y", start=0)+
        ggplot2::scale_fill_brewer(palette="Dark2")+
        blank_theme+
        ggplot2::geom_text(ggplot2::aes(y = .data$label_y, label = .data$label_tx))+
        ggplot2::labs(title="Cell barcode demultiplexing statistics", y="percentage",x="")
    return(p)
}


#' Plot UMI duplication frequency
#'
#' @description Plot the UMI duplication frequency.
#'
#' @param sce a \code{SingleCellExperiment} object
#' @param log10_x whether to use log10 scale for x axis
#'
#' @return a line chart of the UMI duplication frequency
#' @export
#'
#' @examples
#' data("sc_sample_data")
#' data("sc_sample_qc")
#' sce = SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
#' organism(sce) = "mmusculus_gene_ensembl"
#' gene_id_type(sce) = "ensembl_gene_id"
#' QC_metrics(sce) = sc_sample_qc
#' demultiplex_info(sce) = cell_barcode_matching
#' UMI_dup_info(sce) = UMI_duplication
#'
#' plot_UMI_dup(sce)
#'
plot_UMI_dup <- function(sce, log10_x = TRUE){
    sce <- validObject(sce) # check the sce object
    tmp <- UMI_dup_info(sce)
    tmp <- tmp[-nrow(tmp),] # remove 1000
    the_sum <- cumsum(tmp)
    for (i in seq_len(nrow(tmp)-1)){ # cut the zero counts
        if (the_sum$count[nrow(tmp)] - the_sum$count[i]<10){
            tmp <- tmp[seq_len(i),]
            break
        }
    }
    dup_col <- ""
    if ("duplication_number" %in% colnames(tmp)){
        dup_col <- "duplication_number"
    }else if ("duplication.number" %in% colnames(tmp)){
        dup_col <- "duplication.number"
    }else{
        message("cannot find column containing UMI duplication number.")
        return(NULL)
    }
    p <- ggplot2::ggplot(data=tmp, ggplot2::aes(x=tmp[, dup_col], y=tmp[,"count"])) +
        ggplot2::geom_smooth(method = "loess",se = FALSE)+ggplot2::theme_bw()
    if (log10_x){
        p <- p + ggplot2::scale_x_log10()+
            ggplot2::labs(title="Distribution of UMI duplication numbers(log10).",
                x="log10(UMI duplication number)",
                y="frequency")
    }else{
        p <- p + ggplot2::labs(title="Distribution of UMI duplication numbers.",
                x="UMI duplication number",
                y="frequency")
    }
    return(p)
}



#' Remove outliers in \code{SingleCellExperiment}
#'
#' @description Removes outliers flagged by \code{detect_outliers()}
#'
#' @param sce a \code{SingleCellExperiment} object
#' @export
#' @return a \code{SingleCellExperiment} object without outliers
#' @examples
#' data("sc_sample_data")
#' data("sc_sample_qc")
#' sce = SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
#' organism(sce) = "mmusculus_gene_ensembl"
#' gene_id_type(sce) = "ensembl_gene_id"
#' QC_metrics(sce) = sc_sample_qc
#' demultiplex_info(sce) = cell_barcode_matching
#' UMI_dup_info(sce) = UMI_duplication
#' sce = detect_outlier(sce)
#' dim(sce)
#' sce = remove_outliers(sce)
#' dim(sce)
#'
remove_outliers <- function(sce) {
    sce <- validObject(sce) # check the sce object
    if (!("outliers" %in% colnames(QC_metrics(sce)))) {
      stop("No outlier information. Please run `detect_outlier()` to get the outliers.")
    }
    out_cell <- QC_metrics(sce)$outliers == FALSE
    return(sce[, out_cell])
}

###############################
# Demultiplexing FASTQ Reads
###############################

#' @name sc_atac_trim_barcode
#' @title demultiplex raw single-cell ATAC-Seq fastq reads
#' @description  single-cell data need to be demultiplexed in order to retain the information of the cell barcodes
#' the data belong to. Here we reformat fastq files so barcode/s (and if available the UMI sequences) are moved from
#' the sequence into the read name. Since scATAC-Seq data are mostly paired-end, both `r1` and `r2` are demultiplexed in this function.
#' @param r1 read one for pair-end reads.
#' @param r2 read two for pair-end reads, NULL if single read.
#' @param bc_file the barcode information, can be either in a \code{fastq} format (e.g. from 10x-ATAC) or
#' from a \code{.csv} file (here the barcode is expected to be on the second column). 
#' Currently, for the fastq approach, this can be a list of barcode files.
#' @param valid_barcode_file optional file path of the valid (expected) barcode sequences to be found in the bc_file (.txt, can be txt.gz). 
#' Must contain one barcode per line on the second column separated by a comma (default ="").
#' If given, each barcode from bc_file is matched against the barcode of
#' best fit (allowing a hamming distance of 1). If a FASTQ \code{bc_file} is provided, barcodes with a higher mapping quality, as given by
#' the fastq reads quality score are prioritised.
#' @param output_folder the output dir for the demultiplexed fastq file, which will contain 
#' fastq files with reformatted barcode and UMI into the read name. 
#' Files ending in \code{.gz} will be automatically compressed.
#' @param id1_st barcode start position (0-indexed) for read 1, which is an extra parameter that is needed if the
#' \code{bc_file} is in a \code{.csv} format.
#' @param id2_st barcode start position (0-indexed) for read 2, which is an extra parameter that is needed if the
#' \code{bc_file} is in a \code{.csv} format.
#' @param id1_len barcode length for read 1, which is an extra parameter that is needed if the
#' \code{bc_file} is in a \code{.csv} format.
#' @param id2_len barcode length for read 2, which is an extra parameter that is needed if the
#' \code{bc_file} is in a \code{.csv} format.
#' @param umi_start if available, the start position of the molecular identifier.
#' @param umi_length if available, the start position of the molecular identifier.  
#' @param umi_in umi_in
#' @param rmN logical, whether to remove reads that contains N in UMI or cell barcode.
#' @param rmlow logical, whether to remove reads that have low quality barcode sequences
#' @param min_qual the minimum base pair quality that is allowed (default = 20).
#' @param num_below_min the maximum number of base pairs below the quality threshold.
#' @param no_reverse_complement specifies if the reverse complement of the barcode sequence should be 
#' used for barcode error correction (only when barcode sequences are provided as fastq files). FALSE (default)
#' lets the function decide whether to use reverse complement, and TRUE forces the function to
#' use the forward barcode sequences.
#'
#' @returns None (invisible `NULL`)
#'
#' @examples
#' data.folder <- system.file("extdata", package = "scPipe", mustWork = TRUE)
#' r1      <- file.path(data.folder, "small_chr21_R1.fastq.gz") 
#' r2      <- file.path(data.folder, "small_chr21_R3.fastq.gz") 
#' 
#' # Using a barcode fastq file:
#'
#' # barcodes in fastq format
#' barcode_fastq      <- file.path(data.folder, "small_chr21_R2.fastq.gz") 
#' 
#' sc_atac_trim_barcode (
#' r1            = r1, 
#' r2            = r2, 
#' bc_file       = barcode_fastq,
#' rmN           = TRUE,
#' rmlow         = TRUE,
#' output_folder = tempdir())
#'
#' # Using a barcode csv file:
#'
#' # barcodes in .csv format
#' barcode_1000       <- file.path(data.folder, "chr21_modified_barcode_1000.csv")
#' 
#' \dontrun{
#' sc_atac_trim_barcode (
#' r1            = r1, 
#' r2            = r2, 
#' bc_file       = barcode_1000, 
#' id1_st        = 0,
#` id1_len       = 16,
#` id2_st        = 0,
#` id2_len       = 16 
#' rmN           = TRUE,
#' rmlow         = TRUE,
#' output_folder = tempdir())
#' }
#'@export
sc_atac_trim_barcode <- function(
    r1,
    r2,
    bc_file = NULL,
    valid_barcode_file = "",
    output_folder = "",
    umi_start=0,
    umi_length=0,
    umi_in = "both",
    rmN = FALSE,
    rmlow = FALSE,
    min_qual = 20,
    num_below_min = 2,
    id1_st = -0,
    id1_len = 16,
    id2_st = 0,
    id2_len = 16,
    no_reverse_complement=FALSE) {
    
    if(output_folder == ''){
    output_folder <- file.path(getwd(), "scPipe-atac-output")
    }
    if (!dir.exists(output_folder)){
    dir.create(output_folder,recursive=TRUE)
    message("Output Directory Does Not Exist. Created Directory: ", output_folder)
    }

    log_and_stats_folder <- paste0(output_folder, "/scPipe_atac_stats/")
    dir.create(log_and_stats_folder, showWarnings = FALSE)

    log_file             <- paste0(log_and_stats_folder, "log_file.txt")
    stats_file           <- paste0(log_and_stats_folder, "stats_file_trimbarcode.txt")
    if(!file.exists(log_file)) file.create(log_file)
    file.create(stats_file)

    write(
        c("trimbarcode starts at ", as.character(Sys.time()), "\n"), file = log_file, append = TRUE
    )

    if (substr(r1, nchar(r1) - 2, nchar(r1)) == ".gz") {
    	write_gz <- TRUE
    }
    else {
    	write_gz <- FALSE
    }

    if (is.null(bc_file)) {
	stop("Barcode file is mandatory")
	}

	i <- 1;
	for (bc in bc_file) {
		if (!file.exists(bc)) {stop("Barcode file does not exist.")}
		bc_file[i] <- path.expand(bc)
		i          <- i+1;
	}

	if (!file.exists(r1)) {stop("read1 fastq file does not exist.")}
	
	if ((valid_barcode_file != "") && tools::file_ext(valid_barcode_file) != 'csv') {
		stop("Valid Barcode File must be a CSV")
	}

	if(umi_start != 0) {
		if(umi_in %in% c("both", "R1", "R2")) {
			message("UMI Present in: ", umi_in)
		}else{
			stop("Invalid value of umi_in. Possible values are both, R1 and R2")
		}
	}

	# expand tilde to home path for downstream gzopen() call
	r1 <- path.expand(r1)

	if(!is.null(r2)) {
		if (!file.exists(r2)) {stop("read2 file does not exist.")}
		r2 <- path.expand(r2)
	} else{
		r2 <- ""
	}
	message("Saving the output at location: \n", output_folder)

	if(tools::file_ext(bc_file) != "csv"){
		# fastq barcode files are provided, run the FASTQ method
		out_vec <- rcpp_sc_atac_trim_barcode_paired(
			output_folder,
			r1,
			bc_file,
			r2,
			valid_barcode_file,
			write_gz,
			rmN,
			rmlow,
			min_qual,
			num_below_min,
			no_reverse_complement)
	
		write(
			c(
				"Total Reads: ", out_vec[1],
				"\nTotal N's removed: ", out_vec[2],
				"\nremoved_low_qual: ", out_vec[3],
				"\nUnique sequences read in barcode file: ", out_vec[4],
				"\n"
			),
			file = stats_file, append = TRUE)
	
	} else {
		message("Using barcode CSV file, since barcode FastQ file is not passed")
		if(id1_st < 0 || id2_st < 0 || id1_len < 0 || id2_len < 0 ){
			stop("Please pass positive integer values for id1_st, id2_st, id1_len, and id2_len")
		}
		
		
		# trim the barcode csv file (which contains the actual barcodes in the second column)
		# into a file with barcodes on each line and no whitespace
		temp_barcode_file <- paste0(output_folder, "/tempbarcode.csv")
		on.exit(if(file.exists(temp_barcode_file)) {file.remove(temp_barcode_file)})
		
		# change this to handle multiple barcode files!! TODO
		barcodes <- read.csv(bc_file, header=FALSE, strip.white=TRUE)
		write(barcodes$V2, temp_barcode_file)

		# perform the same manipulation for valid
		
		# Check if given barcode start position is valid
		# check_barcode_start_position is expecting a single barcode, of only the barcode sequences, no commas
		message("Checking if id1_st is valid")
		if (!check_barcode_start_position(r1, temp_barcode_file, bc_file, id1_st, id1_len, 10000, .8)) {
			if (tolower(readline(prompt="Continue anyway? (y/n) ")) != "y") {
				stop("Please change id1_st and try again")
			}
			message("Continuing...")
		}
		message("Checking if id2_st is valid")
		if (!check_barcode_start_position(r2, temp_barcode_file, bc_file, id2_st, id2_len, 10000, .8)) {
			if (tolower(readline(prompt="Continue anyway? (y/n) ")) != "y") {
				stop("Please change id2_st and try again")
			}
			message("Continuing...")
		}
		
		out_vec <- rcpp_sc_atac_trim_barcode(
			output_folder,
			r1,
			r2,
			temp_barcode_file, # not urgent but this needs to be changed to bc_file or barcode_file
			valid_barcode_file,
			#bc_start,
			#bc_length,
			umi_start,
			umi_length,
			umi_in,
			write_gz,
			rmN,
			rmlow,
			min_qual,
			num_below_min,
			id1_st,
			id1_len,
			id2_st,
			id2_len)
		
		# concatenate results to stats_file
		write(
			c(
				"Total Reads: ", out_vec[1],
				"\nTotal N's removed: ", out_vec[2],
				"\nremoved_low_qual: ", out_vec[3],
				"\nExact match Reads: ", out_vec[4],
				"\nReads Matched After Correction: ", out_vec[5],
				"\nTotal barcodes: ", out_vec[6],
				"\n"
			),
			file = stats_file, append = TRUE)
	}
    
    
    write(
        c(
            "trimbarcode finishes at ",
            as.character(Sys.time()),
            "\n\n"
        ),
        file = log_file, append = TRUE)
    
    # return(out_vec)
}

#' @importFrom basilisk BasiliskEnvironment
scPipe_env <- BasiliskEnvironment(envname="scPipe_env", 
    pkgname="scPipe",
    packages=c("python==3.9.0"),
    pip=c("sinto==0.7.2.2", "MACS3==3.0.0a6"))

#' Get read structure for particular scRNA-seq protocol
#'
#' The supported protocols are:
#' \itemize{
#'  \item CelSeq
#'  \item CelSeq2
#'  \item DropSeq
#'  \item 10x (also called ChromiumV1)
#'}
#' If you know the structure of a specific protocol and would like it supported,
#' please leave a issue post at www.github.com/luyitian/scPipe.
#'
#' @param protocol name of the protocol
#'
#' @return list of UMI and Barcode locations for use in other scPipe functions
#' @export
#'
#' @examples
#' get_read_str("celseq")
get_read_str <- function(protocol) {
    original_protocol <- protocol
    error_msg <- paste("No read structure for", original_protocol, "found")

    # strip non-alphanumeric characters and convert to lower case
    protocol <- gsub("[^[:alnum:]_]", "", protocol)
    protocol <- tolower(protocol)

    switch(
        protocol,
        "celseq"         = list(bs1 = -1, bl1 = 0, bs2 = 6, bl2 = 8,  us = 0,  ul = 6),
        "celseq2"        = list(bs1 = -1, bl1 = 0, bs2 = 6, bl2 = 8,  us = 0,  ul = 6),
        "dropseq"        = list(bs1 = -1, bl1 = 0, bs2 = 0, bl2 = 12, us = 12, ul = 8),
        "10x"            = list(bs1 = -1, bl1 = 0, bs2 = 0, bl2 = 16, us = 16, ul = 10),
        "chromiumv1"     = list(bs1 = -1, bl1 = 0, bs2 = 0, bl2 = 16, us = 16, ul = 10),
        stop(error_msg)
    )
}

context("Read structure helper")
test_that("Stripping and lower casing is correct", {
    expect_equal(
        get_read_str("celSeq"),
        list(bs1 = -1, bl1 = 0, bs2 = 6, bl2 = 8,  us = 0,  ul = 6)
    )

    expect_equal(
        get_read_str("cel-Seq"),
        list(bs1 = -1, bl1 = 0, bs2 = 6, bl2 = 8,  us = 0,  ul = 6)
    )

    expect_equal(
        get_read_str("CelSeq"),
        list(bs1 = -1, bl1 = 0, bs2 = 6, bl2 = 8,  us = 0,  ul = 6)
    )
})

test_that("Unknown protocols are properly caught", {
    expect_error(get_read_str("FakeSeq"))
    expect_error(get_read_str("errorseq"))
})

#' @name sc_sample_qc
#' @title quality control information for a small sample scRNA-seq dataset to 
#' demonstrate capabilities of scPipe.
#' @description This data.frame contains cell quality control information 
#' for the 100 cells. For each cell it has: \itemize{
#'  \item{unaligned} the number of unaligned reads.
#'  \item{aligned_unmapped} the number of reads that aligned to genome but 
#'  fail to map to any features. 
#'  \item{mapped_to_exon} is the number of reads that mapped
#'  to exon.
#'  \item{mapped_to_intron} is the number of reads that mapped to intron.
#'  \item{ambiguous_mapping} is the number of reads that mapped to multiple
#'  features. They are not considered in the following analysis.
#'  \item{mapped_to_ERCC} is the number of reads that mapped to ERCC 
#'  spike-in controls.
#'  \item{mapped_to_MT} is the number of reads that mapped to mitochondrial
#'  genes.
#'  \item{total_count_per_cell} is the number of reads that mapped to exon
#'  after UMI deduplication. In contrast, `mapped_to_exon` is the 
#'  number of reads mapped to exon before UMI deduplication.
#'  \item{number_of_genes} is the number of genes detected for each cells
#'  \item{non_ERCC_percent} is 1 - (percentage of ERCC reads). Reads are
#'  UMI deduplicated.
#'  \item{non_mt_percent} is 1 - (percentage of mitochondrial reads).
#'  Reads are UMI deduplicated.
#'  \item{non_ribo_percent} is 1- (percentage of ribosomal reads).
#'  Reads are UMI deduplicated.
#'  }
#' @return NULL, but makes a data frame with cell quality control data.frame
#' @docType data
#' @format a data.frame instance, one row per cell.
#' @source Christin Biben (WEHI). She FACS sorted cells from several immune
#' cell types including B cells, granulocyte and some early progenitors. 
#' @author Luyi Tian
#' @examples 
#' data("sc_sample_data")
#' data("sc_sample_qc")
#' sce = SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
#' organism(sce) = "mmusculus_gene_ensembl"
#' gene_id_type(sce) = "ensembl_gene_id"
#' QC_metrics(sce) = sc_sample_qc
#' head(QC_metrics(sce))
#' plot_mapping(sce,percentage=TRUE,dataname="sc_sample")
#' 
NULL

#' @name UMI_duplication
#' @title UMI duplication statistics for a small sample scRNA-seq dataset to 
#' demonstrate capabilities of scPipe
#' @description This data.frame contains UMI duplication statistics, where the
#' first column is the number of duplication, and the second column is the
#' count of UMIs.
#' 
#' @return NULL, but makes a data frame with UMI dulication statistics
#' @docType data
#' @format a data.frame instance, one row per cell.
#' @source Christin Biben (WEHI). She FACS sorted cells from several immune
#' cell types including B cells, granulocyte and some early progenitors. 
#' @author Luyi Tian
#' @examples 
#' data("sc_sample_data")
#' data("sc_sample_qc")
#' sce = SingleCellExperiment(assays = list(counts =as.matrix(sc_sample_data)))
#' organism(sce) = "mmusculus_gene_ensembl"
#' gene_id_type(sce) = "ensembl_gene_id"
#' QC_metrics(sce) = sc_sample_qc
#' demultiplex_info(sce) = cell_barcode_matching
#' UMI_dup_info(sce) = UMI_duplication
#' 
#' head(UMI_dup_info(sce))
#' 
NULL

#' @name cell_barcode_matching
#' @title cell barcode demultiplex statistics for a small sample scRNA-seq 
#' dataset to demonstrate capabilities of scPipe
#' @description This data.frame contains cell barcode demultiplex statistics 
#' with several rows:\itemize{
#' \item{barcode_unmatch_ambiguous_mapping} is the number of reads that do not
#' match any barcode, but aligned to the genome and mapped to multiple 
#' features.
#' \item{barcode_unmatch_mapped_to_intron} is the number of reads that do
#' not match any barcode, but aligned to the genome and mapped to intron.
#' \item{barcode_match} is the number of reads that match the cell barcodes
#' \item{barcode_unmatch_unaligned} is the number of reads that do not 
#' match any barcode, and not aligned to the genome
#' \item{barcode_unmatch_aligned} is the number of reads that do not 
#' match any barcode, but aligned to the genome and do not mapped to any 
#' feature
#' \item{barcode_unmatch_mapped_to_exon} is the number of reads that do not
#' match any barcode, but aligned to the genome and mapped to the exon
#' }
#' @return NULL, but makes a data frame with cell barcode demultiplex statistics
#' @docType data
#' @format a data.frame instance, one row per cell.
#' @source Christin Biben (WEHI). She FACS sorted cells from several immune
#' cell types including B cells, granulocyte and some early progenitors. 
#' @author Luyi Tian
#' @examples 
#' data("sc_sample_data")
#' data("sc_sample_qc")
#' sce = SingleCellExperiment(assays = list(counts =as.matrix(sc_sample_data)))
#' organism(sce) = "mmusculus_gene_ensembl"
#' gene_id_type(sce) = "ensembl_gene_id"
#' QC_metrics(sce) = sc_sample_qc
#' demultiplex_info(sce) = cell_barcode_matching
#' UMI_dup_info(sce) = UMI_duplication
#' 
#' demultiplex_info(sce)
#' 
NULL

.onAttach <- function(...) {
    # packageStartupMessage("Some startup message")
    

    
    # Check for platform
    if(.Platform$OS.type != "unix") {
        packageStartupMessage("Windows platform detected, sc_atac_peak_calling() and sc_atac_remove_duplicates() function may not operate. Please call peaks and remove duplicate reads outside the package.")  
    } else {
        packageStartupMessage("Linux/MacOS detected! \n")
        
        # Check if samtools is installed
        packageStartupMessage("Checking if samtools is installed ... \n")
        tryCatch({
            system2("samtools", stdout = NULL, stderr = NULL)
            packageStartupMessage("samtools was located! \n")
        },
        
        warning = function(w) {
            packageStartupMessage("samtools was not located. Please specify the path of it when running sc_atac_remove_duplicates()")
        }
        )
    }
}

#' @name sc_sample_data
#' @title a small sample scRNA-seq counts dataset to 
#' demonstrate capabilities of scPipe
#' @description This data set contains counts for high variable genes for 
#' 100 cells. The cells have different cell types. The data contains
#' raw read counts. The cells are chosen randomly from 384 cells and
#' they did not go through quality controls. The rows names are 
#' Ensembl gene ids and the columns are cell names, which is the wall
#' position in the 384 plates.
#' @return NULL, but makes a matrix of count data
#' @docType data
#' @format a matrix instance, one row per gene.
#' @source Christin Biben (WEHI). She FACS sorted cells from several immune
#' cell types including B cells, granulocyte and some early progenitors. 
#' @author Luyi Tian
#' @examples 
#' # use the example dataset to perform quality control
#' data("sc_sample_data")
#' data("sc_sample_qc")
#' sce = SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
#' organism(sce) = "mmusculus_gene_ensembl"
#' gene_id_type(sce) = "ensembl_gene_id"
#' QC_metrics(sce) = sc_sample_qc
#' demultiplex_info(sce) = cell_barcode_matching
#' UMI_dup_info(sce) = UMI_duplication
#' sce = detect_outlier(sce)
#' 
#' plot_QC_pairs(sce)
#' 
NULL

#' @name sc_atac_remove_duplicates
#'
#' @title Removing PCR duplicates using samtools
#' @description Takes in a BAM file and removes the PCR duplicates using the samtools markdup function. Requires samtools 1.10 or newer for statistics to be generated. 
#' 
#' @param inbam The tagged, sorted and duplicate-free input BAM file 
#' @param samtools_path The path of the samtools executable (if a custom installation is to be specified)
#' @param output_folder The path of the output folder 
#'
#' @returns file path to a bam file created from samtools markdup
#' @importFrom Rsamtools sortBam
#' @export
#' 
sc_atac_remove_duplicates <- function(inbam, 
                                        samtools_path = NULL,
                                        output_folder = NULL) {
    
    successfully_removed_duplicates <- FALSE
    # Check if samtools is installed
    if (!is.null(samtools_path)) {
        samtools <- samtools_path
    } else {
        samtools <- "samtools"
    }
    
    # Check if samtools is installed
    samtools.installed <- tryCatch(
        {
        system2(samtools, stdout = NULL, stderr = NULL)
        message("samtools was located!")
        TRUE
        },
        
        warning = function(w) {
            if (is.null(samtools_path)) {
                message("samtools was not located, so can't remove duplicates. Please make sure it is installed.")
            } else {
                message("samtools was not located via the specified path. Please make sure it is correct.")
            }
            return(FALSE)
        }
    )

    if (isTRUE(samtools.installed)) {
        return(tryCatch({
            # Check if file exists
            if (!file.exists(inbam)) {
                stop("Couldn't locate the supplied BAM file")
            }
            
            # Check for validity of file
            system2(samtools, c("quickcheck", inbam), stderr = NULL, stdout = NULL)
            
            
            # Check if output directory exists
            if (is.null(output_folder)) {
                output_folder <- file.path(getwd(), "scPipe-atac-output")
            }
            
            if (!dir.exists(output_folder)){
                dir.create(output_folder,recursive=TRUE)
                message("Output Directory Does Not Exist. Created Directory: ", output_folder)
            }
            
            # Create log folder/file
            log_and_stats_folder <- file.path(output_folder, "scPipe_atac_stats")
            dir.create(log_and_stats_folder, showWarnings = FALSE)
            log_file <- file.path(log_and_stats_folder, "log_file.txt")
            if(!file.exists(log_file)) file.create(log_file)
            write(c(
                "sc_atac_remove_duplicates starts at ",
                as.character(Sys.time()),
                "\n"
                ),
                file = log_file, append = TRUE)
            
            
            inbam.name <- substr(inbam, 0, nchar(inbam)-4)
            message("Sorting the BAM file by name")
            system2(samtools, c("sort", "-n", "-o", paste(inbam.name, "namesorted.bam", sep="_"), inbam))
            message("Running fixmate on the sorted BAM File")
            system2(samtools, c("fixmate", "-m", paste(inbam.name, "namesorted.bam", sep="_"), paste(inbam.name, "fixmate.bam", sep="_")))
            message("Sorting the BAM file by coordinates")
            # Remove existing copy if one exists
            output.bam <- paste0(output_folder, "/", basename(inbam.name), "_markdup.bam")
            if (file.exists(output.bam)) {
                system2("rm", output.bam)
            }
            
            Rsamtools::sortBam(paste(inbam.name, "fixmate.bam", sep="_"), paste(inbam.name, "positionsort", sep="_"), indexDestination = TRUE)

            # Note: the output bam file is originally created in the same directory as the input bam file
            
            message("Running samtools markdup now")
            # Check if the version of samtools is 1.10 or greater (to have the stats functionality)
            version.text <- strsplit(strsplit(system2(samtools, "--version", stdout=TRUE)[1], " ")[[1]][2], "\\.")[[1]]
            if (as.numeric(version.text[1]) < 1 || (as.numeric(version.text[1]) >= 1 && as.numeric(version.text[2]) < 10)) {
                message("Version of samtools isn't greater or equal to 1.10. Can't produce duplicate removal stats file.")
                system2(samtools, c("markdup", "-s", "-r", paste(inbam.name, "positionsort.bam", sep="_"), paste(inbam.name, "markdup.bam", sep="_")))
            } else {
                message("Detected samtools with version greater or equal to 1.10. Producing duplicate removal stats file.")
                system2(samtools, c("markdup", "-s", "-f", file.path(log_and_stats_folder, "duplicate_removal_stats.txt"), "-r", paste(inbam.name, "positionsort.bam", sep="_"), paste(inbam.name, "markdup.bam", sep="_")))
            }

            message("Indexing the BAM file")
            Rsamtools::indexBam(paste(inbam.name, "markdup.bam", sep="_"))
            message("Cleaning up intermediary files")
            system2("rm", paste(inbam.name, "namesorted.bam", sep="_"))
            system2("rm", paste(inbam.name, "positionsort.bam", sep="_"))
            system2("rm", paste(inbam.name, "fixmate.bam", sep="_"))
            
            if (file.exists(paste(inbam.name, "markdup.bam", sep="_"))) {
                message("The output BAM file was sent to", output_folder)
                
                # If the new file is already in the destination folder, don't need to move it!
                if (paste(inbam.name, "markdup.bam", sep="_") != output.bam)
                    system2("mv", c("--force", paste(inbam.name, "markdup.bam", sep="_"), output_folder))
                successfully_removed_duplicates <- output.bam
            } else {
                message("Couldn't remove duplicates from the input BAM file. Make sure it is a valid BAM file.")
            }
            write(
            c(
                "sc_atac_remove_duplicates finishes at ",
                as.character(Sys.time()),
                "\n\n"
            ),
            file = log_file, append = TRUE)
            successfully_removed_duplicates
        },
        warning = function(w) {w
            message(w)
            return(FALSE)
        },
        
        error = function(e) {
            message(e)
            return(FALSE)
        }))
    } else {
        return(FALSE)
    }
}

## ----echo=FALSE, warning=FALSE, message=FALSE, out.width="100%"---------------
knitr::include_graphics(system.file("scATAC-seq_workflow.png", package = "scPipe"))

## ----message=FALSE------------------------------------------------------------
library(scPipe)
data.folder <- system.file("extdata", package = "scPipe", mustWork = TRUE)


## -----------------------------------------------------------------------------
output_folder <- paste0(tempdir(), "/scPipeATACVignette")

## -----------------------------------------------------------------------------
# data fastq files
r1      <- file.path(data.folder, "small_chr21_R1.fastq.gz") 
r2      <- file.path(data.folder, "small_chr21_R3.fastq.gz") 

# barcodes in fastq format
barcode_fastq      <- file.path(data.folder, "small_chr21_R2.fastq.gz") 

# barcodes in .csv format
barcode_1000       <- file.path(data.folder, "chr21_modified_barcode_1000.csv")


## -----------------------------------------------------------------------------
sc_atac_trim_barcode (r1            = r1, 
                      r2            = r2, 
                      bc_file       = barcode_fastq, 
                      rmN           = FALSE,
                      rmlow         = FALSE,
                      output_folder = output_folder)

## ----eval=FALSE---------------------------------------------------------------
# sc_atac_trim_barcode (r1            = r1,
#                       r2            = r2,
#                       bc_file       = barcode_1000,
#                       id1_st        = -1,
#                       id1_len       = -1,
#                       id2_st        = -1,
#                       id2_len       = -10,
#                       output_folder = output_folder,
#                       rmN           = TRUE)

## -----------------------------------------------------------------------------
demux_r1        <- file.path(output_folder, "demux_completematch_small_chr21_R1.fastq.gz")
demux_r2        <- file.path(output_folder, "demux_completematch_small_chr21_R3.fastq.gz")
reference = file.path(data.folder, "small_chr21.fa")


aligned_bam <- sc_aligning(ref = reference, 
                R1 = demux_r1, 
                R2 = demux_r2, 
                nthreads  = 12,
                output_folder = output_folder)

## -----------------------------------------------------------------------------

sorted_tagged_bam <- sc_atac_bam_tagging (inbam = aligned_bam, 
                       output_folder =  output_folder, 
                       bam_tags      = list(bc="CB", mb="OX"), 
                       nthreads      =  12)


## ----eval=FALSE---------------------------------------------------------------
# 
# removed <- sc_atac_remove_duplicates(sorted_tagged_bam,
#                           output_folder = output_folder)
# if (!isFALSE(removed))
#   sorted_tagged_bam <- removed
# 

## -----------------------------------------------------------------------------
sc_atac_create_fragments(inbam = sorted_tagged_bam,
                         output_folder = output_folder)


## ----eval=FALSE---------------------------------------------------------------
# # CHECK IF THE .narrowPeak file is small enough to include in the package,
# # otherwise, we need to make this basilisk call work??
# sc_atac_peak_calling(inbam = sorted_tagged_bam,
#                      ref = reference,
#                      genome_size = NULL,
#                      output_folder = output_folder)
# 

## ----eval=FALSE---------------------------------------------------------------
# features          <- file.path(output_folder, "NA_peaks.narrowPeak")
# 
# sc_atac_feature_counting (fragment_file = file.path(output_folder, "fragments.bed"),
#                           feature_input = features,
#                           bam_tags      = list(bc="CB", mb="OX"),
#                           feature_type  = "peak",
#                           organism      = "hg38",
#                           yieldsize     = 1000000,
#                           exclude_regions = TRUE,
#                           output_folder = output_folder,
#                           fix_chr       = "none"
#                           )

## ----eval=FALSE---------------------------------------------------------------
# reference       <- file.path(data.folder, "small_chr21.fa")
# sc_atac_feature_counting (fragment_file = file.path(output_folder, "fragments.bed"),
#                           feature_input = reference,
#                           bam_tags      = list(bc="CB", mb="OX"),
#                           feature_type  = "genome_bin",
#                           organism      = "hg38",
#                           cell_calling  = FALSE,
#                           yieldsize     = 1000000,
#                           exclude_regions = TRUE,
#                           output_folder = output_folder,
#                           fix_chr       = "none"
#                           )
# 

## ----eval=FALSE---------------------------------------------------------------
# features          <- file.path(output_folder, "NA_peaks.narrowPeak")

## ----eval=TRUE, echo=FALSE----------------------------------------------------
features          <- file.path(data.folder, "NA_peaks.narrowPeak")

## -----------------------------------------------------------------------------
sc_atac_feature_counting (fragment_file = file.path(output_folder, "fragments.bed"),
                          feature_input = features, 
                          bam_tags      = list(bc="CB", mb="OX"), 
                          feature_type  = "peak",
                          organism      = "hg38",
                          cell_calling  = "filter",
                          min_uniq_frags = 0,
                          min_frac_peak = 0,
                          min_frac_promoter = 0,
                          yieldsize     = 1000000,
                          exclude_regions = TRUE,
                          output_folder = output_folder,
                          fix_chr       = "none"
                          )

## -----------------------------------------------------------------------------
feature_matrix <- readRDS(file.path(output_folder, "unfiltered_feature_matrix.rds"))
dplyr::glimpse(feature_matrix)

sparseM <- readRDS(file.path(output_folder, "sparse_matrix.rds"))
dplyr::glimpse(sparseM)

## -----------------------------------------------------------------------------
sce <- sc_atac_create_sce(input_folder = output_folder,
                   organism     = "hg38",
                   feature_type = "peak",
                   pheno_data   = NULL,
                   report       = FALSE)

## ----echo=FALSE, warning=FALSE, message=FALSE, out.width="100%"---------------
knitr::include_graphics(system.file("report_example.png", package = "scPipe"))

## ----eval = FALSE-------------------------------------------------------------
# data.folder <- system.file("extdata", package = "scPipe", mustWork = TRUE)
# output_folder <- tempdir()
# sce <- sc_atac_pipeline(r1 = file.path(data.folder, "small_chr21_R1.fastq.gz"),
#                         r2 = file.path(data.folder, "small_chr21_R3.fastq.gz"),
#                         barcode_fastq = file.path(data.folder, "small_chr21_R2.fastq.gz"),
#                         organism = "hg38",
#                         reference = file.path(data.folder, "small_chr21.fa"),
#                         feature_type = "peak",
#                         remove_duplicates = FALSE,
#                         min_uniq_frags = 0, # set to 0 as the sample dataset only has a small number of reads
#                         min_frac_peak = 0,
#                         min_frac_promoter = 0,
#                         output_folder = output_folder)

## ----echo=FALSE, results='hide'-----------------------------------------------
# cleanup tempfiles
unlink(output_folder, recursive=TRUE)

## -----------------------------------------------------------------------------
sessionInfo()

#' scPipe - single cell RNA-seq pipeline
#'
#' @description The scPipe will do cell barcode demultiplexing, UMI deduplication and quality control on
#' fastq data generated from all protocols
#'
#' @author Luyi Tian <tian.l@wehi.edu.au>; Shian Su <su.s@wehi.edu.au>
#' @docType package
#' @name scPipe
#' @import Rhtslib
#' @import SingleCellExperiment
#' @import Rcpp
#' @importFrom methods as
#' @useDynLib scPipe, .registration = TRUE
#' @aliases scPipe scPipe-package
NULL

#' Check Valid Barcode Start Position
#'
#' @description Checks to see if the given barcode start position (\code{bstart})
#' is valid for the fastq file. If the found barcode percentage is less than
#' the given \code{threshold}, a new barcode start position is searched for by
#' checking every position from the start of each read to 10 bases after the \code{bstart}
#' @param fastq file containing reads
#' @param barcode_file csv file
#' @param barcode_file_realname the real name of the csv file
#' @param bstart the start position for barcodes in the given reads
#' @param blength length of each barcode
#' @param search_lines the number of fastq lines to use for the check
#' @param threshold the minimum percentage of found barcodes to accept for the
#' program to continue
#'
#' @return Boolean; TRUE if program can continue execution, FALSE otherwise.
#'
check_barcode_start_position <- function(fastq, barcode_file, barcode_file_realname, bstart, blength, search_lines, threshold) {
    # check if the given bstart param is valid
    continue <- check_barcode_reads(fastq, barcode_file, barcode_file_realname, bstart, blength, search_lines, threshold)

    return (continue)
}

context("Check Barcode Reads before Trimming")
test_that("Correct input positions are accepted", {
    small_r1 <- system.file("extdata", "simu_R1.fastq.gz", package="scPipe")
    small_barcode <- system.file("extdata", "smallbarcode1col.csv", package="scPipe")

    expect_true(check_barcode_start_position(small_r1, small_barcode, "smallbarcode1col.csv", 0, 4, 40, .5))
})

test_that("Bad input position causes rejection", {
    small_r1 <- system.file("extdata", "simu_R1.fastq.gz", package="scPipe")
    small_barcode <- system.file("extdata", "smallbarcode1col.csv", package="scPipe")

    expect_false(check_barcode_start_position(small_r1, small_barcode, "smallbarcode1col.csv", 10, 4, 40, .5))
})

test_that("Bad match rate causes rejection", {
    small_r1 <- system.file("extdata", "simu_R1.fastq.gz", package="scPipe")
    small_barcode <- system.file("extdata", "smallbarcode1col.csv", package="scPipe")

    expect_false(check_barcode_start_position(small_r1, small_barcode, "smallbarcode1col.csv", 0, 4, 100, .5))
})

# This dummy function definition is included with the package to ensure that
# 'tools::package_native_routine_registration_skeleton()' generates the required
# registration info for the 'run_testthat_tests' symbol.
(function() {
    .Call("run_testthat_tests", FALSE, PACKAGE = "scPipe")
})

#' Get ERCC annotation table
#'
#' Helper function to retrieve ERCC annotation as a dataframe in SAF format
#'
#' @return data.frame containing ERCC annotation
#'
#' @examples
#' ercc_anno <- get_ercc_anno()
#'
#' @export
#'
get_ercc_anno <- function() {
    anno_import(system.file("extdata", "ERCC92_anno.gff3", package = "scPipe"))
}

# guess the organism and species from input data
.guess_attr <- function(row_names) {
    hsp_ensembl <- length(grep("^ENSG",row_names))
    mm_ensembl <- length(grep("^ENSMUSG", row_names))
    if ((hsp_ensembl>0) & (hsp_ensembl>mm_ensembl)) {
        return(list(organism="hsapiens_gene_ensembl", gene_id_type="ensembl_gene_id"))
    }
    else if ((mm_ensembl>0) & (mm_ensembl>hsp_ensembl)) {
        return(list(organism="mmusculus_gene_ensembl", gene_id_type="ensembl_gene_id"))
    }
    else {
        return(list(organism=NA, gene_id_type=NA))
    }
}


# check the object, fix empty slot with default.
validObject <- function(object){
    if (!is(object, "SingleCellExperiment")) {
        stop("object must be an `SingleCellExperiment` object.")
    }

    if(!("scPipe" %in% names(object@metadata))){
        object@metadata$scPipe$version <- packageVersion("scPipe")  # set version information
    }

    if(min(dim(object)) == 0){
        stop("The dimension of sce should be larger than zero.")
    }else if(is.null(rownames(object)) | is.null(colnames(object))){
        stop("rowname/colname does not exists for sce.")
    }else if(!all(rownames(QC_metrics(object)) == colnames(object))){
        stop("The rownames of QC metrics is not consistent with column names of the object.")
    }


    if(any(is.null(organism(object)) || is.na(organism(object)))){
        tmp_res <- .guess_attr(rownames(object))
        if((!is.na(tmp_res$organism)) & (!is.na(tmp_res$gene_id_type))){
            gene_id_type(object) <- tmp_res$gene_id_type
            organism(object) <- tmp_res$organism
            message("organism/gene_id_type not provided. Make a guess:",
                    tmp_res$organism,
                    "/",
                    tmp_res$gene_id_type)
        } else {
            gene_id_type(object) <- "NA"
            organism(object) <- "NA"
        }
    }
    return(object)
}


#' Get or set quality control metrics in a SingleCellExperiment object
#' @rdname QC_metrics
#' @param object A \code{\link{SingleCellExperiment}} object.
#' @param value Value to be assigned to corresponding object.
#'
#' @return A DataFrame of quality control metrics.
#' @author Luyi Tian
#'
#' @importFrom S4Vectors DataFrame SimpleList
#'
#' @export
#'
#' @examples
#' data("sc_sample_data")
#' data("sc_sample_qc")
#' sce = SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
#' QC_metrics(sce) = sc_sample_qc
#'
#' head(QC_metrics(sce))
#'
QC_metrics.sce <- function(object) {
    if(!("scPipe" %in% names(object@metadata))){
        warning("`scPipe` not in `metadata`. Cannot identify quality control columns")
        return(NULL)
    } else if (!("QC_cols" %in% names(object@metadata$scPipe))) {
        warning("metadata$scPipe is missing `QC_cols`. Cannot identify quality control columns")
        return(NULL)
    }
    return(colData(object)[, object@metadata$scPipe$QC_cols])
}

#' @rdname QC_metrics
#' @aliases QC_metrics
#' @export
#'
setMethod("QC_metrics", signature(object = "SingleCellExperiment"),
            QC_metrics.sce)

#' @rdname QC_metrics
#' @aliases QC_metrics
#' @export
setReplaceMethod(
    "QC_metrics",
    signature = "SingleCellExperiment",
    function(object, value) {
        if (!("scPipe" %in% names(object@metadata))) {
            object@metadata[["scPipe"]] <- list(QC_cols=colnames(value))
        } else {
            object@metadata$scPipe$QC_cols <- colnames(value)
        }

        colData(object)[, colnames(value)] <- DataFrame(value)

        return(object)
    }
)


#' @title demultiplex_info
#'
#' @description Get or set cell barcode demultiplex results in a SingleCellExperiment object
#' @rdname demultiplex_info
#' @param object A \code{\link{SingleCellExperiment}} object.
#' @param value Value to be assigned to corresponding object.
#'
#' @return A DataFrame of cell barcode demultiplx results.
#' @author Luyi Tian
#'
#' @export
#'
#' @examples
#' data("sc_sample_data")
#' data("sc_sample_qc")
#' sce = SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
#' organism(sce) = "mmusculus_gene_ensembl"
#' gene_id_type(sce) = "ensembl_gene_id"
#' QC_metrics(sce) = sc_sample_qc
#' demultiplex_info(sce) = cell_barcode_matching
#' UMI_dup_info(sce) = UMI_duplication
#'
#' demultiplex_info(sce)
#'
demultiplex_info.sce <- function(object) {
    if(!("scPipe" %in% names(object@metadata))){
        warning("`scPipe` not in `metadata`. Cannot find columns for cell barcode demultiplex results")
        return(NULL)
    }else if(!("demultiplex_info" %in% names(object@metadata$scPipe))){
        warning("The metadata$scPipe does not have `demultiplex_info`.")
        return(NULL)
    }
    return(object@metadata$scPipe$demultiplex_info)
}

#' @rdname demultiplex_info
#' @aliases demultiplex_info
#' @export
#'
setMethod("demultiplex_info", signature(object = "SingleCellExperiment"),
        demultiplex_info.sce)

#' @rdname demultiplex_info
#' @aliases demultiplex_info
#' @export
setReplaceMethod("demultiplex_info",
                signature="SingleCellExperiment",
                function(object, value) {
                    if(!("scPipe" %in% names(object@metadata))){
                        object@metadata[["scPipe"]] <- list(demultiplex_info=value)
                    }else{
                        object@metadata$scPipe$demultiplex_info <- value
                    }
                    object <- validObject(object) # could add other checks
                    return(object)
})




#' Get or set UMI duplication results in a SingleCellExperiment object
#' @rdname UMI_dup_info
#' @param object A \code{\link{SingleCellExperiment}} object.
#' @param value Value to be assigned to corresponding object.
#'
#' @return A DataFrame of UMI duplication results.
#' @author Luyi Tian
#'
#' @export
#'
#' @examples
#' data("sc_sample_data")
#' data("sc_sample_qc")
#' sce = SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
#' organism(sce) = "mmusculus_gene_ensembl"
#' gene_id_type(sce) = "ensembl_gene_id"
#' QC_metrics(sce) = sc_sample_qc
#' demultiplex_info(sce) = cell_barcode_matching
#' UMI_dup_info(sce) = UMI_duplication
#'
#' head(UMI_dup_info(sce))
#'
UMI_dup_info.sce <- function(object) {
    if(!("scPipe" %in% names(object@metadata))){
        warning("`scPipe` not in `metadata`. Cannot find columns for cell barcode demultiplex results")
        return(NULL)
    }else if(!("UMI_dup_info" %in% names(object@metadata$scPipe))){
        warning("The metadata$scPipe does not have `UMI_dup_info`.")
        return(NULL)
    }
    return(object@metadata$scPipe$UMI_dup_info)
}

#' @rdname UMI_dup_info
#' @aliases UMI_dup_info
#' @export
#'
setMethod("UMI_dup_info", signature(object = "SingleCellExperiment"),
        UMI_dup_info.sce)

#' @rdname UMI_dup_info
#' @aliases UMI_dup_info
#' @export
setReplaceMethod("UMI_dup_info",
                signature="SingleCellExperiment",
                function(object, value) {
                    if(!("scPipe" %in% names(object@metadata))){
                        object@metadata[["scPipe"]] <- list(UMI_dup_info=value)
                    }else{
                        object@metadata$scPipe$UMI_dup_info <- value
                    }
                    object <- validObject(object) # could add other checks
                    return(object)
})



#' Get or set \code{organism} from a SingleCellExperiment object
#' @rdname organism
#' @param object A \code{\link{SingleCellExperiment}} object.
#' @param value Value to be assigned to corresponding object.
#'
#' @importFrom BiocGenerics organism organism<-
#' @return organism string
#' @author Luyi Tian
#' @export
#' @examples
#' data("sc_sample_data")
#' data("sc_sample_qc")
#' sce = SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
#' organism(sce) = "mmusculus_gene_ensembl"
#' gene_id_type(sce) = "ensembl_gene_id"
#' QC_metrics(sce) = sc_sample_qc
#' demultiplex_info(sce) = cell_barcode_matching
#' UMI_dup_info(sce) = UMI_duplication
#'
#' organism(sce)
#'
organism.sce <- function(object) {
    return(object@metadata$Biomart$organism)
}


#' @aliases organism
#' @rdname organism
#' @export
setMethod("organism", signature(object="SingleCellExperiment"),
        organism.sce)


#' @aliases organism
#' @rdname organism
#' @export
#' @export
setReplaceMethod("organism",signature="SingleCellExperiment",
                function(object, value) {
                    if(is.null(value)){
                        object@metadata$Biomart$organism <- NA
                        }else if(value == "NA"){
                        object@metadata$Biomart$organism <- NA
                        }else{
                            object@metadata$Biomart$organism <- value
                            }
                    return(object)
                })



#' Get or set \code{gene_id_type} from a SingleCellExperiment object
#' @rdname gene_id_type
#' @param object A \code{\link{SingleCellExperiment}} object.
#' @param value Value to be assigned to corresponding object.
#'
#' @return gene id type string
#' @author Luyi Tian
#'
#' @export
#'
#' @examples
#' data("sc_sample_data")
#' data("sc_sample_qc")
#' sce = SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
#' organism(sce) = "mmusculus_gene_ensembl"
#' gene_id_type(sce) = "ensembl_gene_id"
#' QC_metrics(sce) = sc_sample_qc
#' demultiplex_info(sce) = cell_barcode_matching
#' UMI_dup_info(sce) = UMI_duplication
#'
#' gene_id_type(sce)
#'
gene_id_type.sce <- function(object) {
    return(object@metadata$Biomart$gene_id_type)
}


#' @rdname gene_id_type
#' @aliases gene_id_type
#' @export
setMethod("gene_id_type", signature(object = "SingleCellExperiment"),
            gene_id_type.sce)


#' @aliases gene_id_type
#' @rdname gene_id_type
#' @export
setReplaceMethod("gene_id_type",signature="SingleCellExperiment",
                function(object, value) {
                    if(is.null(value)){
                        object@metadata$Biomart$gene_id_type <- NA
                    }else if(value == "NA"){
                        object@metadata$Biomart$gene_id_type <- NA
                    }else{
                        object@metadata$Biomart$gene_id_type <- value
                    }
                    return(object)
                })

# -------------------------------------------------------- scATAC-seq -----------------------------------------------------------
#' Get or set \code{feature_info} from a SingleCellExperiment object
#' @rdname feature_info
#' @param object A \code{\link{SingleCellExperiment}} object.
#' @param value Value to be assigned to corresponding object.
#' @author Shani Amarasinghe
#' @return A DataFrame of feature information
#' @export
#'
feature_info.sce <- function(object) {
    if(!("scPipe" %in% names(object@metadata))){
        warning("`scPipe` not in `metadata`.")
        return(NULL)
    }else if(!("feature_cols" %in% names(object@metadata$scPipe))){
        warning("The metadata$scPipe does not have `feature_cols`.")
        return(NULL)
    }
    return(rowData(object)[, object@metadata$scPipe$feature_cols])
}


#' @rdname feature_info
#' @aliases feature_info
#' @export
setMethod("feature_info", signature(object = "SingleCellExperiment"),
        feature_info.sce)


#' @aliases feature_info
#' @rdname feature_info
#' @export
setReplaceMethod("feature_info",signature="SingleCellExperiment",
                function(object, value) {
                    feature <- NULL
                    value <- subset(value, select = -c(feature))
                    if (!("scPipe" %in% names(object@metadata))) {
                        object@metadata[["scPipe"]] <= list(feature_cols=colnames(value))
                    } else {
                        object@metadata$scPipe$feature_cols <- colnames(value)
                    }
                    rowData(object)[, colnames(value)] <- DataFrame(value)
                    return(object)
                })


#' Get or set \code{feature_type} from a SingleCellExperiment object
#' @rdname feature_type
#' @param object A \code{\link{SingleCellExperiment}} object.
#' @param value Value to be assigned to corresponding object.
#' @author Shani Amarasinghe 
#' @return A string representing the feature type
#'
#' @export
#'
feature_type.sce <- function(object) {
    return(object@metadata$scPipe$feature_type)
}


#' @rdname feature_type
#' @aliases feature_type
#' @export
setMethod("feature_type", signature(object = "SingleCellExperiment"),
            feature_type.sce)


#' @aliases feature_type
#' @rdname feature_type
#' @export
setReplaceMethod("feature_type",signature="SingleCellExperiment",
                function(object, value) {
                    if(is.null(value) || value == "NA"){
                        object@metadata$scPipe$feature_type <- NA
                    }else{
                        object@metadata$scPipe$feature_type <- value
                    }
                    return(object)
                })

#############################
# Demultiplxing FASTQ Reads
#############################

#' sc_atac_bam_tagging()
#' @name sc_atac_bam_tagging
#' @title BAM tagging
#' @description Demultiplexes the reads
#'
#' @param inbam The input BAM file
#' @param output_folder The path of the output folder
#' @param bc_length The length of the cellular barcodes
#' @param bam_tags The BAM tags
#' @param nthreads The number of threads
#'
#' @returns file path of the resultant demmultiplexed BAM file. 
#' @examples
#' r1 <- system.file("extdata", "small_chr21_R1.fastq.gz", package="scPipe")
#' r2 <- system.file("extdata", "small_chr21_R3.fastq.gz", package="scPipe")
#' barcode_fastq <- system.file("extdata", "small_chr21_R2.fastq.gz", package="scPipe")
#' out <- tempdir()
#'
#' sc_atac_trim_barcode(r1=r1, r2=r2, bc_file=barcode_fastq, output_folder=out)
#'
#' demux_r1 <- file.path(out, "demux_completematch_small_chr21_R1.fastq.gz")
#' demux_r2 <- file.path(out, "demux_completematch_small_chr21_R3.fastq.gz")
#' reference <- system.file("extdata", "small_chr21.fa", package="scPipe")
#'
#' aligned_bam <- sc_aligning(ref=reference, R1=demux_r1, R2=demux_r2, nthreads=6, output_folder=out)
#'
#' out_bam <- sc_atac_bam_tagging(
#'    inbam = aligned_bam,
#'    output_folder = out,
#'    nthreads = 6)
#'
#' @export
sc_atac_bam_tagging <- function(inbam,
                                output_folder = NULL,
                                bc_length = NULL,
                                bam_tags = list(bc="CB", mb="OX"),
                                nthreads = 1) {

    if (any(!file.exists(inbam))) {
        stop("At least one input bam file should be present")
    } else {
        inbam <- path.expand(inbam)
    }

    if(is.null(output_folder)){
        output_folder <- file.path(getwd(), "scPipe-atac-output")
        # output_folder <- paste0(sub('[/][^/]+$', '', inbam)) # same as output_folder <- basename(inbam)
    }

    if (!dir.exists(output_folder)){
        dir.create(output_folder,recursive=TRUE)
        message("Output directory does not exist. Creating ", output_folder)
    }

    get_filename_without_extension <- function(path) {
        vec <- strsplit(basename(path), "\\.")[[1]]
        return(paste(vec[seq_len(length(vec)-1)], collapse = "."))
    }
    fileNameWithoutExtension <- get_filename_without_extension(inbam)

    outbam                   <- file.path(output_folder, paste0(fileNameWithoutExtension, "_tagged_sorted.bam"))
    outsortedbam             <- file.path(output_folder, paste0(fileNameWithoutExtension, "_tagged_sorted")) # don't want extension

    log_and_stats_folder       <- file.path(output_folder, "scPipe_atac_stats")
    dir.create(log_and_stats_folder, showWarnings = FALSE)
    log_file                   <- file.path(log_and_stats_folder, "log_file.txt")
    stats_file                 <- file.path(log_and_stats_folder, "stats_file_bam_tagging.txt")
    if(!file.exists(log_file)) file.create(log_file)

    write(
        c(
            "sc_atac_tagging began at ",
            as.character(Sys.time()),
            "\n"
        ),
        file = log_file, append = TRUE)

    outbam <- path.expand(outbam)
    if(!file.exists(outbam)){
        file.create(outbam)

        rcpp_sc_atac_bam_tagging(inbam, outbam, bam_tags$bc, bam_tags$mb, nthreads)

        # Rsamtools::sortBam(outbam, outsortedbam, indexDestination = TRUE, maxMemory = 1024/32*max_memory)
        Rsamtools::indexBam(paste0(outsortedbam, ".bam"))

        message("Demultiplexed BAM file sorted and indexed ...")
    }
    #cat(paste0(outsortedbam, ".bam"))
    #cat("\n")


    if(is.null(bc_length)){
        message("Using default value for barcode length (bc_length = 16) ")
        bc_length <- 16
    }


    flag_defs <- tibble::tibble(
        type =
            c("one_read_unmapped", "one_read_unmapped", "one_read_unmapped", "one_read_unmapped", "one_read_unmapped", "one_read_unmapped", "one_read_unmapped", "one_read_unmapped", "one_read_unmapped", "one_read_unmapped", "one_read_unmapped", "one_read_unmapped", "both_reads_unmapped", "both_reads_unmapped", "mapped", "mapped", "mapped", "mapped", "mapped_wrong_orientation", "mapped_wrong_orientation", "mapped_wrong_orientation", "mapped_wrong_orientation", "mapped_ambigously", "mapped_ambigously", "mapped_ambigously", "mapped_ambigously", "mapped_ambigously", "mapped_ambigously", "mapped_ambigously", "mapped_ambigously")
        ,
        flag =
            c(73, 133, 89, 121, 165, 181, 101, 117, 153, 185, 69, 137, 77, 141, 99, 147, 83, 163, 67, 131, 115, 179, 81, 161, 97, 145, 65, 129, 113, 177))

    add_matrices <- function(...) {
        a <- list(...)
        cols <- sort(unique(unlist(lapply(a, colnames))))
        rows <- sort(unique(unlist(lapply(a, rownames))))

        nrows <- length(rows)
        ncols <- length(cols)
        newms <- lapply(a, function(m) {
            b <- as(as(m, "dgCMatrix"), "dgTMatrix")
            s <- cbind.data.frame(i = b@i + 1, j = b@j + 1, x = b@x)


            i <- match(rownames(m), rows)[s$i]
            j <- match(colnames(m), cols)[s$j]

            Matrix::sparseMatrix(i=i,
                        j=j,
                        x=s$x,
                        dims=c(nrows, ncols),
                        dimnames=list(rows, cols))
        })
        Reduce(`+`, newms)
    }

    message("Generating mapping statistics per barcode")
    message("Iterating through 5,000,000 reads at a time")
    bamfl <- open(Rsamtools::BamFile(outbam, yieldSize = 5000000))
    params <- Rsamtools::ScanBamParam(what=c("flag"), tag=c("CB"))
    iter <- 1
    full_matrix <- NULL

    # Iterate over the BAM file in chunks of 5 million reads
    while(length((bam0 <- Rsamtools::scanBam(bamfl, param = params)[[1]])$flag)) {
        message("chunk", iter)

        # Create a data.table object with each row representing a read, and the columns as the barcode and flag
        df <- data.table::setDT(data.frame(bam0$tag$CB, bam0$flag) %>% data.table::setnames(c("barcode", "flag")) %>% dplyr::left_join(flag_defs, by = "flag"))[, !"flag"]

        # Count the reads per barcode
        x <- df[, list(count=.N), names(df)]
        x <- data.table::dcast(x, barcode ~ type, value.var = "count")
        for (i in seq_along(x)) data.table::set(x, i=which(is.na(x[[i]])), j=i, value=0)

        if (is.null(full_matrix)) {
            full_matrix <- as.matrix(x, rownames=TRUE)
        } else {
            full_matrix <- add_matrices(full_matrix, as.matrix(x, rownames=TRUE))
        }
        iter <- iter+1
    }
    df <- data.table::setnames(data.table::setDT(as.data.frame(as.matrix(full_matrix)), keep.rownames = TRUE), "rn", "barcode")
    df[ ,count := rowSums(.SD), .SDcols = names(df[,!"barcode"])]
    mapping_stats_path <- file.path(log_and_stats_folder, "mapping_stats_per_barcode.csv")
    data.table::fwrite(df, mapping_stats_path)
    message("Completed generating mapping statistics per barcode, saved to ", file.path(log_and_stats_folder, "mapping_stats_per_barcode.csv"), "\n")
    write(
        c(
            "sc_atac_tagging() completed at ",
            as.character(Sys.time()),
            "\n\n"
        ),
        file = log_file, append = TRUE)

  
    message("The output tagged and sorted BAM file was sent to ", output_folder)
  
    return(outbam)
}

#########################################################################
# Generating the Feature Matrix from a Given BAM file and a Feature FIle
#########################################################################

#' @name sc_atac_feature_counting
#' @title generating the feature by cell matrix 
#' @description feature matrix is created using a given demultiplexed BAM file and 
#' a selected feature type
#' @param fragment_file The fragment file
#' @param feature_input The feature input data e.g. the .narrowPeak file for peaks of a bed file format
#' @param bam_tags The BAM tags
#' @param feature_type The type of feature
#' @param organism The organism type (contains hg19, hg38, mm10)
#' @param cell_calling The desired cell calling method; either \code{cellranger}, \code{emptydrops} or  \code{filter}.
#' @param pheno_data The phenotypic data as a data frame
#' 
#' @param promoters_file The path of the promoter annotation file (if the specified organism isn't recognised).
#' @param tss_file The path of the tss annotation file (if the specified organism isn't recognised).
#' @param enhs_file The path of the enhs annotation file (if the specified organism isn't recognised).
#' @param gene_anno_file The path of the gene annotation file (gtf or gff3 format).
#' @param sample_name The sample name to identify which is the data is analysed for.
#' 
#' @param bin_size The size of the bins
#' @param yieldsize The yield size
#' @param n_filter_cell_counts An integer value to filter the feature matrix on the number of reads per cell (default = 200)
#' @param n_filter_feature_counts An integer value to filter the feature matrix on the number of reads per feature (default = 10).
#' @param exclude_regions Whether or not the regions (specified in the file) should be excluded
#' @param excluded_regions_filename The filename of the file containing the regions to be excluded
#' @param fix_chr Whether chr should be fixed or not
#' 
#' @param lower the lower threshold for the data if using the \code{emptydrops} function for cell calling
#' @param genome_size The size of the genome (used for the \code{cellranger} cell calling method)
#'
#' @param min_uniq_frags The minimum number of required unique fragments required for a cell (used for \code{filter} cell calling)
#' @param max_uniq_frags The maximum number of required unique fragments required for a cell (used for \code{filter} cell calling)
#' @param min_frac_peak The minimum proportion of fragments in a cell to overlap with a peak (used for \code{filter} cell calling)
#' @param min_frac_tss The minimum proportion of fragments in a cell to overlap with a tss (used for \code{filter} cell calling)
#' @param min_frac_enhancer The minimum proportion of fragments in a cell to overlap with a enhancer sequence (used for \code{filter} cell calling)
#' @param min_frac_promoter The minimum proportion of fragments in a cell to overlap with a promoter sequence (used for \code{filter} cell calling)
#' @param max_frac_mito The maximum proportion of fragments in a cell that are mitochondrial (used for \code{filter} cell calling)
#' @param output_folder The output folder
#' @param create_report Logical value to say whether to create the report or not (default = TRUE).
#' 
#' @returns None (invisible `NULL`)
#' 
#'
#' @examples
#' \dontrun{
#' sc_atac_feature_counting(
#'    fragment_file = fragment_file,
#'    cell_calling = "filter",
#'    exclude_regions = TRUE,
#'    feature_input = feature_file)
#' }    
#' @export
#' 
sc_atac_feature_counting <- function(
    fragment_file, 
    feature_input             = NULL, 
    bam_tags                  = list(bc="CB", mb="OX"), 
    feature_type              = "peak", 
    organism                  = "hg38", 
    cell_calling              = "filter",
    sample_name               = "",
    genome_size               = NULL, 
    promoters_file            = NULL,
    tss_file                  = NULL,
    enhs_file                 = NULL,
    gene_anno_file            = NULL,
    pheno_data                = NULL, 
    bin_size                  = NULL, 
    yieldsize                 = 1000000,
    n_filter_cell_counts      = 200,
    n_filter_feature_counts   = 10,
    exclude_regions           = FALSE, 
    excluded_regions_filename = NULL,
    output_folder             = NULL,
    fix_chr                   = "none", 
    lower                     = NULL,
    min_uniq_frags            = 3000,
    max_uniq_frags            = 50000,
    min_frac_peak             = 0.3,
    min_frac_tss              = 0,
    min_frac_enhancer         = 0,
    min_frac_promoter         = 0.1,
    max_frac_mito             = 0.15,
    create_report             = FALSE
    ) {
  
    . <- V1 <- V2 <- V3 <- init <- peakStart <- seqnames <- peakEnd <- CB <- chromosome <- chrS <- feature <- barcode <- NULL
    
    init_time <- Sys.time()
    
    available_organisms <- c("hg19",
                            "hg38",
                            "mm10")
    
    if(!is.null(organism)) stopifnot(organism %in% available_organisms)
    
    stopifnot(fix_chr %in% c("none", "excluded_regions", "feature", "both"))
    
    if(is.null(output_folder)) {
        output_folder <- file.path(getwd(), "scPipe-atac-output")
        message("Output directory has not been provided. Saving output in\n", output_folder)
    }
    
    if (!dir.exists(output_folder)){
        dir.create(output_folder,recursive=TRUE)
        message("Output directory does not exist. Creating...", output_folder)
    }
    
    
    # initiate log file and location for stats in scPipe_atac_stats
    log_and_stats_folder <- paste0(output_folder, "/scPipe_atac_stats/")
    dir.create(log_and_stats_folder, showWarnings = FALSE)
    log_file             <- paste0(log_and_stats_folder, "log_file.txt")
    stats_file           <- paste0(log_and_stats_folder, "stats_file_align.txt")
    if(!file.exists(log_file)) file.create(log_file)
    # file.create(stats_file)
    
    # timer
    write(
        c(
        "sc_atac_feature_counting starts at ",
        as.character(Sys.time()),
        "\n"
        ), 
        file = log_file, append = TRUE)

    
    
    ############# feature type is genome_bin ####################

    if(feature_type == 'genome_bin'){
        # TODO: test the format of the fasta file here and stop if not proper format
        message("`genome bin` feature type is selected for feature input. reading the genome fasta file ...")
        
        # index for feature_input is created if not found in the same directory as the feature_input
        if(!file.exists(paste0(feature_input,".fai"))){
            Rsamtools::indexFa(feature_input)
            message("Index for ", feature_input, " is being created... ")
        }
        
        if(is.null(bin_size)){
            bin_size <- 2000
            message("Default bin size of 2000 is selected")
        }
        
        out_bed_filename <- paste0(output_folder, "/", sub('\\..[^\\.]*$', '', basename(feature_input)), ".bed") # remove extension and append output folder
        
        if(file.exists(out_bed_filename)) {
            message(out_bed_filename, " file already exists. Replacing!")
            file.remove(out_bed_filename)
        }
        
        if(!is.null(feature_input)){
            
            rcpp_fasta_bin_bed_file(feature_input, out_bed_filename, bin_size)
            
            # Check if file was created
            if(file.exists(out_bed_filename)) {
                message("Generated the genome bin file:\n", out_bed_filename)
            }
            else {
                stop("File ", out_bed_filename, "file was not created.")
            }
        
        ## End if is.null
        } else{
            if(!is.null(organism)){
                
                # Use organism data sizes saved in repository
                organism_files     <- list.files(system.file("extdata/annotations/", package = "scPipe", mustWork = TRUE))
                sizes_filename_aux <- grep(pattern = organism, x = organism_files, value = TRUE) %>% 
                    grep(pattern     <- "size", x = ., value = TRUE)
                
                sizes_filename <- system.file(paste0("extdata/annotations/", sizes_filename_aux), package = "scPipe", mustWork = TRUE)
                
                sizes_df <- utils::read.table(sizes_filename, header = FALSE, col.names = c("V1", "V2"))
                
                sizes_df_aux <- sizes_df %>% 
                    dplyr::group_by(V1) %>% 
                    dplyr::mutate(V3 = floor(V2/bin_size),
                                V4 = bin_size*V3) %>% 
                    dplyr::ungroup()
                
                
                
                out_bed <- purrr::map_df(seq_len(nrow(sizes_df_aux)), function(i){
                    
                    aux_i        <- sizes_df_aux %>% dplyr::slice(i)
                    seq_aux_end  <- seq(from = bin_size, to = aux_i$V4, by = bin_size)
                    seq_aux_init <- seq_aux_end - bin_size + 1
                    
                    if(seq_aux_end[length(seq_aux_end)] == aux_i$V2){
                        last_row_init <- c()
                        last_row_end  <- c()
                    } else{
                        last_row_init <- seq_aux_end[length(seq_aux_end)] + 1
                        last_row_end  <- aux_i$V2
                    }
                    
                    options(scipen = 999) # To prevent scientific notation
                    out_df <- dplyr::tibble(init = c(seq_aux_init, last_row_init), 
                                            end = c(seq_aux_end, last_row_end)
                    ) %>% 
                    dplyr::mutate(name = aux_i$V1, .before = init)
                
                    return(out_df)
                })
        
                utils::write.table(out_bed, file = out_bed_filename, row.names = FALSE, col.names = FALSE, quote = FALSE, sep = "\t")
                
                # Check if file was created
                if(file.exists(out_bed_filename)) {
                message("Generated the genome bin file:\n", out_bed_filename)
                }
                else {
                stop("File ", out_bed_filename, "file was not created.")
                }
            } else{
                stop("Either organism or feature_input should be provided.")
            }
        } # End else is.null(genome_size)
    
        # feature_input <- genome_bin
    }
    
    ############################## feature type is peak #####################
    if(feature_type == 'peak' || feature_type == 'tss' || feature_type == 'gene'){
        message("`peak`, `tss` or `gene` feature_type is selected for feature input")
        
        ############################### fix_chr in feature file
        if(fix_chr %in% c("feature", "both")){
            
            out_bed_filename_feature <- paste0(output_folder, "/", 
                                                sub('\\..[^\\.]*$', 
                                                    '', 
                                                    basename(feature_input)
                                                ), "_fixedchr.bed"
            ) # remove extension and append output folder
            
            # Try to read first 5 rows of feature_input file to see if the format is correct
            feature_head <- utils::read.table(feature_input, nrows = 5)
            if(ncol(feature_head) < 3){
                stop("Feature file provided does not contain 3 columns. Cannot append chr")
                # warning("Feature file provided does not contain 3 columns. Cannot append chr")
                # break;
            }
            
            
            rcpp_append_chr_to_bed_file(feature_input, out_bed_filename_feature)
            
            # Check if file was created
            if(file.exists(out_bed_filename_feature)) {
                message("Appended 'chr' to feature file and output created in:", out_bed_filename_feature)
                feature_input <- out_bed_filename_feature
            }
            else {
                stop("File ", out_bed_filename_feature, "file was not created.")
            }
            
        }
        
        
        
        ############################### fix_chr in excluded regions
        if(fix_chr %in% c("excluded_regions", "both")){
            if(!is.null(excluded_regions_filename)){
                
                out_bed_filename_excluded_regions <- paste0(output_folder, "/", 
                                                            sub('\\..[^\\.]*$', 
                                                                '', 
                                                                basename(excluded_regions_filename)
                                                            ), "_fixedchr.bed"
                ) # remove extension and append output folder
                
                # Try to read first 5 rows of excluded_regions file to see if the format is correct
                excluded_regions_head <- utils::read.table(excluded_regions_filename, nrows = 5)
                if(ncol(excluded_regions_head) < 3){
                    stop("excluded_regions file provided does not contain 3 columns. Cannot append chr")
                #   warning("excluded_regions file provided does not contain 3 columns. Cannot append chr")
                #   break
                }
                
                rcpp_append_chr_to_bed_file(excluded_regions_filename, out_bed_filename_excluded_regions)
                
                # Check if file was created
                if(file.exists(out_bed_filename_excluded_regions)) {
                    message("Appended 'chr' to excluded_regions file and output created in:", out_bed_filename_excluded_regions)
                    excluded_regions_filename <- out_bed_filename_excluded_regions
                }
                else {
                    stop("File ", out_bed_filename_excluded_regions, "file was not created.")
                }
            }
        
        } # end if excluded_regions
    }
    
    ############################### end fix_chr
    
    ###################### generate the GAlignment objects from BAM and features ######################
    

    ############## generate the GAalignment file from the feature_input file
    message("Creating Galignment object for the feature input ...")
    if(feature_type != 'genome_bin'){
        feature.gr <- rtracklayer::import(feature_input)
    } else {
        feature.gr <- rtracklayer::import(out_bed_filename)
    }
  
    ############################### exclude regions 
    
    number_of_lines_to_remove <- 0
    
    if(exclude_regions){
        
        if(is.null(excluded_regions_filename) & !is.null(organism)){
            ## If excluded_regions_filename is null but organism is not, then read the file from system
            
            organism_files                <- list.files(system.file("extdata/annotations/", package = "scPipe", mustWork = TRUE))
            excluded_regions_filename_aux <- grep(pattern = organism, x = organism_files, value = TRUE) %>% 
                grep(pattern = "blacklist", x = ., value = TRUE)
            excluded_regions_filename     <- system.file(paste0("extdata/annotations/", excluded_regions_filename_aux), package = "scPipe", mustWork = TRUE)
        } 
        
        if(!is.null(excluded_regions_filename)){
            excluded_regions.gr               <- rtracklayer::import(excluded_regions_filename)
            overlaps_excluded_regions_feature <- IRanges::findOverlaps(excluded_regions.gr, feature.gr, maxgap = -1L, minoverlap = 0L) # Find overlaps
            lines_to_remove                   <- as.data.frame(overlaps_excluded_regions_feature)$subjectHits # Lines to remove in feature file
            number_of_lines_to_remove         <- length(lines_to_remove)
            
            if(number_of_lines_to_remove > 0){ # If there are lines to remove
                feature.gr.df            <- as.data.frame(feature.gr)
                lines_to_keep            <- setdiff(seq_len(nrow(feature.gr.df)), lines_to_remove)
                feature.gr               <- feature.gr[lines_to_keep, ]
            } # End if(number_of_lines_to_remove > 0)
        
        } else{
            warning("Parameter exclude_regions was TRUE but no known organism or excluded_regions_filename provided. Proceding without excluding regions.")
        }
        
        
    } # End if(exclude_regions)
  

    # Check if organism is pre-recognized and if so then use the package's annotation files
    if (organism %in% c("hg19", "hg38", "mm10")) {
        message(organism, "is a recognized organism. Using annotation files in repository.")
        anno_paths <- system.file("extdata/annotations/", package = "scPipe", mustWork = TRUE)
        
        promoters_file <- file.path(anno_paths, paste0(organism, "_promoter.bed.gz"))
        tss_file <- file.path(anno_paths, paste0(organism, "_tss.bed.gz"))
        enhs_file <- file.path(anno_paths, paste0(organism, "_enhancer.bed.gz"))
    } else if (!all(file.exists(c(promoters_file, tss_file, enhs_file)))) {
        stop("One of the annotation files could not be located. Please make sure their paths are valid.")
    }

    # Create bins used for TSS enrichment plot
    tss_df <- data.table::fread(tss_file, select=c(seq_len(3)), header = FALSE, col.names = c("chr", "start", "end"))
    range <- 4000
    bin_size <- 100
    n_bins <- range/bin_size-1
    bins_df_all <- get_all_TSS_bins(tss_df, range, bin_size)
    bins_df_all.gr <- GenomicRanges::makeGRangesFromDataFrame(bins_df_all, keep.extra.columns=TRUE) 
    bin_hits <- c() # used to store all overlaps with bins
    
    ############## read in the aligned and demultiplexed BAM file
    
    # Helper utility function for adding matrices
    add_matrices <- function(...) {
        a <- list(...)
        cols <- sort(unique(unlist(lapply(a, colnames))))
        rows <- sort(unique(unlist(lapply(a, rownames))))

        nrows <- length(rows)
        ncols <- length(cols)
        newms <- lapply(a, function(m) {
        b <- as(as(m, "dgCMatrix"), "dgTMatrix")
        s <- cbind.data.frame(i = b@i + 1, j = b@j + 1, x = b@x)
        
        i <- match(rownames(m), rows)[s$i]
        j <- match(colnames(m), cols)[s$j]
        
        Matrix::sparseMatrix(i=i,
                    j=j,
                    x=s$x,
                    dims=c(nrows, ncols),
                    dimnames=list(rows, cols))
        })
        Reduce(`+`, newms)
    }
    
    unique_feature.gr <- GenomicRanges::makeGRangesFromDataFrame(as.data.frame(dplyr::distinct(data.frame(feature.gr), seqnames, start, end)))
    #   unique_feature.gr <- data.frame(feature.gr) %>% dplyr::distinct(seqnames, start, end) %>% as.data.frame() %>% GenomicRanges::makeGRangesFromDataFrame()
    min_feature_width <- min(GenomicAlignments::ranges(feature.gr)@width)
    
    fragments <- data.table::fread(fragment_file, select=seq_len(5), header = FALSE, col.names = c("seqnames", "start", "end", "barcode", "count")) 
    fragments.gr <- fragments %>% GenomicRanges::makeGRangesFromDataFrame(keep.extra.columns = TRUE)
    
    # Compute overlaps with feature
    peak.overlaps <- GenomicRanges::findOverlaps(query = fragments.gr,
                                                subject = unique_feature.gr,
                                                type = "any",
                                                ignore.strand = TRUE)
    
    # Compute overlaps with bins for TSS enrichment plot
    message("Generating TSS plot data")
    median_feature_overlaptss <- stats::median(GenomicAlignments::ranges(bins_df_all.gr)@width)
    maxgaptss                 <- 0.51*median_feature_overlaptss
    tss_bin_overlaps <- GenomicAlignments::findOverlaps(query = bins_df_all.gr,
                                                        subject = fragments.gr,
                                                        type = "any",
                                                        ignore.strand = TRUE)
    bin_hits <- S4Vectors::queryHits(tss_bin_overlaps) 
    
    num_overlaps <- length(peak.overlaps)
    
    barcodes <- fragments[S4Vectors::queryHits(peak.overlaps), ]$barcode
    features <- data.frame(unique_feature.gr[S4Vectors::subjectHits(peak.overlaps), ]) %>% 
        dplyr::select(seqnames, start, end) %>% 
        tidyr::unite("range", start:end, sep="-") %>% 
        tidyr::unite("feature", seqnames:range, sep=":") 
    
    # Chunk size
    chunk_size <- yieldsize
    i <- 1
    
    message("Generating feature-barcode fragment count matrix")
    # Iterate over chunks
    feature_matrix <- NULL
    while (i <= num_overlaps) {
        temp_mat <- as.matrix(table(feature = features$feature[i:(i + chunk_size-1)], 
                                    barcode = barcodes[i:(i + chunk_size-1)]))
        if (is.null(feature_matrix)) {
            feature_matrix <- Matrix::Matrix(temp_mat)
        } else {
            feature_matrix <- add_matrices(feature_matrix, Matrix::Matrix(temp_mat))
        }
        i <- i + chunk_size
    }
    
    average_number_of_lines_per_CB <- sum(feature_matrix)/ncol(feature_matrix)

    message("Average no. of fragments per barcode: ", average_number_of_lines_per_CB, "\n")

    
    saveRDS(feature_matrix, file = file.path(output_folder, "unfiltered_feature_matrix.rds"))
    message("Raw feature matrix generated: ", file.path(output_folder, "unfiltered_feature_matrix.rds") )
    
    
    ################ Initiate log file
    log_and_stats_folder       <- paste0(output_folder, "/scPipe_atac_stats/")
    dir.create(log_and_stats_folder, showWarnings = FALSE)
    log_file                   <- paste0(log_and_stats_folder, "log_file.txt")
    if(!file.exists(log_file)) file.create(log_file)

    write(c("Average number of reads per cell barcode:", average_number_of_lines_per_CB, "\n"),
        file = log_file, append = TRUE)

    write(c("Number of regions removed from feature input due to being invalid:", number_of_lines_to_remove, "\n"),
        file = log_file, append = TRUE)

    write(
        c(
        "Raw matrix generated at ",
        as.character(Sys.time()),
        "\n"
        ),
        file = log_file, append = TRUE)
    
    message("Calculating TSS enrichment scores")
    # Create a matrix where the rows are TSSs and the columns are the bins
    mat <- matrix(0, nrow(tss_df), n_bins)
    for (i in seq_len(length(bin_hits))) { # populate matrix with hits
        tss_index <- (bin_hits[i]-1) %/% n_bins
        bin <- (bin_hits[i]-1) %% n_bins
        mat[tss_index+1, bin+1] <- mat[tss_index+1, bin+1]+1
    }
    
    mat_non_zero <- mat[rowSums(mat) != 0,]
    
    # Calculate TSS enrichment scores
    tss_dists <- seq(-range/2+bin_size, range/2-bin_size, bin_size)
    flank_read_depth <- (mat_non_zero[,1]+mat_non_zero[,ncol(mat_non_zero)])/2 # Used to normalise
    norm_read_depths <- stats::na.omit(mat_non_zero/flank_read_depth) # also ignore rows where flanks have no overlaps
    
    aggregate_tss_scores <- colMeans(norm_read_depths)
    tsse <- max(aggregate_tss_scores)
    tss_plot_data <- data.frame(dists = tss_dists, 
                                agg_tss_scores = aggregate_tss_scores)
    
    utils::write.csv(tss_plot_data, file = file.path(log_and_stats_folder, "tss_plot_data.csv"))
    
    # generate quality control metrics for cells
    message("Generating QC metrics for cells\n")
    sc_atac_create_cell_qc_metrics(frags_file = fragment_file,
                                    peaks_file = feature_input,
                                    promoters_file = promoters_file,
                                    tss_file = tss_file,
                                    enhs_file = enhs_file,
                                    output_folder = output_folder)


    cell_qc_metrics_file <- file.path(output_folder, "cell_qc_metrics.csv")

    write(
        c(
        "Cell QC metrics generated at ",
        as.character(Sys.time()),
        "\n"
        ),
        file = log_file, append = TRUE)

    # Cell calling
    matrixData <- sc_atac_cell_calling(mat                  = feature_matrix,
                                        cell_calling         = cell_calling,
                                        output_folder        = output_folder,
                                        genome_size          = genome_size,
                                        cell_qc_metrics_file = cell_qc_metrics_file,
                                        lower                = lower,
                                        min_uniq_frags       = min_uniq_frags,
                                        max_uniq_frags       = max_uniq_frags,
                                        min_frac_peak        = min_frac_peak,
                                        min_frac_tss         = min_frac_tss,
                                        min_frac_enhancer    = min_frac_enhancer,
                                        min_frac_promoter    = min_frac_promoter,
                                        max_frac_mito        = max_frac_mito)

    
    write(
        c(
        "Cell calling completed at ",
        as.character(Sys.time()),
        "\n"
        ),
        file = log_file, append = TRUE)
    
    # filter the matrix based on counts per cell
    if (n_filter_cell_counts > 0) {
        message("Cells with less than ", n_filter_cell_counts, " counts are filtered out.")
        write(
            c(
                "cells with less than ",
                n_filter_cell_counts," counts are filtered out.",
                "\n"
            ),
            file = log_file, append = TRUE)
        filtered_indices  <- base::colSums(Matrix::as.matrix(matrixData), na.rm=TRUE) > n_filter_cell_counts
        message("Number of cells to remove:", sum(!filtered_indices))
        if (length(filtered_indices[filtered_indices == TRUE]) >= 10) { # only use if resulting matrix isn't too small
            matrixData <- as.matrix(matrixData[, filtered_indices]) # all the remaining columns
        } else {
            message("No cells were filtered out since otherwise there would be too few left.")
        }
    } else {
        message("No cells were filtered out based on counts.")
    }
    
    
    # filter the matrix based on counts per feature
    if (n_filter_feature_counts > 0) {
        message("features with less than ", n_filter_feature_counts, " counts are filtered out.")
        write(
            c(
                "features with less than ",
                n_filter_feature_counts," counts are filtered out.",
                "\n"
            ),
            file = log_file, append = TRUE)
        filtered_indices  <- base::rowSums(Matrix::as.matrix(matrixData), na.rm=TRUE) > n_filter_feature_counts
        message("Number of features to remove:", sum(!filtered_indices))
        if (length(filtered_indices[filtered_indices == TRUE]) >= 10) {
            matrixData <- matrixData[filtered_indices,] # all the remaining rows
        } else {
            message("No features were filtered out since otherwise there would be too few left.")
        }

    } else {
        message("No features were filtered out based on counts.")
    }

    # Update cell QC metrics to include whether the cell was kept or not
    cqc <- read.csv(file.path(output_folder, "cell_qc_metrics.csv"))
    cqc$cell_called <- cqc$bc %in% colnames(matrixData)
    write.csv(cqc, file.path(output_folder, "cell_qc_metrics.csv"), row.names = FALSE)
    
    
    # converting the NAs to 0s if the sparse option to create the sparse Matrix properly
    message("making sparse")
    sparseM <- Matrix::Matrix(matrixData, sparse=TRUE)
    # # add dimensions of the sparse matrix if available
    # if(cell_calling != FALSE){
    #   barcodes <- utils::read.table(paste0(output_folder, '/non_empty_barcodes.txt'))$V1
    #   features <- utils::read.table(paste0(output_folder, '/non_empty_features.txt'))$V1
    #   dimnames(sparseM) <- list(features, barcodes)
    # }

    message("Sparse matrix generated")
    saveRDS(sparseM, file = paste(output_folder, "/sparse_matrix.rds", sep = ""))
    # Matrix::writeMM(obj = sparseM, file=paste(output_folder, "/sparse_matrix.mtx", sep =""))
    message("Sparse count matrix is saved in\n", output_folder,"/sparse_matrix.mtx")

    # generate and save jaccard matrix
    # jaccardM <- locStra::jaccardMatrix(sparseM)
    # message("Jaccard matrix generated")
    # saveRDS(jaccardM, file = paste(output_folder,"/jaccard_matrix.rds",sep = ""))
    # message("Jaccard matrix is saved in\n", paste(output_folder,"/jaccard_matrix.rds",sep = "") )

    # generate and save the binary matrix
    matrixData[matrixData>0] <- 1
    saveRDS(matrixData, file = paste(output_folder,"/binary_matrix.rds",sep = ""))
    message("Binary matrix is saved in:\n", output_folder,"/binary_matrix.rds")
    
    ###### calculate the TF-IDF matrices here : TO-DO

    # following can be used to plot the stats and load it into sce object ########
    # (from https://broadinstitute.github.io/2020_scWorkshop/data-wrangling-scrnaseq.html)
    counts_per_cell    <- Matrix::colSums(sparseM)
    counts_per_feature <- Matrix::rowSums(sparseM)
    features_per_cell  <- Matrix::colSums(sparseM>0)
    cells_per_feature  <- Matrix::rowSums(sparseM>0)

    # generating the data frames for downstream use
    info_per_cell <- data.frame(counts_per_cell = counts_per_cell) %>%
        tibble::rownames_to_column(var = "cell") %>%
        dplyr::full_join(
        data.frame(features_per_cell = features_per_cell) %>%
            tibble::rownames_to_column(var = "cell"),
        by = "cell"
        )

    info_per_feature <- data.frame(counts_per_feature = counts_per_feature) %>%
        tibble::rownames_to_column(var = "feature") %>%
        dplyr::full_join(
        data.frame(cells_per_feature = cells_per_feature) %>%
            tibble::rownames_to_column(var = "feature"),
        by = "feature"
        )


    # Add annotation overlap information to the feature information data frame
    message("Computing feature QC metrics")
    features_in_matrix <- unique_feature.gr[paste(GenomicRanges::seqnames(unique_feature.gr), GenomicRanges::ranges(unique_feature.gr), sep=":") %in% info_per_feature$feature]

    pro.gr <- rtracklayer::import(promoters_file)
    enhs.gr <- rtracklayer::import(enhs_file)
    tss_df <- data.table::fread(tss_file, select=c(seq_len(3)), header = FALSE, col.names = c("chr", "start", "end"))
    tss.gr <- GenomicRanges::makeGRangesFromDataFrame(tss_df)

    pro.overlaps <- GenomicRanges::findOverlaps(query = features_in_matrix,
                                                    subject = pro.gr,
                                                    type = "any",
                                                    ignore.strand = TRUE)
    enhs.overlaps <- GenomicRanges::findOverlaps(query = features_in_matrix,
                                                subject = enhs.gr,
                                                type = "any",
                                                ignore.strand = TRUE)
    tss.overlaps <- GenomicRanges::findOverlaps(query = features_in_matrix,
                                                subject = tss.gr,
                                                type = "any",
                                                ignore.strand = TRUE)


    pro.hits <- seq(length(features_in_matrix)) %in% S4Vectors::queryHits(pro.overlaps)
    enhs.hits <- seq(length(features_in_matrix)) %in% S4Vectors::queryHits(enhs.overlaps)
    tss.hits <- seq(length(features_in_matrix)) %in% S4Vectors::queryHits(tss.overlaps)

    info_per_feature <- cbind(info_per_feature,
                                promoter_overlaps = pro.hits,
                                enhancer_overlaps = enhs.hits,
                                tss_overlaps = tss.hits)
    if (!is.null(gene_anno_file) && file.exists(gene_anno_file)) {
        gene_anno.gr <- rtracklayer::import(gene_anno_file)
        gene.overlaps <- GenomicRanges::findOverlaps(query = features_in_matrix,
                                                    subject = gene_anno.gr,
                                                    type = "any",
                                                    ignore.strand = TRUE)
        gene.hits <- seq(length(features_in_matrix)) %in% S4Vectors::queryHits(gene.overlaps)
        intergenic <- !(pro.hits | enhs.hits | tss.hits | gene.hits)
        info_per_feature <- cbind(info_per_feature, gene_overlaps = gene.hits, intergenic = intergenic)
    }
    

    write(
        c(
        "Feature quality control metrics produced at ",
        as.character(Sys.time()),
        "\n"
        ),
        file = log_file, append = TRUE)

    utils::write.csv(info_per_cell, paste0(log_and_stats_folder, "filtered_stats_per_cell.csv"), row.names = FALSE)
    utils::write.csv(info_per_feature, paste0(log_and_stats_folder, "filtered_stats_per_feature.csv"), row.names = FALSE)
    
    message("writing to csv")
    
    write(
        c(
        "sc_atac_feature_counting completed at ",
        as.character(Sys.time()),
        "\n\n"
        ), 
        file = log_file, append = TRUE)
    
    # create the sce object
    sc_atac_create_sce(input_folder = output_folder,
                                organism     = organism,
                                sample       = sample_name,
                                feature_type = feature_type,
                                pheno_data   = pheno_data,
                                report       = create_report)

    
    end_time <- Sys.time()
    message("sc_atac_feature_counting completed in ", difftime(end_time, init_time, units = "secs")[[1]], " seconds")
}

#' @name sc_atac_create_cell_qc_metrics
#' @title generating a file useful for producing the qc plots
#' @description uses the peak file and annotation files for features
#' 
#' @param frags_file The fragment file
#' @param peaks_file The peak file
#' @param promoters_file The path of the promoter annotation file 
#' @param tss_file The path of the tss annotation file 
#' @param enhs_file The path of the enhs annotation file 
#' @param output_folder The path of the output folder for resultant files
#' 
#' @importFrom data.table fread setkey copy :=

#' 
#' @returns Nothing (Invisible 'NULL')
#' @export
#' 
sc_atac_create_cell_qc_metrics <- function(frags_file,
                                            peaks_file,
                                            promoters_file,
                                            tss_file,
                                            enhs_file,
                                            output_folder) {
    
    pro.gr <- rtracklayer::import(promoters_file)
    tss_df <- data.table::fread(tss_file, select=c(seq_len(3)), header = FALSE, col.names = c("chr", "start", "end"))
    tss.gr <- GenomicRanges::makeGRangesFromDataFrame(tss_df)
    enhs.gr <- rtracklayer::import(enhs_file)
    
    # Peak file
    peaks.gr <-  rtracklayer::import(peaks_file)
    unique_peak.gr <- data.frame(peaks.gr) %>% dplyr::distinct(seqnames, start, end) %>% as.data.frame() %>% GenomicRanges::makeGRangesFromDataFrame()
    
    
    # Sinto fragments
    fragments <- data.table::fread(frags_file, select=seq_len(5), header = FALSE, col.names = c("seqnames", "start", "end", "barcode", "count")) 
    fragments.gr <- fragments %>% GenomicRanges::makeGRangesFromDataFrame(keep.extra.columns = TRUE)
    
    # Get all barcodes
    barcodes <- unique(fragments$barcode)
    
    mito_counts <- fragments %>% dplyr::group_by(barcode) %>% dplyr::mutate(count = sum(seqnames == "chrM")) %>% dplyr::select(barcode, count)
    
    # Compute overlaps of fragments with each feature
    peak.overlaps <- GenomicRanges::findOverlaps(query = fragments.gr,
                                                subject = unique_peak.gr,
                                                type = "any",
                                                ignore.strand = TRUE)
    pro.overlaps <- GenomicRanges::findOverlaps(query = fragments.gr,
                                                subject = pro.gr,
                                                type = "any",
                                                ignore.strand = TRUE)
    enhs.overlaps <- GenomicRanges::findOverlaps(query = fragments.gr,
                                                subject = enhs.gr,
                                                type = "any",
                                                ignore.strand = TRUE)
    tss.overlaps <- GenomicRanges::findOverlaps(query = fragments.gr,
                                                subject = tss.gr,
                                                type = "any",
                                                maxgap = 1000,
                                                ignore.strand = TRUE)
    
    total_counts <- table(fragments$barcode)
    peak_counts <- table(fragments.gr[unique(S4Vectors::queryHits(peak.overlaps))]$barcode)
    promoter_counts <- table(fragments.gr[unique(S4Vectors::queryHits(pro.overlaps))]$barcode)
    enhancer_counts <- table(fragments.gr[unique(S4Vectors::queryHits(enhs.overlaps))]$barcode)
    tss_counts <- table(fragments.gr[unique(S4Vectors::queryHits(tss.overlaps))]$barcode)
    
    qc_table <- data.frame(total_frags = rep(0, length(barcodes)),
                            frac_peak = rep(0, length(barcodes)),
                            frac_promoter = rep(0, length(barcodes)),
                            frac_tss = rep(0, length(barcodes)),
                            frac_enhancer = rep(0, length(barcodes)),
                            frac_mito = rep(0, length(barcodes)))
    
    rownames(qc_table) <- barcodes
    
    qc_table[names(total_counts), "total_frags"] <- total_counts
    qc_table[names(peak_counts), "frac_peak"] <- peak_counts
    qc_table[names(promoter_counts), "frac_promoter"] <- promoter_counts  
    qc_table[names(enhancer_counts), "frac_enhancer"] <- enhancer_counts
    qc_table[names(tss_counts), "frac_tss"] <- tss_counts
    qc_table[mito_counts$barcode, "frac_mito"] <- mito_counts$count
    
    qc_table$frac_peak <- qc_table$frac_peak/qc_table$total_frags
    qc_table$frac_promoter <- qc_table$frac_promoter/qc_table$total_frags
    qc_table$frac_enhancer <- qc_table$frac_enhancer/qc_table$total_frags
    qc_table$frac_tss <- qc_table$frac_tss/qc_table$total_frags
    qc_table$frac_mito <- qc_table$frac_mito/qc_table$total_frags
    
    qc_table <- qc_table[qc_table$total_frags > 5, ] %>% tibble::rownames_to_column("bc")
    
    utils::write.csv(qc_table, file = file.path(output_folder, "cell_qc_metrics.csv"), row.names = FALSE)
}

#' create a SingleCellExperiment object from data folder generated by preprocessing step
#'
#' after we run \code{sc_gene_counting} and finish the preprocessing step. \code{create_sce_by_dir}
#' can be used to generate the \link{SingleCellExperiment} object from the folder that contains gene count matrix and QC statistics.
#' it can also generate the html report based on the gene count and quality control statistics
#'
#' @param datadir the directory that contains all the data and `stat` subfolder.
#' @param organism the organism of the data. List of possible names can be retrieved using the function
#' `listDatasets`from `biomaRt` package. (i.e `mmusculus_gene_ensembl` or `hsapiens_gene_ensembl`)
#' @param gene_id_type gene id type of the data A possible list of ids can be retrieved using the function `listAttributes` from `biomaRt` package.
#' the commonly used id types are `external_gene_name`, `ensembl_gene_id` or `entrezgene`
#' @param pheno_data the external phenotype data that linked to each single cell. This should be an \code{AnnotatedDataFrame} object
#' @param report whether to generate the html report in the data folder
#'
#' @details after we run \code{sc_gene_counting} and finish the preprocessing step. \code{create_sce_by_dir}
#' can be used to generate the SingleCellExperiment object from the folder that contains gene count matrix and QC statistics.
#'
#' @return a SingleCellExperiment object
#'
#' @importFrom utils read.csv
#' @import SingleCellExperiment
#' @importClassesFrom SingleCellExperiment SingleCellExperiment
#' @importFrom utils packageVersion
#'
#' @export
#'
#' @examples
#' \dontrun{
#' # the sce can be created fron the output folder of scPipe
#' # please refer to the vignettes
#' sce = create_sce_by_dir(datadir="output_dir_of_scPipe",
#'     organism="mmusculus_gene_ensembl",
#'     gene_id_type="ensembl_gene_id")
#' }
#' # or directly from the gene count and quality control matrix:
#' data("sc_sample_data")
#' data("sc_sample_qc")
#' sce = SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
#' organism(sce) = "mmusculus_gene_ensembl"
#' gene_id_type(sce) = "ensembl_gene_id"
#' QC_metrics(sce) = sc_sample_qc
#' demultiplex_info(sce) = cell_barcode_matching
#' UMI_dup_info(sce) = UMI_duplication
#' dim(sce)
#'

create_sce_by_dir <- function(datadir, organism=NULL, gene_id_type=NULL, pheno_data=NULL, report=FALSE) {
    gene_cnt <- read.csv(file.path(datadir, "gene_count.csv"), row.names=1)
    cell_stat <- read.csv(file.path(datadir, "stat", "cell_stat.csv"), row.names=1)
    demultiplex_stat <- read.csv(file.path(datadir, "stat", "overall_stat.csv"))
    UMI_dup_stat <- read.csv(file.path(datadir, "stat", "UMI_duplication_count.csv"))

    gene_cnt <- gene_cnt[, order(colnames(gene_cnt))]
    cell_stat <- cell_stat[order(rownames(cell_stat)), ]


    sce <- SingleCellExperiment(assays = list(counts =as.matrix(gene_cnt)))
    sce@metadata$scPipe$version <- packageVersion("scPipe")  # set version information
    if(!is.null(organism)){
        organism(sce) <- organism
    }
    if(!is.null(gene_id_type)){
        gene_id_type(sce) <- gene_id_type
    }
    QC_metrics(sce) <- cell_stat
    if(!is.null(pheno_data)){
        colData(sce) <- cbind(colData(sce), pheno_data[order(rownames(pheno_data)),])
    }

    demultiplex_info(sce) <- demultiplex_stat
    UMI_dup_info(sce) <- UMI_dup_stat
    #if(any(grepl("^ERCC-", rownames(sce)))){
    #  isSpike(sce, "ERCC") <- grepl("^ERCC-", rownames(sce))
    #}


    if(report){
        create_report(sample_name=basename(datadir),
                    outdir=datadir,
                    organism=organism,
                    gene_id_type=gene_id_type)
    }


    return(sce)
}


#' create_report
#'
#' create an HTML report using data generated by proprocessing step.
#'
#' @param sample_name sample name
#' @param outdir output folder
#' @param r1 file path of read1
#' @param r2 file path of read2 default to be NULL
#' @param outfq file path of the output of \code{sc_trim_barcode}
#' @param read_structure a list contains read structure configuration. For more help see `?sc_trim_barcode`
#' @param filter_settings a list contains read filter settings for more help see `?sc_trim_barcode`
#' @param align_bam the aligned bam file
#' @param genome_index genome index used for alignment
#' @param map_bam the mapped bam file
#' @param exon_anno the gff exon annotation used. Can have multiple files
#' @param stnd whether to perform strand specific mapping
#' @param fix_chr add `chr` to chromosome names, fix inconsistent names.
#' @param barcode_anno cell barcode annotation file path.
#' @param max_mis maximum mismatch allowed in barcode. Default to be 1
#' @param UMI_cor correct UMI sequence error: 0 means no correction, 1 means simple correction and merge UMI with distance 1.
#' @param gene_fl whether to remove low abundant gene count. Low abundant is defined as only one copy of one UMI for this gene
#' @param organism the organism of the data. List of possible names can be retrieved using the function
#' `listDatasets`from `biomaRt` package. (i.e `mmusculus_gene_ensembl` or `hsapiens_gene_ensembl`)
#' @param gene_id_type gene id type of the data A possible list of ids can be retrieved using the function `listAttributes` from `biomaRt` package.
#' the commonly used id types are `external_gene_name`, `ensembl_gene_id` or `entrezgene`
#'
#' @return no return
#' @export
#'
#' @examples
#' \dontrun{
#' create_report(sample_name="sample_001",
#'        outdir="output_dir_of_scPipe",
#'        r1="read1.fq",
#'        r2="read2.fq",
#'        outfq="trim.fq",
#'        read_structure=list(bs1=-1, bl1=2, bs2=6, bl2=8, us=0, ul=6),
#'        filter_settings=list(rmlow=TRUE, rmN=TRUE, minq=20, numbq=2),
#'        align_bam="align.bam",
#'        genome_index="mouse.index",
#'        map_bam="aligned.mapped.bam",
#'        exon_anno="exon_anno.gff3",
#'        stnd=TRUE,
#'        fix_chr=FALSE,
#'        barcode_anno="cell_barcode.csv",
#'        max_mis=1,
#'        UMI_cor=1,
#'        gene_fl=FALSE,
#'        organism="mmusculus_gene_ensembl",
#'        gene_id_type="ensembl_gene_id")
#' }
#'
create_report <- function(sample_name,
                            outdir,
                            r1="NA",
                            r2="NA",
                            outfq="NA",
                            read_structure=list(bs1=0, bl1=0, bs2=0, bl2=0, us=0, ul=0),
                            filter_settings=list(rmlow = TRUE, rmN = TRUE, minq = 20, numbq = 2),
                            align_bam="NA",
                            genome_index="NA",
                            map_bam="NA",
                            exon_anno="NA",
                            stnd=TRUE,
                            fix_chr=FALSE,
                            barcode_anno="NA",
                            max_mis=1,
                            UMI_cor=1,
                            gene_fl=FALSE,
                            organism,
                            gene_id_type) {
    fn <- system.file("extdata", "report_template.Rmd", package = "scPipe")
    tx <- readLines(fn)

    tx <- gsub(pattern = "SAMPLE_NAME__", replacement = sample_name, x = tx)
    tx <- gsub(pattern = "FQ1__", replacement = r1, x = tx)
    if (!is.null(r2)) {
        tx <- gsub(pattern = "FQ2__", replacement = r2, x = tx)
    }
    else {
        tx <- gsub(pattern = "FQ2__", replacement = "NA", x = tx)
    }

    tx <- gsub(pattern = "FQOUT__", replacement = outfq, x = tx)
    if (read_structure$bs1<0) {
        tx <- gsub(pattern = "BC1_INFO__", replacement = "NA", x = tx)
    }
    else {
        tx <- gsub(pattern = "BC1_INFO__", replacement =
                    paste0("start at position ", read_structure$bs1, ", length ", read_structure$bl1), x = tx)
    }

    tx <- gsub(pattern = "BC2_INFO__", replacement =
                paste0("start at position ", read_structure$bs2, ", length ", read_structure$bl2), x = tx)
    tx <- gsub(pattern = "UMI_INFO__", replacement =
                paste0("start at position ", read_structure$us, ", length ", read_structure$ul), x = tx)

    tx <- gsub(pattern = "RM_N__", replacement = as.character(filter_settings$rmN), x = tx)
    tx <- gsub(pattern = "RM_LOW__", replacement = as.character(filter_settings$rmlow), x = tx)
    tx <- gsub(pattern = "MIN_Q__", replacement = filter_settings$minq, x = tx)
    tx <- gsub(pattern = "NUM_BQ__", replacement = filter_settings$numbq, x = tx)

    tx <- gsub(pattern = "BAM_ALIGN__", replacement = align_bam, x = tx)
    tx <- gsub(pattern = "G_INDEX__", replacement = genome_index, x = tx)
    tx <- gsub(pattern = "BAM_MAP__", replacement = map_bam, x = tx)

    tx <- gsub(pattern = "OUTDIR__", replacement = outdir, x = tx)
    tx <- gsub(pattern = "ANNO_GFF__", replacement = paste(exon_anno, collapse=", "), x = tx)

    tx <- gsub(pattern = "STND__", replacement = as.character(stnd), x = tx)
    tx <- gsub(pattern = "FIX_CHR__", replacement = as.character(fix_chr), x = tx)
    tx <- gsub(pattern = "BC_ANNO__", replacement = barcode_anno, x = tx)
    tx <- gsub(pattern = "MAX_MIS__", replacement = max_mis, x = tx)

    if (UMI_cor == 1) {
        tx <- gsub(pattern = "UMI_COR__", replacement = "simple correction and merge UMI with distance 1", x = tx)
    }
    else if (UMI_cor == 0) {
        tx <- gsub(pattern = "UMI_COR__", replacement = "no correction", x = tx)
    }
    else {
        tx <- gsub(pattern = "UMI_COR__", replacement = "unknown", x = tx)
    }

    # If organism and gene id type are not provided, delete them from param list
    # of rmd. param$organism and param$gene_id_type will then return NULL when
    # when used in code.
    tx <- gsub(pattern = "GENE_FL__", replacement = as.character(gene_fl), x = tx)
    if(!missing(organism) && !is.null(organism)){
        tx <- gsub(pattern = "ORGANISM__", replacement = organism, x = tx)
    }else{
        tx <- tx[!grepl(pattern = "ORGANISM__", x = tx)]
    }

    if(!missing(gene_id_type) && !is.null(gene_id_type)){
        tx <- gsub(pattern = "GENE_ID_TYPE__", replacement = gene_id_type, x = tx)
    }else{
        tx <- tx[!grepl(pattern = "GENE_ID_TYPE__", x = tx)]
    }

    writeLines(tx, con=file.path(outdir, "report.Rmd"))
    knitr::wrap_rmd(file.path(outdir, "report.Rmd"), width = 120, backup = NULL)
    rmarkdown::render(file.path(outdir, "report.Rmd"), output_file = file.path(outdir, "report.html"), knit_root_dir = ".")
}

#' create_processed_report
#'
#' Create an HTML report summarising pro-processed data. This is an alternative to the more verbose \code{create_report} that requires only the processed counts and stats folders.
#' @param outdir output folder.
#' @param organism the organism of the data. List of possible names can be retrieved using the function
#' `listDatasets`from `biomaRt` package. (e.g. `mmusculus_gene_ensembl` or `hsapiens_gene_ensembl`).
#' @param gene_id_type gene id type of the data A possible list of ids can be retrieved using the function `listAttributes` from `biomaRt` package.
#' the commonly used id types are `external_gene_name`, `ensembl_gene_id` or `entrezgene`.
#' @param report_name the name of the report .Rmd and .html files.
#'
#' @returns file path of the created compiled document.
#' @examples
#' \dontrun{
#' create_report(
#'        outdir="output_dir_of_scPipe",
#'        organism="mmusculus_gene_ensembl",
#'        gene_id_type="ensembl_gene_id")
#' }
#'
#' @export

create_processed_report <- function(
            outdir = ".",
            organism,
            gene_id_type,
            report_name = "report"
            ) {

    if (!requireNamespace("rmarkdown", quietly=TRUE)) {
        stop("Install 'rmarkdown' to use this function.")
    }
    fn <-  system.file("extdata", "report_template_slim.Rmd", package = "scPipe")
    tx <- readLines(fn)
    fill_report_field <- function(field, value) {
        pattern <- paste0(field, "__")

        if (is.na(value)) {
        tx <- tx[-grep(pattern, tx)]
        } else {
        gsub(pattern, value, tx)
        }
    }

    if (!missing(organism) && !is.null(organism)) {
        tx <- fill_report_field("ORGANISM", organism)
    } else {
        tx <- fill_report_field("ORGANISM", NA)
    }

    if (!missing(gene_id_type) && !is.null(gene_id_type)) {
        tx <- fill_report_field("GENE_ID_TYPE", gene_id_type)
    } else {
        tx <- fill_report_field("GENE_ID_TYPE", NA)
    }

    report_path <- file.path(outdir, paste0(report_name, ".Rmd"))
    tx <- tx[!is.na(tx)]
    writeLines(tx, con = report_path)
    rmarkdown::render(
        input = report_path,
        envir = new.env(),
        knit_root_dir = "."
    )
}

#' @name sc_atac_pipeline
#' 
#' @title A convenient function for running the entire pipeline
#' 
#' @param r1 The first read fastq file
#' @param r2 The second read fastq file
#' @param bc_file the barcode information, can be either in a \code{fastq} format (e.g. from 10x-ATAC) or
#' from a \code{.csv} file (here the barcode is expected to be on the second column). 
#' Currently, for the fastq approach, this can be a list of barcode files.
#' @param valid_barcode_file optional file path of the valid (expected) barcode sequences to be found in the bc_file (.txt, can be txt.gz). 
#' Must contain one barcode per line on the second column separated by a comma (default ="").
#' If given, each barcode from bc_file is matched against the barcode of
#' best fit (allowing a hamming distance of 1). If a FASTQ \code{bc_file} is provided, barcodes with a higher mapping quality, as given by
#' the fastq reads quality score are prioritised.
#' @param id1_st barcode start position (0-indexed) for read 1, which is an extra parameter that is needed if the
#' \code{bc_file} is in a \code{.csv} format.
#' @param id2_st barcode start position (0-indexed) for read 2, which is an extra parameter that is needed if the
#' \code{bc_file} is in a \code{.csv} format.
#' @param id1_len barcode length for read 1, which is an extra parameter that is needed if the
#' \code{bc_file} is in a \code{.csv} format.
#' @param id2_len barcode length for read 2, which is an extra parameter that is needed if the
#' \code{bc_file} is in a \code{.csv} format.
#' @param rmN ogical, whether to remove reads that contains N in UMI or cell barcode.
#' @param rmlow logical, whether to remove reads that have low quality barcode sequences.
#' @param organism The name of the organism e.g. hg38
#' @param reference The reference genome file
#' @param feature_type The feature type (either `genome_bin` or `peak`)
#' @param remove_duplicates Whether or not to remove duplicates (samtools is required) 
#' @param samtools_path A custom path of samtools to use for duplicate removal
#' @param bin_size The size of the bins for feature counting with the `genome_bin` feature type
#' @param yieldsize The number of reads to read in for feature counting
#' @param exclude_regions Whether or not the regions should be excluded
#' @param excluded_regions_filename The filename of the file containing the regions to be excluded
#' @param cell_calling The desired cell calling method either \code{cellranger}, \code{emptydrops} or  \code{filter}
#' @param promoters_file The path of the promoter annotation file (if the specified organism isn't recognised)
#' @param tss_file The path of the tss annotation file (if the specified organism isn't recognised)
#' @param enhs_file The path of the enhs annotation file (if the specified organism isn't recognised)
#' @param gene_anno_file The path of the gene annotation file (gtf or gff3 format)
#' @param fix_chr Specify `none`, `exclude_regions`, `feature` or `both` to prepend the string "chr" to the start of the associated file
#' @param lower the lower threshold for the data if using the \code{emptydrops} function for cell calling.
#' @param genome_size The size of the genome (used for the \code{cellranger} cell calling method)
#' @param min_uniq_frags The minimum number of required unique fragments required for a cell (used for \code{filter} cell calling)
#' @param max_uniq_frags The maximum number of required unique fragments required for a cell (used for \code{filter} cell calling)
#' @param min_frac_peak The minimum proportion of fragments in a cell to overlap with a peak (used for \code{filter} cell calling)
#' @param min_frac_tss The minimum proportion of fragments in a cell to overlap with a tss (used for \code{filter} cell calling)
#' @param min_frac_enhancer The minimum proportion of fragments in a cell to overlap with a enhancer sequence (used for \code{filter} cell calling)
#' @param min_frac_promoter The minimum proportion of fragments in a cell to overlap with a promoter sequence (used for \code{filter} cell calling)
#' @param max_frac_mito The maximum proportion of fragments in a cell that are mitochondrial (used for \code{filter} cell calling)
#' @param report Whether or not a HTML report should be produced
#' @param nthreads The number of threads to use for alignment (sc_align) and demultiplexing (sc_atac_bam_tagging)
#' @param output_folder The path of the output folder
#'
#' @returns None (invisible `NULL`)
#' @examples
#' data.folder <- system.file("extdata", package = "scPipe", mustWork = TRUE)
#' r1      <- file.path(data.folder, "small_chr21_R1.fastq.gz") 
#' r2      <- file.path(data.folder, "small_chr21_R3.fastq.gz") 
#' 
#' # Using a barcode fastq file:
#'
#' # barcodes in fastq format
#' barcode_fastq      <- file.path(data.folder, "small_chr21_R2.fastq.gz") 
#' 
#' \dontrun{
#' sc_atac_pipeline(
#'   r1 = r1,
#'   r2 = r2,
#'   bc_file = barcode_fastq
#' )
#' }
#'
#' @export
#' 
sc_atac_pipeline <- function(r1,
                            r2,
                            bc_file,
                            valid_barcode_file = "",
                            id1_st = -0,
                            id1_len = 16,
                            id2_st = 0,
                            id2_len = 16,
                            rmN           = TRUE,
                            rmlow         = TRUE,
                            organism = NULL,
                            reference = NULL,
                            feature_type = NULL,
                            remove_duplicates = FALSE,
                            samtools_path = NULL,
                            genome_size   = NULL,
                            bin_size      = NULL,
                            yieldsize     = 1000000,
                            exclude_regions = TRUE,
                            excluded_regions_filename = NULL,
                            fix_chr = "none",
                            lower = NULL,
                            cell_calling = "filter",
                            promoters_file = NULL,
                            tss_file       = NULL,
                            enhs_file      = NULL,
                            gene_anno_file = NULL,
                            min_uniq_frags = 3000,
                            max_uniq_frags = 50000,
                            min_frac_peak = 0.3,
                            min_frac_tss = 0,
                            min_frac_enhancer = 0,
                            min_frac_promoter = 0.1,
                            max_frac_mito = 0.15,
                            report = TRUE,
                            nthreads = 12,
                            output_folder = NULL) {
                                
    get_filename_without_extension <- function(path, extension_length = 1) {
        vec <- strsplit(basename(path), "\\.")[[1]]
        name.size <- length(vec) - extension_length
        return(paste(vec[seq_len(name.size)], collapse = "."))
    }

    r1_name <- get_filename_without_extension(r1, extension_length = 2)
    r2_name <- get_filename_without_extension(r2, extension_length = 2)

    sc_atac_trim_barcode (r1            = r1,
                        r2            = r2,
                        bc_file = bc_file,
                        valid_barcode_file = valid_barcode_file,
                        id1_st = id1_st,
                        id1_len = id1_len,
                        id2_st = id2_st,
                        id2_len = id2_len,
                        rmN           = TRUE,
                        rmlow         = TRUE,
                        output_folder = output_folder)
    
    
    demux_r1        <- file.path(output_folder, paste0("demux_completematch_", r1_name, ".fastq.gz"))
    demux_r2        <- file.path(output_folder, paste0("demux_completematch_", r2_name, ".fastq.gz"))
    
    reference       <- reference
    bam_to_tag <- sc_aligning(ref = reference,
                    tech = "atac",
                    R1 = demux_r1,
                    R2 = demux_r2,
                    nthreads  = nthreads,
                    output_folder = output_folder)

    sorted_tagged_bam <- sc_atac_bam_tagging(inbam = bam_to_tag,
                        output_folder = output_folder,
                        bam_tags      = list(bc="CB", mb="OX"),
                        nthreads      =  nthreads)
    
    if (isTRUE(remove_duplicates)) {
        removed <- sc_atac_remove_duplicates(inbam = sorted_tagged_bam,
                                samtools_path = samtools_path,
                                output_folder = output_folder)
        if (!isFALSE(removed))
        sorted_tagged_bam <- removed
    }
    sc_atac_create_fragments(inbam = sorted_tagged_bam,
                            output_folder = output_folder)

    features <- NULL
    if (feature_type == "peak") {
        sc_atac_peak_calling(inbam = sorted_tagged_bam,
                            output_folder = output_folder,
                            reference)

        features <- file.path(output_folder, "NA_peaks.narrowPeak")
    } else { # otherwise is genome bins
        features <- reference
    }

    sc_atac_feature_counting (fragment_file = file.path(output_folder, "fragments.bed"),
                                feature_input = features,
                                bam_tags      = list(bc="CB", mb="OX"),
                                feature_type  = feature_type,
                                organism      = organism,
                                cell_calling  = cell_calling,
                                genome_size   = genome_size,
                                promoters_file = promoters_file,
                                tss_file       = tss_file,
                                enhs_file      = enhs_file,
                                gene_anno_file = gene_anno_file,
                                bin_size      = bin_size,
                                yieldsize     = yieldsize,
                                exclude_regions = exclude_regions,
                                excluded_regions_filename = excluded_regions_filename,
                                output_folder = output_folder,
                                fix_chr       = fix_chr,
                                lower         = lower,
                                min_uniq_frags = min_uniq_frags,
                                max_uniq_frags = max_uniq_frags,
                                min_frac_peak = min_frac_peak,
                                min_frac_tss = min_frac_tss,
                                min_frac_enhancer = min_frac_enhancer,
                                min_frac_promoter = min_frac_promoter,
                                max_frac_mito = max_frac_mito)
    
    # sce <- sc_atac_create_sce(input_folder = output_folder,
    #                           organism     = organism,
    #                           feature_type = feature_type,
    #                           pheno_data   = NULL,
    #                           report       = report)
    # return(sce)
}

#' @name sc_atac_pipeline_quick_test
#' @title A function that tests the pipeline on a small test sample (without duplicate removal)
#' @returns None (invisible `NULL`)
sc_atac_pipeline_quick_test <- function() {
    data.folder <- system.file("extdata", package = "scPipe", mustWork = TRUE)
    output_folder <- file.path(getwd(), "scPipe-atac-output")
    out <- tryCatch({
        sce <- sc_atac_pipeline(r1 = file.path(data.folder, "small_chr21_R1.fastq.gz"),
                                r2 = file.path(data.folder, "small_chr21_R3.fastq.gz"),
                                bc_file = file.path(data.folder, "small_chr21_R2.fastq.gz"),
                                organism = "hg38",
                                reference = file.path(data.folder, "small_chr21.fa"),
                                feature_type = "peak",
                                remove_duplicates = FALSE,
                                min_uniq_frags = 0,
                                min_frac_peak = 0,
                                min_frac_promoter = 0,
                                output_folder = output_folder)
        message("Successfully ran pipeline.")
        },
        error = function(e) {
        message(e)
        },
        finally = {
        # system2("rm", c("-rf", output_folder))
        }
    )
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

check_barcode_reads <- function(fastq, barcodeseqs, barcodeRealname, barcode_start, barcode_length, lines_to_search, threshold) {
    .Call(`_scPipe_check_barcode_reads`, fastq, barcodeseqs, barcodeRealname, barcode_start, barcode_length, lines_to_search, threshold)
}

sc_atac_getOverlaps_read2AnyRegion <- function(reads, regions) {
    .Call(`_scPipe_sc_atac_getOverlaps_read2AnyRegion`, reads, regions)
}

sc_atac_getOverlaps_tss2Reads <- function(regions, left_flank, reads) {
    .Call(`_scPipe_sc_atac_getOverlaps_tss2Reads`, regions, left_flank, reads)
}

rcpp_sc_trim_barcode_paired <- function(outfq, r1, r2, bs1, bl1, bs2, bl2, us, ul, rmlow, rmN, minq, numbq, write_gz) {
    invisible(.Call(`_scPipe_rcpp_sc_trim_barcode_paired`, outfq, r1, r2, bs1, bl1, bs2, bl2, us, ul, rmlow, rmN, minq, numbq, write_gz))
}

rcpp_sc_exon_mapping <- function(inbam, outbam, annofn, am, ge, bc, mb, bc_len, bc_vector, UMI_len, stnd, fix_chr, nthreads) {
    invisible(.Call(`_scPipe_rcpp_sc_exon_mapping`, inbam, outbam, annofn, am, ge, bc, mb, bc_len, bc_vector, UMI_len, stnd, fix_chr, nthreads))
}

rcpp_sc_exon_mapping_df_anno <- function(inbam, outbam, anno, am, ge, bc, mb, bc_len, bc_vector, UMI_len, stnd, fix_chr, nthreads) {
    invisible(.Call(`_scPipe_rcpp_sc_exon_mapping_df_anno`, inbam, outbam, anno, am, ge, bc, mb, bc_len, bc_vector, UMI_len, stnd, fix_chr, nthreads))
}

rcpp_sc_demultiplex <- function(inbam, outdir, bc_anno, max_mis, am, ge, bc, mb, mito, has_UMI, nthreads) {
    invisible(.Call(`_scPipe_rcpp_sc_demultiplex`, inbam, outdir, bc_anno, max_mis, am, ge, bc, mb, mito, has_UMI, nthreads))
}

rcpp_sc_clean_bam <- function(inbam, outbam, bc_anno, max_mis, am, ge, bc, mb, mito, nthreads) {
    invisible(.Call(`_scPipe_rcpp_sc_clean_bam`, inbam, outbam, bc_anno, max_mis, am, ge, bc, mb, mito, nthreads))
}

rcpp_sc_gene_counting <- function(outdir, bc_anno, UMI_cor, gene_fl) {
    invisible(.Call(`_scPipe_rcpp_sc_gene_counting`, outdir, bc_anno, UMI_cor, gene_fl))
}

rcpp_sc_detect_bc <- function(infq, outcsv, prefix, bc_len, max_reads, number_of_cells, min_count, max_mismatch, white_list) {
    invisible(.Call(`_scPipe_rcpp_sc_detect_bc`, infq, outcsv, prefix, bc_len, max_reads, number_of_cells, min_count, max_mismatch, white_list))
}

#' @useDynLib scPipe, .registration = TRUE
sc_atac_create_fragments_cpp <- function(inbam, output, contigs, ends, min_mapq, nproc, cellbarcode, chromosomes, readname_barcodeN, cellsN, max_distance, min_distance, chunksize) {
    invisible(.Call(`_scPipe_sc_atac_create_fragments_cpp`, inbam, output, contigs, ends, min_mapq, nproc, cellbarcode, chromosomes, readname_barcodeN, cellsN, max_distance, min_distance, chunksize))
}

rcpp_sc_atac_trim_barcode <- function(outfq, r1, r3, barcode_file, valid_barcode_file, umi_start, umi_len, umi_in, write_gz, rmN, rmlow, min_qual, num_below_min, id1_st, id1_len, id2_st, id2_len) {
    .Call(`_scPipe_rcpp_sc_atac_trim_barcode`, outfq, r1, r3, barcode_file, valid_barcode_file, umi_start, umi_len, umi_in, write_gz, rmN, rmlow, min_qual, num_below_min, id1_st, id1_len, id2_st, id2_len)
}

rcpp_sc_atac_trim_barcode_paired <- function(outfq, r1, r2_list, r3, valid_barcode_file, write_gz, rmN, rmlow, min_qual, num_below_min, no_reverse_complement) {
    .Call(`_scPipe_rcpp_sc_atac_trim_barcode_paired`, outfq, r1, r2_list, r3, valid_barcode_file, write_gz, rmN, rmlow, min_qual, num_below_min, no_reverse_complement)
}

rcpp_sc_atac_bam_tagging <- function(inbam, outbam, bc, mb, nthreads) {
    invisible(.Call(`_scPipe_rcpp_sc_atac_bam_tagging`, inbam, outbam, bc, mb, nthreads))
}

rcpp_fasta_bin_bed_file <- function(in_filename, out_filename, bin_size) {
    invisible(.Call(`_scPipe_rcpp_fasta_bin_bed_file`, in_filename, out_filename, bin_size))
}

rcpp_append_chr_to_bed_file <- function(in_filename, out_filename) {
    invisible(.Call(`_scPipe_rcpp_append_chr_to_bed_file`, in_filename, out_filename))
}

get_all_TSS_bins <- function(tss_df, range, bin_size) {
    .Call(`_scPipe_get_all_TSS_bins`, tss_df, range, bin_size)
}

###############################################
# Create the final report for scATAC-Seq data
###############################################

#' @name sc_atac_create_report
#' @title HTML report generation
#' @description Generates a HTML report using the output folder produced by the pipeline
#'
#' @param input_folder The path of the folder produced by the pipeline
#' @param output_folder The path of the output folder to store the HTML report in
#' @param organism A string indicating the name of the organism being analysed
#' @param sample_name A string indicating the name of the sample
#' @param feature_type A string indicating the type of the feature (`genome_bin` or `peak`)
#' @param n_barcode_subset if you require only to visualise stats for a sample of barcodes to improve processing time (integer)
#'
#' @returns the path of the output file
#' @export
sc_atac_create_report <- function(input_folder, 
                                    output_folder    = NULL, 
                                    organism         = NULL,
                                    sample_name      = NULL,
                                    feature_type     = NULL,
                                    n_barcode_subset = 500){
    
    if (!dir.exists(input_folder)){
        stop("The input folder could not be found at ", input_folder);
    }

    if (!requireNamespace("rmarkdown", quietly=TRUE)) {
        stop("Install 'rmarkdown' to use this function")
    }
    
    if (!dir.exists(file.path(input_folder, "scPipe_atac_stats"))) {
        stop("The input stats folder could not be found at ", 
            file.path(input_folder, "scPipe_atac_stats"), 
            ". Either run the pipeline to generate it or create a `scPipe_atac_stats` folder that contains the required files.");
    }
    
    available_organisms <- c("hg19",
                            "hg38",
                            "mm10")


    if(is.null(output_folder)) {
        output_folder <- file.path(input_folder, "scPipe_atac_stats")
    }
    
    if (!dir.exists(output_folder)){
        message("Output folder could not be found at " , output_folder,  "Creating it now...");
        dir.create(output_folder, recursive=TRUE)
    }
    
    report.file <- system.file("extdata/rmd_report_skeleton.Rmd", package = "scPipe") 
    if (report.file == "") # Used if installed via the repo
        report.file <- system.file("inst/extdata/rmd_report_skeleton.Rmd", package = "scPipe", mustWork = TRUE)
    
    output_file <- file.path(output_folder, "scPipe_atac_report.html")
    rmarkdown::render(
        input  = report.file,
        params = list(
        input_folder     = input_folder,
        organism         = organism,
        sample           = sample_name,
        feature_type     = feature_type,
        n_barcode_subset = n_barcode_subset
        ),
        output_file = output_file 
    )

    return(output_file)
}

#####################################################
# Integrate multi-omic scRNA-Seq and scATAC-Seq data
#####################################################

#' @name sc_integrate
#' @title Integrate multi-omic scRNA-Seq and scATAC-Seq data into a MultiAssayExperiment
#' @description Generates an integrated SCE object with scRNA-Seq and scATAC-Seq data produced by the scPipe pipelines
#' @param sce_list A list of SCE objects, named with the corresponding technologies
#' @param sce_column_to_barcode_files A list of files containing the barcodes for each tech (if not needed then give a `NULL` entry)
#' @param rev_comp A named list of technologies and logical flags specifying if reverse complements should be applyed for sequences (if not needed then provide a `NULL` entry)
#' @param cell_line_info A list of files, each of which contains 2 columns corresponding to the barcode and cell line for each tech (if not needed then provide a `NULL` entry)
#' @param barcode_match_file A .csv file with columns corresponding to the barcodes for each tech
#' @param output_folder The path to the output folder
#' @returns Returns a MultiAssayExperiment containing the scRNA-Seq and scATAC-Seq data produced by the scPipe pipelines
#'
#' @examples
#' \dontrun{
#' sc_integrate(
#'    sce_list = list("RNA" = sce.rna, "ATAC" = sce.atac),
#'    barcode_match_file = bc_match_file,
#'    sce_column_to_barcode_files = list("RNA" = rna_bc_anno, "ATAC" = NULL),
#'    rev_comp = list("RNA" = FALSE, "ATAC" = TRUE),
#'    cell_line_info = list("RNA" = rna_cell_line_info, "ATAC" = atac_cell_line_info,)
#'    output_folder = output_folder
#'    )
#' }  
#' 
#' @export
#'
sc_integrate <- function(sce_list,
                            barcode_match_file = NULL,
                            sce_column_to_barcode_files = NULL,
                            rev_comp = NULL,
                            cell_line_info = NULL,
                            output_folder = NULL) {
    
    if (!is.null(output_folder) && !dir.exists(output_folder)){
        dir.create(output_folder,recursive=TRUE)
        message("Output directory is not provided. Created directory: ", output_folder)
    }
    techs <- names(sce_list)
    
    # Apply reverse complement if specified
    if (!is.null(rev_comp)) {
        message("Applying reverse complement to columns of SCE objects where required.")
        sce_list <- lapply(seq_along(sce_list), function(i) {
            sce <- sce_list[[i]]
            if (isTRUE(rev_comp[[i]]))
                colnames(sce) <- Biostrings::reverseComplement(Biostrings::DNAStringSet(colnames(sce))) %>% as.character()
            sce
        })
    }
    names(sce_list) <- techs # add names back
    
    if (!is.null(sce_column_to_barcode_files)) {
        message("Updating columns of SCE objects to barcodes where required.")
        sce_list <- lapply(seq_along(sce_list), function(i) {
            tech <- names(sce_list)[[i]]
            sce <- sce_list[[i]]
            column_to_bc_file <- sce_column_to_barcode_files[tech][[1]]
            if(!is.null(column_to_bc_file)) {
                column_to_bc <- read.csv(column_to_bc_file, header = TRUE)
                colnames(column_to_bc)[seq_len(2)] <- c("cell_name", "barcode_sequence")
                if (!all(colnames(sce) %in% column_to_bc$cell_name)) 
                stop("Columns of SCE object not present in annotation file!")
                colnames(sce) <- column_to_bc$barcode_sequence[match(colnames(sce), column_to_bc$cell_name)]
            }
            sce
        })
        names(sce_list) <- techs # add names back
    }
    
    if (!is.null(barcode_match_file)) {
        barcode_match_df <- read.csv(barcode_match_file, header = TRUE)
        # Check if the barcodes match up with those in the match file, and if not, try taking the reverse complement to see if it helps
        sce_list <- lapply(seq_along(sce_list), function(i) {
            sce <- sce_list[[i]]
            tech <- names(sce_list)[[i]]
            matched <- stats::na.omit(match(colnames(sce), barcode_match_df[[tech]]))
            if (length(matched)/length(colnames(sce)) < 0.2) {
                message("Less than 20% of", tech, "barcodes match with the barcodes in the barcode match file. Trying the reverse complement.")
                rev_comp <- Biostrings::reverseComplement(Biostrings::DNAStringSet(colnames(sce))) %>% as.character()
                new_matched <- stats::na.omit(match(rev_comp, barcode_match_df[[tech]]))
                if (length(new_matched)/length(colnames(sce)) <= length(matched)/length(colnames(sce))) {
                message("Reverse complement didn't yield higher proportion of matched barcodes so using existing barcodes.")
                } else {
                message("Reverse complement yielded higher proportion of matched barcodes so using reverse complement")
                colnames(sce) <- rev_comp
                }
            }
            sce
        })
        names(sce_list) <- techs # add names back
    }
    
    
    # Left join the cell line info
    if (!is.null(cell_line_info)) {
        message("Attaching cell line information provided.")
        sce_list <- lapply(seq_along(sce_list), function(i) {
            sce <- sce_list[[i]]
            tech <- names(sce_list)[[i]]
            cell_line_file <- cell_line_info[[i]]
            
            if (!is.null(cell_line_file)) {
                cli <- read.csv(cell_line_file, header=FALSE)
                colnames(cli) <- c("barcode", "cell_line")
                merged_cell_line_info <- base::merge(data.frame(barcode = colnames(sce)), cli, all.x = TRUE) %>% tibble::column_to_rownames("barcode")
                reordered_cli <- merged_cell_line_info[colnames(sce), , drop=FALSE]
                colData(sce) <- cbind(colData(sce), cell_line = reordered_cli$cell_line)
            } 
            sce
        })
        names(sce_list) <- techs # add names back
    }
    
    # Outer join the column data
    if (!is.null(barcode_match_file)) { 
        message("Merging qc metrics")
        qc_dfs <- lapply(seq_along(sce_list), function(i) {
            sce <- sce_list[[i]]
            tech <- techs[[i]]
            qc <- data.frame(colData(sce)) %>% dplyr::rename_with( ~ paste0(tech, "_", .x))
            base::merge(qc, barcode_match_df, by.x = "row.names", by.y = techs[[i]], all.x = TRUE) %>% dplyr::rename(!!tech := "Row.names")
        })
        merged_qc <- Reduce(function(x, y) base::merge(x, y, all = TRUE), qc_dfs)
    }
    
    # Mark in coldata of each experiment if the barcode is shared
    shared_bc <- stats::na.omit(merged_qc)
    for (i in seq_len(length(sce_list))) {
        tech <- techs[[i]]
        colData(sce_list[[tech]])[, "shared"] <- rownames(colData(sce_list[[tech]])) %in% shared_bc[[tech]]
    }
    
    # Create MultiAssayExperiment
    message("Creating MultiAssayExperiment")
    mae <- MultiAssayExperiment::MultiAssayExperiment(experiments = sce_list)
    mae@metadata$scPipe$version <- packageVersion("scPipe") 
    if (!is.null(barcode_match_file)) mae@metadata$scPipe$integrated_qc <- merged_qc
    
    # Save the MAE object
    if (!is.null(output_folder)) saveRDS(mae, file.path(output_folder, "scPipe_MAE_object.rds"))
    
    message("sc_integrate_complete.")
    return(mae)
}

#' @name sc_mae_plot_umap
#' @title Generates UMAP of multiomic data
#' @description Uses feature count data from multiple experiment objects to produce UMAPs for each assay and then overlay them on the same pair of axes
#' @param mae The MultiAssayExperiment object
#' @param by What to colour the points by. Needs to be in colData of all experiments.
#' @param output_file The path of the output file
#' @returns A ggplot2 object representing the UMAP plot
#'
#' @export
sc_mae_plot_umap <- function(mae,
                                by = NULL,
                                output_file = NULL) {
    #set.seed(123)
    
    if (!requireNamespace("RANN", quietly=TRUE)) {
        stop("Install 'RANN' to use this function")
    }
    if (!requireNamespace("igraph", quietly=TRUE)) {
        stop("Install 'igraph' to use this function")
    }
    if (!requireNamespace("umap", quietly=TRUE)) {
        stop("Install 'umap' to use this function")
    }
    umap_dfs <- lapply(seq_along(mae), function(i) {
        tech <- names(mae)[[i]]
        counts <- assays(mae)[[tech]]
        bin_mat <- as.matrix((counts>0)+0)
        binary.mat <- TF.IDF.custom(bin_mat)
        n_bcs <- max(min(50, ncol(binary.mat), nrow(binary.mat))-1,0)
        mat.lsi          <- irlba(binary.mat, n_bcs)
        d_diagtsne       <- matrix(0, n_bcs, n_bcs)
        diag(d_diagtsne) <- mat.lsi$d
        mat_pcs          <- t(d_diagtsne %*% t(mat.lsi$v))
        rownames(mat_pcs)<- colnames(binary.mat)
        
        # clustering in the PCA space using KNN --------------
        knn.info<- RANN::nn2(mat_pcs, k = 30)
    
        ## convert to adjacency matrix
        knn           <- knn.info$nn.idx
        adj           <- matrix(0, nrow(mat_pcs), nrow(mat_pcs))
        rownames(adj) <- colnames(adj) <- rownames(mat_pcs)
        for(i in seq_len(nrow(mat_pcs))) {
            adj[i,rownames(mat_pcs)[knn[i,]]] <- 1
        }
        
        ## convert to graph
        g <- igraph::graph.adjacency(adj, mode="undirected")
        g <- purrr::simplify(g) ## remove self loops
        
        # identify communities, many algorithms. Use the Louvain clustering ------------
        km         <- igraph::cluster_louvain(g)
        com        <- km$membership
        names(com) <- km$names
    
        # running UMAP ------------------------------
        norm.data.umap    <- umap::umap(mat_pcs)
        
        df_umap           <- as.data.frame(norm.data.umap$layout)
        colnames(df_umap) <- c("UMAP1", "UMAP2")
        df_umap$barcode   <- rownames(mat_pcs)
        
        df_umap           <- dplyr::left_join(df_umap, tibble::enframe(com), by = c("barcode" = "name")) %>%
        dplyr::rename(cluster = value) %>%
        dplyr::mutate(cluster = as.factor(cluster))
        
        df_umap$source <- tech
        
        if (!is.null(by) && !is.null(colData(experiments(mae)[[tech]])[[by]])) {
            sce_coldata <- colData(experiments(mae)[[tech]])[, c(by), drop=FALSE]
            df_umap <- base::merge(df_umap, sce_coldata, by.x = "barcode", by.y = "row.names", all.x = TRUE) 
        } else if (!is.null(by)) {
            df_umap[[by]] <- NA
        }
        df_umap
    })
    names(umap_dfs) <- names(mae)
    
    umap_data <- do.call(rbind, umap_dfs)
    if (!is.null(by)) { 
        g <- ggplot2::ggplot(umap_data, ggplot2::aes(x = UMAP1, y = UMAP2)) +
        ggplot2::geom_point(ggplot2::aes(col = .data[[by]]), size = 2, alpha = 0.5) +
        ggplot2::theme_bw(base_size = 14)
        if (is.numeric(umap_data[[by]])) {
            g <- ggplot2::ggplot(umap_data, ggplot2::aes(x = UMAP1, y = UMAP2)) +
                ggplot2::geom_point(ggplot2::aes(col = .data[[by]]), size = 2, alpha = 0.5) +
                ggplot2::scale_colour_gradientn(colours=c("green","black")) +
                ggplot2::theme_bw(base_size = 14)
        }
    } else {
        g <- ggplot2::ggplot(umap_data, ggplot2::aes(x = UMAP1, y = UMAP2)) +
        ggplot2::geom_point(ggplot2::aes(col = source), size = 2, alpha = 0.5) +
        ggplot2::theme_bw(base_size = 14)
    }
    
    if (!is.null(output_file)) {
        if (file.exists(output_file)) {
            ggplot2::ggsave(output_file)
            message("Saved plot to", output_file)
        } else {
            message("The supplied output file path was invalid.")
        }
    }
    
    # plotly::ggplotly(g)
    
    return(g)
}

#' @name TF.IDF.custom
#' @title Returns the TF-IDF normalised version of a binary matrix
#' @param binary.mat The binary matrix
#' @param verbose boolean flag to print status messages
#' @returns Returns the TF-IDF normalised version of a binary matrix
#' @export
TF.IDF.custom <- function(binary.mat, verbose = TRUE) {
    if (verbose) {
        message("Performing TF-IDF normalization")
    }
    object <- binary.mat
    npeaks       <- Matrix::colSums(x = object)
    tf           <- Matrix::tcrossprod(x = as.matrix(object), y = Matrix::Diagonal(x = 1 / npeaks))
    rsums        <- Matrix::rowSums(x = object)
    idf          <- ncol(x = object) / rsums
    norm.data    <- Matrix::Diagonal(n = length(x = idf), x = idf) %*% tf
    scale.factor <- 1e4
    methods::slot(object = norm.data, name = "x") <- log1p(x = methods::slot(object = norm.data, name = "x") * scale.factor)
    norm.data[which(x = is.na(x = norm.data))] <- 0
    return(norm.data)
}

#' @name QC_metrics
#' @aliases QC_metrics
#' @export
#' @docType methods
#' @return a dataframe of quality control matrics
#' @rdname QC_metrics
setGeneric("QC_metrics", function(object) standardGeneric("QC_metrics"))

#' @name QC_metrics
#' @aliases QC_metrics<-
#' @export
#' @docType methods
#' @rdname QC_metrics
setGeneric("QC_metrics<-", function(object, value) standardGeneric("QC_metrics<-"))


#' @name demultiplex_info
#' @aliases demultiplex_info
#' @export
#' @docType methods
#' @return a dataframe of cell barcode demultiplex information
#' @rdname demultiplex_info
setGeneric("demultiplex_info", function(object) standardGeneric("demultiplex_info"))

#' @name demultiplex_info
#' @aliases demultiplex_info<-
#' @export
#' @docType methods
#' @rdname demultiplex_info
setGeneric("demultiplex_info<-", function(object, value) standardGeneric("demultiplex_info<-"))



#' @name UMI_dup_info
#' @aliases UMI_dup_info
#' @export
#' @docType methods
#' @return a dataframe of cell UMI duplication information
#' @rdname UMI_dup_info
setGeneric("UMI_dup_info", function(object) standardGeneric("UMI_dup_info"))

#' @name UMI_dup_info
#' @aliases UMI_dup_info<-
#' @export
#' @docType methods
#' @rdname UMI_dup_info
setGeneric("UMI_dup_info<-", function(object, value) standardGeneric("UMI_dup_info<-"))

#' @name gene_id_type
#' @aliases gene_id_type
#' @export
#' @docType methods
#' @return the gene id type used by Biomart
#' @rdname gene_id_type
setGeneric("gene_id_type", function(object) standardGeneric("gene_id_type"))

#' @name gene_id_type
#' @aliases gene_id_type<-
#' @export
#' @docType methods
#' @rdname gene_id_type
setGeneric("gene_id_type<-", function(object, value) standardGeneric("gene_id_type<-"))

#' @name feature_info
#' @aliases feature_info
#' @export
#' @docType methods
#' @return a dataframe of feature info for scATAC-seq data
#' @rdname feature_info
setGeneric("feature_info", function(object) standardGeneric("feature_info"))

#' @name feature_info
#' @aliases feature_info<-
#' @export
#' @docType methods
#' @rdname feature_info
setGeneric("feature_info<-", function(object, value) standardGeneric("feature_info<-"))

#' @name feature_type
#' @aliases feature_type
#' @export
#' @docType methods
#' @return the feature type used in feature counting for scATAC-Seq data
#' @rdname feature_type
setGeneric("feature_type", function(object) standardGeneric("feature_type"))

#' @name feature_type
#' @aliases feature_type<-
#' @export
#' @docType methods
#' @rdname feature_type
setGeneric("feature_type<-", function(object, value) standardGeneric("feature_type<-"))

##########################################################
# Aligning Demultiplxed FASTQ Reads to a Known Reference
##########################################################

#' @name sc_aligning
#' @title aligning the demultiplexed FASTQ reads using the Rsubread:align()
#' @description after we run the \code{sc_trim_barcode} or \code{sc_atac_trim_barcode} to demultiplex the fastq files, we are using this
#' function to align those fastq files to a known reference.
#' @param ref a character string specifying the path to reference genome file (.fasta, .fa format)
#' @param index_path character string specifying the path/basename of the index files, if the Rsubread genome build is available 
#' @param tech a character string giving the sequencing technology. Possible value includes "atac" or "rna"
#' @param R1 a mandatory character vector including names of files that include sequence reads to be aligned. For paired-end reads, this gives the list of files including first reads in each library. File format is FASTQ/FASTA by default.
#' @param R2 a character vector, the second fastq file, which is required if the data is paired-end
#' @param output_folder a character string, the name of the output folder
#' @param output_file a character vector specifying names of output files. By default, names of output files are set as the file names provided in R1 added with an suffix string
#' @param type type of sequencing data (`RNA` or `DNA`)
#' @param input_format a string indicating the input format
#' @param output_format a string indicating the output format
#' @param nthreads numeric value giving the number of threads used for mapping. 
#'
#' @returns the file path of the output aligned BAM file
#'
#' @examples
#' \dontrun{
#' sc_aligning(index_path,
#'     tech = 'atac',
#'     R1,  
#'     R2, 
#'     nthreads  = 6) 
#' }
#'@export
sc_aligning <- function (
    R1, 
    R2            = NULL,
    tech          = "atac",
    index_path    = NULL,
    ref           = NULL,
    output_folder = NULL, 
    output_file   = NULL,
    input_format  = "FASTQ",
    output_format = "BAM",
    type          = "dna",
    nthreads      = 1){

    if(!all(file.exists(R1))){
        stop("At least one of the input files for R1 does not exist")    
    }
    
    if(!is.null(R2) && !all(file.exists(R2))){
        stop("At least one of input file for R2 does not exist")
    }
    
    if(tech == "atac") {
        message("ATAC-Seq mode is selected...")
    
        if(is.null(output_folder)) {
            output_folder        <- file.path(getwd(), "scPipe-atac-output")
        }
        
        if (!dir.exists(output_folder)){
            dir.create(output_folder,recursive=TRUE)
            message("Output directory is not provided. Created directory: ", output_folder)
        }
        
        log_and_stats_folder <- paste0(output_folder, "/scPipe_atac_stats/")
        type                 <- "dna"
        
    } else if(tech == "rna") {
        message("RNA-Seq mode is selected...")
        
        if(is.null(output_folder)) {
            stop("output_folder cannot be NULL for rna mode. Aborting...\n")
        }
        log_and_stats_folder <- output_folder
        type                 <- "rna"  
    } #else

    dir.create(log_and_stats_folder, showWarnings = FALSE)
    log_file             <- paste0(log_and_stats_folder, "log_file.txt")
    stats_file           <- paste0(log_and_stats_folder, "stats_file_align.txt")
    if(!file.exists(log_file)) file.create(log_file)

    write(
        c(
        "sc_aligning starts at ",
        as.character(Sys.time()),
        "\n"
        ), 
        file = log_file, append = TRUE
    )


    # creating an index if not available
    if (is.null(index_path) && is.null(ref)) {
        stop("either a subread index path or a reference.fa path needs to be added \n")
    } else {
        if (!is.null(index_path)) {
            indexPath <- index_path
            if (!file.exists(paste0(indexPath, ".log"))) {
                stop("Genome index does not exist in the specificed location. Please check the full index path again.\n")   
            }
        } else {
            message("Genome index location not specified. Looking for the index in", output_folder)
            indexPath <-  file.path(output_folder, "genome_index") 
            if (file.exists(paste0(indexPath, ".log"))) {
                message("Genome index found in ", output_folder, "...")
            } else {
                message("Genome index not found. Creating one in ", output_folder, " ...")
                if(file.exists(ref)){
                    Rsubread::buildindex(basename=indexPath, reference=ref)
                } else {
                    stop("reference file does not exist. Please check the path and retry. \n")
                }
                
            }
        }
    }

    # Check for partial/nomatch files
    if(tech == "atac") {
        containing_folder <- dirname(R1) # Assume partial and nomatch files are also in the same directory as supplied input fastq files
        input_folder_files <- list.files(containing_folder)

        # Initialise demultiplexing stats
        barcode_completematch_count <- length(readLines(R1))/2
        demux_stats <- data.frame(status = c("barcode_completematch_count"),
                                count = c(barcode_completematch_count))
        # Concatenate the complete and partial matches

        partial_matches_R1 <- file.path(containing_folder, input_folder_files[grep("dem.+partialmatch.+R1.+fastq", input_folder_files)])
        partial_matches_R3 <- file.path(containing_folder, input_folder_files[grep("dem.+partialmatch.+R3.+fastq", input_folder_files)])

        if (all(file.exists(partial_matches_R1, partial_matches_R3)) && !identical(partial_matches_R1, character(0)) && !identical(partial_matches_R3, character(0))) {
            if (length(readLines(partial_matches_R1)) > 0 && length(readLines(partial_matches_R3)) > 0) {
                message("Found partial match fastq files, proceeding to concatenate with complete match fastq files.")
                barcode_partialmatch_count <- length(readLines(partial_matches_R1))/2
                demux_stats <- demux_stats %>% tibble::add_row(status = "barcode_partialmatch_count", count = barcode_partialmatch_count)
                concat_filename_R1 <- paste0("demultiplexed_complete_partialmatch_", stringr::str_remove(basename(R1), stringr::regex("dem.+completematch_")))
                concat_file_R1 <- file.path(containing_folder, concat_filename_R1)
                concat_filename_R3 <- paste0("demultiplexed_complete_partialmatch_", stringr::str_remove(basename(R2), stringr::regex("dem.+completematch_")))
                concat_file_R3 <- file.path(containing_folder, concat_filename_R3)
                system2("zcat", c(R1, partial_matches_R1, "|", "gzip", "-c", ">", concat_file_R1))
                system2("zcat", c(R2, partial_matches_R3, "|", "gzip", "-c", ">", concat_file_R3))

                if (!all(file.exists(concat_file_R1, concat_file_R3))) {
                    stop("Couldn't concatenate files!\n")
                }

                message("Output concatenated read files to:")
                message("R1:", concat_file_R1)
                message("R3:", concat_file_R3)

                # Replace original fastq files with concatenated files for aligning
                R1 <- concat_file_R1
                R2 <- concat_file_R3

            } else {
                message("No partial matches, checking for reads with non-matched barcodes.")
            }

        }
        # ------------ Align the nomatch file -------

        no_matches_R1 <- file.path(containing_folder, input_folder_files[grep("nomatch.+R1.+fastq", input_folder_files)])
        no_matches_R3 <- file.path(containing_folder, input_folder_files[grep("nomatch.+R3.+fastq", input_folder_files)])
        if (all(file.exists(no_matches_R1, no_matches_R3)) && !identical(no_matches_R1, character(0)) && !identical(no_matches_R3, character(0))) {
            if (length(readLines(no_matches_R1)) > 0 && length(readLines(no_matches_R3)) > 0) {
                message("Found barcode non-matched demultiplexed FASTQ files. Proceeding to align them.")

                fileNameWithoutExtension <- paste0(output_folder, "/", strsplit(basename(no_matches_R1), "\\.")[[1]][1])
                nomatch_bam <- paste0(fileNameWithoutExtension, "_aligned.bam")
                Rsubread::align(
                index = file.path(output_folder, "genome_index"),
                readfile1 = no_matches_R1,
                readfile2 = no_matches_R3,
                sortReadsByCoordinates = TRUE,
                type = "DNA",
                nthreads = 12,
                output_file = nomatch_bam)

                # Extract columns
                bam_tags <-list(bc="CB", mb="OX")
                param <- Rsamtools::ScanBamParam(tag = as.character(bam_tags),  mapqFilter=20)
                bamfl <- open(Rsamtools::BamFile(nomatch_bam))
                params <- Rsamtools::ScanBamParam(what=c("flag"), tag=c("CB"))
                bam0 <- Rsamtools::scanBam(bamfl, param = params)

                flag_defs <- tibble::tibble(
                type =
                    paste0("barcode_unmatch_", c("one_read_unmapped", "one_read_unmapped", "one_read_unmapped", "one_read_unmapped", "one_read_unmapped", "one_read_unmapped", "one_read_unmapped", "one_read_unmapped", "one_read_unmapped", "one_read_unmapped", "one_read_unmapped", "one_read_unmapped", "both_reads_unmapped", "both_reads_unmapped", "mapped", "mapped", "mapped", "mapped", "mapped_wrong_orientation", "mapped_wrong_orientation", "mapped_wrong_orientation", "mapped_wrong_orientation", "mapped_ambigously", "mapped_ambigously", "mapped_ambigously", "mapped_ambigously", "mapped_ambigously", "mapped_ambigously", "mapped_ambigously", "mapped_ambigously"))
                ,
                flag =
                    c(73, 133, 89, 121, 165, 181, 101, 117, 153, 185, 69, 137, 77, 141, 99, 147, 83, 163, 67, 131, 115, 179, 81, 161, 97, 145, 65, 129, 113, 177))

                # Create stats data frame
                demux_stats <- rbind(demux_stats, as.data.frame(table((data.frame(flag = bam0[[1]]$flag) %>% dplyr::left_join(flag_defs, by = "flag"))[,c('type')])) %>%
                                    dplyr::rename(status = Var1, count = Freq))

            } else {
                message("No reads found with non-matching barcodes.")
            }
        }
        utils::write.csv(demux_stats, file.path(log_and_stats_folder, "demultiplexing_stats.csv"), row.names = FALSE)
        message("Outputted demultiplexing stats file to", file.path(log_and_stats_folder, "demultiplexing_stats.csv"), "\n")

    } 

    # Generate the output filename
    if (is.null(output_file)) {
        # Only exception is if filename (excluding directory name) contains '.' then will only extract the first part
        fileNameWithoutExtension <- paste0(output_folder, "/", strsplit(basename(R1), "\\.")[[1]][1])
        outbam                   <- paste0(fileNameWithoutExtension, "_aligned.bam")
        message("Output file name is not provided. Aligned reads are saved in ", outbam)
    }
    else {
        fileNameWithoutExtension <- paste(output_folder, strsplit(output_file, "\\.")[[1]][1], sep = "/")
        outbam                   <- paste0(output_folder, "/", output_file)
    }
    
    #execute Rsubread align()
    
    if(!is.null(R2) && file.exists(R2)){       # paired-end
        align_output_df <- Rsubread::align(
        index       = indexPath,
        readfile1          = R1,
        readfile2          = R2,
        sortReadsByCoordinates = TRUE,
        type = type,
        nthreads = nthreads,
        output_file = outbam)
    } else {                                   # single-end
        align_output_df <- Rsubread::align(
        index       = indexPath,
        readfile1          = R1,
        sortReadsByCoordinates = TRUE,
        type = type,
        nthreads = nthreads,
        output_file = outbam)
        
    }
    
    utils::write.csv(align_output_df, file = stats_file, row.names = TRUE, quote = FALSE)
    
    #generating the bam index
    Rsamtools::indexBam(paste0(fileNameWithoutExtension, "_aligned.bam"))
    
    # get the unmapped mapped stats to be output and stored in a log file
    bamstats <- Rsamtools::idxstatsBam(paste0(fileNameWithoutExtension, "_aligned.bam"))
    utils::write.csv(bamstats, file = paste0(log_and_stats_folder, "stats_file_align_per_chrom.csv"), row.names = FALSE, quote = FALSE)

    write(
        c(
        "sc_aligning finishes at ",
        as.character(Sys.time()),
        "\n\n"
        ), 
        file = log_file, append = TRUE)
    
    return(outbam)
}

###########################################################
# sub-functions for report plot generation for scATAC-Seq data
###########################################################

#' @name sc_atac_plot_fragments_per_cell
#' @title A histogram of the log-number of fragments per cell
#'
#' @param sce The SingleExperimentObject produced by the sc_atac_create_sce function at the end of the pipeline
#'
#' @return returns NULL
#' @export
#'
sc_atac_plot_fragments_per_cell <- function(sce) {
    cell_stats <- as.data.frame(QC_metrics(sce))
    cell_stats$log_counts_per_cell <- log(cell_stats$counts_per_cell+1)
    
    ggplot2::ggplot(cell_stats, ggplot2::aes(x=log_counts_per_cell, y = ..count..)) +
        ggplot2::geom_histogram(color = "#E6AB02", fill = "#E6AB02", bins = 10) +
        ggplot2::stat_density(geom = "line", color = "#E6AB02") +
        ggplot2::ggtitle("Counts per cell") +
        ggplot2::xlab("log_counts_per_cell") + 
        ggplot2::ylab("count") 
}

#' @name sc_atac_plot_fragments_per_feature
#' @title A histogram of the log-number of fragments per feature
#'
#' @param sce The SingleExperimentObject produced by the sc_atac_create_sce function at the end of the pipeline
#'
#' @return returns NULL
#' @export
sc_atac_plot_fragments_per_feature <- function(sce) {
    feature_stats <- as.data.frame(feature_info(sce))
    feature_stats$log_counts_per_feature <- log(feature_stats$counts_per_feature+1)
    
    ggplot2::ggplot(feature_stats, ggplot2::aes(x=log_counts_per_feature, y = ..count..)) +
        ggplot2::geom_histogram(color = "#E6AB02", fill = "#E6AB02", bins = 10) +
        ggplot2::stat_density(geom = "line", color = "#E6AB02") +
        ggplot2::ggtitle("Counts per feature") +
        ggplot2::xlab("log_counts_per_feature") + 
        ggplot2::ylab("count") 
}


#' @name sc_atac_plot_features_per_cell
#' @title A histogram of the log-number of features per cell
#'
#' @param sce The SingleExperimentObject produced by the sc_atac_create_sce function at the end of the pipeline
#'
#' @return returns NULL
#' @export
#'
sc_atac_plot_features_per_cell <- function(sce) {
    cell_stats <- as.data.frame(QC_metrics(sce))
    cell_stats$log_features_per_cell <- log(cell_stats$features_per_cell+1)
    
    ggplot2::ggplot(cell_stats, ggplot2::aes(x=log_features_per_cell, y = ..count..)) +
        ggplot2::geom_histogram(color = "#7570B3", fill = "#7570B3", bins = 10) +
        ggplot2::stat_density(geom = "line", color = "#7570B3") +
        ggplot2::ggtitle("Features per cell") +
        ggplot2::xlab("log_features_per_cell") + 
        ggplot2::ylab("count") 
}

#' @name sc_atac_plot_features_per_cell_ordered
#' @title Plot showing the number of features per cell in ascending order
#'
#' @param sce The SingleExperimentObject produced by the sc_atac_create_sce function at the end of the pipeline
#'
#' @return returns NULL
#' @export
#'
sc_atac_plot_features_per_cell_ordered <- function(sce) {
    cell_stats <- QC_metrics(sce)
    plot(sort(cell_stats$features_per_cell), 
        xlab= 'cell', 
        log= 'y', 
        ylab = "features", 
        main= 'features per cell (ordered)',
        col = "#1B9E77")
}

#' @name sc_atac_plot_cells_per_feature
#' @title A histogram of the log-number of cells per feature
#'
#' @param sce The SingleExperimentObject produced by the sc_atac_create_sce function at the end of the pipeline
#'
#' @return returns NULL
#' @export
#'
sc_atac_plot_cells_per_feature <- function(sce) {
    feature_stats <- as.data.frame(feature_info(sce))
    feature_stats$log_cells_per_feature <- log(feature_stats$cells_per_feature+1)
    
   ggplot2::ggplot(feature_stats, ggplot2::aes(x=log_cells_per_feature, y = ..count..)) +
        ggplot2::geom_histogram(color = "#7570B3", fill = "#7570B3", bins = 10) +
        ggplot2::stat_density(geom = "line", color = "#7570B3") +
        ggplot2::ggtitle("Cells per feature") +
        ggplot2::xlab("log_cells_per_feature") + 
        ggplot2::ylab("count") 
    
}

#' @name sc_atac_plot_fragments_features_per_cell
#' @title A scatter plot of the log-number of fragments and log-number of features per cell
#'
#' @param sce The SingleExperimentObject produced by the sc_atac_create_sce function at the end of the pipeline
#'
#' @return returns NULL
#' @export
#'
sc_atac_plot_fragments_features_per_cell <- function(sce) {
    cell_stats <- as.data.frame(QC_metrics(sce))
    cell_stats$log_counts_per_cell <- log(cell_stats$counts_per_cell+1)
    cell_stats$log_features_per_cell <- log(cell_stats$features_per_cell+1)
    
    ggplot2::ggplot(cell_stats, ggplot2::aes(x=log_counts_per_cell, y=log_features_per_cell)) +
        ggplot2::geom_point(color = "#E6AB02") +
        ggplot2::xlab("log_counts_per_cell") + 
        ggplot2::ylab("log_features_per_cell") +
        ggplot2::geom_smooth(formula = y ~ x, method='lm', color = "#E6AB02", fill = "#E6AB02")
}

#' @name sc_atac_plot_fragments_cells_per_feature
#' @title A scatter plot of the log-number of fragments and log-number of cells per feature
#'
#' @param sce The SingleExperimentObject produced by the sc_atac_create_sce function at the end of the pipeline
#'
#' @return returns NULL
#' @export
#'
sc_atac_plot_fragments_cells_per_feature <- function(sce) {
    feature_stats <- as.data.frame(feature_info(sce))
    feature_stats$log_counts_per_feature <- log(feature_stats$counts_per_feature+1)
    feature_stats$log_cells_per_feature <- log(feature_stats$cells_per_feature+1)
    
    ggplot2::ggplot(feature_stats, ggplot2::aes(x=log_counts_per_feature, y=log_cells_per_feature)) +
        ggplot2::geom_point(color = "#7570B3") +
        ggplot2::xlab("log_counts_per_feature") + 
        ggplot2::ylab("log_cells_per_feature") +
        ggplot2::geom_smooth(formula = y ~ x, method='lm', color = "#7570B3", fill = "#7570B3")
}

#######################################
# Cell Calling on a matrix (scATAC-Seq)
#######################################

#' @name sc_atac_cell_calling
#' @title identifying true vs empty cells
#' @description the methods to call true cells are of various ways.
#' implement (i.e. \code{filtering} from \code{scATAC-Pro} as default
#' @param mat the feature by cell matrix.
#' @param cell_calling the cell calling approach, possible options were "emptydrops" , "cellranger" and "filter".
#' But we opten to using "filter" as it was most robust. "emptydrops" is still an opition for data with large umber of cells.
#' @param output_folder output directory for the cell called matrix.
#' @param genome_size genome size for the data in feature by cell matrix.
#' @param cell_qc_metrics_file quality per barcode file for the barcodes in the matrix if using the \code{cellranger} or \code{filter} options.
#'
#' @param lower the lower threshold for the data if using the \code{emptydrops} function for cell calling.
#'
#' @param min_uniq_frags The minimum number of required unique fragments required for a cell (used for \code{filter} cell calling)
#' @param max_uniq_frags The maximum number of required unique fragments required for a cell (used for \code{filter} cell calling)
#' @param min_frac_peak The minimum proportion of fragments in a cell to overlap with a peak (used for \code{filter} cell calling)
#' @param min_frac_tss The minimum proportion of fragments in a cell to overlap with a tss (used for \code{filter} cell calling)
#' @param min_frac_enhancer The minimum proportion of fragments in a cell to overlap with a enhancer sequence (used for \code{filter} cell calling)
#' @param min_frac_promoter The minimum proportion of fragments in a cell to overlap with a promoter sequence (used for \code{filter} cell calling)
#' @param max_frac_mito The maximum proportion of fragments in a cell that are mitochondrial (used for \code{filter} cell calling)
#'
#' @importFrom utils write.table
#' @examples
#' \dontrun{
#' sc_atac_cell_calling <- function(mat,
#'  cell_calling,
#'  output_folder,
#'  genome_size    = NULL,
#'  cell_qc_metrics_file = NULL,
#'  lower          = NULL)
#' }
#'
#'
#' @export
#'
sc_atac_cell_calling <- function(mat,
                                cell_calling = 'filter',
                                output_folder,
                                genome_size    = NULL,
                                cell_qc_metrics_file = NULL,
                                lower          = NULL,
                                min_uniq_frags = 3000,
                                max_uniq_frags = 50000,
                                min_frac_peak = 0.3,
                                min_frac_tss = 0,
                                min_frac_enhancer = 0,
                                min_frac_promoter = 0.1,
                                max_frac_mito = 0.15){

    message("`", cell_calling, "` function is used for cell calling ... ")

   selected_cells <- tryCatch({
        if(cell_calling == "emptydrops") {
          sc_atac_emptydrops_cell_calling(mat = mat, output_folder = output_folder, lower = lower)
        }
   #     else if(cell_calling == 'cellranger') {
   #       sc_atac_cellranger_cell_calling(mat = mat, genome_size = genome_size, cell_qc_metrics_file = cell_qc_metrics_file)
   #     }
    }, error = function(e) {
       message(e)
       message("\nWill now default to the filter method with the default qc cutoffs.")
       message("min_uniq_frags = ", min_uniq_frags)
       message("max_uniq_frags = ", max_uniq_frags)
      message("min_frac_peak = ", min_frac_peak)
       message("min_frac_tss = ", min_frac_tss)
       message("min_frac_enhancer = ", min_frac_enhancer)
       message("min_frac_promoter = ", min_frac_promoter)
       message("max_frac_mito = ", max_frac_mito)
       message("Running the filter method...")
         sc_atac_filter_cell_calling(mtx = mat,
                                   cell_qc_metrics_file = cell_qc_metrics_file,
                                   min_uniq_frags = min_uniq_frags,
                                   max_uniq_frags = max_uniq_frags,
                                   min_frac_peak = min_frac_peak,
                                   min_frac_tss = min_frac_tss,
                                   min_frac_enhancer = min_frac_enhancer,
                                   min_frac_promoter = min_frac_promoter,
                                   max_frac_mito = max_frac_mito)
   })

    if(cell_calling == 'filter') {
        selected_cells <- sc_atac_filter_cell_calling(mtx = mat,
                                                    cell_qc_metrics_file = cell_qc_metrics_file,
                                                    min_uniq_frags = min_uniq_frags,
                                                    max_uniq_frags = max_uniq_frags,
                                                    min_frac_peak = min_frac_peak,
                                                    min_frac_tss = min_frac_tss,
                                                    min_frac_enhancer = min_frac_enhancer,
                                                    min_frac_promoter = min_frac_promoter,
                                                    max_frac_mito = max_frac_mito)
    }
    else if(isFALSE(cell_calling)) {
        message("No cell calling method was selected.")
    }
    else if (!cell_calling %in% c("emptydrops", "filter")) { # no legitimate cell calling method chosen, so just return the original matrix
        message(cell_calling, "was not an implemented cell calling method")
    }
    out_mat <- mat
    barcodes     <- colnames(out_mat)
    features     <- rownames(out_mat)
    if (cell_calling %in% c("emptydrops", "filter")) {
        # Only keep selected cells in matrix
        out_mat      <- mat[, colnames(mat) %in% selected_cells]
        barcodes     <- colnames(out_mat)
        features     <- rownames(out_mat)

        # cat("Number of called barcodes: ")
        # cat(length(selected_cells))
        #
        # cat("\nNumber of columns: ")
        # cat(length(barcodes))
        # cat("\n")
        if (length(barcodes) == 0) {
            stop("No cells were called...")
        }
    }

    # Store output matrix
    # Matrix::writeMM(Matrix::Matrix(out_mat), file =file.path(output_folder, 'cell_called_matrix.mtx'))
    message("cell called and stored in ", output_folder)
    utils::write.table(barcodes, file = file.path(output_folder, 'non_empty_barcodes.txt'), sep = '\t',
                row.names = FALSE, quote = FALSE, col.names = FALSE)
    utils::write.table(features, file = file.path(output_folder, 'non_empty_features.txt'), sep = '\t',
                row.names = FALSE, quote = FALSE, col.names = FALSE)


    return(out_mat)
}


# ############## Modified testEmptyDrops function ######################
# testEmptyDrops2 <- function (m, lower = 100, niters = 10000, test.ambient = FALSE,
#                             ignore = NULL, alpha = NULL, BPPARAM = SerialParam())
# {
#   discard      <- rowSums(m) == 0
#   m            <- m[!discard, , drop = FALSE]
#   ncells       <- ncol(m)
#   umi.sum      <- as.integer(round(colSums(m)))
#   ambient      <- umi.sum <= lower
#   ambient.cells<- m[, ambient, drop = FALSE]
#   ambient.prof <- rowSums(ambient.cells)
#
#   if (sum(ambient.prof) == 0) {
#     stop("no counts available to estimate the ambient profile")
#   }
#   ambient.prop <- edgeR::goodTuringProportions(ambient.prof)
#   if (!test.ambient) {
#     keep <- !ambient
#   } else {
#     keep <- umi.sum > 0L
#   }
#   if (!is.null(ignore)) {
#     keep <- keep & umi.sum > ignore
#   }
#   obs <- m[, keep, drop = FALSE]
#   obs.totals <- umi.sum[keep]
#   if (is.null(alpha)) {
#     alpha <- DropletUtils:::.estimate_alpha(m[, ambient, drop = FALSE],
#                                             ambient.prop, umi.sum[ambient])
#   }
#   obs.P <-DropletUtils:::.compute_multinom_prob_data(obs, ambient.prop, alpha = alpha)
#   rest.P <-DropletUtils:::.compute_multinom_prob_rest(obs.totals, alpha = alpha)
#   n.above <-DropletUtils:::.permute_counter(totals = obs.totals, probs = obs.P,
#                                             ambient = ambient.prop, iter = niters, BPPARAM = BPPARAM,
#                                             alpha = alpha)
#   limited <- n.above == 0L
#   pval <- (n.above + 1)/(niters + 1)
#   all.p <- all.lr <- all.exp <- rep(NA_real_, ncells)
#   all.lim <- rep(NA, ncells)
#   all.p[keep] <- pval
#   all.lr[keep] <- obs.P + rest.P
#   all.lim[keep] <- limited
#   output <- DataFrame(Total = umi.sum, LogProb = all.lr, PValue = all.p,
#                       Limited = all.lim, row.names = colnames(m))
#   metadata(output) <- list(lower = lower, niters = niters,
#                            ambient = ambient.prop, alpha = alpha)
#   output
# }
#
# ############## Rounded to integer function #####################
# .rounded_to_integer <- function(m, round=TRUE) {
#   if (round) {
#     m <- round(m)
#   }
#   m
# }
#
# ############## Modified emptyDrops function ######################
# emptyDrops2 <- function (m, lower = 100, retain = -1, barcode.args = list(),
#                         ...)
# {
#   m <- .rounded_to_integer(m)
#   stats <- testEmptyDrops2(m, lower = lower, ...)
#   tmp <- stats$PValue
#   if (is.null(retain)) {
#     br.out <- do.call(barcodeRanks, c(list(m, lower = lower),
#                                       barcode.args))
#     retain <- metadata(br.out)$knee
#   }
#   always <- stats$Total >= retain
#   tmp[always] <- 0
#   metadata(stats)$retain <- retain
#   stats$FDR <- p.adjust(tmp, method = "BH")
#   return(stats)
# }

#' @name sc_atac_filter_cell_calling
#' @title filter cell calling
#' @description specify various qc cutoffs to select the desired cells
#'
#' @param mtx The input matrix
#' @param cell_qc_metrics_file A file containing qc statistics for each cell
#' @param min_uniq_frags The minimum number of required unique fragments required for a cell
#' @param max_uniq_frags The maximum number of required unique fragments required for a cell
#' @param min_frac_peak The minimum proportion of fragments in a cell to overlap with a peak
#' @param min_frac_tss The minimum proportion of fragments in a cell to overlap with a tss
#' @param min_frac_enhancer The minimum proportion of fragments in a cell to overlap with a enhancer sequence
#' @param min_frac_promoter The minimum proportion of fragments in a cell to overlap with a promoter sequence
#' @param max_frac_mito The maximum proportion of fragments in a cell that are mitochondrial
#'
#'
#' @export
#'
sc_atac_filter_cell_calling <- function(
    mtx,
    cell_qc_metrics_file,
    min_uniq_frags = 0,
    max_uniq_frags = 50000,
    min_frac_peak = 0.05,
    min_frac_tss = 0,
    min_frac_enhancer = 0,
    min_frac_promoter = 0,
    max_frac_mito = 0.2) {

    # https://github.com/wbaopaul/scATAC-pro/blob/master/scripts/src/filter_barcodes.R

    total_frags <- frac_mito <- frac_peak <- frac_tss <- frac_promoter <- frac_enhancer <- NULL

    qc_bc_stat <- data.table::fread(cell_qc_metrics_file)

    qc_sele <- qc_bc_stat[total_frags >= min_uniq_frags & total_frags <= max_uniq_frags &
                frac_mito <= max_frac_mito &
                frac_peak >= min_frac_peak &
                frac_tss >= min_frac_tss &
                frac_promoter >= min_frac_promoter &
                frac_enhancer >= min_frac_enhancer]

    return(qc_sele$bc)
}


#' @name sc_atac_emptydrops_cell_calling
#' @title empty drops cell calling
#' @description The empty drops cell calling method
#'
#' @param mat The input matrix
#' @param output_folder The path of the output folder
#' @param lower The lower threshold for the data if using the \code{emptydrops} function for cell calling.
#' @export
#'
sc_atac_emptydrops_cell_calling <- function(
    mat,
    output_folder,
    lower = NULL) {

    #set.seed(2019)

    # generating the knee plot
    my.counts <- Matrix::Matrix(mat)
    br.out    <- DropletUtils::barcodeRanks(my.counts)

    # Making a plot
    while (!is.null(grDevices::dev.list()))  grDevices::dev.off()
    grDevices::png(file=paste0(output_folder, "/scPipe_atac_stats/knee_plot.png"))
    plot(br.out$rank, br.out$total, log="xy", xlab="Rank", ylab="Total")
    o <- order(br.out$rank)
    graphics::lines(br.out$rank[o], br.out$fitted[o], col="red")

    graphics::abline(h=S4Vectors::metadata(br.out)$knee, col="dodgerblue", lty=2)
    graphics::abline(h=S4Vectors::metadata(br.out)$inflection, col="forestgreen", lty=2)
    graphics::legend("bottomleft", lty=2, col=c("dodgerblue", "forestgreen"),
            legend=c("knee", "inflection"))
    grDevices::dev.off()

    if(is.null(lower)){
        #lower <- floor(0.1*ncol(mat))
        lower <- 1
    }
    cell.out <- DropletUtils::emptyDrops(mat, lower = lower)
    # cell.out <- emptyDrops2(mat, lower = lower)

    filter.out <- cell.out[stats::complete.cases(cell.out), ]

    #saveRDS(filter.out, file = paste0(output_folder, '/EmptyDrop_obj.rds'))
    #cat("Empty cases are removed and saved in ", output_folder, "\n")

    if(length(filter.out$FDR) > 0){
        fdr <- 0.01
        message("FDR of 0.01 is assigned...")
        filter.out <- filter.out[filter.out$FDR <= fdr, ]
    } else{
        message("insufficient unique points for computing knee/inflection points ... Use the output matrices with caution!")
    }

    select.cells <- rownames(filter.out)

    return(select.cells)
}

#####################################################################
# Generating Fragments for Aligned and Demultiplexed scATAC-Seq Reads
#####################################################################
#' @name sc_atac_create_fragments
#' @title Generating the popular fragments for scATAC-Seq data
#' @description Takes in a tagged and sorted BAM file and outputs the associated fragments in a .bed file
#'
#' @param inbam The tagged, sorted and duplicate-free input BAM file
#' @param output_folder The path of the output folder
#' @param min_mapq : int
#'    Minimum MAPQ to retain fragment
#' @param nproc : int, optional
#'    Number of processors to use. Default is 1.
#' @param cellbarcode : str
#'   Tag used for cell barcode. Default is CB (used by cellranger)
#' @param chromosomes : str, optional
#'    Regular expression used to match chromosome names to include in the
#'    output file. Default is "(?i)^chr" (starts with "chr", case-insensitive).
#'    If None, use all chromosomes in the BAM file.
#' @param readname_barcode : str, optional
#'    Regular expression used to match cell barocde stored in read name.
#'    If None (default), use read tags instead. Use "[^:]*" to match all characters
#'    before the first colon (":").
#' @param cells : str
#'    File containing list of cell barcodes to retain. If None (default), use all cell barcodes
#'    found in the BAM file.
#' @param max_distance : int, optional
#'    Maximum distance between integration sites for the fragment to be retained.
#'    Allows filtering of implausible fragments that likely result from incorrect
#'    mapping positions. Default is 5000 bp.
#' @param min_distance : int, optional
#'    Minimum distance between integration sites for the fragment to be retained.
#'    Allows filtering implausible fragments that likely result from incorrect
#'    mapping positions. Default is 10 bp.
#' @param chunksize : int
#'    Number of BAM entries to read through before collapsing and writing
#'    fragments to disk. Higher chunksize will use more memory but will be
#'    faster.
#'
#' @returns returns NULL
#'
#' @export
sc_atac_create_fragments <- function(
        inbam,
        output_folder="",
        min_mapq=30,
        nproc=1,
        cellbarcode="CB",
        chromosomes="^chr",
        readname_barcode=NULL,
        cells=NULL,
        max_distance=5000,
        min_distance=10,
        chunksize=500000) {
    if(output_folder == ''){
        output_folder <- file.path(getwd(), "scPipe-atac-output")
    }

    if (!dir.exists(output_folder)){
        dir.create(output_folder,recursive=TRUE)
        message("Output Directory Does Not Exist. Created Directory: ", output_folder)
    }

    log_and_stats_folder <- file.path(output_folder, "scPipe_atac_stats")
    dir.create(log_and_stats_folder, showWarnings = FALSE)
    log_file <- file.path(log_and_stats_folder, "log_file.txt")
    if(!file.exists(log_file)) file.create(log_file)
    write(
    c(
        "sc_atac_create_fragments starts at ",
        as.character(Sys.time()),
        "\n"
    ),
    file = log_file, append = TRUE)

    output = paste0(output_folder, "/fragments.bed")
    if (file.exists(output)) {
        message("Fragment output file already exists: ", output)
        return(invisible())
    }
    if (file.exists(paste0(output_folder, "tmpDebug0"))) file.remove(paste0(output_folder, "tmpDebug0"))
    
    chrom <- get_chromosomes(inbam, keep_contigs=chromosomes) # List with names as contigs and elements as lengths
    cells <- read_cells(cells) # character vector / StringVector

    sc_atac_create_fragments_cpp(inbam,
                                output_folder,
                                names(chrom),
                                as.integer(chrom),
                                min_mapq,
                                nproc,
                                cellbarcode,
                                chromosomes,
                                readname_barcode,
                                cells,
                                max_distance,
                                min_distance,
                                chunksize)

    invisible()
}

#' Get Chromosomes
#'
#' Gets a list of NamedList of chromosomes and the reference length
#' acquired through the bam index file.
#'
#' @param bam file path to the bam file to get data from
#' @param keep_contigs regular expression used with grepl to filter reference names
#'
#' @return a named list where element names are chromosomes reference names and elements are integer lengths
get_chromosomes <- function(bam, keep_contigs="^chr") {
    if (is.null(keep_contigs)) {
        keep_contigs <- "."
    }
    if (!file.exists(paste0(bam, ".bai"))) {
        Rsamtools::indexBam(bam)
    }

    idxstats <- Rsamtools::idxstatsBam(bam)
    # regex = (?i)^chr
    # above matches chr at start of string ignoring case ((?i) means ignore case)
    # R version: grepl("^chr", ignore.case=TRUE)
    contigs <- grepl(keep_contigs, idxstats$seqnames, ignore.case=TRUE)
    mapped <- idxstats$mapped > 0
    keep_contigs <- idxstats[contigs & mapped, ]
    conlen <- stats::setNames(as.list(keep_contigs$seqlength), keep_contigs$seqnames)

    return(conlen)
}

#' Read Cell barcode file
#'
#' @param cells the file path to the barcode file. Assumes
#' one barcode per line or barcode csv.
#' Or, cells can be a comma delimited string of barcodes
#' @return a character vector of the provided barcodes
#' @importFrom data.table fread
read_cells <- function(cells) {
    if (is.null(cells)) return(NULL)

    if (file.exists(cells)) {
        barcodes <- data.table::fread(cells, header=FALSE, data.table=FALSE)[[1]]
    } else {
        barcodes <- strsplit(cells, ",")[[1]]
    }

    return (barcodes)

}

###########################################################
# TF-IDF normalisation and UMAP for scATAC-Seq data
###########################################################

#' @name sc_atac_tfidf
#'
#' @title generating the UMAPs for sc-ATAC-Seq preprocessed data
#' @description Takes the binary matrix and generate a TF-IDF so the clutering can take place on the reduced dimentions. 
#' 
#' @param binary.mat The final, filtered feature matrix in binary format 
#' @param output_folder The path of the output folder 
#'
#' @returns None (invisible `NULL`)
#' @examples
#' \dontrun{
#' sc_atac_tfidf(binary.mat = final_binary_matrix) 
#' }
#' 
#' @export
#' 
sc_atac_tfidf <- function(binary.mat, output_folder = NULL) {

    # Check if output directory exists
    if (is.null(output_folder)) {
        output_folder <- file.path(getwd(), "scPipe-atac-output")
    }
    
    if (!dir.exists(output_folder)){
        dir.create(output_folder,recursive=TRUE)
        message("Output directory does not exist. Created directory: ", output_folder)
    }
    
    #Create log folder/file
    log_and_stats_folder <- file.path(output_folder, "scPipe_atac_stats")
    dir.create(log_and_stats_folder, showWarnings = FALSE)
    log_file            <- file.path(log_and_stats_folder, "log_file.txt")
    if(!file.exists(log_file)) file.create(log_file)
    write(c(
        "sc_atac_tfidf starts at ",
        as.character(Sys.time()),
        "\n"
        ),
        file = log_file, append = TRUE)
    
        
#   TF.IDF.custom <- function(binary.mat, verbose = TRUE) {
#     if (is(binary.mat, "data.frame")) {
#       binary.mat <- as.matrix(x = binary.mat)
#     }
#     if (!is(binary.mat, "dgCMatrix")) {
#       binary.mat <- as(object = binary.mat, Class = "dgCMatrix")
#     }
#     if (verbose) {
#       message("Performing TF-IDF normalization")
#     }
    
#     npeaks       <- Matrix::colSums(x = object)
#     tf           <- Matrix::tcrossprod(x = as.matrix(object), y = Matrix::Diagonal(x = 1 / npeaks))
#     rsums        <- Matrix::rowSums(x = object)
#     idf          <- ncol(x = object) / rsums
#     norm.data    <- Matrix::Diagonal(n = length(x = idf), x = idf) %*% tf
#     scale.factor <- 1e4
#     methods::slot(object = norm.data, name = "x") <- log1p(x = methods::slot(object = norm.data, name = "x") * scale.factor)
#     norm.data[which(x = is.na(x = norm.data))] <- 0
#     return(norm.data)
#   }
    
    
    binary.mat <- TF.IDF.custom(binary.mat)
    
    saveRDS(binary.mat, file = file.path(output_folder, "tf-idf_mat.rds"))
}

#' @name sc_interactive_umap_plot
#' @title Produces an interactive UMAP plot via Shiny
#' @description Can colour the UMAP by any of the colData columns in the SCE object
#' @param sce The SingleCellExperiment object
#'
#' @returns A shiny object which represents the app. Printing the object or passing it to `shiny::runApp(...)` will run the app.
#' @export
sc_interactive_umap_plot <- function(sce) {
    umap_data <- sc_get_umap_data(sce)
    
    #   library(shiny)
    if(!requireNamespace("shiny", quietly=TRUE)) {
        stop("Install 'shiny' to use this function.")
    }
    if(!requireNamespace("plotly", quietly=TRUE)) {
        stop("Install 'plotly' to use this function.")
    }
    
    ui <- shiny::fluidPage(
        shiny::titlePanel("Interactive UMAP plot"),
        
        shiny::sidebarLayout(
            shiny::sidebarPanel( # INPUT
                
                shiny::selectInput("by", "Colour by",
                            colnames(umap_data)[!colnames(umap_data) %in% c("barcode", "UMAP1", "UMAP2")]),
            ),
            shiny::mainPanel( # OUTPUT      
                plotly::plotlyOutput(outputId = "umapPlot")
            )
        )
    )
    
    # Define server logic required to draw a histogram ----
    server <- function(input, output) {
        output$umapPlot <- plotly::renderPlotly({ 
            g <- ggplot2::ggplot(umap_data, ggplot2::aes(x = UMAP1, y = UMAP2, text = paste("barcode: ", barcode))) +
                ggplot2::geom_point(ggplot2::aes(col = .data[[input$by]]), size = 2, alpha = 0.5) +
                ggplot2::theme_bw(base_size = 14)
            if (is.numeric(umap_data[[input$by]])) {
                g <- ggplot2::ggplot(umap_data, ggplot2::aes(x = UMAP1, y = UMAP2, text = paste("barcode: ", barcode))) +
                ggplot2::geom_point(ggplot2::aes(col = .data[[input$by]]), size = 2, alpha = 0.5) +
                ggplot2::scale_colour_gradientn(colours = c("green", "red")) +
                ggplot2::theme_bw(base_size = 14)
            }
            plotly::ggplotly(g)
        })
    }
    
    shiny::shinyApp(ui = ui, server = server)
}


#' @name sc_get_umap_data
#' @title Generates UMAP data from sce object
#' @description Produces a DataFrame containing the UMAP dimensions, as well as all the colData of the sce object for each cell
#' @param sce The SingleCellExperiment object
#' @param n_neighbours No. of neighbours for KNN 
#'
#' @returns A dataframe containing the UMAP dimensions, as well as all the colData of the sce object for each cell
#' @export
sc_get_umap_data <- function(sce,
                                n_neighbours = 30) {
    #set.seed(123)
    
    if (!requireNamespace("RANN", quietly=TRUE)) {
        stop("Install 'RANN' to use this function")
    }
    if (!requireNamespace("igraph", quietly=TRUE)) {
        stop("Install 'igraph' to use this function")
    }

    counts <- assay(sce)
    bin_mat <- as.matrix((counts>0)+0)
    
    binary.mat <- TF.IDF.custom(bin_mat)
    n_bcs <- max(min(50, ncol(binary.mat), nrow(binary.mat))-1,0)
    mat.lsi          <- irlba(binary.mat, n_bcs)
    d_diagtsne       <- matrix(0, n_bcs, n_bcs)
    diag(d_diagtsne) <- mat.lsi$d
    mat_pcs          <- t(d_diagtsne %*% t(mat.lsi$v))
    rownames(mat_pcs)<- colnames(binary.mat)
    
    # clustering in the PCA space using KNN --------------
    knn.info<- RANN::nn2(mat_pcs, k = n_neighbours)
    
    ## convert to adjacency matrix
    knn           <- knn.info$nn.idx
    adj           <- matrix(0, nrow(mat_pcs), nrow(mat_pcs))
    rownames(adj) <- colnames(adj) <- rownames(mat_pcs)
    for(i in seq_len(nrow(mat_pcs))) {
        adj[i,rownames(mat_pcs)[knn[i,]]] <- 1
    }
  
    ## convert to graph
    g <- igraph::graph.adjacency(adj, mode="undirected")
    g <- purrr::simplify(g) ## remove self loops
    
    # identify communities, many algorithms. Use the Louvain clustering ------------
    km         <- igraph::cluster_louvain(g)
    com        <- km$membership
    names(com) <- km$names
    
    # running UMAP ------------------------------
    norm.data.umap    <- umap::umap(mat_pcs)
    
    df_umap           <- as.data.frame(norm.data.umap$layout)
    colnames(df_umap) <- c("UMAP1", "UMAP2")
    df_umap$barcode   <- rownames(mat_pcs)
    
    df_umap           <- dplyr::left_join(df_umap, tibble::enframe(com), by = c("barcode" = "name")) %>%
        dplyr::rename(cluster = value) %>%
        dplyr::mutate(cluster = as.factor(cluster))
    
    
    sce_coldata <- colData(sce)
    df_umap <- base::merge(df_umap, sce_coldata, by.x = "barcode", by.y = "row.names", all.x = TRUE) 
    
    
    return(df_umap)
}

###########################################################
# Create a SingleCell Experiment Object for scATAC-Seq data
###########################################################

#' sc_atac_create_sce()
#'
#' @param input_folder The output folder produced by the pipeline
#' @param organism The type of the organism
#' @param sample_name The name of the sample
#' @param feature_type The type of the feature
#' @param pheno_data The pheno data
#' @param report Whether or not a HTML report should be produced
#'
#' @returns a SingleCellExperiment object created from the scATAC-Seq data provided
#'
#' @importFrom utils read.csv
#' @importFrom tibble column_to_rownames
#' @importFrom SingleCellExperiment SingleCellExperiment
#' @examples
#' \dontrun{
#' sc_atac_create_sce(
#'    input_folder = input_folder,
#'    organism = "hg38",
#'    feature_type = "peak",
#'    report = TRUE)
#' }    
#' 
#' @export
#'
sc_atac_create_sce <- function(input_folder  = NULL, 
                                organism      = NULL, 
                                sample_name   = NULL,
                                feature_type  = NULL, 
                                pheno_data    = NULL, 
                                report        = FALSE) {
    
    # if(is.null(input_folder)){
    #   input_folder <- file.path(getwd(), "scPipe-atac-output")
    #   input_stats_folder <- file.path(getwd(), "scPipe-atac-output/scPipe_atac_stats")
    # } else {
    #   cat("input location ", input_folder, " is not valid. Please enter the full path to proceed. \n")
    #   break;
    # }
    
    if(is.null(input_folder)){
        input_folder <- file.path(getwd(), "scPipe-atac-output")
        input_stats_folder <- file.path(getwd(), "scPipe-atac-output/scPipe_atac_stats")
    }
    
    if (!dir.exists(input_folder)){
        message("Default input folder could not be found at " , input_folder,  "\nPlease enter the full input path to proceed");
    } else {
        input_stats_folder <- file.path(input_folder, "scPipe_atac_stats")
    }
    
    #feature_cnt   <- readMM(file.path(input_folder, "sparse_matrix.mtx"))
    feature_cnt   <- readRDS(file.path(input_folder, "sparse_matrix.rds"))
    cell_stats    <- utils::read.csv(file.path(input_stats_folder, "filtered_stats_per_cell.csv"), row.names=1)
    feature_stats <- utils::read.csv(file.path(input_stats_folder, "filtered_stats_per_feature.csv"))
    

    # need to change from here.... (check whether I need to filter before saving to the SCE object)
    
    # can I order a matrix like a csv file like below? test...
    feature_cnt      <- feature_cnt[, order(c(colnames(feature_cnt)))]
    
    
    qc <- utils::read.csv(file.path(input_folder, "cell_qc_metrics.csv")) %>% tibble::column_to_rownames(var = "bc")
    cell_stats <- merge(x = cell_stats, y = qc, by = 0, all.x = TRUE) %>% tibble::column_to_rownames(var = "Row.names")
    cell_stats       <- cell_stats[order(c(rownames(cell_stats))), ]
    

    # generating the SCE object
    sce                         <- SingleCellExperiment::SingleCellExperiment(assays = list(counts = feature_cnt))

    sce@metadata$scPipe$version <- packageVersion("scPipe")  # set version information
    
    if(!is.null(organism)){
        organism(sce) <- organism
    }
    
    if(!is.null(feature_type)){
        feature_type(sce) <- feature_type
    }
    
    # Saving demultiplexing stats to sce object
    stats_file <- file.path(input_stats_folder, "mapping_stats_per_barcode.csv")
    raw <- read.csv(stats_file, row.names = "barcode")
    
    
    QC_metrics(sce) <- cell_stats
    demultiplex_info(sce) <- raw
    
    
    
    if(!is.null(pheno_data)){
        colData(sce) <- cbind(colData(sce), pheno_data[order(rownames(pheno_data)),])
    }

    feature_info(sce) <- feature_stats
    saveRDS(sce, file = file.path(input_folder, "scPipe_atac_SCEobject.rds"))
    
    message("SCE object is saved in:\n\t", input_folder,"/scPipe_atac_SCEobject.rds")
    
    if(report){
        sc_atac_create_report(input_folder  = file.path(input_folder),
                            output_folder = file.path(input_folder, "scPipe_atac_stats"),
                            sample_name   = sample_name,
                            organism      = organism,
                            feature_type  = feature_type)
    }
    
    
    return(sce)
    
}

#' Import gene annotation
#'
#' Imports and GFF3 or GTF gene annotation file and transforms it into a SAF
#' formatted data.frame. SAF described at
#' http://bioinf.wehi.edu.au/featureCounts/. SAF contains positions for exons,
#' strand and the GeneID they are associated with.
#'
#'
#' @param filename The name of the annotation gff3 or gtf file. File can be
#'   gzipped.
#'
#' @description Because of the variations in data format depending on annotation
#'   source, this function has only been tested with human annotation from
#'   ENSEMBL, RefSeq and Gencode. If it behaves unexpectedly with any annotation
#'   please submit an issue at www.github.com/LuyiTian/scPipe with details.
#'
#' @return data.frame containing exon information in SAF format
#'
#' @importFrom magrittr %>%
#' @export
#'
#' @examples
#' ens_chrY <- anno_import(system.file("extdata", "ensembl_hg38_chrY.gtf.gz", package = "scPipe"))
#'
anno_import <- function(filename) {
    accepted_formats <- c("gff", "gff3", "gtf")
    file_format <- tools::file_ext(stringr::str_remove(filename, ".gz$"))
    all_valid_format <- any(file_format %in% accepted_formats)

    if (!all_valid_format) {
        stop("only files the following annotation formats are accepted: ", paste(accepted_formats, collapse = ", "), " and their gzipped variants")
    }

    anno <- lapply(filename, rtracklayer::import)
    anno <- lapply(anno, anno_to_saf)

    # gene_id column is present and contains necessary information
    # return SAF converted data.frame
    return(do.call(rbind, anno))
}

#' Convert annotation from GenomicRanges to Simple Annotation Format (SAF)
#'
#' Convert a GRanges object containing type and gene_id information into a SAF
#' format data.frame. SAF described at http://bioinf.wehi.edu.au/featureCounts/.
#' SAF contains positions for exons, strand and the GeneID they are associated
#' with.
#'
#' @param anno The GRanges object containing exon information
#'
#' @description This function converts a GRanges object into a data.frame of the
#'      SAF format for scPipe's consumption. The GRanges object should contain a
#'      "type" column where at least some features are annotated as "exon", in
#'      addition there should be a gene_id column specifying the gene to which the
#'      exon belongs. In the SAF only the gene ID, chromosome, start, end and
#'      strand are recorded, this is a gene-exon centric format, with all entries
#'      containing the same gene ID treated as exons of that gene. It is possible
#'      to count alternative features by setting the gene_id column to an arbitrary
#'      feature name and having alternative features in the SAF table, the main
#'      caveat is that the features are still treated as exons, and the mapping
#'      statistics for exon and intron will not reflect biological exons and
#'      introns but rather the annotation features.
#'
#' @return data.frame containing exon information in SAF format
#'
#' @importFrom GenomicRanges mcols
#' @importFrom magrittr %>%
#' @importFrom rlang .data
#' @export
#'
#' @examples
#'  \dontrun{
#'      anno <- system.file("extdata", "ensembl_hg38_chrY.gtf.gz", package = "scPipe")
#'      saf_chrY <- anno_to_saf(rtracklayer::import(anno))
#'  }
#'
anno_to_saf <- function(anno) {
    stopifnot(is(anno, "GRanges"))

    meta_cols <- colnames(GenomicRanges::mcols(anno))
    if (!"type" %in% meta_cols) {
        stop("'type' column missing from GRanges metadata")
    }

    anno <- infer_gene_ids(anno)
    meta_cols <- colnames(GenomicRanges::mcols(anno))

    if (!"gene_id" %in% meta_cols) {
        stop("'gene_id' column missing from GRanges metadata and could not be inferred")
    }

    anno_df <- as.data.frame(anno)

    n_exons <- anno_df %>%
        dplyr::filter(.data$type == "exon") %>%
        nrow()

    if (n_exons == 0) {
        stop("no exons found in annotation, must be labelled 'exon' under 'type' column")
    }

    saf <- anno_df %>%
        dplyr::filter(.data$type == "exon") %>%
        dplyr::select("gene_id", "seqnames", "start", "end", "strand") %>%
        dplyr::rename(
            GeneID = "gene_id",
            Chr = "seqnames",
            Start = "start",
            End = "end",
            Strand = "strand"
        )

    if (anyNA(saf$GeneID)) {
        orig_rows <- nrow(saf)
        saf <- saf %>%
            dplyr::filter(!is.na("GeneID"))
        filt_rows <- nrow(saf)
        message(glue::glue("NA found in GeneID of {orig_rows - filt_rows} of {orig_rows} entries, automatically removing these entries"))
    }

    saf %>% dplyr::select("GeneID", dplyr::everything())
}

infer_gene_ids <- function(anno) {
    # check if gene_id column is present, this is missing from RefSeq annotations
    no_gene_ids <- is.null(anno$gene_id)
    has_dbx <- !is.null(anno$Dbxref)

    if (!no_gene_ids) {
        exons_missing_gene_ids <- anno %>%
            as.data.frame() %>%
            dplyr::filter(.data$type == "exon") %>%
            dplyr::pull("gene_id") %>%
            is.na() %>%
            any()

        if (!exons_missing_gene_ids) {
            return(anno)
        }
    }
    if (!no_gene_ids && all(!is.na(anno$gene_id))) {
        return(anno)
    }

    if (no_gene_ids && has_dbx) {
        anno <- infer_gene_id_from_dbx(anno)
        return(anno)
    }

    # check if every entry has a gene_id value, this is not true for ENSEMBL gff3
    incomplete_gene_ids <- anyNA(anno$gene_id[anno$type == "exon"])
    has_parent <- !is.null(anno$Parent)

    if ((no_gene_ids || incomplete_gene_ids) && has_parent) {
        anno <- infer_gene_id_from_parent(anno)
        return(anno)
    }

    if (incomplete_gene_ids) {
        stop("some exons have missing gene_id, and gene_id could not be inferred")
    }

    return(anno)
}

# shorthand for stringr string interpolation
fmt_str <- stringr::str_interp

infer_gene_id_from_dbx <- function(anno) {
    extract_gene_id <- function(anno) {
        sapply(anno, function(x) x[stringr::str_detect(x, "GeneID")][1]) %>%
            stringr::str_extract("GeneID:[^,]+") %>%
            stringr::str_remove("GeneID:")
    }

    gene_ids <- extract_gene_id(anno$Dbxref)

    anno$gene_id <- gene_ids

    anno
}

infer_gene_id_from_parent <- function(anno) {
    # create hash map from transcript id to parent gene id
    transcript_hash <- local({
        transcripts <- anno %>%
            as.data.frame() %>%
            dplyr::filter(!is.na(.data$transcript_id)) %>%
            dplyr::select("transcript_id", "Parent") %>%
            dplyr::mutate(
                transcript_id = paste0("transcript:", .data$transcript_id),
                Parent = stringr::str_remove(.data$Parent, "gene:")
            )

        if (any(is.na(transcripts$Parent))) {
            warning("there are entries in annotation with transcript_id but no parent")
        }

        hash::hash(transcripts$transcript_id, transcripts$Parent)
    })

    # get parent ids of all exons
    parents <- anno %>%
        as.data.frame() %>%
        dplyr::filter(.data$type == "exon") %>%
        dplyr::pull("Parent") %>%
        unlist()
    
    gene_ids <- character(length(parents))
    
    for (i in seq_along(parents)) {
        gene_ids[i] <- transcript_hash[[parents[i]]]
    }

    anno$gene_id[anno$type == "exon"] <- gene_ids

    anno
}

validate_saf <- function(saf_df) {
    if (!is(saf_df, "data.frame")) {
        stop("annotation must object of class data.frame")
    }

    expected_colnames <- c("GeneID", "Chr", "Start", "End", "Strand")
    if (!identical(colnames(saf_df), expected_colnames)) {
        stop("columns of SAF data.frame must be: ", paste(expected_colnames, collapse = ", "))
    }

    if (anyNA(saf_df)) {
        stop("SAF data.frame must not contain any NA")
    }
}

context("Annotation import")
test_that("ENSEMBL gtf annotation can be imported", {
  anno_path <- system.file("extdata", "ensembl_y_tiny.gtf.gz", package = "scPipe")
  
  expected <- data.frame(
    GeneID = c("ENSG00000251841", "ENSG00000184895", "ENSG00000237659", "ENSG00000232195"),
    Chr = factor(c("Y", "Y", "Y", "Y")),
    Start = as.integer(c(2784749, 2786855, 2789827, 2827982)),
    End = as.integer(c(2784853, 2787699, 2790328, 2828218)),
    Strand = factor(c("+", "-", "+", "+"), levels = c("+", "-", "*")),
    stringsAsFactors = FALSE
  )
  
  expect_identical(anno_import(anno_path), expected)
})

test_that("Common annotation sources can be imported", {
    expect_silent(anno_import(system.file("extdata", "ens_tiny_anno.gff3.gz", package = "scPipe")))
    expect_silent(anno_import(system.file("extdata", "ens_tiny_anno.gtf.gz", package = "scPipe")))
    expect_equal(
        anno_import(system.file("extdata", "ens_tiny_anno.gff3.gz", package = "scPipe")),
        anno_import(system.file("extdata", "ens_tiny_anno.gtf.gz", package = "scPipe"))
    )
    expect_silent(anno_import(system.file("extdata", "gen_tiny_anno.gff3.gz", package = "scPipe")))
    expect_silent(anno_import(system.file("extdata", "gen_tiny_anno.gtf.gz", package = "scPipe")))
    expect_equal(
        anno_import(system.file("extdata", "gen_tiny_anno.gff3.gz", package = "scPipe")),
        anno_import(system.file("extdata", "gen_tiny_anno.gtf.gz", package = "scPipe"))
    )
    expect_silent(anno_import(system.file("extdata", "ref_tiny_anno.gff3.gz", package = "scPipe")))
})

test_that("Errors are properly reported", {
  anno <- rtracklayer::import(system.file("extdata", "ensembl_y_tiny.gtf.gz", package = "scPipe"))
  
  anno_missing_gene_id <- anno
  GenomicRanges::mcols(anno_missing_gene_id) <- subset(GenomicRanges::mcols(anno_missing_gene_id), select = -gene_id) # delete gene_id column
  expect_error(anno_to_saf(anno_missing_gene_id), "'gene_id' column missing from GRanges metadata and could not be inferred")
})

test_that("SAF validation works", {
  expect_error(validate_saf(1), "annotation must object of class data.frame")
  
  anno <- anno_import(system.file("extdata", "ens_tiny_anno.gff3.gz", package = "scPipe"))
  expect_error(validate_saf(anno[, 5:1]), "columns of SAF data.frame must be: GeneID, Chr, Start, End, Strand")
  
  anno_na <- anno
  anno_na[4, 1] <- NA
  expect_error(validate_saf(anno_na), "SAF data.frame must not contain any NA")
})

context("scPipe testing")

test_that("new SingleCellExperiment does not contain QC information", {
  data("sc_sample_data")
  data("sc_sample_qc")
  sce = SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
  expect_that(sce, is_a("SingleCellExperiment"))
  expect_warning(val <- QC_metrics(sce),
                 "`scPipe` not in `metadata`. Cannot identify quality control columns")
  expect_null(val)

  expect_warning(val <- demultiplex_info(sce),
                 "`scPipe` not in `metadata`. Cannot find columns for cell barcode demultiplex results")
  expect_null(val)

  expect_warning(val <- UMI_dup_info(sce),
                 "`scPipe` not in `metadata`. Cannot find columns for cell barcode demultiplex results")
  expect_null(val)

})

test_that("new SingleCellExperiment should work properly", {
  data("sc_sample_data")
  data("sc_sample_qc")
  sce = SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
  QC_metrics(sce) = sc_sample_qc
  demultiplex_info(sce) = cell_barcode_matching
  UMI_dup_info(sce) = UMI_duplication
  expect_equal(dim(sce), c(1000, 383))
  expect_equal(dim(QC_metrics(sce)), c(383, 12))
  expect_equal(dim(UMI_dup_info(sce)), c(1001, 2))
  expect_equal(dim(demultiplex_info(sce)), c(6, 2))
})

---
title: "scPipe: a flexible data preprocessing pipeline for scATAC-seq data"
author: "Shani Amarasinghe, Phil Yang"
date: "`r Sys.Date()`"
output: 
    BiocStyle::html_document:
        toc_float: true
vignette: >
  %\VignetteIndexEntry{scPipe: a flexible data preprocessing pipeline for scATAC-seq data}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

# Introduction

`scPipe` is a package initially designed to process single-cell RNA-sequencing (scRNA-seq) data generated by different protocols. We have modified it to accommodate pre-processing capability of single-cell ATAC-Seq (Assay for Transposase-Accessible Chromatin using sequencing) data pre-processing. `scPipe` ATAC-Seq module is designed for protocols without UMIs, but can also adapt to any UMI protocols if the need arise.

`scPipe` ATAC-Seq module consists of two major components. The first is data pre-processing with barcodes as raw fastq as input and a feature count matrix as output. Second is the data pre-processing with barcode CSV file as input and a feature count matrix as output.

`10X ATAC` method currently is the most popular method to generate scATAC-Seq data with higher sensitivity and lower cost. The structure of the 10X ATAC library is shown below. 

The output fastq files from a `10X ATAC` experiment is paired-ended and data is contained within both reads.

# Workflow

The pipeline is visually described via the workflow diagram depicted below. The functions will be explained in greater detail.

```{r, echo=FALSE, warning=FALSE, message=FALSE, out.width="100%"}
knitr::include_graphics(system.file("scATAC-seq_workflow.png", package = "scPipe"))
```

# Getting started

It is not mandatory to specify an output folder even though it can be specified. If no `output_folder` is defined a folder named `scPipe-atac-output` will get created in the working directory.

We begin by loading the library.

```{r, message=FALSE}
library(scPipe)
data.folder <- system.file("extdata", package = "scPipe", mustWork = TRUE)

```

Specify the output folder.

```{r}
output_folder <- paste0(tempdir(), "/scPipeATACVignette")
```

# Fastq reformatting

To process the data, we need the `fastq` files (both compressed and uncompressed versions are accepted) and a cell barcode annotation. We have supplied very small sample FASTQ files of chr21.  The barcode annotation could be either in a `.fastq` format or a `.csv` file with at least two columns, where the first column has the cell id and the second column contains the barcode sequence. All these files can be found in the `data` folder of the `scPipe` package:

```{r}
# data fastq files
r1      <- file.path(data.folder, "small_chr21_R1.fastq.gz") 
r2      <- file.path(data.folder, "small_chr21_R3.fastq.gz") 

# barcodes in fastq format
barcode_fastq      <- file.path(data.folder, "small_chr21_R2.fastq.gz") 

# barcodes in .csv format
barcode_1000       <- file.path(data.folder, "chr21_modified_barcode_1000.csv")

```

The pipeline starts with fastq file reformatting. We move the barcode and UMI sequences (if available) to the read name and leave the transcript sequence as is. This outputs a read name that looks like `@[barcode_sequence]_[UMI_sequence]#[readname]`. Usually the scATAC-Seq data is paired-end and a 16bp long barcode is located on both reads. Here the barcode information is located on a separate fastq file and the length of the barcode fastq file matches the length of the reads files. Therefore, you need a minimal example like below to generate the output.

```{r}
sc_atac_trim_barcode (r1            = r1, 
                      r2            = r2, 
                      bc_file       = barcode_fastq, 
                      rmN           = FALSE,
                      rmlow         = FALSE,
                      output_folder = output_folder)
```

This generates two output fastq files that are appended by the prefix `demux_` to notify that the new files are the reformatted (a.k.a. demultiplexed) `.fastq` files. These will get saved in the `scPipe-atac-output` directory if the user has not specified an `output_folder`.

However, if the barcodes are in the form of a `.csv` file, some extra information on 0-indexed barcode start (`id1_st`, `id2_st`) and the barcode length (`id1_len`, `id2_len`) are also required to be entered by the user. The algorithm is flexible to do a "look-around" to identify whether the correct parameters are used for a subset of data (hence saving time) and report back to the console if it believes the barcode position is incorrect and/or should be shifted.

```{r,eval=FALSE}
sc_atac_trim_barcode (r1            = r1, 
                      r2            = r2, 
                      bc_file       = barcode_1000, 
                      id1_st        = -1,  
                      id1_len       = -1,  
                      id2_st        = -1,  
                      id2_len       = -10,  
                      output_folder = output_folder, 
                      rmN           = TRUE)
```

To accommodate combinatorial indexing in some scATAC-seq protocols, the `bc_file` parameter will accept a list of barcode files as well (currently only implemented for the `fastq` approach).

Additionally, `rmN`, `rmlow` and `min_qual` parameters define the quality thresholds for the reads. If there are `Ns` in the barcode or UMI positions those will be discarded by `rmN = TRUE`. `rmlow =TRUE` will remove reads having lower quality than what is defined by `min_qual` (default: 20).

Completion of this function will output three different outputs depending on the findings;
* complete matches: When the barcode is completely matched and identified in the correct position
* partial matches: When the barcode is identified in the location specified but corrected with hamming distance approach
* unmatched: no barcode match is found in the given position even after hamming distance corrections are applied

**NOTE**: we use a zero based index system, so the indexing of the sequence starts at zero.

# Aligning reads to a reference genome

Next, we align reads to the genome. This example uses `Rsubread` but any aligner that support RNA-seq alignment and gives standard BAM output can be used here.

```{r}
demux_r1        <- file.path(output_folder, "demux_completematch_small_chr21_R1.fastq.gz")
demux_r2        <- file.path(output_folder, "demux_completematch_small_chr21_R3.fastq.gz")
reference = file.path(data.folder, "small_chr21.fa")


aligned_bam <- sc_aligning(ref = reference, 
                R1 = demux_r1, 
                R2 = demux_r2, 
                nthreads  = 12,
                output_folder = output_folder)
```

# Demultiplexing the BAM file

Next, the BAM file needs to be modified in a way one or two new columns are generated for the cellular barcode tag and the molecular barcode (i.e. UMI) tag denoted by `CB:Z:` and `OX:Z:`, respectively. 

```{r}

sorted_tagged_bam <- sc_atac_bam_tagging (inbam = aligned_bam, 
                       output_folder =  output_folder, 
                       bam_tags      = list(bc="CB", mb="OX"), 
                       nthreads      =  12)

```

# Remove duplicates

Next, PCR duplicate reads are removed from the BAM file. If `samtools` is installed, then `sc_atac_remove_duplicates` can be used, with the installation path of `samtools` being an optional argument if a particular version is preferred. However, if `samtools` isn't installed, the PCR duplicate removal should be performed externally.

```{r, eval=FALSE}

removed <- sc_atac_remove_duplicates(sorted_tagged_bam,
                          output_folder = output_folder)
if (!isFALSE(removed))
  sorted_tagged_bam <- removed

```


# Gemerating a fragment file

Next, a `fragment file` in a `.bed` format is generated, where each line represents a unique fragment generated by the assay. This file is used to generate useful quality control statistics, as well as for the `filter` and `cellranger` cell calling methods. 

```{r}
sc_atac_create_fragments(inbam = sorted_tagged_bam,
                         output_folder = output_folder)

```


# Peak calling

Similar to many other tools, the the ability to call peaks on a pseudo-bulk level on the demultiplexed reads has also been incorporated. `MACS3` is used underneath to achieve this functionality. If the genome size is not provided then it will be roughly estimated.

```{r, eval=FALSE}
# CHECK IF THE .narrowPeak file is small enough to include in the package,
# otherwise, we need to make this basilisk call work??
sc_atac_peak_calling(inbam = sorted_tagged_bam, 
                     ref = reference,
                     genome_size = NULL,
                     output_folder = output_folder)

```


# Assigning reads to features and feature counting

After the read alignment and BAM file demultiplexing, a feature set can be used to find the overlap between the aligned reads and the features using the `sc_atac_feature_counting` function. 

Accepted format of the `feature_input` should be either a BED format (i.e. format should have three main columns; chromosome, feature start and feature end, extension of the file is not considered) or a `genome.fasta` file. If using a BED format as the `feature_input`, the `feature_type` should be either "peak" or "tss". If using a `.fasta` for the `feature_input`, the `feature_type` needs to be defined as `genome_bin`. This `genome.fasta` file will be used within the function to generate a `genome_bin` that defines the array of features. The size of the bins can be set using the `bin_size` parameter (default: 2000). 

Note: avoid using input BAM files larger than 8GB for best performance

```{r, eval=FALSE}
features          <- file.path(output_folder, "NA_peaks.narrowPeak")

sc_atac_feature_counting (fragment_file = file.path(output_folder, "fragments.bed"),
                          feature_input = features, 
                          bam_tags      = list(bc="CB", mb="OX"), 
                          feature_type  = "peak", 
                          organism      = "hg38",
                          yieldsize     = 1000000,
                          exclude_regions = TRUE,
                          output_folder = output_folder,
                          fix_chr       = "none"
                          )
```

If genome_bin approach is selected, the following can be used:

```{r,eval=FALSE}
reference       <- file.path(data.folder, "small_chr21.fa")
sc_atac_feature_counting (fragment_file = file.path(output_folder, "fragments.bed"),
                          feature_input = reference,
                          bam_tags      = list(bc="CB", mb="OX"),
                          feature_type  = "genome_bin",
                          organism      = "hg38",
                          cell_calling  = FALSE,
                          yieldsize     = 1000000,
                          exclude_regions = TRUE,
                          output_folder = output_folder,
                          fix_chr       = "none"
                          )

```

Call calling is performed on the data prior to finding overlaps. The cell calling methods available are the `emptyDrops` method from [DropletUtils](https://bioconductor.org/packages/release/DropletUtils), `filter` which filters the cells based on various QC cut-offs, and `cellranger` which models the signal and noise as a mixture of two negative binomial distributions, though the `filter` method is recommended as it is most commonly used and the best-suited for scATAC-seq data.

For the `filter` cell-calling method, there are a variety of QC metrics that can be used, including:

* `min_uniq_frags`
* `max_uniq_frags`
* `min_frac_peak`
* `min_frac_tss`
* `min_frac_enhancer`
* `min_frac_promoter`
* `max_frac_mito`


```{r,eval=FALSE}
features          <- file.path(output_folder, "NA_peaks.narrowPeak")
```
```{r, eval=TRUE, echo=FALSE}
features          <- file.path(data.folder, "NA_peaks.narrowPeak")
```
```{r}
sc_atac_feature_counting (fragment_file = file.path(output_folder, "fragments.bed"),
                          feature_input = features, 
                          bam_tags      = list(bc="CB", mb="OX"), 
                          feature_type  = "peak",
                          organism      = "hg38",
                          cell_calling  = "filter",
                          min_uniq_frags = 0,
                          min_frac_peak = 0,
                          min_frac_promoter = 0,
                          yieldsize     = 1000000,
                          exclude_regions = TRUE,
                          output_folder = output_folder,
                          fix_chr       = "none"
                          )
```

Mapping quality (MAPQ) value can be set to filter data further in step (default: 30). Additionally, regions that are considered anomalous, unstructured, or high signal in next-generation sequencing experiments are excluded using an inbuilt `excluded_regions` file (available are for `organism` `hg19`, `hg38`, and `mm10`) or a user provided `excluded_regions_filename` file in BED format.

The discrepancy of `chr` between the alignment file and the feature file/excluded regions file can also be fixed (using `fix_char` parameter) if needed by adding the string `chr` to the beginning of either the features, and/or excluded regions. So the options for `fix_char` are "none", "excluded_regions", "feature", "both".

**NOTE** `genome_bin` approach may be more reliable in detecting sensitive features than using a pseudo-bulk peak calling approach, hence it would make the function slower as well.

The `sc_atac_feature_counting` function generates a matrix format of the feature by cell matrix (features as rows, cell barcodes as columns) in several formats (raw, sparse, binary, jaccard) that can be used downstream to generate a `singleCellExperiment, SCE` object.

```{r}
feature_matrix <- readRDS(file.path(output_folder, "unfiltered_feature_matrix.rds"))
dplyr::glimpse(feature_matrix)

sparseM <- readRDS(file.path(output_folder, "sparse_matrix.rds"))
dplyr::glimpse(sparseM)
```

# Generating the *Single-cell Experiment (SCE)* object

Easiest way to generate a *SCE* object is to run `sc_atac_create_sce` with the `input_folder` parameter. However, if the default dir name was used for previous steps simply running `sc_atac_create_sce()` would produce a `sce` object in the default location (i.e. `scPipe-atac-output`)

```{r}
sce <- sc_atac_create_sce(input_folder = output_folder,
                   organism     = "hg38",
                   feature_type = "peak",
                   pheno_data   = NULL,
                   report       = FALSE)
```

We have now completed the preprocessing steps. The feature count matrix is available as a `.rds` file in `scPipe-atac-output/feature_matrix.rds` and quality control statistics are saved in the `scPipe-atac-output/scPipe_atac_stats` folder. This data is useful for later quality control as well (QC).

If the report parameter is set to `TRUE` then an HTML report with various statistics and plots is generated. A partial screenshot of the report is shown below.

```{r, echo=FALSE, warning=FALSE, message=FALSE, out.width="100%"}
knitr::include_graphics(system.file("report_example.png", package = "scPipe"))
```

# A convenience function for running the whole pipeline

A function `sc_atac_pipeline` has been included to make it easier to run the whole pipeline.

```{r eval = FALSE}
data.folder <- system.file("extdata", package = "scPipe", mustWork = TRUE)
output_folder <- tempdir()
sce <- sc_atac_pipeline(r1 = file.path(data.folder, "small_chr21_R1.fastq.gz"),
                        r2 = file.path(data.folder, "small_chr21_R3.fastq.gz"),
                        barcode_fastq = file.path(data.folder, "small_chr21_R2.fastq.gz"),
                        organism = "hg38",
                        reference = file.path(data.folder, "small_chr21.fa"),
                        feature_type = "peak",
                        remove_duplicates = FALSE,
                        min_uniq_frags = 0, # set to 0 as the sample dataset only has a small number of reads
                        min_frac_peak = 0,
                        min_frac_promoter = 0,
                        output_folder = output_folder)
```


# Downstream analysis

Since the [scater](https://bioconductor.org/packages/release/scater) and [scran](https://bioconductor.org/packages/release/scran) packages both use the [SingleCellExperiment](https://bioconductor.org/packages/release/SingleCellExperiment) class, it will be easy to further process this data using these packages for normalization and visualization. Other packages such as [SC3](https://bioconductor.org/packages/release/SC3) may be useful for clustering and [MAST](https://bioconductor.org/packages/release/MAST) and [edgeR](https://bioconductor.org/packages/release/edgeR) for differential expression analysis.

```{r, echo=FALSE, results='hide'}
# cleanup tempfiles
unlink(output_folder, recursive=TRUE)
```
# Session Information {-}
```{r}
sessionInfo()
```

# Minimal example -----------------

Open RStudio

Load the project

```r
devtools::load_all()
```

# Demultiplexing 

## define inputs

```r
data.folder <- system.file("data", package = "scPipe", mustWork = TRUE)

# original read files
r1 <- file.path(data.folder,"testfastq_S1_L001_R1_001.fastq.gz") 
r2 <- file.path(data.folder,"testfastq_S1_L001_R3_001.fastq.gz") 
  
# read files with barcode starting from 1st position
r1_barcode <- file.path(data.folder,"testfastq_S1_L001_R1_001_withbarcode.fastq.gz") 
r2_barcode <- file.path(data.folder,"testfastq_S1_L001_R3_001_withbarcode.fastq.gz") 

# barcodes in fastq format
barcode_fastq      <- file.path(data.folder, "testfastq_S1_L001_R2_001.fastq.gz") 

# barcodes in .csv format
barcode_csv_sample <- file.path(data.folder,"barcode.csv")
barcode_csv        <- file.path(data.folder, "testfastq_modified_barcode.csv") 
barcode_csv_small  <- file.path(data.folder, "testfastq_modified_barcode_small.csv") 
barcode_csv_incorr <- file.path(data.folder, "testfastq_modified_barcode_corrupt.csv") 

output_folder <- NULL

```
### if using a barcode fastq file 

```r
sc_atac_trim_barcode (r1            = r1, 
                      r2            = r2, 
                      bc_file       = barcode_fastq, 
                      output_folder = output_folder)
```
### if using a barcode .csv file 

program expects the barcode to be on the second column of the .csv file

```r
sc_atac_trim_barcode(r1 = r1_barcode, 
                     r2 = r2_barcode, 
                     bc_file = barcode_csv_small, 
                     bc_start = 3, bc_length = 16, 
                     output_folder = output_folder, 
                     rmN = FALSE)
```
# Aligning to reference 

## define inputs

```r
reference       <- file.path(data.folder, "genome.fa")
demux_r1        <- file.path(output_folder, "demux_testfastq_S1_L001_R1_001.fastq.gz")
demux_r2        <- file.path(output_folder, "demux_testfastq_S1_L001_R3_001.fastq.gz")
```
## run function

```r
sc_atac_aligning(ref = reference, 
                 R1 = demux_r1, 
                 R2 = demux_r2, 
                 output_folder = output_folder,
                 nthreads = 6)
```
# Tagging the aligned BAM file

```r
bam_to_tag  <- file.path(output_folder, "demux_testfastq_S1_L001_R1_001_aligned.bam")

sc_atac_bam_tagging (inbam         = bam_to_tag, 
                     output_folder = output_folder, 
                     bam_tags      = list(bc="CB", mb="OX"), 
                     nthreads      = 6)
```

# Remove PCR duplicates (requires samtools)
```r
sorted_tagged_bam <- file.path(output_folder, "demux_testfastq_S1_L001_R1_001_aligned_tagged_sorted.bam")

sc_atac_remove_duplicates(inbam         = sorted_tagged_bam,
                          samtools_path = NULL, # can specify custom path here
                          output_folder = output_folder)

sorted_tagged_bam <- file.path(output_folder, "demux_testfastq_S1_L001_R1_001_aligned_tagged_sorted_markdup.bam")
```

# Gemerating a fragment file
```r
sc_atac_create_fragments(inbam = sorted_tagged_bam,
                         output_folder = output_folder)
```

# Peak calling
```r
sc_atac_peak_calling(inbam = sorted_tagged_bam, 
                     ref = reference,
                     genome_size = NULL,
                     output_folder = output_folder)
```

# Feature counting

```r
sorted_tagged_bam <- file.path(output_folder, "demux_testfastq_S1_L001_R1_001_aligned_tagged_sorted.bam")
features          <- file.path(data.folder, "extdata", "NA_peaks.narrowPeak")

sc_atac_feature_counting (insortedbam   = sorted_tagged_bam,
                          feature_input = features, 
                          bam_tags      = list(bc="CB", mb="OX"), 
                          feature_type  = "peak",
                          organism      = "hg38",
                          cell_calling  = "filter",
                          genome_size   = NULL,
                          bin_size      = NULL, 
                          yieldsize     = 1000000,
                          mapq          = 30,
                          exclude_regions = TRUE,
                          output_folder = output_folder,
                          fix_chr       = "none"
                          )
```

# Generating the *Single-cell Experiment (SCE)* object

```r
sce <- sc_atac_create_sce(input_folder = output_folder,
                          organism     = "hg38",
                          feature_type = "peak",
                          pheno_data   = NULL,
                          report       = TRUE
                          )
```

---
title: "scPipe report for scRNA-Seq sample `r params$samplename`"
output:
  html_notebook:
    code_folding: hide
    toc: true
    toc_float: true
    collapsed: false
params:
  samplename: "SAMPLE_NAME__"
  fq1: "FQ1__"
  fq2: "FQ2__"
  fqout: "FQOUT__"
  bc1_info: "BC1_INFO__"
  bc2_info: "BC2_INFO__"
  umi_info: "UMI_INFO__"
  rm_n: RM_N__
  rm_low: RM_LOW__
  min_q: MIN_Q__
  num_bq: NUM_BQ__
  bam_align: "BAM_ALIGN__"
  g_index: "G_INDEX__"
  bam_map: "BAM_MAP__"
  outdir: "OUTDIR__"
  anno_gff: "ANNO_GFF__"
  stnd: STND__
  fix_chr: FIX_CHR__
  bc_anno: "BC_ANNO__"
  max_mis: MAX_MIS__
  UMI_cor: "UMI_COR__"
  gene_fl: GENE_FL__
  organism: "ORGANISM__"
  gene_id_type: "GENE_ID_TYPE__"

---
```{r, include=FALSE}
getVal <- function(x, default = NULL) {
  ifelse(is.null(x), default, x)
}
```

```{r, warning=FALSE, echo=FALSE, message=FALSE}
library(scales)
library(readr)
library(ggplot2)
library(plotly)
library(DT)
library(scater)
library(scran)
library(scPipe)
library(Rtsne)
```

# Parameters for each preprocessing step

## Parameters for `sc_trim_barcode`

### File paths

* input fastq1: `r params$fq1`
* input fastq2: `r params$fq2`
* output fastq: `r params$fqout`

### Read structure

assume read1 contains the transcript

* barcode in read1: `r params$bc1_info`
* barcode in read2: `r params$bc2_info`
* UMI in read2: `r params$umi_info`

### Read filter

* remove reads that have `N` in its barcode or UMI: `r params$rm_n`
* remove reads with low quality: `r params$rm_low`
`r if (params$rm_low){paste("\t* minimum read quality:", params$min_q, "\n", "\t* maximum number of base below minimum read quality:", params$num_bq, "\n")}`
  
## Parameters for alignment

* input fastq: `r params$fqout`
* output bam file: `r params$bam_align`
* genome index: `r params$g_index`

## Parameters for `sc_exon_mapping`

* input bam file: `r params$bam_align`
* output bam file: `r params$bam_map`
* transctiptome annotations: `r params$anno_gff`
* do strand specific mapping: `r params$stnd`
* fix chromosome names: `r params$fix_chr`

## Parameters for `sc_demultiplex`

* input bam file: `r params$bam_map`
* output folder: `r params$outdir`
* barcode annotation file: `r params$bc_anno`
* maximum mismatch allowed in barcode: `r params$max_mis`

## Parameters for `sc_gene_counting`

* output folder: `r params$outdir`
* barcode annotation file: `r params$bc_anno`
* UMI correction: `r params$UMI_cor`
* gene filtering: `r params$gene_fl`

# Data summary

The organism is "`r getVal(params$organism, "unknown")`", and gene id type is "`r getVal(params$gene_id_type, "unknown")`".

## Overall barcode statistics

```{r, warning=FALSE, message=FALSE}
if (is.null(params$organism) || is.na(params$organism)) {
  sce = create_sce_by_dir(params$outdir)
} else {
  sce = create_sce_by_dir(params$outdir, organism=params$organism, gene_id_type=params$gene_id_type)
}
overall_stat = demultiplex_info(sce)
datatable(overall_stat, width=800)
```

Plot barcode match statistics in pie chart:
```{r}
plot_demultiplex(sce)
```

## Read alignment statistics

```{r}
ggplotly(plot_mapping(sce, dataname=params$samplename, percentage = FALSE))
```

```{r}
ggplotly(plot_mapping(sce, dataname=params$samplename, percentage = TRUE))
```

## Summary and distributions of QC metrics

```{r}
if (any(colSums(counts(sce)) == 0)) {
  zero_cells = sum(colSums(counts(sce)) == 0)
  sce = sce[, colSums(counts(sce)) > 0]
} else {
  zero_cells = 0
}
```

`r if (zero_cells > 0){paste(zero_cells, "cells have zero read counts, remove them.")}`

Datatable of all QC metrics:
```{r}
sce = calculate_QC_metrics(sce)
if(!all(colSums(as.data.frame(QC_metrics(sce)))>0)){
  QC_metrics(sce) = QC_metrics(sce)[, colSums(as.data.frame(QC_metrics(sce)))>0]
}
datatable(as.data.frame(QC_metrics(sce)), width=800, options=list(scrollX= TRUE))
```

Summary of all QC metrics:
```{r}
datatable(do.call(cbind, lapply(QC_metrics(sce), summary)), width=800, options=list(scrollX= TRUE))
```

Number of reads mapped to exon before UMI deduplication VS number of genes detected:
```{r}
ggplotly(ggplot(as.data.frame(QC_metrics(sce)), aes(x=mapped_to_exon, y=number_of_genes))+geom_point(alpha=0.8))
```

# Quality control

## Detect outlier cells

A robustified Mahalanobis Distance is calculated for each cell then outliers are detected based on the distance.
However, due to the complex nature of single cell transcriptomes and protocol used, such a method can only be used to
assist the quality control process. Visual inspection of the quality control metrics is still required. By default we
use `comp = 2` and the algorithm will try to separate the quality control metrics into two gaussian clusters.

The number of outliers:
```{r}
sce_qc = detect_outlier(sce, type="low", comp = 2)
table(QC_metrics(sce_qc)$outliers)
```

Pairwise plot for QC metrics, colored by outliers:
```{r, warning=FALSE, message=FALSE}
plot_QC_pairs(sce_qc)
```

## plot highest expression genes

Remove low quality cells and plot highest expression genes.

```{r}
sce_qc = remove_outliers(sce_qc)
sce_qc = convert_geneid(sce_qc, returns="external_gene_name")

sce_qc <- calculate_QC_metrics(sce_qc)
plotHighestExprs(sce_qc, n=20)
```

## remove low abundant genes

Plot the average count for each genes:
```{r}
ave.counts <- rowMeans(counts(sce_qc))
hist(log10(ave.counts), breaks=100, main="", col="grey80",
     xlab=expression(Log[10]~"average count"))
```

As a loose filter we keep genes that are expressed in at least two cells and for cells that express that gene, the average count larger than two. However this is not the gold standard and the filter may variy depending on the data.

```{r}
keep1 = (apply(counts(sce_qc), 1, function(x) mean(x[x>0])) > 1.1)  # average count larger than 1.1
keep2 = (rowSums(counts(sce_qc)>0) > 5)  # expressed in at least 5 cells

sce_qc = sce_qc[(keep1 & keep2), ]
dim(sce_qc)
```

We got `r nrow(sce_qc)` genes left after removing low abundant genes.

# Data normalization

## Normalization by `scran` and `scater`

Compute the normalization size factor

```{r}
ncells = ncol(sce_qc)
if (ncells > 200) {
  sce_qc <- computeSumFactors(sce_qc)
} else {
  sce_qc <- computeSumFactors(sce_qc, sizes=as.integer(c(ncells/7, ncells/6, ncells/5, ncells/4, ncells/3)))
}
summary(sizeFactors(sce_qc))
```

`r if (min(sizeFactors(sce_qc)) <= 0){paste("We have negative size factors in the data. They indicate low quality cells and we have removed them. To avoid negative size factors, the best solution is to increase the stringency of the filtering.")}`

```{r}
if (min(sizeFactors(sce_qc)) <= 0) {
  sce_qc = sce_qc[, sizeFactors(sce_qc)>0]
}
```

PCA plot using gene expressions as input, colored by the number of genes.

```{r}
cpm(sce_qc) = calculateCPM(sce_qc, size.factors=NULL)
sce_qc <- runPCA(sce_qc, exprs_values = "cpm")
plotPCA(sce_qc, colour_by="number_of_genes")
```

### Normalize the data using size factor and get high variable genes

The highly variable genes are chosen based on `trendVar` from `scran` with `FDR > 0.05` and biological variation larger
than `0.5`. If the number of highly variable genes is smaller than 100 we will select the top 100 genes by
biological variation. If the number is larger than 500 we will only keep top 500 genes by biological variation.

```{r, warning=FALSE, message=FALSE}
sce_qc <- logNormCounts(sce_qc)

var.out <- modelGeneVar(sce_qc)

if (length(which(var.out$FDR <= 0.05 & var.out$bio >= 0.5)) < 500){
  hvg.out <- var.out[order(var.out$bio, decreasing=TRUE)[1:500], ]
}else if(length(which(var.out$FDR <= 0.05 & var.out$bio >= 0.5)) > 1000){
  hvg.out <- var.out[order(var.out$bio, decreasing=TRUE)[1:1000], ]
}else{
  hvg.out <- var.out[which(var.out$FDR <= 0.05 & var.out$bio >= 0.5), ]
  hvg.out <- hvg.out[order(hvg.out$bio, decreasing=TRUE), ]
}

plot(var.out$mean, var.out$total, pch=16, cex=0.6, xlab="Mean log-expression",
     ylab="Variance of log-expression")
o <- order(var.out$mean)
lines(var.out$mean[o], var.out$tech[o], col="dodgerblue", lwd=2)
points(var.out$mean[rownames(var.out) %in% rownames(hvg.out)], var.out$total[rownames(var.out) %in% rownames(hvg.out)], col="red", pch=16)
```

## Heatmap of top100 high variable genes

```{r}
gene_exp = exprs(sce_qc)

gene_exp = gene_exp[rownames(hvg.out)[1:100], ]

hc.rows <- hclust(dist(gene_exp))
hc.cols <- hclust(dist(t(gene_exp)))

gene_exp = gene_exp[hc.rows$order, hc.cols$order]

m = list(
  l = 100,
  r = 40,
  b = 10,
  t = 10,
  pad = 0
) 

plot_ly(
    x = colnames(gene_exp), y = rownames(gene_exp),
    z = gene_exp, type = "heatmap")%>% 
layout(autosize = F, margin = m)
```

# Dimensionality reduction using high variable genes

## Dimensionality reduction by PCA

```{r}
sce_qc <- runPCA(sce_qc, exprs_values = "logcounts")
plotPCA(sce_qc, colour_by="number_of_genes")
```

## Dimensionality reduction by t-SNE

```{r, fig.width=10, fig.height=3.5}
set.seed(100)
if (any(duplicated(t(logcounts(sce_qc)[rownames(hvg.out), ])))) {
  sce_qc = sce_qc[, !duplicated(t(logcounts(sce_qc)[rownames(hvg.out), ]))]
}
sce_qc <- runTSNE(sce_qc, exprs_values="logcounts", perplexity=10,feature_set=rownames(hvg.out))
out5 <- plotTSNE(sce_qc, colour_by="number_of_genes") + ggtitle("Perplexity = 10")
sce_qc <- runTSNE(sce_qc, exprs_values="logcounts", perplexity=20,feature_set=rownames(hvg.out))
out10 <- plotTSNE(sce_qc, colour_by="number_of_genes") + ggtitle("Perplexity = 20")
sce_qc <- runTSNE(sce_qc, exprs_values="logcounts", perplexity=30,feature_set=rownames(hvg.out))
out20 <- plotTSNE(sce_qc, colour_by="number_of_genes") + ggtitle("Perplexity = 30")
gridExtra::grid.arrange(out5, out10, out20, ncol = 3)
```

---
title: "scPipe experiment report"
date: "`r format(Sys.time(), '%b %d %Y')`"
output:
  html_document:
    toc: true
    toc_float:
      collapsed: no

params:
  outdir: "."
  organism: "ORGANISM__"
  gene_id_type: "GENE_ID_TYPE__"
---
```{r, include=FALSE}
getVal <- function(x, default = NULL) {
  ifelse(is.null(x), default, x)
}
```

```{r, warning=FALSE, echo=FALSE, message=FALSE}
library(scales)
library(readr)
library(ggplot2)
library(plotly)
library(DT)
library(scater)
library(scran)
library(scPipe)
library(Rtsne)
knitr::opts_chunk$set(echo = FALSE)
```

# Data summary

The organism is `r getVal(params$organism, "unknown")`, and gene id type is "`r getVal(params$organism, "unknown")`".

## Cell barcode statistics

```{r, warning=FALSE, message=FALSE}
sce <- create_sce_by_dir(
  params$outdir,
  organism = params$organism,
  gene_id_type = params$gene_id_type
)
overall_stat <- demultiplex_info(sce)
datatable(overall_stat, width=800)
```

Plot barcode match statistics in pie chart:
```{r}
plot_demultiplex(sce) +
  theme(axis.text.x = element_text(size = 10))
```

## Read alignment statistics

```{r}
ggplotly(plot_mapping(sce, percentage=FALSE))
```

```{r}
ggplotly(plot_mapping(sce, percentage=TRUE))
```

## Summary and distributions of QC metrics

```{r}
if (any(colSums(counts(sce)) == 0)) {
  zero_cells <- sum(colSums(counts(sce)) == 0)
  sce <- sce[, colSums(counts(sce)) > 0]
} else {
  zero_cells = 0
}
```

`r if (zero_cells > 0){paste(zero_cells, "cells have zero read counts, remove them.")}`

Datatable of all QC metrics:
```{r}
sce <- calculate_QC_metrics(sce)

rounded_qc_metrics <- round(as.data.frame(QC_metrics(sce)), 2)
datatable(rounded_qc_metrics, width=800, options=list(scrollX=TRUE))
```

Summary of all QC metrics:
```{r}
summary_table <- round(do.call(cbind, lapply(QC_metrics(sce), summary)), 2)
datatable(summary_table, width=800, options=list(scrollX=TRUE))
```

Number of reads mapped to exon before UMI deduplication VS number of genes detected:
```{r}
ggplotly(ggplot(as.data.frame(QC_metrics(sce)), aes(x=mapped_to_exon, y=number_of_genes))+geom_point(alpha=0.8))
```

# Quality control

## Detect outlier cells

A robustified Mahalanobis Distance is calculated for each cell then outliers are detected based on the distance.
However, due to the complex nature of single cell transcriptomes and protocol used, such a method can only be used to
assist the quality control process. Visual inspection of the quality control metrics is still required. By default we
use `comp = 1` and the algorithm will try to separate the quality control metrics into two gaussian clusters.

The number of outliers:
```{r}
sce_qc = detect_outlier(sce, type="low", comp=1)
table(QC_metrics(sce_qc)$outliers)
```

Pairwise plot for QC metrics, colored by outliers:
```{r, warning=FALSE, message=FALSE, fig.width=10, fig.height=10}
plot_QC_pairs(sce_qc) + theme_bw()
```

## Plot high expression genes

Remove low quality cells and plot highest expression genes.

```{r}
sce_qc <- remove_outliers(sce_qc)
sce_qc <- convert_geneid(sce_qc, returns="external_gene_name")
sce_qc <- calculateQCMetrics(sce_qc)
plotHighestExprs(sce_qc, n=20)
```

## Remove low abundance genes

Plot the log10 average count for each gene:
```{r}
ave.counts <- rowMeans(counts(sce_qc))
hist(log10(ave.counts), breaks=100, main="", col="grey80",
     xlab=expression(Log[10]~"average count"))
```

As a loose filter we keep genes that are expressed in at least two cells and for cells that express that gene, the
average count larger than 1. This is not 

```{r}
keep1 <- rowMeans(counts(sce_qc)) > 1 # average count larger than one
keep2 <- rowSums(counts(sce_qc) > 0) > 2 # expressed in at least three cells

sce_qc <- sce_qc[(keep1 & keep2), ]
dim(sce_qc)
```

We have `r nrow(sce_qc)` genes left after removing low abundance genes.

# Data normalization

## Sample normalization

We perform normalization using `scater` and `scran`, 

5-point summary of size factors:
```{r}
ncells <- ncol(sce_qc)
if (ncells > 200) {
  sce_qc <- computeSumFactors(sce_qc)
} else {
  sizes <- as.integer(c(ncells/7, ncells/6, ncells/5, ncells/4, ncells/3))
}

has_spikes_ins <- any( str_detect(rownames(sce_qc), "ERCC") )
if (has_spikes_ins) {
  sce_qc <- computeSpikeFactors(sce_qc, general.use=FALSE)
}

summary(sizeFactors(sce_qc))
```

`r if (min(sizeFactors(sce_qc)) <= 0) { print("We have negative size factors in the data. They indicate low quality cells
and we have removed them. To avoid negative size factors, the best solution is to increase the stringency of the
filtering.") }`

```{r}
# filter out samples with negative size factors
if (min(sizeFactors(sce_qc)) <= 0) {
  sce_qc <- sce_qc[, sizeFactors(sce_qc) > 0]
}
```

PCA plot using gene expressions as input, colored by the number of genes.

```{r}
cpm(sce_qc) = calculateCPM(sce_qc, use_size_factors=TRUE)
plotPCA(sce_qc, exprs_values="cpm", colour_by="total_features")
```

### Normalize the data using size factor and get high variable genes

The highly variable genes are chosen based on `trendVar` from `scran` with `FDR > 0.05` and biological variation larger
than `0.5`. If the number of highly variable genes is smaller than 100 we will select the top 100 genes by
biological variation. If the number is larger than 500 we will only keep top 500 genes by biological variation.

```{r, warning=FALSE, message=FALSE}
sce_qc <- normalize(sce_qc)

var.fit <- trendVar(sce_qc, method="loess", use.spikes=FALSE, span=0.2)
var.out <- decomposeVar(sce_qc, var.fit)
var.out <- var.out[order(var.out$bio, decreasing=TRUE), ]

signif.genes <- which(var.out$FDR <= 0.05 & var.out$bio >= 0.5)

if (length(signif.genes) < 100) {
  hvg.out <- var.out[1:100, ]
} else if (length(signif.genes) > 500){
  hvg.out <- var.out[1:500, ]
} else {
  hvg.out <- var.out[signif.genes, ]
}

plot(var.out$mean, var.out$total, pch=16, cex=0.6,
     xlab="Mean log-expression", ylab="Variance of log-expression")
o <- order(var.out$mean)
lines(var.out$mean[o], var.out$tech[o], col="dodgerblue", lwd=2)
points(var.out$mean[rownames(var.out) %in% rownames(hvg.out)],
       var.out$total[rownames(var.out) %in% rownames(hvg.out)],
       col="red", pch=16)
```

## Heatmap of high variable genes

```{r}
gene_exp <- exprs(sce_qc)

gene_exp <- gene_exp[rownames(hvg.out), ]

hc.rows <- hclust(dist(gene_exp))
hc.cols <- hclust(dist(t(gene_exp)))

gene_exp = gene_exp[hc.rows$order, hc.cols$order]

m = list(
  l = 100,
  r = 40,
  b = 10,
  t = 10,
  pad = 0
)

plot_ly(
  x = colnames(gene_exp), y = rownames(gene_exp),
  z = gene_exp, type = "heatmap"
) %>%
  layout(autosize = F, margin = m)
```

# Dimensionality reduction using high variable genes

## Dimensionality reduction by PCA

```{r}
plotPCA(sce_qc, exprs_values="logcounts", colour_by="total_features")
```

## Dimensionality reduction by t-SNE

```{r, fig.width=10, fig.height=3.5}
set.seed(100)
if (any(duplicated(t(logcounts(sce_qc)[rownames(hvg.out), ])))) {
  sce_qc <- sce_qc[, !duplicated(t(logcounts(sce_qc)[rownames(hvg.out), ]))]
}

out5 <- plotTSNE(sce_qc, exprs_values="logcounts", perplexity=5,
                 colour_by="total_features", feature_set=rownames(hvg.out)) + 
                    ggtitle("Perplexity = 5")
out10 <- plotTSNE(sce_qc, exprs_values="logcounts", perplexity=10,
                  colour_by="total_features", feature_set=rownames(hvg.out))  + 
                    ggtitle("Perplexity = 10")
out20 <- plotTSNE(sce_qc, exprs_values="logcounts", perplexity=20,
                  colour_by="total_features", feature_set=rownames(hvg.out))  + 
                    ggtitle("Perplexity = 20")

multiplot(out5, out10, out20, cols=3)
```

# Session information

```{r}
sessionInfo()
```

---
title: 'scPipe_atac report `r params$organism` scPipe `r Biobase::package.version("scPipe")`'
date: "`r Sys.Date()`"
output:
  html_document:
    toc: true
    toc_float:
      collapsed: no
runtime: shiny
params:
  input_folder: "FOLDER__"
  organism: "ORGANISM__"
  sample: "SAMPLE__"
  feature_type: "FEATURE_TYPE__"
  n_barcode_subset: "N_BARCODE_SUBSET__"
editor_options:
  chunk_output_type: console
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = FALSE,
                       # results = "hide",
                       message = FALSE,
                       warning = FALSE)
```


```{r}
library(data.table)
library(dplyr)
library(ggplot2)
library(kableExtra)
library(RColorBrewer)
library(grid)
library(knitr)
library(stringr)
library(scales)
library(tibble)
library(readr)
library(plotly)


theme_set(theme_bw())

# if not defined downstream specific for each plot use this order of the colours
custom_colours <- c("#E6AB02", "#1B9E77", "#7570B3", "#E7298A", "#66A61E", "#A6761D")
```

```{r inputs}

# final SCE object
sce                            <- readRDS(file.path(params$input_folder, "scPipe_atac_SCEobject.rds"))

# unfiltered feature-count matrix
unfiltered_mtx                 <- readRDS(file.path(params$input_folder, "unfiltered_feature_matrix.rds"))
unfiltered_mtx_bcs             <- colnames(unfiltered_mtx)

# filtered feature-count matrix
filtered_mtx   <- readRDS(file.path(params$input_folder, "sparse_matrix.rds"))

# binary matrix
binary.mat                     <- readRDS(file.path(params$input_folder, "binary_matrix.rds"))

# quality values for the complete dataset 
trimbarcode_stats_filename    <- file.path(params$input_folder, "scPipe_atac_stats", "stats_file_trimbarcode.txt")
trimbarcode_stats             <- read.csv(trimbarcode_stats_filename, header=FALSE) %>% as.data.frame()

# for alignment statistics for the complete dataset
alignment_stats_filename      <- file.path(params$input_folder, "scPipe_atac_stats", "stats_file_align.txt")
alignment_stats               <- read.csv(alignment_stats_filename, header = FALSE, skip = 1) %>% as.data.frame()

# for alignment statistics per chromosome
alignment_stats_chrom_filename <- file.path(params$input_folder, "scPipe_atac_stats", "stats_file_align_per_chrom.csv")
alignment_stats_chrom          <- read.csv(alignment_stats_chrom_filename)

# for demultiplexing statistics plotting
data_file_path                 <- file.path(params$input_folder, "scPipe_atac_stats", "demultiplexing_stats.csv")
data                           <- read.csv(data_file_path)

# for mapping rates across barcodes
stats_file                     <- file.path(params$input_folder, "scPipe_atac_stats", "mapping_stats_per_barcode.csv")
barcode_stats                  <- read.csv(stats_file, row.names = "barcode")[-5] 

# duplicate removal statistics
duplicate_stats.file           <- file.path(params$input_folder, "scPipe_atac_stats", "duplicate_removal_stats.txt")

# for plotting fragments
frags_file                     <- file.path(params$input_folder, "fragments.bed")
frags                          <- fread(frags_file)[V4 %in% unfiltered_mtx_bcs, ]

# read in called cell barcodes
cell_barcode_file              <- file.path(params$input_folder, "non_empty_barcodes.txt")
cell_barcodes                  <- fread(cell_barcode_file, header=F)$V1

# read in qc_per_bc file
cell_qc_metrics_file           <- file.path(params$input_folder, "cell_qc_metrics.csv")
bc_stat                        <- fread(cell_qc_metrics_file)[bc %in% unfiltered_mtx_bcs, ]

qc_sele                        <- bc_stat[bc %in% cell_barcodes, ]
qc_nonsele                     <- bc_stat[!bc %in% cell_barcodes, ]

```


# Global quality statistics

```{r}
trimbarcode_stats             <- stringr::str_replace_all(trimbarcode_stats[,1], "\\: +", "\\,") %>% as.data.frame()
trimbarcode_stats             <- stringr::str_replace_all(trimbarcode_stats[,1],"removed_low_qual","Number of low quality cells removed") %>% as.data.frame()
trimbarcode_stats             <- stringr::str_split_fixed(trimbarcode_stats[,1], ",", n=2) %>% as.data.frame()
```

```{r}
knitr::kable(trimbarcode_stats,
             col.names = c('Statistic', 'Value'),
             caption = "Global quality statistics") %>%
    kableExtra::kable_styling()
```

# Alignment statistics

## Global level

```{r}
knitr::kable(alignment_stats, 
             col.names = c('Statistic', 'Value'), 
             caption = "") %>%
    kableExtra::kable_styling()
```

<!-- Below are bar plots showing the distribution of how the fragments have been mapped. The top plot shows the number of uniquely and multi-mapped fragments, while the bottom plot provides a more detailed breakdown of the mapped fragments. -->

```{r alignment-stats-plot, fig.wide = TRUE, fig.asp = 0.9, fig.width = 14}
as.1              <- alignment_stats[alignment_stats$V1 %in% c("Unmapped_fragments", 
                                                  "Uniquely_mapped_fragments", 
                                                  "Multi_mapping_fragments"),] %>% rename(Mapping_type = "V1")
x                 <- c("Uniquely_mapped_fragments", "Multi_mapping_fragments", "Unmapped_fragments")
as.1              <- as.1 %>% slice(match(x, Mapping_type))
as.1$Mapping_type <- factor(as.1$Mapping_type, levels = as.1$Mapping_type)

as.2              <- alignment_stats[alignment_stats$V1 %in% c("Properly_paired_fragments", 
                                                  "Singleton_fragments",
                                                  "More_than_one_chr_fragments",
                                                  "Unexpected_strandness_fragments",
                                                  "Unexpected_template_length",
                                                  "Inversed_mapping"),] %>% rename(Mapping_type = "V1")

y                 <- c("Properly_paired_fragments",
                       "Singleton_fragments",
                       "More_than_one_chr_fragments",
                       "Unexpected_strandness_fragments",
                       "Unexpected_template_length",
                       "Inversed_mapping") 
as.2              <- as.2 %>% slice(match(y, Mapping_type))
as.2$Mapping_type <- factor(as.2$Mapping_type, levels = as.2$Mapping_type)


as.1$Proportion <- as.1$V2/sum(as.1$V2)
as.2$Proportion <- as.2$V2/sum(as.2$V2)

as.combined <- rbind(as.1, as.2)
g1 <- ggplot(as.1, aes(x = Mapping_type, y = Proportion, fill = Mapping_type)) + 
  geom_bar(stat="identity", width=1) +
  scale_fill_manual(values=custom_colours) +
  ylab("Proportion of total fragments") + 
  ggtitle("Uniquely mapped, multimapped and \nunmapped fragments") +
  theme(axis.text.x=element_blank(),
        title = element_text(size=10)) +
  scale_y_continuous(labels = scales::percent_format(), expand = expansion(mult = c(0, 0)))

g2 <- ggplot(as.2, aes(x = Mapping_type, y = Proportion, fill = Mapping_type)) + 
  geom_bar(stat="identity", width=1) +
  scale_fill_manual(values=custom_colours) +
  ylab("Proportion of total fragments") + 
  ggtitle("Breakdown of mapped fragments") +
  theme(axis.text.x=element_blank(),
        title = element_text(size=10)) +
  scale_y_continuous(labels = scales::percent_format(), expand = expansion(mult = c(0, 0)))

gridExtra::grid.arrange(g1, g2, ncol=1)

```

## Chromosome level

```{r}
alignment_stats_chrom          <- alignment_stats_chrom[!(alignment_stats_chrom$seqnames %in% c("*")),  
                                                        !(names(alignment_stats_chrom) %in% c("unmapped"))]
DT::datatable(alignment_stats_chrom)
```

# Barcode demultiplexing

```{r}
if(nrow(data) > 1)
{
  DT::datatable(data, options=list(paging=FALSE, searching=FALSE))
} else {
  # add missing columns before plotting
  status <- c("barcode_partial_match_count", 
              "barcode_unmatched_both_reads_unmapped", 
              "barcode_unmapped_matched", 
              "barcode_unmatched_mapped_ambigously",
              "barcode_unmatch_one_read_unmapped")
  count  <- c(0,0,0,0,0) 
  extra_cols <- data.frame(status = status, count = count)
  data <- base::rbind (data, extra_cols)
}
```

```{r barcode-alignment-barplot, fig.asp = 0.9, fig.align = 'center',fig.width=10}
data$prop <- data$count/sum(data$count) 

plotly::ggplotly(ggplot(data, aes_string(x="status", y="prop", fill="status")) + 
  scale_fill_manual(values=custom_colours) +
  geom_bar(stat="identity", width=1) +
  geom_text(aes(y = prop, label = percent(prop)), vjust = -0.5) +
  scale_y_continuous(expand = expansion(mult = c(0, 0))) +
  theme(axis.text.x=element_text(angle = 50, hjust = 1),
        axis.ticks.x=element_blank(),
        panel.border = element_blank()) +  
  xlab("status") +
  ylab("Percentage") +
  expand_limits(y = 1) +
  ggtitle(paste0("Overall alignment mapping statistics of demultiplexed unfiltered reads"))) 
```

# Duplicate removal 

```{r}
if (file.exists(duplicate_stats.file)) {
  duplicate.stats <- read.table(duplicate_stats.file, sep=":", skip=1)
  DT::datatable(duplicate.stats)
} else {
  cat("duplicate_removals_stats.txt couldn't be located in the stats folder. Did you run the duplicate removal function with samtools >= 1.10?\n")
}

```

# Insert size distribution

```{r insert-size-dist, fig.wide = TRUE, fig.asp = 0.9, fig.align='center'}
  frags[, 'isize' := V3 - V2]
  if (nrow(frags) >= 100000) {
    frags = frags[sort(sample(1:nrow(frags), 100000)), ]
  }
  
  plotly::ggplotly(ggplot(data = frags[isize < 800], aes(x = isize)) +
    geom_density(fill = '#1B9E77') + xlab('Insert Size (bp)') + ylab('Density') + theme_bw() +
    scale_fill_manual(values=custom_colours) +
    theme(legend.title=element_blank(), 
          legend.background = NULL, 
          axis.text = element_text(size = 15, family = "Helvetica"), 
          axis.title = element_text(size = 18, family = "Helvetica")))
```

# Cell calling, feature selection and barcode and feature level filtering 

## Summary

```{r}
cell_calling_stats          <- data.frame(row.names = c("No. of cells prior to filtering", "No. of cells retained after filtering"), count = c(ncol(unfiltered_mtx), ncol(filtered_mtx)))
kable(cell_calling_stats, row.names = TRUE, col.names = NULL) %>%
  kable_styling(full_width = TRUE, position = 'center', font_size = 24)

```

:::: {style="display: flex;"}
:::{}

```{r  frag-peak, fig.wide = TRUE, fig.asp = 0.9}
bc_stat[, 'group' := ifelse(bc %in% cell_barcodes, 'cell', 'non-cell')]
nsub_frags    <- min(15000, nrow(bc_stat))  ## downsample for scatter plot
bc_stat_down  <- bc_stat[sort(sample(1:nrow(bc_stat), nsub_frags)), ]


g             <- ggplot(data = bc_stat_down, 
                                         aes(x = total_frags, y = frac_peak, col = group)) + 
                                    geom_point(size = 0.5) + scale_x_continuous(trans='log10') + 
                                    theme_bw() +
                                    theme(legend.position = 'none', 
                                          legend.title=element_blank(),
                                          axis.text = element_text(size = 15, family = "Helvetica"),
                                          axis.title = element_text(size = 18, family = "Helvetica")) +
                                    xlab('Total #Unique Fragments') + 
                                    ylab('Fraction in Peak')

text1          <- grobTree(textGrob("Cell", x=0.8,  y=0.93, hjust=0,
                           gp=gpar(col='#E6AB02', fontsize=15, fontface = 'bold', fontfamily = "Helvetica")))
text2          <- grobTree(textGrob("Non-cell", x=0.8,  y=0.83, hjust=0,
                           gp=gpar(col='#7570B3', fontsize=15, fontface = 'bold', fontfamily = "Helvetica")))

g + annotation_custom(text1) + annotation_custom(text2) + scale_color_manual(values = c('#E6AB02', '#7570B3'))
```

:::
:::{}

```{r frag-mito, fig.wide = TRUE, fig.asp = 0.9}
nsub_frags    <- min(15000, nrow(bc_stat))  ## downsample for scatter plot
bc_stat_down  <- bc_stat[sort(sample(1:nrow(bc_stat), nsub_frags)), ]


g             <- ggplot(data = bc_stat_down, 
                        aes(x = total_frags, y = frac_mito, col = group)) + 
  geom_point(size = 0.5) + scale_x_continuous(trans='log10') + 
  theme_bw() +
  theme(legend.position = 'none', 
        legend.title=element_blank(),
        axis.text = element_text(size = 15, family = "Helvetica"),
        axis.title = element_text(size = 18, family = "Helvetica")) +
  xlab('Total #Unique Fragments') + 
  ylab('Fraction in mitochondrial genome')

text1          <- grobTree(textGrob("Cell", x=0.8,  y=0.93, hjust=0,
                                    gp=gpar(col='#E6AB02', fontsize=15, fontface = 'bold', fontfamily = "Helvetica")))
text2          <- grobTree(textGrob("Non-cell", x=0.8,  y=0.83, hjust=0,
                                    gp=gpar(col='#7570B3', fontsize=15, fontface = 'bold', fontfamily = "Helvetica")))

g + annotation_custom(text1) + annotation_custom(text2) + scale_color_manual(values = c('#E6AB02', '#7570B3'))
```

:::
:::{}

```{r frag-density, fig.wide = TRUE, fig.asp = 0.9}
bc_stat[, 'group' := ifelse(bc %in% cell_barcodes, 'cell', 'non-cell')]

p     <- ggplot(data = bc_stat, aes(x = total_frags, fill = group)) +
                            geom_density() +
                            scale_x_continuous(trans = 'log10') +
                            theme_bw() +
                            theme(legend.position='none',
                                  legend.title=element_blank(),
                                  axis.title = element_text(size = 18, family = "Helvetica"),
                                  axis.text = element_text(size = 15, family = "Helvetica")) +
                            xlab('Total #Unique Fragments') + ylab('Density')

text1 <- grobTree(textGrob("Cell", x=0.8,  y=0.93, hjust=0,
                           gp=gpar(col='#E6AB02', fontsize=15, fontface = 'bold', fontfamily = "Helvetica")))
text2 <- grobTree(textGrob("Non-cell", x=0.8,  y=0.83, hjust=0,
                           gp=gpar(col='#7570B3', fontsize=15, fontface = 'bold', fontfamily = "Helvetica")))

p + annotation_custom(text1) + annotation_custom(text2) +
  scale_fill_manual(values = c('#E6AB02', '#7570B3'))

```

:::
::::

## Reference mapping rate for filtered barcodes 

```{r barcode-alignment-dist, fig.asp = 0.9, fig.align = 'center'}

# subset the cell barcodes
if (!is.null(n_barcode_subset)){
  barcode_stats                      <- barcode_stats[row.names(barcode_stats) %in% qc_sele$bc, ]
  n <- nrow(barcode_stats)
  if (n < n_barcode_subset) n_barcode_subset <- n
  barcode_stats                      <-barcode_stats %>% sample_n(n_barcode_subset)
} else {
  barcode_stats                      <- barcode_stats[row.names(barcode_stats) %in% qc_sele$bc, ] 
}

# rearranging barcode_stats for plotting
mapping_stat              <- barcode_stats %>% dplyr::arrange(desc(mapped))
mapping_stat$barcode      <- stats::reorder(rownames(mapping_stat), mapping_stat$mapped)


#subset to order
mapped_subset             <- mapping_stat[mapping_stat$mapped>0 & 
                                            mapping_stat$mapped_ambigously==0 & 
                                            mapping_stat$one_read_unmapped==0 & 
                                            mapping_stat$both_reads_unmapped==0, ]

mapped_subset             <- mapped_subset[order(-mapped_subset[,2]),]

mapped_amb_subset         <- mapping_stat[mapping_stat$mapped>0 & 
                                            mapping_stat$mapped_ambigously>0 & 
                                            mapping_stat$one_read_unmapped==0 & 
                                            mapping_stat$both_reads_unmapped==0, ]

mapped_amb_subset         <- mapped_amb_subset[order(-mapped_amb_subset[,3]),]

mapped_one_subset         <- mapping_stat[mapping_stat$mapped>0 & 
                                            mapping_stat$mapped_ambigously==0 & 
                                            mapping_stat$one_read_unmapped>0 & 
                                            mapping_stat$both_reads_unmapped==0, ]

mapped_one_subset         <- mapped_one_subset[order(-mapped_one_subset[,4]),]

unmapped_subset           <- mapping_stat[mapping_stat$mapped>0 & 
                                            mapping_stat$mapped_ambigously==0 & 
                                            mapping_stat$one_read_unmapped==0 & 
                                            mapping_stat$both_reads_unmapped>0, ]

unmapped_subset           <- unmapped_subset[order(-unmapped_subset[,1]),]

all_subset                <- mapping_stat[mapping_stat$mapped>0 & 
                                            mapping_stat$mapped_ambigously>0 & 
                                            mapping_stat$one_read_unmapped>0 & 
                                            mapping_stat$both_reads_unmapped>0, ]

all_subset                <- all_subset[order(-all_subset[,2]),]

mapping_stat_ordered      <- rbind(mapped_subset,
                                   mapped_amb_subset,
                                   mapped_one_subset,
                                   unmapped_subset,
                                   all_subset)

remaining_subset          <- mapping_stat[!mapping_stat$barcode %in% mapping_stat_ordered$barcode, ]
remaining_subset          <- remaining_subset[order(-remaining_subset[,2]),]

mapping_stat_ordered      <- rbind(mapping_stat_ordered, remaining_subset)


# proportion convert to see the differences clearly
mapping_stat_prop         <- as.data.frame(prop.table(as.matrix(mapping_stat_ordered[, sapply(mapping_stat_ordered, is.numeric)]), 1))
mapping_stat_prop$barcode <- mapping_stat_ordered$barcode

# reshape to plot
dat.m1                    <- reshape2::melt(mapping_stat_prop, id.vars="barcode")
colnames(dat.m1)[2]       <- "type"

# order by category
dat.m1 <- dat.m1 %>% 
  mutate(type = factor(type, levels=c("mapped", "mapped_ambigously", "one_read_unmapped", "both_reads_unmapped")))

#order by barcode to show the mapped ones first
dat.m1 <- dat.m1 %>% 
  mutate(barcode = factor(barcode, levels=c(mapping_stat_ordered$barcode)))

# plotly::ggplotly(ggplot(dat.m1, 
#        aes(x = as.factor(barcode), y = as.numeric(value), fill = type)) + 
#   geom_bar(stat='identity', width =1) +
#   guides(fill = guide_legend(title = "")) +   
#   xlab("Sample of 500 barcodes") +   
#   ylab("Percentage of reads")+
#   theme(axis.text.x=element_blank(),
#           axis.ticks.x=element_blank()) +
#   scale_y_continuous(labels = percent_format(), expand = expansion(mult = c(0, 0))))

plotly::ggplotly(ggplot(dat.m1, 
       aes(x = as.factor(barcode), y = as.numeric(value), fill = type)) + 
  scale_fill_manual(values=custom_colours) + 
  geom_bar(stat='identity', width =1) +
  guides(fill = guide_legend(title = "")) +   
  xlab("Sample of 500 barcodes") +   
  ylab("Percentage of reads")+
  theme(axis.text.x=element_blank(),
          axis.ticks.x=element_blank()) +
  scale_y_continuous(labels = percent_format(), expand = expansion(mult = c(0, 0))))
```

## Distribution of fragment and feature counts for filtered cells
:::: {style="display: flex;"}
::: {}


```{r percell-counts, fig.wide = TRUE, fig.asp = 0.9}
sc_atac_plot_fragments_per_cell(sce)
```

:::
::: {}

```{r percell-features, fig.wide = TRUE, fig.asp = 0.9}
sc_atac_plot_features_per_cell(sce)
```
:::
::::

```{r, fig.align='center'}
sc_atac_plot_features_per_cell_ordered(sce)
```

## Distribution of fragment and cell counts for features
:::: {style="display: flex;"}
::: {}

```{r perfeature-counts, fig.wide = TRUE, fig.asp = 0.9}
sc_atac_plot_fragments_per_feature(sce)
```
:::
::: {}

```{r perfeature-cells, fig.wide = TRUE, fig.asp = 0.9}
sc_atac_plot_cells_per_feature(sce)
```
:::
::::

:::: {style="display: flex;"}
::: {}



```{r corr-cells, fig.wide = TRUE, fig.asp = 0.9}
sc_atac_plot_fragments_features_per_cell(sce)
```
:::
::: {}

```{r corr-features, fig.wide = TRUE, fig.asp = 0.9}
sc_atac_plot_fragments_cells_per_feature(sce)
```
:::
::::

# Feature overlap

:::: {style="display: flex;"}
::: {}

```{r overall-overlap, fig.wide = TRUE, fig.asp = 0.9}
# Get fraction of all fragments overlapping with each category (essentially weighted sum of fractions, weighted by number of total frags/cell)
frac_peak     <- sum(qc_sele$total_frags * qc_sele$frac_peak)/sum(qc_sele$total_frags)
frac_mito     <- sum(qc_sele$total_frags * qc_sele$frac_mito)/sum(qc_sele$total_frags)
frac_promoter <- sum(qc_sele$total_frags * qc_sele$frac_promoter)/sum(qc_sele$total_frags)
frac_enh      <- sum(qc_sele$total_frags * qc_sele$frac_enhancer)/sum(qc_sele$total_frags)
frac_tss      <- sum(qc_sele$total_frags * qc_sele$frac_tss)/sum(qc_sele$total_frags)

fracs         <- data.frame(c(frac_peak,  frac_promoter, frac_enh, frac_tss, frac_mito))
row.names(fracs) = c('Fraction in peaks', 
                     'Fraction in promoters', 
                     'Fraction in Enhancers(ENCODE)', 
                     'Fraction in TSS',
                     'Fraction in mitochondrial genome')
colnames(fracs) <- 'pr'
fracs$pr        <- round(fracs$pr, 3)
fracs$pr        <- paste0(100*fracs$pr, '%')

kable(fracs, row.names = T, col.names = NULL) %>%
  kable_styling(full_width = F, position = 'left', font_size = 15)
```

:::

::: {}

```{r anno-overlap-boxes, fig.wide = TRUE, fig.asp = 0.9}
qc_sele_df      <- data.table(frac = c(qc_sele$frac_peak, qc_sele$frac_tss, qc_sele$frac_promoter, qc_sele$frac_enh, qc_sele$frac_mito), 
                         'type' = rep(c('Peaks', 'TSS', 'Promoter', 'Enhancer', 'Mito'), each = nrow(qc_sele)))
  
qc_sele_df$type <- factor(qc_sele_df$type, levels = c('Peaks', 'TSS', 'Promoter', 'Enhancer', 'Mito'))

ggplot(data = qc_sele_df, aes(y = frac, x = type, fill = type)) + ylab('Fraction') + theme_bw() +
  geom_boxplot(outlier.size = 0.01, show.legend = FALSE) + 
  scale_fill_manual(values=custom_colours) +
  theme(legend.position = 'none', 
        axis.text = element_text(size = 18, family = "Helvetica"), 
        axis.title.x = element_blank(), 
        axis.title.y = element_text(size = 18, family = "Helvetica")) + xlab('')
```

:::
::::

# Aggregate TSS enrichment plot

TSSE: the maximum TSS score among all 100bp windows within 1000bp from each side of the TSS, where the TSS score is the read depth of the window normalised by the read depth of the end flanks.

```{r tss-enrichment, fig.wide = TRUE, fig.asp = 0.9}
tss_plot_data <- utils::read.csv(file.path(params$input_folder, "scPipe_atac_stats", "tss_plot_data.csv"))
if (any(is.na(tss_plot_data$agg_tss_scores))) {
  message("Could not generate plot; perhaps no overlaps.")
} else {
  tsse <- max(tss_plot_data$agg_tss_scores)
  max_x <- tss_plot_data$dists[which(tss_plot_data$agg_tss_scores==tsse)]
  ggplot(data=tss_plot_data, aes(x=dists, y=agg_tss_scores)) +
    geom_line()+
    geom_point() + 
    geom_text(aes(max_x, tsse+3, label=paste("TSSE:", signif(tsse,3)))) +
    ggtitle("Aggregate TSS enrichment") +
    xlab("Distance from TSS (bp)") + 
    ylab("Aggregate TSS score") 
}


```


# Visualisation of the preprocessed dataset - UMAP

```{r}
TF.IDF.custom <- function(binary.mat, verbose = TRUE) {
    object <- binary.mat
    npeaks       <- Matrix::colSums(x = object)
    tf           <- Matrix::tcrossprod(x = as.matrix(object), y = Matrix::Diagonal(x = 1 / npeaks))
    rsums        <- Matrix::rowSums(x = object)
    idf          <- ncol(x = object) / rsums
    norm.data    <- Matrix::Diagonal(n = length(x = idf), x = idf) %*% tf
    scale.factor <- 1e4
    slot(object = norm.data, name = "x") <- log1p(x = slot(object = norm.data, name = "x") * scale.factor)
    norm.data[which(x = is.na(x = norm.data))] <- 0
    return(norm.data)
  }

message("Generating UMAP data")

sce <-readRDS(file.path(params$input_folder, "scPipe_atac_SCEobject.rds"))
counts <- assay(sce)
  bin_mat <- as.matrix((counts>0)+0)
  binary.mat <- TF.IDF.custom(bin_mat)
  
library(irlba)
set.seed(123)
n_bcs <- max(min(50, ncol(binary.mat), nrow(binary.mat))-1,0)
mat.lsi          <- irlba(binary.mat, n_bcs)
d_diagtsne       <- matrix(0, n_bcs, n_bcs)
diag(d_diagtsne) <- mat.lsi$d
mat_pcs          <- t(d_diagtsne %*% t(mat.lsi$v))
rownames(mat_pcs)<- colnames(binary.mat)

# clustering in the PCA space using KNN --------------

library(RANN)
knn.info<- RANN::nn2(mat_pcs, k = 30)

## convert to adjacency matrix
knn           <- knn.info$nn.idx
adj           <- matrix(0, nrow(mat_pcs), nrow(mat_pcs))
rownames(adj) <- colnames(adj) <- rownames(mat_pcs)
for(i in seq_len(nrow(mat_pcs))) {
  adj[i,rownames(mat_pcs)[knn[i,]]] <- 1
}

## convert to graph
library(igraph)
g <- igraph::graph.adjacency(adj, mode="undirected")
g <- simplify(g) ## remove self loops

# identify communities, many algorithums. Use the Louvain clustering ------------
km         <- igraph::cluster_louvain(g)
com        <- km$membership
names(com) <- km$names

# running UMAP ------------------------------

library(umap)
norm.data.umap    <- umap::umap(mat_pcs)

df_umap           <- as.data.frame(norm.data.umap$layout)
colnames(df_umap) <- c("UMAP1", "UMAP2")
df_umap$barcode   <- rownames(mat_pcs)

df_umap           <- dplyr::left_join(df_umap, enframe(com), by = c("barcode" = "name")) %>%
  dplyr::rename(cluster = value) %>%
  dplyr::mutate(cluster = as.factor(cluster))

sce_coldata <- colData(sce)
  df_umap <- base::merge(df_umap, sce_coldata, by.x = "barcode", by.y = "row.names", all.x = TRUE) 

```

```{r interactive-umap, fig.wide = TRUE, fig.asp = 0.9, echo = FALSE}

g <- ggplot(df_umap, aes(x = UMAP1, y = UMAP2, text = paste("barcode: ", barcode))) +
        geom_point(aes(col = cluster), size = 0.5) +
        theme_bw(base_size = 14)

plotly::ggplotly(g)

```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_atac_tfidf.R
\name{sc_interactive_umap_plot}
\alias{sc_interactive_umap_plot}
\title{Produces an interactive UMAP plot via Shiny}
\usage{
sc_interactive_umap_plot(sce)
}
\arguments{
\item{sce}{The SingleCellExperiment object}
}
\value{
A shiny object which represents the app. Printing the object or passing it to `shiny::runApp(...)` will run the app.
}
\description{
Can colour the UMAP by any of the colData columns in the SCE object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_barcode_start.R
\name{check_barcode_start_position}
\alias{check_barcode_start_position}
\title{Check Valid Barcode Start Position}
\usage{
check_barcode_start_position(
  fastq,
  barcode_file,
  barcode_file_realname,
  bstart,
  blength,
  search_lines,
  threshold
)
}
\arguments{
\item{fastq}{file containing reads}

\item{barcode_file}{csv file}

\item{barcode_file_realname}{the real name of the csv file}

\item{bstart}{the start position for barcodes in the given reads}

\item{blength}{length of each barcode}

\item{search_lines}{the number of fastq lines to use for the check}

\item{threshold}{the minimum percentage of found barcodes to accept for the
program to continue}
}
\value{
Boolean; TRUE if program can continue execution, FALSE otherwise.
}
\description{
Checks to see if the given barcode start position (\code{bstart})
is valid for the fastq file. If the found barcode percentage is less than
the given \code{threshold}, a new barcode start position is searched for by
checking every position from the start of each read to 10 bases after the \code{bstart}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_atac_bam_tagging.R
\name{sc_atac_bam_tagging}
\alias{sc_atac_bam_tagging}
\title{BAM tagging}
\usage{
sc_atac_bam_tagging(
  inbam,
  output_folder = NULL,
  bc_length = NULL,
  bam_tags = list(bc = "CB", mb = "OX"),
  nthreads = 1
)
}
\arguments{
\item{inbam}{The input BAM file}

\item{output_folder}{The path of the output folder}

\item{bc_length}{The length of the cellular barcodes}

\item{bam_tags}{The BAM tags}

\item{nthreads}{The number of threads}
}
\value{
file path of the resultant demmultiplexed BAM file.
}
\description{
Demultiplexes the reads
}
\details{
sc_atac_bam_tagging()
}
\examples{
r1 <- system.file("extdata", "small_chr21_R1.fastq.gz", package="scPipe")
r2 <- system.file("extdata", "small_chr21_R3.fastq.gz", package="scPipe")
barcode_fastq <- system.file("extdata", "small_chr21_R2.fastq.gz", package="scPipe")
out <- tempdir()

sc_atac_trim_barcode(r1=r1, r2=r2, bc_file=barcode_fastq, output_folder=out)

demux_r1 <- file.path(out, "demux_completematch_small_chr21_R1.fastq.gz")
demux_r2 <- file.path(out, "demux_completematch_small_chr21_R3.fastq.gz")
reference <- system.file("extdata", "small_chr21.fa", package="scPipe")

aligned_bam <- sc_aligning(ref=reference, R1=demux_r1, R2=demux_r2, nthreads=6, output_folder=out)

out_bam <- sc_atac_bam_tagging(
   inbam = aligned_bam,
   output_folder = out,
   nthreads = 6)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_read_str.R
\name{get_read_str}
\alias{get_read_str}
\title{Get read structure for particular scRNA-seq protocol}
\usage{
get_read_str(protocol)
}
\arguments{
\item{protocol}{name of the protocol}
}
\value{
list of UMI and Barcode locations for use in other scPipe functions
}
\description{
The supported protocols are:
\itemize{
 \item CelSeq
 \item CelSeq2
 \item DropSeq
 \item 10x (also called ChromiumV1)
}
If you know the structure of a specific protocol and would like it supported,
please leave a issue post at www.github.com/luyitian/scPipe.
}
\examples{
get_read_str("celseq")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/biomart_methods.R
\name{get_genes_by_GO}
\alias{get_genes_by_GO}
\title{Get genes related to certain GO terms from biomart database}
\usage{
get_genes_by_GO(
  returns = "ensembl_gene_id",
  dataset = "mmusculus_gene_ensembl",
  go = NULL
)
}
\arguments{
\item{returns}{the gene id which is set as return. Default to be ensembl id
A possible list of attributes can be retrieved using the
function \code{listAttributes} from \code{biomaRt} package. The commonly used
id types are `external_gene_name`, `ensembl_gene_id` or `entrezgene`.}

\item{dataset}{Dataset you want to use. List of possible datasets can be
retrieved using the function \code{listDatasets} from \code{biomaRt} package.}

\item{go}{a vector of GO terms}
}
\value{
a vector of gene ids.
}
\description{
Get genes related to certain GO terms from biomart database
}
\details{
Get genes related to certain GO terms from biomart database
}
\examples{
# get all genes under GO term GO:0005739 in mouse, return ensembl gene id
get_genes_by_GO(returns="ensembl_gene_id",
    dataset="mmusculus_gene_ensembl",
    go=c('GO:0005739'))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_atac_create_sce.R
\name{sc_atac_create_sce}
\alias{sc_atac_create_sce}
\title{sc_atac_create_sce()}
\usage{
sc_atac_create_sce(
  input_folder = NULL,
  organism = NULL,
  sample_name = NULL,
  feature_type = NULL,
  pheno_data = NULL,
  report = FALSE
)
}
\arguments{
\item{input_folder}{The output folder produced by the pipeline}

\item{organism}{The type of the organism}

\item{sample_name}{The name of the sample}

\item{feature_type}{The type of the feature}

\item{pheno_data}{The pheno data}

\item{report}{Whether or not a HTML report should be produced}
}
\value{
a SingleCellExperiment object created from the scATAC-Seq data provided
}
\description{
sc_atac_create_sce()
}
\examples{
\dontrun{
sc_atac_create_sce(
   input_folder = input_folder,
   organism = "hg38",
   feature_type = "peak",
   report = TRUE)
}    

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/qc.R
\name{.qq_outliers_robust}
\alias{.qq_outliers_robust}
\title{Detect outliers based on robust linear regression of QQ plot}
\usage{
.qq_outliers_robust(x, df, conf)
}
\arguments{
\item{x}{a vector of mahalanobis distance}

\item{df}{degree of freedom for chi-square distribution}

\item{conf}{confidence for linear regression}
}
\value{
cell names of outliers
}
\description{
Detect outliers based on robust linear regression of QQ plot
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_sample_qc.R
\docType{data}
\name{sc_sample_qc}
\alias{sc_sample_qc}
\title{quality control information for a small sample scRNA-seq dataset to 
demonstrate capabilities of scPipe.}
\format{
a data.frame instance, one row per cell.
}
\source{
Christin Biben (WEHI). She FACS sorted cells from several immune
cell types including B cells, granulocyte and some early progenitors.
}
\value{
NULL, but makes a data frame with cell quality control data.frame
}
\description{
This data.frame contains cell quality control information 
for the 100 cells. For each cell it has: \itemize{
 \item{unaligned} the number of unaligned reads.
 \item{aligned_unmapped} the number of reads that aligned to genome but 
 fail to map to any features. 
 \item{mapped_to_exon} is the number of reads that mapped
 to exon.
 \item{mapped_to_intron} is the number of reads that mapped to intron.
 \item{ambiguous_mapping} is the number of reads that mapped to multiple
 features. They are not considered in the following analysis.
 \item{mapped_to_ERCC} is the number of reads that mapped to ERCC 
 spike-in controls.
 \item{mapped_to_MT} is the number of reads that mapped to mitochondrial
 genes.
 \item{total_count_per_cell} is the number of reads that mapped to exon
 after UMI deduplication. In contrast, `mapped_to_exon` is the 
 number of reads mapped to exon before UMI deduplication.
 \item{number_of_genes} is the number of genes detected for each cells
 \item{non_ERCC_percent} is 1 - (percentage of ERCC reads). Reads are
 UMI deduplicated.
 \item{non_mt_percent} is 1 - (percentage of mitochondrial reads).
 Reads are UMI deduplicated.
 \item{non_ribo_percent} is 1- (percentage of ribosomal reads).
 Reads are UMI deduplicated.
 }
}
\examples{
data("sc_sample_data")
data("sc_sample_qc")
sce = SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
organism(sce) = "mmusculus_gene_ensembl"
gene_id_type(sce) = "ensembl_gene_id"
QC_metrics(sce) = sc_sample_qc
head(QC_metrics(sce))
plot_mapping(sce,percentage=TRUE,dataname="sc_sample")

}
\author{
Luyi Tian
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_sample_data.R
\docType{data}
\name{sc_sample_data}
\alias{sc_sample_data}
\title{a small sample scRNA-seq counts dataset to 
demonstrate capabilities of scPipe}
\format{
a matrix instance, one row per gene.
}
\source{
Christin Biben (WEHI). She FACS sorted cells from several immune
cell types including B cells, granulocyte and some early progenitors.
}
\value{
NULL, but makes a matrix of count data
}
\description{
This data set contains counts for high variable genes for 
100 cells. The cells have different cell types. The data contains
raw read counts. The cells are chosen randomly from 384 cells and
they did not go through quality controls. The rows names are 
Ensembl gene ids and the columns are cell names, which is the wall
position in the 384 plates.
}
\examples{
# use the example dataset to perform quality control
data("sc_sample_data")
data("sc_sample_qc")
sce = SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
organism(sce) = "mmusculus_gene_ensembl"
gene_id_type(sce) = "ensembl_gene_id"
QC_metrics(sce) = sc_sample_qc
demultiplex_info(sce) = cell_barcode_matching
UMI_dup_info(sce) = UMI_duplication
sce = detect_outlier(sce)

plot_QC_pairs(sce)

}
\author{
Luyi Tian
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/qc.R
\name{detect_outlier}
\alias{detect_outlier}
\title{Detect outliers based on QC metrics}
\usage{
detect_outlier(
  sce,
  comp = 1,
  sel_col = NULL,
  type = c("low", "both", "high"),
  conf = c(0.9, 0.99),
  batch = FALSE
)
}
\arguments{
\item{sce}{a \code{SingleCellExperiment} object containing QC metrics.}

\item{comp}{the number of component used in GMM. Depending on the quality of
the experiment.}

\item{sel_col}{a vector of column names which indicate the columns to use
 for QC.
By default it will be the statistics generated by `calculate_QC_metrics()`}

\item{type}{only looking at low quality cells (`low`) or possible
doublets (`high`) or both (`both`)}

\item{conf}{confidence interval for linear regression at lower and
upper tails.Usually, this is smaller for lower tail because we hope
to pick out more low quality cells than doublets.}

\item{batch}{whether to perform quality control separately for each batch.
Default is FALSE. If set to TRUE then you should have a column called
`batch` in the `colData(sce)`.}
}
\value{
an updated \code{SingleCellExperiment} object with an `outlier`
column in \code{colData}
}
\description{
This algorithm will try to find \code{comp} number of components
  in quality control metrics using a Gaussian mixture model. Outlier
  detection is performed on the component with the most genes detected. The
  rest of the components will be considered poor quality cells. More cells
  will be classified low quality as you increase \code{comp}.
}
\details{
detect outlier using Mahalanobis distances
}
\examples{
data("sc_sample_data")
data("sc_sample_qc")
sce = SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
organism(sce) = "mmusculus_gene_ensembl"
gene_id_type(sce) = "ensembl_gene_id"
QC_metrics(sce) = sc_sample_qc
demultiplex_info(sce) = cell_barcode_matching
UMI_dup_info(sce) = UMI_duplication
# the sample qc data already run through function `calculate_QC_metrics`
# for a new sce please run `calculate_QC_metrics` before `detect_outlier`
sce = detect_outlier(sce)
table(QC_metrics(sce)$outliers)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/help.R
\docType{package}
\name{scPipe}
\alias{scPipe}
\alias{scPipe-package}
\title{scPipe - single cell RNA-seq pipeline}
\description{
The scPipe will do cell barcode demultiplexing, UMI deduplication and quality control on
fastq data generated from all protocols
}
\author{
Luyi Tian <tian.l@wehi.edu.au>; Shian Su <su.s@wehi.edu.au>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_workflow.R
\name{create_processed_report}
\alias{create_processed_report}
\title{create_processed_report}
\usage{
create_processed_report(
  outdir = ".",
  organism,
  gene_id_type,
  report_name = "report"
)
}
\arguments{
\item{outdir}{output folder.}

\item{organism}{the organism of the data. List of possible names can be retrieved using the function
`listDatasets`from `biomaRt` package. (e.g. `mmusculus_gene_ensembl` or `hsapiens_gene_ensembl`).}

\item{gene_id_type}{gene id type of the data A possible list of ids can be retrieved using the function `listAttributes` from `biomaRt` package.
the commonly used id types are `external_gene_name`, `ensembl_gene_id` or `entrezgene`.}

\item{report_name}{the name of the report .Rmd and .html files.}
}
\value{
file path of the created compiled document.
}
\description{
Create an HTML report summarising pro-processed data. This is an alternative to the more verbose \code{create_report} that requires only the processed counts and stats folders.
}
\examples{
\dontrun{
create_report(
       outdir="output_dir_of_scPipe",
       organism="mmusculus_gene_ensembl",
       gene_id_type="ensembl_gene_id")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_atac_pipeline.R
\name{sc_atac_pipeline_quick_test}
\alias{sc_atac_pipeline_quick_test}
\title{A function that tests the pipeline on a small test sample (without duplicate removal)}
\usage{
sc_atac_pipeline_quick_test()
}
\value{
None (invisible `NULL`)
}
\description{
A function that tests the pipeline on a small test sample (without duplicate removal)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_integrate.R
\name{sc_mae_plot_umap}
\alias{sc_mae_plot_umap}
\title{Generates UMAP of multiomic data}
\usage{
sc_mae_plot_umap(mae, by = NULL, output_file = NULL)
}
\arguments{
\item{mae}{The MultiAssayExperiment object}

\item{by}{What to colour the points by. Needs to be in colData of all experiments.}

\item{output_file}{The path of the output file}
}
\value{
A ggplot2 object representing the UMAP plot
}
\description{
Uses feature count data from multiple experiment objects to produce UMAPs for each assay and then overlay them on the same pair of axes
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_atac_tfidf.R
\name{sc_atac_tfidf}
\alias{sc_atac_tfidf}
\title{generating the UMAPs for sc-ATAC-Seq preprocessed data}
\usage{
sc_atac_tfidf(binary.mat, output_folder = NULL)
}
\arguments{
\item{binary.mat}{The final, filtered feature matrix in binary format}

\item{output_folder}{The path of the output folder}
}
\value{
None (invisible `NULL`)
}
\description{
Takes the binary matrix and generate a TF-IDF so the clutering can take place on the reduced dimentions.
}
\examples{
\dontrun{
sc_atac_tfidf(binary.mat = final_binary_matrix) 
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_workflow.R
\name{create_report}
\alias{create_report}
\title{create_report}
\usage{
create_report(
  sample_name,
  outdir,
  r1 = "NA",
  r2 = "NA",
  outfq = "NA",
  read_structure = list(bs1 = 0, bl1 = 0, bs2 = 0, bl2 = 0, us = 0, ul = 0),
  filter_settings = list(rmlow = TRUE, rmN = TRUE, minq = 20, numbq = 2),
  align_bam = "NA",
  genome_index = "NA",
  map_bam = "NA",
  exon_anno = "NA",
  stnd = TRUE,
  fix_chr = FALSE,
  barcode_anno = "NA",
  max_mis = 1,
  UMI_cor = 1,
  gene_fl = FALSE,
  organism,
  gene_id_type
)
}
\arguments{
\item{sample_name}{sample name}

\item{outdir}{output folder}

\item{r1}{file path of read1}

\item{r2}{file path of read2 default to be NULL}

\item{outfq}{file path of the output of \code{sc_trim_barcode}}

\item{read_structure}{a list contains read structure configuration. For more help see `?sc_trim_barcode`}

\item{filter_settings}{a list contains read filter settings for more help see `?sc_trim_barcode`}

\item{align_bam}{the aligned bam file}

\item{genome_index}{genome index used for alignment}

\item{map_bam}{the mapped bam file}

\item{exon_anno}{the gff exon annotation used. Can have multiple files}

\item{stnd}{whether to perform strand specific mapping}

\item{fix_chr}{add `chr` to chromosome names, fix inconsistent names.}

\item{barcode_anno}{cell barcode annotation file path.}

\item{max_mis}{maximum mismatch allowed in barcode. Default to be 1}

\item{UMI_cor}{correct UMI sequence error: 0 means no correction, 1 means simple correction and merge UMI with distance 1.}

\item{gene_fl}{whether to remove low abundant gene count. Low abundant is defined as only one copy of one UMI for this gene}

\item{organism}{the organism of the data. List of possible names can be retrieved using the function
`listDatasets`from `biomaRt` package. (i.e `mmusculus_gene_ensembl` or `hsapiens_gene_ensembl`)}

\item{gene_id_type}{gene id type of the data A possible list of ids can be retrieved using the function `listAttributes` from `biomaRt` package.
the commonly used id types are `external_gene_name`, `ensembl_gene_id` or `entrezgene`}
}
\value{
no return
}
\description{
create an HTML report using data generated by proprocessing step.
}
\examples{
\dontrun{
create_report(sample_name="sample_001",
       outdir="output_dir_of_scPipe",
       r1="read1.fq",
       r2="read2.fq",
       outfq="trim.fq",
       read_structure=list(bs1=-1, bl1=2, bs2=6, bl2=8, us=0, ul=6),
       filter_settings=list(rmlow=TRUE, rmN=TRUE, minq=20, numbq=2),
       align_bam="align.bam",
       genome_index="mouse.index",
       map_bam="aligned.mapped.bam",
       exon_anno="exon_anno.gff3",
       stnd=TRUE,
       fix_chr=FALSE,
       barcode_anno="cell_barcode.csv",
       max_mis=1,
       UMI_cor=1,
       gene_fl=FALSE,
       organism="mmusculus_gene_ensembl",
       gene_id_type="ensembl_gene_id")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_atac_create_fragments.R
\name{get_chromosomes}
\alias{get_chromosomes}
\title{Get Chromosomes}
\usage{
get_chromosomes(bam, keep_contigs = "^chr")
}
\arguments{
\item{bam}{file path to the bam file to get data from}

\item{keep_contigs}{regular expression used with grepl to filter reference names}
}
\value{
a named list where element names are chromosomes reference names and elements are integer lengths
}
\description{
Gets a list of NamedList of chromosomes and the reference length
acquired through the bam index file.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrapper_scPipeCPP.R
\name{sc_demultiplex}
\alias{sc_demultiplex}
\title{sc_demultiplex}
\usage{
sc_demultiplex(
  inbam,
  outdir,
  bc_anno,
  max_mis = 1,
  bam_tags = list(am = "YE", ge = "GE", bc = "BC", mb = "OX"),
  mito = "MT",
  has_UMI = TRUE,
  nthreads = 1
)
}
\arguments{
\item{inbam}{input bam file. This should be the output of
\code{sc_exon_mapping}}

\item{outdir}{output folder}

\item{bc_anno}{barcode annotation, first column is cell id, second column
is cell barcode sequence}

\item{max_mis}{maximum mismatch allowed in barcode. (default: 1)}

\item{bam_tags}{list defining BAM tags where mapping information is
stored.
\itemize{
  \item "am": mapping status tag
  \item "ge": gene id
  \item "bc": cell barcode tag
  \item "mb": molecular barcode tag
}}

\item{mito}{mitochondrial chromosome name.
This should be consistent with the chromosome names in the bam file.}

\item{has_UMI}{whether the protocol contains UMI (default: TRUE)}

\item{nthreads}{number of threads to use. (default: 1)}
}
\value{
no return
}
\description{
Process bam file by cell barcode,
output to outdir/count/[cell_id].csv.
the output contains information for all reads that can be mapped to exons.
including the gene id, UMI of that read and the distance to transcript end
position.
}
\examples{
data_dir="celseq2_demo"
barcode_annotation_fn = system.file("extdata", "barcode_anno.csv",
    package = "scPipe")
\dontrun{
# refer to the vignettes for the complete workflow
...
sc_demultiplex(file.path(data_dir, "out.map.bam"),
    data_dir,
    barcode_annotation_fn,has_UMI=FALSE)
...
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_atac_create_cell_qc_metrics.R
\name{sc_atac_create_cell_qc_metrics}
\alias{sc_atac_create_cell_qc_metrics}
\title{generating a file useful for producing the qc plots}
\usage{
sc_atac_create_cell_qc_metrics(
  frags_file,
  peaks_file,
  promoters_file,
  tss_file,
  enhs_file,
  output_folder
)
}
\arguments{
\item{frags_file}{The fragment file}

\item{peaks_file}{The peak file}

\item{promoters_file}{The path of the promoter annotation file}

\item{tss_file}{The path of the tss annotation file}

\item{enhs_file}{The path of the enhs annotation file}

\item{output_folder}{The path of the output folder for resultant files}
}
\value{
Nothing (Invisible 'NULL')
}
\description{
uses the peak file and annotation files for features
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_atac_cell_calling.R
\name{sc_atac_cell_calling}
\alias{sc_atac_cell_calling}
\title{identifying true vs empty cells}
\usage{
sc_atac_cell_calling(
  mat,
  cell_calling = "filter",
  output_folder,
  genome_size = NULL,
  cell_qc_metrics_file = NULL,
  lower = NULL,
  min_uniq_frags = 3000,
  max_uniq_frags = 50000,
  min_frac_peak = 0.3,
  min_frac_tss = 0,
  min_frac_enhancer = 0,
  min_frac_promoter = 0.1,
  max_frac_mito = 0.15
)
}
\arguments{
\item{mat}{the feature by cell matrix.}

\item{cell_calling}{the cell calling approach, possible options were "emptydrops" , "cellranger" and "filter".
But we opten to using "filter" as it was most robust. "emptydrops" is still an opition for data with large umber of cells.}

\item{output_folder}{output directory for the cell called matrix.}

\item{genome_size}{genome size for the data in feature by cell matrix.}

\item{cell_qc_metrics_file}{quality per barcode file for the barcodes in the matrix if using the \code{cellranger} or \code{filter} options.}

\item{lower}{the lower threshold for the data if using the \code{emptydrops} function for cell calling.}

\item{min_uniq_frags}{The minimum number of required unique fragments required for a cell (used for \code{filter} cell calling)}

\item{max_uniq_frags}{The maximum number of required unique fragments required for a cell (used for \code{filter} cell calling)}

\item{min_frac_peak}{The minimum proportion of fragments in a cell to overlap with a peak (used for \code{filter} cell calling)}

\item{min_frac_tss}{The minimum proportion of fragments in a cell to overlap with a tss (used for \code{filter} cell calling)}

\item{min_frac_enhancer}{The minimum proportion of fragments in a cell to overlap with a enhancer sequence (used for \code{filter} cell calling)}

\item{min_frac_promoter}{The minimum proportion of fragments in a cell to overlap with a promoter sequence (used for \code{filter} cell calling)}

\item{max_frac_mito}{The maximum proportion of fragments in a cell that are mitochondrial (used for \code{filter} cell calling)}
}
\description{
the methods to call true cells are of various ways.
implement (i.e. \code{filtering} from \code{scATAC-Pro} as default
}
\examples{
\dontrun{
sc_atac_cell_calling <- function(mat,
 cell_calling,
 output_folder,
 genome_size    = NULL,
 cell_qc_metrics_file = NULL,
 lower          = NULL)
}


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/biomart_methods.R
\name{convert_geneid}
\alias{convert_geneid}
\title{convert the gene ids of a SingleCellExperiment object}
\usage{
convert_geneid(sce, returns = "external_gene_name", all = TRUE)
}
\arguments{
\item{sce}{a SingleCellExperiment object}

\item{returns}{the gene id which is set as return. Default to be `external_gene_name`
A possible list of attributes can be retrieved using the
function \code{listAttributes} from \code{biomaRt} package. The commonly used
id types are `external_gene_name`, `ensembl_gene_id` or `entrezgene`.}

\item{all}{logic. For genes that cannot covert to new gene id, keep them with the old
id or delete them. The default is keep them.}
}
\value{
sce with converted id
}
\description{
convert the gene ids of a SingleCellExperiment object
}
\details{
convert the gene id of all datas in the SingleCellExperiment object
}
\examples{
# the gene id in example data are `external_gene_name`
# the following example will convert it to `external_gene_name`.
data("sc_sample_data")
data("sc_sample_qc")
sce = SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
organism(sce) = "mmusculus_gene_ensembl"
gene_id_type(sce) = "ensembl_gene_id"
QC_metrics(sce) = sc_sample_qc
demultiplex_info(sce) = cell_barcode_matching
UMI_dup_info(sce) = UMI_duplication
head(rownames(sce))
sce = convert_geneid(sce, return="external_gene_name")
head(rownames(sce))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_atac_create_fragments.R
\name{read_cells}
\alias{read_cells}
\title{Read Cell barcode file}
\usage{
read_cells(cells)
}
\arguments{
\item{cells}{the file path to the barcode file. Assumes
one barcode per line or barcode csv.
Or, cells can be a comma delimited string of barcodes}
}
\value{
a character vector of the provided barcodes
}
\description{
Read Cell barcode file
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrapper_scPipeCPP.R
\name{sc_detect_bc}
\alias{sc_detect_bc}
\title{sc_detect_bc}
\usage{
sc_detect_bc(
  infq,
  outcsv,
  prefix = "CELL_",
  bc_len,
  max_reads = 1e+06,
  min_count = 10,
  number_of_cells = 10000,
  max_mismatch = 1,
  white_list_file = NULL
)
}
\arguments{
\item{infq}{input fastq file, should be the output file of
\code{sc_trim_barcode}}

\item{outcsv}{output barcode annotation}

\item{prefix}{the prefix of cell name (default: `CELL_`)}

\item{bc_len}{the length of cell barcode, should be consistent with bl1+bl2
in \code{sc_trim_barcode}}

\item{max_reads}{the maximum of reads processed (default: 1,000,000)}

\item{min_count}{minimum counts to keep, barcode will be discarded if
it has lower count. Default value is 10. This should be set according
to \code{max_reads}.}

\item{number_of_cells}{number of cells kept in result. (default: 10000)}

\item{max_mismatch}{the maximum mismatch allowed. Barcodes within this
number will be considered as sequence error and merged. (default: 1)}

\item{white_list_file}{a file that list all the possible barcodes
each row is a barcode sequence. the list for 10x can be found at:
https://community.10xgenomics.com/t5/Data-Sharing/List-of-valid-cellular-barcodes/td-p/527
(default: NULL)}
}
\value{
no return
}
\description{
Detect cell barcode and generate the barcode annotation
}
\examples{
\dontrun{
# `sc_detect_bc`` should run before `sc_demultiplex` for
# Drop-seq or 10X protocols
sc_detect_bc("input.fastq","output.cell_index.csv",bc_len=8)
sc_demultiplex(...,"output.cell_index.csv")
}


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/qc.R
\name{remove_outliers}
\alias{remove_outliers}
\title{Remove outliers in \code{SingleCellExperiment}}
\usage{
remove_outliers(sce)
}
\arguments{
\item{sce}{a \code{SingleCellExperiment} object}
}
\value{
a \code{SingleCellExperiment} object without outliers
}
\description{
Removes outliers flagged by \code{detect_outliers()}
}
\examples{
data("sc_sample_data")
data("sc_sample_qc")
sce = SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
organism(sce) = "mmusculus_gene_ensembl"
gene_id_type(sce) = "ensembl_gene_id"
QC_metrics(sce) = sc_sample_qc
demultiplex_info(sce) = cell_barcode_matching
UMI_dup_info(sce) = UMI_duplication
sce = detect_outlier(sce)
dim(sce)
sce = remove_outliers(sce)
dim(sce)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/qc.R
\name{plot_QC_pairs}
\alias{plot_QC_pairs}
\title{Plot GGAlly pairs plot of QC statistics from \code{SingleCellExperiment}
object}
\usage{
plot_QC_pairs(sce, sel_col = NULL)
}
\arguments{
\item{sce}{a \code{SingleCellExperiment} object}

\item{sel_col}{a vector of column names which indicate the columns to use for
plot. By default it will be the statistics generated by
`calculate_QC_metrics()`}
}
\value{
a ggplot2 object
}
\description{
Plot GGAlly pairs plot of QC statistics from \code{SingleCellExperiment}
object
}
\examples{
data("sc_sample_data")
data("sc_sample_qc")
sce = SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
organism(sce) = "mmusculus_gene_ensembl"
gene_id_type(sce) = "ensembl_gene_id"
QC_metrics(sce) = sc_sample_qc
demultiplex_info(sce) = cell_barcode_matching
UMI_dup_info(sce) = UMI_duplication
sce = detect_outlier(sce)

plot_QC_pairs(sce)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/anno_import.R
\name{anno_import}
\alias{anno_import}
\title{Import gene annotation}
\usage{
anno_import(filename)
}
\arguments{
\item{filename}{The name of the annotation gff3 or gtf file. File can be
gzipped.}
}
\value{
data.frame containing exon information in SAF format
}
\description{
Because of the variations in data format depending on annotation
  source, this function has only been tested with human annotation from
  ENSEMBL, RefSeq and Gencode. If it behaves unexpectedly with any annotation
  please submit an issue at www.github.com/LuyiTian/scPipe with details.
}
\details{
Imports and GFF3 or GTF gene annotation file and transforms it into a SAF
formatted data.frame. SAF described at
http://bioinf.wehi.edu.au/featureCounts/. SAF contains positions for exons,
strand and the GeneID they are associated with.
}
\examples{
ens_chrY <- anno_import(system.file("extdata", "ensembl_hg38_chrY.gtf.gz", package = "scPipe"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_atac_tfidf.R
\name{sc_get_umap_data}
\alias{sc_get_umap_data}
\title{Generates UMAP data from sce object}
\usage{
sc_get_umap_data(sce, n_neighbours = 30)
}
\arguments{
\item{sce}{The SingleCellExperiment object}

\item{n_neighbours}{No. of neighbours for KNN}
}
\value{
A dataframe containing the UMAP dimensions, as well as all the colData of the sce object for each cell
}
\description{
Produces a DataFrame containing the UMAP dimensions, as well as all the colData of the sce object for each cell
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_aligning.R
\name{sc_aligning}
\alias{sc_aligning}
\title{aligning the demultiplexed FASTQ reads using the Rsubread:align()}
\usage{
sc_aligning(
  R1,
  R2 = NULL,
  tech = "atac",
  index_path = NULL,
  ref = NULL,
  output_folder = NULL,
  output_file = NULL,
  input_format = "FASTQ",
  output_format = "BAM",
  type = "dna",
  nthreads = 1
)
}
\arguments{
\item{R1}{a mandatory character vector including names of files that include sequence reads to be aligned. For paired-end reads, this gives the list of files including first reads in each library. File format is FASTQ/FASTA by default.}

\item{R2}{a character vector, the second fastq file, which is required if the data is paired-end}

\item{tech}{a character string giving the sequencing technology. Possible value includes "atac" or "rna"}

\item{index_path}{character string specifying the path/basename of the index files, if the Rsubread genome build is available}

\item{ref}{a character string specifying the path to reference genome file (.fasta, .fa format)}

\item{output_folder}{a character string, the name of the output folder}

\item{output_file}{a character vector specifying names of output files. By default, names of output files are set as the file names provided in R1 added with an suffix string}

\item{input_format}{a string indicating the input format}

\item{output_format}{a string indicating the output format}

\item{type}{type of sequencing data (`RNA` or `DNA`)}

\item{nthreads}{numeric value giving the number of threads used for mapping.}
}
\value{
the file path of the output aligned BAM file
}
\description{
after we run the \code{sc_trim_barcode} or \code{sc_atac_trim_barcode} to demultiplex the fastq files, we are using this
function to align those fastq files to a known reference.
}
\examples{
\dontrun{
sc_aligning(index_path,
    tech = 'atac',
    R1,  
    R2, 
    nthreads  = 6) 
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R, R/SCE_methods.R
\docType{methods}
\name{feature_type}
\alias{feature_type}
\alias{feature_type<-}
\alias{feature_type.sce}
\alias{feature_type,SingleCellExperiment-method}
\alias{feature_type<-,SingleCellExperiment-method}
\title{Get or set \code{feature_type} from a SingleCellExperiment object}
\usage{
feature_type(object)

feature_type(object) <- value

feature_type.sce(object)

\S4method{feature_type}{SingleCellExperiment}(object)

\S4method{feature_type}{SingleCellExperiment}(object) <- value
}
\arguments{
\item{object}{A \code{\link{SingleCellExperiment}} object.}

\item{value}{Value to be assigned to corresponding object.}
}
\value{
the feature type used in feature counting for scATAC-Seq data

A string representing the feature type
}
\description{
Get or set \code{feature_type} from a SingleCellExperiment object
}
\author{
Shani Amarasinghe
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_atac_create_report.R
\name{sc_atac_create_report}
\alias{sc_atac_create_report}
\title{HTML report generation}
\usage{
sc_atac_create_report(
  input_folder,
  output_folder = NULL,
  organism = NULL,
  sample_name = NULL,
  feature_type = NULL,
  n_barcode_subset = 500
)
}
\arguments{
\item{input_folder}{The path of the folder produced by the pipeline}

\item{output_folder}{The path of the output folder to store the HTML report in}

\item{organism}{A string indicating the name of the organism being analysed}

\item{sample_name}{A string indicating the name of the sample}

\item{feature_type}{A string indicating the type of the feature (`genome_bin` or `peak`)}

\item{n_barcode_subset}{if you require only to visualise stats for a sample of barcodes to improve processing time (integer)}
}
\value{
the path of the output file
}
\description{
Generates a HTML report using the output folder produced by the pipeline
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/qc.R
\name{plot_mapping}
\alias{plot_mapping}
\title{Plot mapping statistics for \code{SingleCellExperiment} object.}
\usage{
plot_mapping(sce, sel_col = NULL, percentage = FALSE, dataname = "")
}
\arguments{
\item{sce}{a \code{SingleCellExperiment} object}

\item{sel_col}{a vector of column names, indicating the columns to use for
plot. by default it will be the mapping result.}

\item{percentage}{TRUE to convert the number of reads to percentage}

\item{dataname}{the name of this dataset, used as plot title}
}
\value{
a ggplot2 object
}
\description{
Plot mapping statistics for \code{SingleCellExperiment} object.
}
\examples{
data("sc_sample_data")
data("sc_sample_qc")
sce = SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
organism(sce) = "mmusculus_gene_ensembl"
gene_id_type(sce) = "ensembl_gene_id"
QC_metrics(sce) = sc_sample_qc
demultiplex_info(sce) = cell_barcode_matching
UMI_dup_info(sce) = UMI_duplication

plot_mapping(sce,percentage=TRUE,dataname="sc_sample")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R, R/SCE_methods.R
\docType{methods}
\name{QC_metrics}
\alias{QC_metrics}
\alias{QC_metrics<-}
\alias{QC_metrics.sce}
\alias{QC_metrics,SingleCellExperiment-method}
\alias{QC_metrics<-,SingleCellExperiment-method}
\title{Get or set quality control metrics in a SingleCellExperiment object}
\usage{
QC_metrics(object)

QC_metrics(object) <- value

QC_metrics.sce(object)

\S4method{QC_metrics}{SingleCellExperiment}(object)

\S4method{QC_metrics}{SingleCellExperiment}(object) <- value
}
\arguments{
\item{object}{A \code{\link{SingleCellExperiment}} object.}

\item{value}{Value to be assigned to corresponding object.}
}
\value{
a dataframe of quality control matrics

A DataFrame of quality control metrics.
}
\description{
Get or set quality control metrics in a SingleCellExperiment object
}
\examples{
data("sc_sample_data")
data("sc_sample_qc")
sce = SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
QC_metrics(sce) = sc_sample_qc

head(QC_metrics(sce))

}
\author{
Luyi Tian
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/anno_import.R
\name{anno_to_saf}
\alias{anno_to_saf}
\title{Convert annotation from GenomicRanges to Simple Annotation Format (SAF)}
\usage{
anno_to_saf(anno)
}
\arguments{
\item{anno}{The GRanges object containing exon information}
}
\value{
data.frame containing exon information in SAF format
}
\description{
This function converts a GRanges object into a data.frame of the
     SAF format for scPipe's consumption. The GRanges object should contain a
     "type" column where at least some features are annotated as "exon", in
     addition there should be a gene_id column specifying the gene to which the
     exon belongs. In the SAF only the gene ID, chromosome, start, end and
     strand are recorded, this is a gene-exon centric format, with all entries
     containing the same gene ID treated as exons of that gene. It is possible
     to count alternative features by setting the gene_id column to an arbitrary
     feature name and having alternative features in the SAF table, the main
     caveat is that the features are still treated as exons, and the mapping
     statistics for exon and intron will not reflect biological exons and
     introns but rather the annotation features.
}
\details{
Convert a GRanges object containing type and gene_id information into a SAF
format data.frame. SAF described at http://bioinf.wehi.edu.au/featureCounts/.
SAF contains positions for exons, strand and the GeneID they are associated
with.
}
\examples{
 \dontrun{
     anno <- system.file("extdata", "ensembl_hg38_chrY.gtf.gz", package = "scPipe")
     saf_chrY <- anno_to_saf(rtracklayer::import(anno))
 }

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_atac_pipeline.R
\name{sc_atac_pipeline}
\alias{sc_atac_pipeline}
\title{A convenient function for running the entire pipeline}
\usage{
sc_atac_pipeline(
  r1,
  r2,
  bc_file,
  valid_barcode_file = "",
  id1_st = -0,
  id1_len = 16,
  id2_st = 0,
  id2_len = 16,
  rmN = TRUE,
  rmlow = TRUE,
  organism = NULL,
  reference = NULL,
  feature_type = NULL,
  remove_duplicates = FALSE,
  samtools_path = NULL,
  genome_size = NULL,
  bin_size = NULL,
  yieldsize = 1e+06,
  exclude_regions = TRUE,
  excluded_regions_filename = NULL,
  fix_chr = "none",
  lower = NULL,
  cell_calling = "filter",
  promoters_file = NULL,
  tss_file = NULL,
  enhs_file = NULL,
  gene_anno_file = NULL,
  min_uniq_frags = 3000,
  max_uniq_frags = 50000,
  min_frac_peak = 0.3,
  min_frac_tss = 0,
  min_frac_enhancer = 0,
  min_frac_promoter = 0.1,
  max_frac_mito = 0.15,
  report = TRUE,
  nthreads = 12,
  output_folder = NULL
)
}
\arguments{
\item{r1}{The first read fastq file}

\item{r2}{The second read fastq file}

\item{bc_file}{the barcode information, can be either in a \code{fastq} format (e.g. from 10x-ATAC) or
from a \code{.csv} file (here the barcode is expected to be on the second column). 
Currently, for the fastq approach, this can be a list of barcode files.}

\item{valid_barcode_file}{optional file path of the valid (expected) barcode sequences to be found in the bc_file (.txt, can be txt.gz). 
Must contain one barcode per line on the second column separated by a comma (default ="").
If given, each barcode from bc_file is matched against the barcode of
best fit (allowing a hamming distance of 1). If a FASTQ \code{bc_file} is provided, barcodes with a higher mapping quality, as given by
the fastq reads quality score are prioritised.}

\item{id1_st}{barcode start position (0-indexed) for read 1, which is an extra parameter that is needed if the
\code{bc_file} is in a \code{.csv} format.}

\item{id1_len}{barcode length for read 1, which is an extra parameter that is needed if the
\code{bc_file} is in a \code{.csv} format.}

\item{id2_st}{barcode start position (0-indexed) for read 2, which is an extra parameter that is needed if the
\code{bc_file} is in a \code{.csv} format.}

\item{id2_len}{barcode length for read 2, which is an extra parameter that is needed if the
\code{bc_file} is in a \code{.csv} format.}

\item{rmN}{ogical, whether to remove reads that contains N in UMI or cell barcode.}

\item{rmlow}{logical, whether to remove reads that have low quality barcode sequences.}

\item{organism}{The name of the organism e.g. hg38}

\item{reference}{The reference genome file}

\item{feature_type}{The feature type (either `genome_bin` or `peak`)}

\item{remove_duplicates}{Whether or not to remove duplicates (samtools is required)}

\item{samtools_path}{A custom path of samtools to use for duplicate removal}

\item{genome_size}{The size of the genome (used for the \code{cellranger} cell calling method)}

\item{bin_size}{The size of the bins for feature counting with the `genome_bin` feature type}

\item{yieldsize}{The number of reads to read in for feature counting}

\item{exclude_regions}{Whether or not the regions should be excluded}

\item{excluded_regions_filename}{The filename of the file containing the regions to be excluded}

\item{fix_chr}{Specify `none`, `exclude_regions`, `feature` or `both` to prepend the string "chr" to the start of the associated file}

\item{lower}{the lower threshold for the data if using the \code{emptydrops} function for cell calling.}

\item{cell_calling}{The desired cell calling method either \code{cellranger}, \code{emptydrops} or  \code{filter}}

\item{promoters_file}{The path of the promoter annotation file (if the specified organism isn't recognised)}

\item{tss_file}{The path of the tss annotation file (if the specified organism isn't recognised)}

\item{enhs_file}{The path of the enhs annotation file (if the specified organism isn't recognised)}

\item{gene_anno_file}{The path of the gene annotation file (gtf or gff3 format)}

\item{min_uniq_frags}{The minimum number of required unique fragments required for a cell (used for \code{filter} cell calling)}

\item{max_uniq_frags}{The maximum number of required unique fragments required for a cell (used for \code{filter} cell calling)}

\item{min_frac_peak}{The minimum proportion of fragments in a cell to overlap with a peak (used for \code{filter} cell calling)}

\item{min_frac_tss}{The minimum proportion of fragments in a cell to overlap with a tss (used for \code{filter} cell calling)}

\item{min_frac_enhancer}{The minimum proportion of fragments in a cell to overlap with a enhancer sequence (used for \code{filter} cell calling)}

\item{min_frac_promoter}{The minimum proportion of fragments in a cell to overlap with a promoter sequence (used for \code{filter} cell calling)}

\item{max_frac_mito}{The maximum proportion of fragments in a cell that are mitochondrial (used for \code{filter} cell calling)}

\item{report}{Whether or not a HTML report should be produced}

\item{nthreads}{The number of threads to use for alignment (sc_align) and demultiplexing (sc_atac_bam_tagging)}

\item{output_folder}{The path of the output folder}
}
\value{
None (invisible `NULL`)
}
\description{
A convenient function for running the entire pipeline
}
\examples{
data.folder <- system.file("extdata", package = "scPipe", mustWork = TRUE)
r1      <- file.path(data.folder, "small_chr21_R1.fastq.gz") 
r2      <- file.path(data.folder, "small_chr21_R3.fastq.gz") 

# Using a barcode fastq file:

# barcodes in fastq format
barcode_fastq      <- file.path(data.folder, "small_chr21_R2.fastq.gz") 

\dontrun{
sc_atac_pipeline(
  r1 = r1,
  r2 = r2,
  bc_file = barcode_fastq
)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/qc.R
\name{calculate_QC_metrics}
\alias{calculate_QC_metrics}
\title{Calculate QC metrics from gene count matrix}
\usage{
calculate_QC_metrics(sce)
}
\arguments{
\item{sce}{a \code{SingleCellExperiment} object containing gene counts}
}
\value{
an \code{SingleCellExperiment} with updated QC metrics
}
\description{
Calculate QC metrics from gene count matrix
}
\details{
get QC metrics using gene count matrix.
The QC statistics added are \itemize{
 \item{number_of_genes} number of genes detected.
 \item{total_count_per_cell} sum of read number after UMI deduplication.
 \item{non_mt_percent} 1 - percentage of mitochondrial gene counts.
 Mitochondrial genes are retrived by GO term GO:0005739
 \item{non_ERCC_percent} ratio of exon counts to ERCC counts
 \item{non_ribo_percent} 1 - percentage of ribosomal gene counts
ribosomal genes are retrived by GO term GO:0005840.
}
}
\examples{
data("sc_sample_data")
data("sc_sample_qc")
sce <- SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
organism(sce) <- "mmusculus_gene_ensembl"
gene_id_type(sce) <- "ensembl_gene_id"
QC_metrics(sce) <- sc_sample_qc
demultiplex_info(sce) <- cell_barcode_matching
UMI_dup_info(sce) <- UMI_duplication

# The sample qc data already run through function `calculate_QC_metrics`.
# So we delete these columns and run `calculate_QC_metrics` to get them again:
colnames(colnames(QC_metrics(sce)))
QC_metrics(sce) <- QC_metrics(sce)[,c("unaligned","aligned_unmapped","mapped_to_exon")]
sce = calculate_QC_metrics(sce)
colnames(QC_metrics(sce))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrapper_scPipeCPP.R
\name{sc_gene_counting}
\alias{sc_gene_counting}
\title{sc_gene_counting}
\usage{
sc_gene_counting(outdir, bc_anno, UMI_cor = 2, gene_fl = FALSE)
}
\arguments{
\item{outdir}{output folder containing \code{sc_demultiplex} output}

\item{bc_anno}{barcode annotation comma-separated-values, first column is
cell id, second column is cell barcode sequence}

\item{UMI_cor}{correct UMI sequencing error: 0 means no correction, 1 means
simple correction and merge UMI with distance 1. 2 means merge on both UMI
alignment position match.}

\item{gene_fl}{whether to remove low abundance genes. A gene is considered to
have low abundance if only one copy of one UMI is associated with it.}
}
\value{
no return
}
\description{
Generate gene counts matrix with UMI deduplication
}
\examples{
data_dir="celseq2_demo"
barcode_annotation_fn = system.file("extdata", "barcode_anno.csv",
package = "scPipe")
\dontrun{
# refer to the vignettes for the complete workflow
...
sc_gene_counting(data_dir, barcode_annotation_fn)
...
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrapper_scPipeCPP.R
\name{sc_exon_mapping}
\alias{sc_exon_mapping}
\title{sc_exon_mapping}
\usage{
sc_exon_mapping(
  inbam,
  outbam,
  annofn,
  bam_tags = list(am = "YE", ge = "GE", bc = "BC", mb = "OX"),
  bc_len = 8,
  barcode_vector = "",
  UMI_len = 6,
  stnd = TRUE,
  fix_chr = FALSE,
  nthreads = 1
)
}
\arguments{
\item{inbam}{input aligned bam file. can have multiple files as input}

\item{outbam}{output bam filename}

\item{annofn}{single string or vector of gff3 annotation filenames,
data.frame in SAF format or GRanges object containing complete gene_id
metadata column.}

\item{bam_tags}{list defining BAM tags where mapping information is
stored.
\itemize{
  \item "am": mapping status tag
  \item "ge": gene id
  \item "bc": cell barcode tag
  \item "mb": molecular barcode tag
}}

\item{bc_len}{total barcode length}

\item{barcode_vector}{a list of barcode if each individual bam is a single
cell. (default: NULL). The barcode should be of the same length for each
cell.}

\item{UMI_len}{UMI length}

\item{stnd}{TRUE to perform strand specific mapping. (default: TRUE)}

\item{fix_chr}{TRUE to add `chr` to chromosome names, MT to chrM. (default: FALSE)}

\item{nthreads}{number of threads to use. (default: 1)}
}
\value{
generates a bam file with exons assigned
}
\description{
Map aligned reads to exon annotation.
The result will be written into optional fields in bam file with different
tags. Following this link for more information regarding to bam file format:
http://samtools.github.io/hts-specs

The function can accept multiple bam file as input, if multiple bam file is
provided and the `bc_len` is zero, then the function will use the barcode in
the `barcode_vector` to insert into the `bc` bam tag. So the length of
`barcode_vector` and the length of `inbam` should be the same
If this is the case then the `max_mis` argument in `sc_demultiplex`
should be zero. If `be_len` is larger than zero, then the function will
still seek for barcode in fastq headers with given length. In this case
each bam file is not treated as from a single cell.
}
\examples{
data_dir="celseq2_demo"
ERCCanno_fn = system.file("extdata", "ERCC92_anno.gff3",
    package = "scPipe")
\dontrun{
# for the complete workflow, refer to the vignettes
...
sc_exon_mapping(file.path(data_dir, "out.aln.bam"),
                file.path(data_dir, "out.map.bam"),
                ERCCanno_fn)
...
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R, R/SCE_methods.R
\docType{methods}
\name{UMI_dup_info}
\alias{UMI_dup_info}
\alias{UMI_dup_info<-}
\alias{UMI_dup_info.sce}
\alias{UMI_dup_info,SingleCellExperiment-method}
\alias{UMI_dup_info<-,SingleCellExperiment-method}
\title{Get or set UMI duplication results in a SingleCellExperiment object}
\usage{
UMI_dup_info(object)

UMI_dup_info(object) <- value

UMI_dup_info.sce(object)

\S4method{UMI_dup_info}{SingleCellExperiment}(object)

\S4method{UMI_dup_info}{SingleCellExperiment}(object) <- value
}
\arguments{
\item{object}{A \code{\link{SingleCellExperiment}} object.}

\item{value}{Value to be assigned to corresponding object.}
}
\value{
a dataframe of cell UMI duplication information

A DataFrame of UMI duplication results.
}
\description{
Get or set UMI duplication results in a SingleCellExperiment object
}
\examples{
data("sc_sample_data")
data("sc_sample_qc")
sce = SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
organism(sce) = "mmusculus_gene_ensembl"
gene_id_type(sce) = "ensembl_gene_id"
QC_metrics(sce) = sc_sample_qc
demultiplex_info(sce) = cell_barcode_matching
UMI_dup_info(sce) = UMI_duplication

head(UMI_dup_info(sce))

}
\author{
Luyi Tian
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{get_ercc_anno}
\alias{get_ercc_anno}
\title{Get ERCC annotation table}
\usage{
get_ercc_anno()
}
\value{
data.frame containing ERCC annotation
}
\description{
Helper function to retrieve ERCC annotation as a dataframe in SAF format
}
\examples{
ercc_anno <- get_ercc_anno()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_sample_qc.R
\docType{data}
\name{UMI_duplication}
\alias{UMI_duplication}
\title{UMI duplication statistics for a small sample scRNA-seq dataset to 
demonstrate capabilities of scPipe}
\format{
a data.frame instance, one row per cell.
}
\source{
Christin Biben (WEHI). She FACS sorted cells from several immune
cell types including B cells, granulocyte and some early progenitors.
}
\value{
NULL, but makes a data frame with UMI dulication statistics
}
\description{
This data.frame contains UMI duplication statistics, where the
first column is the number of duplication, and the second column is the
count of UMIs.
}
\examples{
data("sc_sample_data")
data("sc_sample_qc")
sce = SingleCellExperiment(assays = list(counts =as.matrix(sc_sample_data)))
organism(sce) = "mmusculus_gene_ensembl"
gene_id_type(sce) = "ensembl_gene_id"
QC_metrics(sce) = sc_sample_qc
demultiplex_info(sce) = cell_barcode_matching
UMI_dup_info(sce) = UMI_duplication

head(UMI_dup_info(sce))

}
\author{
Luyi Tian
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_atac_cell_calling.R
\name{sc_atac_emptydrops_cell_calling}
\alias{sc_atac_emptydrops_cell_calling}
\title{empty drops cell calling}
\usage{
sc_atac_emptydrops_cell_calling(mat, output_folder, lower = NULL)
}
\arguments{
\item{mat}{The input matrix}

\item{output_folder}{The path of the output folder}

\item{lower}{The lower threshold for the data if using the \code{emptydrops} function for cell calling.}
}
\description{
The empty drops cell calling method
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/qc.R
\name{plot_demultiplex}
\alias{plot_demultiplex}
\title{plot_demultiplex}
\usage{
plot_demultiplex(sce)
}
\arguments{
\item{sce}{a \code{SingleCellExperiment} object}
}
\value{
a ggplot2 bar chart
}
\description{
Plot cell barcode demultiplexing result for the
 \code{SingleCellExperiment}. The barcode demultiplexing result is shown
 using a barplot, with the bars indicating proportions of total reads.
 Barcode matches and mismatches are summarised along with whether or not the
 read mapped to the genome. High proportion of genome aligned reads with no
 barcode match may indicate barcode integration failure.
}
\examples{
data("sc_sample_data")
data("sc_sample_qc")
sce = SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
organism(sce) = "mmusculus_gene_ensembl"
gene_id_type(sce) = "ensembl_gene_id"
QC_metrics(sce) = sc_sample_qc
demultiplex_info(sce) = cell_barcode_matching
UMI_dup_info(sce) = UMI_duplication

plot_demultiplex(sce)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_atac_trim_barcode.R
\name{sc_atac_trim_barcode}
\alias{sc_atac_trim_barcode}
\title{demultiplex raw single-cell ATAC-Seq fastq reads}
\usage{
sc_atac_trim_barcode(
  r1,
  r2,
  bc_file = NULL,
  valid_barcode_file = "",
  output_folder = "",
  umi_start = 0,
  umi_length = 0,
  umi_in = "both",
  rmN = FALSE,
  rmlow = FALSE,
  min_qual = 20,
  num_below_min = 2,
  id1_st = -0,
  id1_len = 16,
  id2_st = 0,
  id2_len = 16,
  no_reverse_complement = FALSE
)
}
\arguments{
\item{r1}{read one for pair-end reads.}

\item{r2}{read two for pair-end reads, NULL if single read.}

\item{bc_file}{the barcode information, can be either in a \code{fastq} format (e.g. from 10x-ATAC) or
from a \code{.csv} file (here the barcode is expected to be on the second column). 
Currently, for the fastq approach, this can be a list of barcode files.}

\item{valid_barcode_file}{optional file path of the valid (expected) barcode sequences to be found in the bc_file (.txt, can be txt.gz). 
Must contain one barcode per line on the second column separated by a comma (default ="").
If given, each barcode from bc_file is matched against the barcode of
best fit (allowing a hamming distance of 1). If a FASTQ \code{bc_file} is provided, barcodes with a higher mapping quality, as given by
the fastq reads quality score are prioritised.}

\item{output_folder}{the output dir for the demultiplexed fastq file, which will contain 
fastq files with reformatted barcode and UMI into the read name. 
Files ending in \code{.gz} will be automatically compressed.}

\item{umi_start}{if available, the start position of the molecular identifier.}

\item{umi_length}{if available, the start position of the molecular identifier.}

\item{umi_in}{umi_in}

\item{rmN}{logical, whether to remove reads that contains N in UMI or cell barcode.}

\item{rmlow}{logical, whether to remove reads that have low quality barcode sequences}

\item{min_qual}{the minimum base pair quality that is allowed (default = 20).}

\item{num_below_min}{the maximum number of base pairs below the quality threshold.}

\item{id1_st}{barcode start position (0-indexed) for read 1, which is an extra parameter that is needed if the
\code{bc_file} is in a \code{.csv} format.}

\item{id1_len}{barcode length for read 1, which is an extra parameter that is needed if the
\code{bc_file} is in a \code{.csv} format.}

\item{id2_st}{barcode start position (0-indexed) for read 2, which is an extra parameter that is needed if the
\code{bc_file} is in a \code{.csv} format.}

\item{id2_len}{barcode length for read 2, which is an extra parameter that is needed if the
\code{bc_file} is in a \code{.csv} format.}

\item{no_reverse_complement}{specifies if the reverse complement of the barcode sequence should be 
used for barcode error correction (only when barcode sequences are provided as fastq files). FALSE (default)
lets the function decide whether to use reverse complement, and TRUE forces the function to
use the forward barcode sequences.}
}
\value{
None (invisible `NULL`)
}
\description{
single-cell data need to be demultiplexed in order to retain the information of the cell barcodes
the data belong to. Here we reformat fastq files so barcode/s (and if available the UMI sequences) are moved from
the sequence into the read name. Since scATAC-Seq data are mostly paired-end, both `r1` and `r2` are demultiplexed in this function.
}
\examples{
data.folder <- system.file("extdata", package = "scPipe", mustWork = TRUE)
r1      <- file.path(data.folder, "small_chr21_R1.fastq.gz") 
r2      <- file.path(data.folder, "small_chr21_R3.fastq.gz") 

# Using a barcode fastq file:

# barcodes in fastq format
barcode_fastq      <- file.path(data.folder, "small_chr21_R2.fastq.gz") 

sc_atac_trim_barcode (
r1            = r1, 
r2            = r2, 
bc_file       = barcode_fastq,
rmN           = TRUE,
rmlow         = TRUE,
output_folder = tempdir())

# Using a barcode csv file:

# barcodes in .csv format
barcode_1000       <- file.path(data.folder, "chr21_modified_barcode_1000.csv")

\dontrun{
sc_atac_trim_barcode (
r1            = r1, 
r2            = r2, 
bc_file       = barcode_1000, 
id1_st        = 0,
rmN           = TRUE,
rmlow         = TRUE,
output_folder = tempdir())
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_integrate.R
\name{TF.IDF.custom}
\alias{TF.IDF.custom}
\title{Returns the TF-IDF normalised version of a binary matrix}
\usage{
TF.IDF.custom(binary.mat, verbose = TRUE)
}
\arguments{
\item{binary.mat}{The binary matrix}

\item{verbose}{boolean flag to print status messages}
}
\value{
Returns the TF-IDF normalised version of a binary matrix
}
\description{
Returns the TF-IDF normalised version of a binary matrix
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_sample_qc.R
\docType{data}
\name{cell_barcode_matching}
\alias{cell_barcode_matching}
\title{cell barcode demultiplex statistics for a small sample scRNA-seq 
dataset to demonstrate capabilities of scPipe}
\format{
a data.frame instance, one row per cell.
}
\source{
Christin Biben (WEHI). She FACS sorted cells from several immune
cell types including B cells, granulocyte and some early progenitors.
}
\value{
NULL, but makes a data frame with cell barcode demultiplex statistics
}
\description{
This data.frame contains cell barcode demultiplex statistics 
with several rows:\itemize{
\item{barcode_unmatch_ambiguous_mapping} is the number of reads that do not
match any barcode, but aligned to the genome and mapped to multiple 
features.
\item{barcode_unmatch_mapped_to_intron} is the number of reads that do
not match any barcode, but aligned to the genome and mapped to intron.
\item{barcode_match} is the number of reads that match the cell barcodes
\item{barcode_unmatch_unaligned} is the number of reads that do not 
match any barcode, and not aligned to the genome
\item{barcode_unmatch_aligned} is the number of reads that do not 
match any barcode, but aligned to the genome and do not mapped to any 
feature
\item{barcode_unmatch_mapped_to_exon} is the number of reads that do not
match any barcode, but aligned to the genome and mapped to the exon
}
}
\examples{
data("sc_sample_data")
data("sc_sample_qc")
sce = SingleCellExperiment(assays = list(counts =as.matrix(sc_sample_data)))
organism(sce) = "mmusculus_gene_ensembl"
gene_id_type(sce) = "ensembl_gene_id"
QC_metrics(sce) = sc_sample_qc
demultiplex_info(sce) = cell_barcode_matching
UMI_dup_info(sce) = UMI_duplication

demultiplex_info(sce)

}
\author{
Luyi Tian
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_atac_remove_duplicates.R
\name{sc_atac_remove_duplicates}
\alias{sc_atac_remove_duplicates}
\title{Removing PCR duplicates using samtools}
\usage{
sc_atac_remove_duplicates(inbam, samtools_path = NULL, output_folder = NULL)
}
\arguments{
\item{inbam}{The tagged, sorted and duplicate-free input BAM file}

\item{samtools_path}{The path of the samtools executable (if a custom installation is to be specified)}

\item{output_folder}{The path of the output folder}
}
\value{
file path to a bam file created from samtools markdup
}
\description{
Takes in a BAM file and removes the PCR duplicates using the samtools markdup function. Requires samtools 1.10 or newer for statistics to be generated.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/qc.R
\name{plot_UMI_dup}
\alias{plot_UMI_dup}
\title{Plot UMI duplication frequency}
\usage{
plot_UMI_dup(sce, log10_x = TRUE)
}
\arguments{
\item{sce}{a \code{SingleCellExperiment} object}

\item{log10_x}{whether to use log10 scale for x axis}
}
\value{
a line chart of the UMI duplication frequency
}
\description{
Plot the UMI duplication frequency.
}
\examples{
data("sc_sample_data")
data("sc_sample_qc")
sce = SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
organism(sce) = "mmusculus_gene_ensembl"
gene_id_type(sce) = "ensembl_gene_id"
QC_metrics(sce) = sc_sample_qc
demultiplex_info(sce) = cell_barcode_matching
UMI_dup_info(sce) = UMI_duplication

plot_UMI_dup(sce)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_atac_cell_calling.R
\name{sc_atac_filter_cell_calling}
\alias{sc_atac_filter_cell_calling}
\title{filter cell calling}
\usage{
sc_atac_filter_cell_calling(
  mtx,
  cell_qc_metrics_file,
  min_uniq_frags = 0,
  max_uniq_frags = 50000,
  min_frac_peak = 0.05,
  min_frac_tss = 0,
  min_frac_enhancer = 0,
  min_frac_promoter = 0,
  max_frac_mito = 0.2
)
}
\arguments{
\item{mtx}{The input matrix}

\item{cell_qc_metrics_file}{A file containing qc statistics for each cell}

\item{min_uniq_frags}{The minimum number of required unique fragments required for a cell}

\item{max_uniq_frags}{The maximum number of required unique fragments required for a cell}

\item{min_frac_peak}{The minimum proportion of fragments in a cell to overlap with a peak}

\item{min_frac_tss}{The minimum proportion of fragments in a cell to overlap with a tss}

\item{min_frac_enhancer}{The minimum proportion of fragments in a cell to overlap with a enhancer sequence}

\item{min_frac_promoter}{The minimum proportion of fragments in a cell to overlap with a promoter sequence}

\item{max_frac_mito}{The maximum proportion of fragments in a cell that are mitochondrial}
}
\description{
specify various qc cutoffs to select the desired cells
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_integrate.R
\name{sc_integrate}
\alias{sc_integrate}
\title{Integrate multi-omic scRNA-Seq and scATAC-Seq data into a MultiAssayExperiment}
\usage{
sc_integrate(
  sce_list,
  barcode_match_file = NULL,
  sce_column_to_barcode_files = NULL,
  rev_comp = NULL,
  cell_line_info = NULL,
  output_folder = NULL
)
}
\arguments{
\item{sce_list}{A list of SCE objects, named with the corresponding technologies}

\item{barcode_match_file}{A .csv file with columns corresponding to the barcodes for each tech}

\item{sce_column_to_barcode_files}{A list of files containing the barcodes for each tech (if not needed then give a `NULL` entry)}

\item{rev_comp}{A named list of technologies and logical flags specifying if reverse complements should be applyed for sequences (if not needed then provide a `NULL` entry)}

\item{cell_line_info}{A list of files, each of which contains 2 columns corresponding to the barcode and cell line for each tech (if not needed then provide a `NULL` entry)}

\item{output_folder}{The path to the output folder}
}
\value{
Returns a MultiAssayExperiment containing the scRNA-Seq and scATAC-Seq data produced by the scPipe pipelines
}
\description{
Generates an integrated SCE object with scRNA-Seq and scATAC-Seq data produced by the scPipe pipelines
}
\examples{
\dontrun{
sc_integrate(
   sce_list = list("RNA" = sce.rna, "ATAC" = sce.atac),
   barcode_match_file = bc_match_file,
   sce_column_to_barcode_files = list("RNA" = rna_bc_anno, "ATAC" = NULL),
   rev_comp = list("RNA" = FALSE, "ATAC" = TRUE),
   cell_line_info = list("RNA" = rna_cell_line_info, "ATAC" = atac_cell_line_info,)
   output_folder = output_folder
   )
}  

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrapper_scPipeCPP.R
\name{sc_correct_bam_bc}
\alias{sc_correct_bam_bc}
\title{sc_correct_bam_bc}
\usage{
sc_correct_bam_bc(
  inbam,
  outbam,
  bc_anno,
  max_mis = 1,
  bam_tags = list(am = "YE", ge = "GE", bc = "BC", mb = "OX"),
  mito = "MT",
  nthreads = 1
)
}
\arguments{
\item{inbam}{input bam file. This should be the output of
\code{sc_exon_mapping}}

\item{outbam}{output bam file with updated cell barcode}

\item{bc_anno}{barcode annotation, first column is cell id, second column
is cell barcode sequence}

\item{max_mis}{maximum mismatch allowed in barcode. (default: 1)}

\item{bam_tags}{list defining BAM tags where mapping information is
stored.
\itemize{
  \item "am": mapping status tag
  \item "ge": gene id
  \item "bc": cell barcode tag
  \item "mb": molecular barcode tag
}}

\item{mito}{mitochondrial chromosome name.
This should be consistent with the chromosome names in the bam file.}

\item{nthreads}{number of threads to use. (default: 1)}
}
\value{
no return
}
\description{
update the cell barcode tag in bam file with corrected barcode
output to a new bam file. the function will be useful for methods
that use the cell barcode information from bam file, such as `Demuxlet`
}
\examples{
data_dir="celseq2_demo"
barcode_annotation_fn = system.file("extdata", "barcode_anno.csv",
    package = "scPipe")
\dontrun{
# refer to the vignettes for the complete workflow
...
sc_correct_bam_bc(file.path(data_dir, "out.map.bam"),
    file.path(data_dir, "out.map.clean.bam"),
    barcode_annotation_fn)
...
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_atac_peak_calling.R
\name{sc_atac_peak_calling}
\alias{sc_atac_peak_calling}
\title{sc_atac_peak_calling()}
\usage{
sc_atac_peak_calling(
  inbam,
  ref = NULL,
  genome_size = NULL,
  output_folder = NULL
)
}
\arguments{
\item{inbam}{The input tagged, sorted, duplicate-free input BAM file}

\item{ref}{The reference genome file}

\item{genome_size}{The size of the genome}

\item{output_folder}{The path of the output folder}
}
\value{
None (invisible `NULL`)
}
\description{
sc_atac_peak_calling()
}
\examples{
\dontrun{
sc_atac_peak_calling(
    inbam,
    reference) 
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_atac_feature_counting.R
\name{sc_atac_feature_counting}
\alias{sc_atac_feature_counting}
\title{generating the feature by cell matrix}
\usage{
sc_atac_feature_counting(
  fragment_file,
  feature_input = NULL,
  bam_tags = list(bc = "CB", mb = "OX"),
  feature_type = "peak",
  organism = "hg38",
  cell_calling = "filter",
  sample_name = "",
  genome_size = NULL,
  promoters_file = NULL,
  tss_file = NULL,
  enhs_file = NULL,
  gene_anno_file = NULL,
  pheno_data = NULL,
  bin_size = NULL,
  yieldsize = 1e+06,
  n_filter_cell_counts = 200,
  n_filter_feature_counts = 10,
  exclude_regions = FALSE,
  excluded_regions_filename = NULL,
  output_folder = NULL,
  fix_chr = "none",
  lower = NULL,
  min_uniq_frags = 3000,
  max_uniq_frags = 50000,
  min_frac_peak = 0.3,
  min_frac_tss = 0,
  min_frac_enhancer = 0,
  min_frac_promoter = 0.1,
  max_frac_mito = 0.15,
  create_report = FALSE
)
}
\arguments{
\item{fragment_file}{The fragment file}

\item{feature_input}{The feature input data e.g. the .narrowPeak file for peaks of a bed file format}

\item{bam_tags}{The BAM tags}

\item{feature_type}{The type of feature}

\item{organism}{The organism type (contains hg19, hg38, mm10)}

\item{cell_calling}{The desired cell calling method; either \code{cellranger}, \code{emptydrops} or  \code{filter}.}

\item{sample_name}{The sample name to identify which is the data is analysed for.}

\item{genome_size}{The size of the genome (used for the \code{cellranger} cell calling method)}

\item{promoters_file}{The path of the promoter annotation file (if the specified organism isn't recognised).}

\item{tss_file}{The path of the tss annotation file (if the specified organism isn't recognised).}

\item{enhs_file}{The path of the enhs annotation file (if the specified organism isn't recognised).}

\item{gene_anno_file}{The path of the gene annotation file (gtf or gff3 format).}

\item{pheno_data}{The phenotypic data as a data frame}

\item{bin_size}{The size of the bins}

\item{yieldsize}{The yield size}

\item{n_filter_cell_counts}{An integer value to filter the feature matrix on the number of reads per cell (default = 200)}

\item{n_filter_feature_counts}{An integer value to filter the feature matrix on the number of reads per feature (default = 10).}

\item{exclude_regions}{Whether or not the regions (specified in the file) should be excluded}

\item{excluded_regions_filename}{The filename of the file containing the regions to be excluded}

\item{output_folder}{The output folder}

\item{fix_chr}{Whether chr should be fixed or not}

\item{lower}{the lower threshold for the data if using the \code{emptydrops} function for cell calling}

\item{min_uniq_frags}{The minimum number of required unique fragments required for a cell (used for \code{filter} cell calling)}

\item{max_uniq_frags}{The maximum number of required unique fragments required for a cell (used for \code{filter} cell calling)}

\item{min_frac_peak}{The minimum proportion of fragments in a cell to overlap with a peak (used for \code{filter} cell calling)}

\item{min_frac_tss}{The minimum proportion of fragments in a cell to overlap with a tss (used for \code{filter} cell calling)}

\item{min_frac_enhancer}{The minimum proportion of fragments in a cell to overlap with a enhancer sequence (used for \code{filter} cell calling)}

\item{min_frac_promoter}{The minimum proportion of fragments in a cell to overlap with a promoter sequence (used for \code{filter} cell calling)}

\item{max_frac_mito}{The maximum proportion of fragments in a cell that are mitochondrial (used for \code{filter} cell calling)}

\item{create_report}{Logical value to say whether to create the report or not (default = TRUE).}
}
\value{
None (invisible `NULL`)
}
\description{
feature matrix is created using a given demultiplexed BAM file and 
a selected feature type
}
\examples{
\dontrun{
sc_atac_feature_counting(
   fragment_file = fragment_file,
   cell_calling = "filter",
   exclude_regions = TRUE,
   feature_input = feature_file)
}    
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_atac_create_fragments.R
\name{sc_atac_create_fragments}
\alias{sc_atac_create_fragments}
\title{Generating the popular fragments for scATAC-Seq data}
\usage{
sc_atac_create_fragments(
  inbam,
  output_folder = "",
  min_mapq = 30,
  nproc = 1,
  cellbarcode = "CB",
  chromosomes = "^chr",
  readname_barcode = NULL,
  cells = NULL,
  max_distance = 5000,
  min_distance = 10,
  chunksize = 5e+05
)
}
\arguments{
\item{inbam}{The tagged, sorted and duplicate-free input BAM file}

\item{output_folder}{The path of the output folder}

\item{min_mapq}{: int
Minimum MAPQ to retain fragment}

\item{nproc}{: int, optional
Number of processors to use. Default is 1.}

\item{cellbarcode}{: str
Tag used for cell barcode. Default is CB (used by cellranger)}

\item{chromosomes}{: str, optional
Regular expression used to match chromosome names to include in the
output file. Default is "(?i)^chr" (starts with "chr", case-insensitive).
If None, use all chromosomes in the BAM file.}

\item{readname_barcode}{: str, optional
Regular expression used to match cell barocde stored in read name.
If None (default), use read tags instead. Use "[^:]*" to match all characters
before the first colon (":").}

\item{cells}{: str
File containing list of cell barcodes to retain. If None (default), use all cell barcodes
found in the BAM file.}

\item{max_distance}{: int, optional
Maximum distance between integration sites for the fragment to be retained.
Allows filtering of implausible fragments that likely result from incorrect
mapping positions. Default is 5000 bp.}

\item{min_distance}{: int, optional
Minimum distance between integration sites for the fragment to be retained.
Allows filtering implausible fragments that likely result from incorrect
mapping positions. Default is 10 bp.}

\item{chunksize}{: int
Number of BAM entries to read through before collapsing and writing
fragments to disk. Higher chunksize will use more memory but will be
faster.}
}
\value{
returns NULL
}
\description{
Takes in a tagged and sorted BAM file and outputs the associated fragments in a .bed file
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrapper_scPipeCPP.R
\name{sc_count_aligned_bam}
\alias{sc_count_aligned_bam}
\title{sc_count_aligned_bam}
\usage{
sc_count_aligned_bam(
  inbam,
  outbam,
  annofn,
  bam_tags = list(am = "YE", ge = "GE", bc = "BC", mb = "OX"),
  bc_len = 8,
  UMI_len = 6,
  stnd = TRUE,
  fix_chr = FALSE,
  outdir,
  bc_anno,
  max_mis = 1,
  mito = "MT",
  has_UMI = TRUE,
  UMI_cor = 1,
  gene_fl = FALSE,
  keep_mapped_bam = TRUE,
  nthreads = 1
)
}
\arguments{
\item{inbam}{input aligned bam file. can have multiple files as input}

\item{outbam}{output bam filename}

\item{annofn}{single string or vector of gff3 annotation filenames,
data.frame in SAF format or GRanges object containing complete gene_id
metadata column.}

\item{bam_tags}{list defining BAM tags where mapping information is
stored.
\itemize{
  \item "am": mapping status tag
  \item "ge": gene id
  \item "bc": cell barcode tag
  \item "mb": molecular barcode tag
}}

\item{bc_len}{total barcode length}

\item{UMI_len}{UMI length}

\item{stnd}{TRUE to perform strand specific mapping. (default: TRUE)}

\item{fix_chr}{TRUE to add `chr` to chromosome names, MT to chrM. (default: FALSE)}

\item{outdir}{output folder}

\item{bc_anno}{barcode annotation, first column is cell id, second column
is cell barcode sequence}

\item{max_mis}{maximum mismatch allowed in barcode. (default: 1)}

\item{mito}{mitochondrial chromosome name.
This should be consistent with the chromosome names in the bam file.}

\item{has_UMI}{whether the protocol contains UMI (default: TRUE)}

\item{UMI_cor}{correct UMI sequencing error: 0 means no correction, 1 means
simple correction and merge UMI with distance 1. 2 means merge on both UMI
alignment position match.}

\item{gene_fl}{whether to remove low abundance genes. A gene is considered to
have low abundance if only one copy of one UMI is associated with it.}

\item{keep_mapped_bam}{TRUE if feature mapped bam file should be retained.}

\item{nthreads}{number of threads to use. (default: 1)}
}
\value{
no return
}
\description{
Wrapper to run \code{\link{sc_exon_mapping}},
  \code{\link{sc_demultiplex}} and \code{\link{sc_gene_counting}} with a
  single command
}
\examples{
\dontrun{
sc_count_aligned_bam(
  inbam = "aligned.bam",
  outbam = "mapped.bam",
  annofn = c("MusMusculus-GRCm38p4-UCSC.gff3", "ERCC92_anno.gff3"),
  outdir = "output",
  bc_anno = "barcodes.csv"
)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_workflow.R
\name{create_sce_by_dir}
\alias{create_sce_by_dir}
\title{create a SingleCellExperiment object from data folder generated by preprocessing step}
\usage{
create_sce_by_dir(
  datadir,
  organism = NULL,
  gene_id_type = NULL,
  pheno_data = NULL,
  report = FALSE
)
}
\arguments{
\item{datadir}{the directory that contains all the data and `stat` subfolder.}

\item{organism}{the organism of the data. List of possible names can be retrieved using the function
`listDatasets`from `biomaRt` package. (i.e `mmusculus_gene_ensembl` or `hsapiens_gene_ensembl`)}

\item{gene_id_type}{gene id type of the data A possible list of ids can be retrieved using the function `listAttributes` from `biomaRt` package.
the commonly used id types are `external_gene_name`, `ensembl_gene_id` or `entrezgene`}

\item{pheno_data}{the external phenotype data that linked to each single cell. This should be an \code{AnnotatedDataFrame} object}

\item{report}{whether to generate the html report in the data folder}
}
\value{
a SingleCellExperiment object
}
\description{
after we run \code{sc_gene_counting} and finish the preprocessing step. \code{create_sce_by_dir}
can be used to generate the \link{SingleCellExperiment} object from the folder that contains gene count matrix and QC statistics.
it can also generate the html report based on the gene count and quality control statistics
}
\details{
after we run \code{sc_gene_counting} and finish the preprocessing step. \code{create_sce_by_dir}
can be used to generate the SingleCellExperiment object from the folder that contains gene count matrix and QC statistics.
}
\examples{
\dontrun{
# the sce can be created fron the output folder of scPipe
# please refer to the vignettes
sce = create_sce_by_dir(datadir="output_dir_of_scPipe",
    organism="mmusculus_gene_ensembl",
    gene_id_type="ensembl_gene_id")
}
# or directly from the gene count and quality control matrix:
data("sc_sample_data")
data("sc_sample_qc")
sce = SingleCellExperiment(assays = list(counts = as.matrix(sc_sample_data)))
organism(sce) = "mmusculus_gene_ensembl"
gene_id_type(sce) = "ensembl_gene_id"
QC_metrics(sce) = sc_sample_qc
demultiplex_info(sce) = cell_barcode_matching
UMI_dup_info(sce) = UMI_duplication
dim(sce)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sc_atac_subfunctions.R
\name{sc_atac_plot_fragments_per_feature}
\alias{sc_atac_plot_fragments_per_feature}
\title{A histogram of the log-number of fragments per feature}
\usage{
sc_atac_plot_fragments_per_feature(sce)
}
\arguments{
\item{sce}{The SingleExperimentObject produced by the sc_atac_create_sce function at the end of the pipeline}
}
\value{
returns NULL
}
\description{
A histogram of the log-number of fragments per feature
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrapper_scPipeCPP.R
\name{sc_trim_barcode}
\alias{sc_trim_barcode}
\title{sc_trim_barcode}
\usage{
sc_trim_barcode(
  outfq,
  r1,
  r2 = NULL,
  read_structure = list(bs1 = -1, bl1 = 0, bs2 = 6, bl2 = 8, us = 0, ul = 6),
  filter_settings = list(rmlow = TRUE, rmN = TRUE, minq = 20, numbq = 2)
)
}
\arguments{
\item{outfq}{the output fastq file, which reformat the barcode and UMI into
the read name. Files ending in \code{.gz} will be automatically compressed.}

\item{r1}{read one for pair-end reads. This read should contain
the transcript.}

\item{r2}{read two for pair-end reads, NULL if single read.
(default: NULL)}

\item{read_structure}{a list containing the read structure configuration:
\itemize{
  \item{bs1}: starting position of barcode in read one. -1 if no barcode in
    read one.
  \item{bl1}: length of barcode in read one, if there is no
    barcode in read one this number is used for trimming beginning of read
    one.
  \item{bs2}: starting position of barcode in read two
  \item{bl2}: length of barcode in read two
  \item{us}: starting position of UMI
  \item{ul}: length of UMI
}}

\item{filter_settings}{A list contains read filter settings:\itemize{
\item{rmlow} whether to remove the low quality reads.
\item{rmN} whether to remove reads that contains N in UMI or cell barcode.
\item{minq} the minimum base pair quality that we allowed
\item{numbq} the maximum number of base pair that have quality
below \code{numbq}
}}
}
\value{
generates a trimmed fastq file named \code{outfq}
}
\description{
Reformat fastq files so barcode and UMI sequences are moved from
  the sequence into the read name.
}
\details{
Positions used in this function are 0-indexed, so they start from 0
 rather than 1. The default read structure in this function represents
 CEL-seq paired-ended reads. This contains a transcript in the first read, a
 UMI in the first 6bp of the second read followed by a 8bp barcode. So the
 read structure will be : \code{list(bs1=-1, bl1=0, bs2=6, bl2=8, us=0,
 ul=6)}. \code{bs1=-1, bl1=0} indicates negative start position and zero
 length for the barcode on read one, this is used to denote "no barcode" on
 read one. \code{bs2=6, bl2=8} indicates there is a barcode in read two that
 starts at the 7th base with length 8bp. \code{us=0, ul=6} indicates a UMI
 from first base of read two and the length in 6bp.

 For a typical Drop-seq experiment the read structure will be
 \code{list(bs1=-1, bl1=0, bs2=0, bl2=12, us=12, ul=8)}, which means the read
 one only contains transcript, the first 12bp in read two are cell barcode, followed
 by a 8bp UMI.
}
\examples{
data_dir="celseq2_demo"
\dontrun{
# for the complete workflow, refer to the vignettes
...
sc_trim_barcode(file.path(data_dir, "combined.fastq"),
   file.path(data_dir, "simu_R1.fastq"),
   file.path(data_dir, "simu_R2.fastq"))
...
}
}
